/*
 * Copyright (c) 2007, Novell Inc.
 *
 * This program is licensed under the BSD license, read LICENSE.BSD
 * for further information
 */

#ifndef COMMON_WRITE_H
#define COMMON_WRITE_H

#include "repo.h"

void tool_write(Repo *repo, FILE *fp);

#endif
