/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040401
#define SWIGPERL
#define SWIG_CASTRANK_MODE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#ifdef __cplusplus
/* Needed on some windows machines---since MS plays funny games with the header files under C++ */
#include <math.h>
#include <stdlib.h>
extern "C" {
#endif

#if __GNUC__ >= 10
#if defined(__cplusplus)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wvolatile"
#endif
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#if __GNUC__ >= 10
#if defined(__cplusplus)
#pragma GCC diagnostic pop
#endif
#endif

/* PERL_REVISION was added in Perl 5.6. */
#if !defined PERL_REVISION || (PERL_REVISION-0 == 5 && PERL_VERSION-0 < 8)
# error SWIG requires Perl >= 5.8.0
#endif

#if defined(WIN32) && defined(PERL_OBJECT) && !defined(PerlIO_exportFILE)
#define PerlIO_exportFILE(fh,fl) (FILE*)(fh)
#endif

#ifndef SvIOK_UV
# define SvIOK_UV(sv)       (SvIOK(sv) && (SvUVX(sv) == SvIVX(sv)))
#endif

#ifndef SvUOK
# define SvUOK(sv)           SvIOK_UV(sv)
#endif

#ifndef IVSIZE
#  ifdef LONGSIZE
#    define IVSIZE LONGSIZE
#  else
#    define IVSIZE 4 /* A bold guess, but the best we can make. */
#  endif
#endif

#ifndef INT2PTR
#  if (IVSIZE == PTRSIZE) && (UVSIZE == PTRSIZE)
#    define PTRV                  UV
#    define INT2PTR(any,d)        (any)(d)
#  else
#    if PTRSIZE == LONGSIZE
#      define PTRV                unsigned long
#    else
#      define PTRV                unsigned
#    endif
#    define INT2PTR(any,d)        (any)(PTRV)(d)
#  endif

#  define NUM2PTR(any,d)  (any)(PTRV)(d)
#  define PTR2IV(p)       INT2PTR(IV,p)
#  define PTR2UV(p)       INT2PTR(UV,p)
#  define PTR2NV(p)       NUM2PTR(NV,p)

#  if PTRSIZE == LONGSIZE
#    define PTR2ul(p)     (unsigned long)(p)
#  else
#    define PTR2ul(p)     INT2PTR(unsigned long,p)
#  endif
#endif /* !INT2PTR */

#ifndef SvPV_nolen
# define SvPV_nolen(x) SvPV(x,PL_na)
#endif

#ifndef get_sv
#  define get_sv perl_get_sv
#endif

#ifndef ERRSV
#  define ERRSV get_sv("@",FALSE)
#endif

#ifndef pTHX_
#define pTHX_
#endif   

#include <string.h>
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGINTERN const char*
SWIG_Perl_ErrorType(int code) {
  switch(code) {
  case SWIG_MemoryError:
    return "MemoryError";
  case SWIG_IOError:
    return "IOError";
  case SWIG_RuntimeError:
    return "RuntimeError";
  case SWIG_IndexError:
    return "IndexError";
  case SWIG_TypeError:
    return "TypeError";
  case SWIG_DivisionByZero:
    return "ZeroDivisionError";
  case SWIG_OverflowError:
    return "OverflowError";
  case SWIG_SyntaxError:
    return "SyntaxError";
  case SWIG_ValueError:
    return "ValueError";
  case SWIG_SystemError:
    return "SystemError";
  case SWIG_AttributeError:
    return "AttributeError";
  case SWIG_NullReferenceError:
    return "NullReferenceError";
  default:
    return "RuntimeError";
  }
}


/* -----------------------------------------------------------------------------
 * perlrun.swg
 *
 * This file contains the runtime support for Perl modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

#define SWIG_PERL_OBJECT_DECL
#define SWIG_PERL_OBJECT_CALL

/* Common SWIG API */

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pp, type, flags)           SWIG_Perl_ConvertPtr(SWIG_PERL_OBJECT_CALL obj, pp, type, flags)
#define SWIG_ConvertPtrAndOwn(obj, pp, type, flags,own) SWIG_Perl_ConvertPtrAndOwn(SWIG_PERL_OBJECT_CALL obj, pp, type, flags, own)
#define SWIG_NewPointerObj(p, type, flags)              SWIG_Perl_NewPointerObj(SWIG_PERL_OBJECT_CALL p, type, flags)
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Perl_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, p, s, type)             SWIG_Perl_ConvertPacked(SWIG_PERL_OBJECT_CALL obj, p, s, type)
#define SWIG_NewPackedObj(p, s, type)	                SWIG_Perl_NewPackedObj(SWIG_PERL_OBJECT_CALL p, s, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Perl_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Perl_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Perl_ErrorType(code)               
#define SWIG_Error(code, msg)            		sv_setpvf(get_sv("@", GV_ADD), "%s %s", SWIG_ErrorType(code), msg)
#define SWIG_fail                        		goto fail						    

/* Perl-specific SWIG API */

#define SWIG_MakePtr(sv, ptr, type, flags)              SWIG_Perl_MakePtr(SWIG_PERL_OBJECT_CALL sv, ptr, type, flags)
#define SWIG_MakePackedObj(sv, p, s, type)	        SWIG_Perl_MakePackedObj(SWIG_PERL_OBJECT_CALL sv, p, s, type)
#define SWIG_SetError(str)                              SWIG_Error(SWIG_RuntimeError, str)


#define SWIG_PERL_DECL_ARGS_1(arg1)                     (SWIG_PERL_OBJECT_DECL arg1)
#define SWIG_PERL_CALL_ARGS_1(arg1)                     (SWIG_PERL_OBJECT_CALL arg1)
#define SWIG_PERL_DECL_ARGS_2(arg1, arg2)               (SWIG_PERL_OBJECT_DECL arg1, arg2)
#define SWIG_PERL_CALL_ARGS_2(arg1, arg2)               (SWIG_PERL_OBJECT_CALL arg1, arg2)

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

#ifdef __cplusplus
extern "C" {
#endif

#define SWIG_OWNER   SWIG_POINTER_OWN
#define SWIG_SHADOW  SWIG_OWNER << 1

#define SWIG_MAYBE_PERL_OBJECT SWIG_PERL_OBJECT_DECL

/* SWIG Perl macros */

/* Macro to declare an XS function */
#ifndef XSPROTO
#   define XSPROTO(name) void name(pTHX_ CV* cv)
#endif

/* Macro to call an XS function */
#ifndef MULTIPLICITY
#  define SWIG_CALLXS(_name) _name(cv)
#else
#  define SWIG_CALLXS(_name) _name(PERL_GET_THX, cv)
#endif

#define MAGIC_PPERL
#define SWIGCLASS_STATIC static SWIGUNUSED

#ifndef MULTIPLICITY
#define SWIG_MAGIC(a,b) (SV *a, MAGIC *b)

#ifdef __cplusplus
extern "C" {
#endif
typedef int (*SwigMagicFunc)(SV *, MAGIC *);
#ifdef __cplusplus
}
#endif

#else /* MULTIPLICITY */

#define SWIG_MAGIC(a,b) (struct interpreter *interp, SV *a, MAGIC *b)

#ifdef __cplusplus
extern "C" {
#endif
typedef int (*SwigMagicFunc)(struct interpreter *, SV *, MAGIC *);
#ifdef __cplusplus
}
#endif

#endif /* MULTIPLICITY */

static void SWIGUNUSED SWIG_croak_null(void)
{
  SV *err = get_sv("@", GV_ADD);
  if (sv_isobject(err))
    croak(0);
  else
    croak("%s", SvPV_nolen(err));
}


/* 
   Define how strict is the cast between strings and integers/doubles
   when overloading between these types occurs.
   
   The default is making it as strict as possible by using SWIG_AddCast
   when needed.
   
   You can use -DSWIG_PERL_NO_STRICT_STR2NUM at compilation time to
   disable the SWIG_AddCast, making the casting between string and
   numbers less strict.

   In the end, we try to solve the overloading between strings and
   numerical types in the more natural way, but if you can avoid it,
   well, avoid it using %rename, for example.
*/
#ifndef SWIG_PERL_NO_STRICT_STR2NUM
# ifndef SWIG_PERL_STRICT_STR2NUM
#  define SWIG_PERL_STRICT_STR2NUM
# endif
#endif
#ifdef SWIG_PERL_STRICT_STR2NUM
/* string takes precedence */
#define SWIG_Str2NumCast(x) SWIG_AddCast(x)  
#else
/* number takes precedence */
#define SWIG_Str2NumCast(x) x
#endif



#include <stdlib.h>

SWIGRUNTIME const char *
SWIG_Perl_TypeProxyName(const swig_type_info *type) {
  if (type) {
    if (type->clientdata != NULL) {
      return (const char *) type->clientdata;
    }
    else {
      return type->name;
    }
  }
  return NULL;
}

/* SWIG_TypeCheck without binary search by hash value */
SWIGRUNTIME swig_cast_info *
SWIG_TypeProxyCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head, *cast;
    head = ty->cast;
    while (head) {
      for (cast = head; (cast - head) <= head->value; cast++) {
        if (strcmp(SWIG_Perl_TypeProxyName(cast->type), c) == 0) {
          return cast;
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Perl_AcquirePtr(SWIG_MAYBE_PERL_OBJECT SV *SWIGUNUSEDPARM(sv), int SWIGUNUSEDPARM(own)) {
  /* TODO */
  return 0;
}

/* Function for getting a pointer value */

SWIGRUNTIME int
SWIG_Perl_ConvertPtrAndOwn(SWIG_MAYBE_PERL_OBJECT SV *sv, void **ptr, swig_type_info *_t, int flags, int *own) {
  swig_cast_info *tc;
  void *voidptr = (void *)0;
  SV *tsv = 0;
  int check_owned_pointer_release = (flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE;

  if (own)
    *own = 0;

  /* If magical, apply more magic */
  if (SvGMAGICAL(sv))
    mg_get(sv);

  /* Check to see if this is an object */
  if (sv_isobject(sv)) {
    IV tmp = 0;
    tsv = (SV*) SvRV(sv);
    if ((SvTYPE(tsv) == SVt_PVHV || SvTYPE(tsv) == SVt_PVAV)) {
      MAGIC *mg;
      if (SvMAGICAL(tsv)) {
        mg = mg_find(tsv,'P');
        if (mg) {
          sv = mg->mg_obj;
          if (sv_isobject(sv)) {
	    tsv = (SV*)SvRV(sv);
            tmp = SvIV(tsv);
          }
        }
      } else {
        return SWIG_ERROR;
      }
    } else {
      tmp = SvIV(tsv);
    }
    voidptr = INT2PTR(void *,tmp);
  } else if (! SvOK(sv)) {            /* Check for undef */
    *(ptr) = (void *) 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  } else if (SvTYPE(sv) == SVt_RV) {  /* Check for NULL pointer */
    if (!SvROK(sv)) {
      /* In Perl 5.12 and later, SVt_RV == SVt_IV, so sv could be a valid integer value.  */
      if (SvIOK(sv)) {
        return SWIG_ERROR;
      } else {
        /* NULL pointer (reference to undef). */
        *(ptr) = (void *) 0;
        return SWIG_OK;
      }
    } else {
      return SWIG_ERROR;
    }
  } else {                            /* Don't know what it is */
    return SWIG_ERROR;
  }
  if (_t) {
    /* Now see if the types match */
    char *_c = HvNAME(SvSTASH(SvRV(sv)));
    tc = SWIG_TypeProxyCheck(_c,_t);
#ifdef SWIG_DIRECTORS
    if (!tc && !sv_derived_from(sv,SWIG_Perl_TypeProxyName(_t))) {
#else
    if (!tc) {
#endif
      return SWIG_ERROR;
    }
    {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,voidptr,&newmemory);
      if (newmemory == SWIG_CAST_NEW_MEMORY) {
        assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
        if (own)
          *own = *own | SWIG_CAST_NEW_MEMORY;
      }
    }
  } else {
    *ptr = voidptr;
  }

  /* 
   *  DISOWN implementation: we need a perl guru to check this one.
   */
  if (tsv && ((flags & SWIG_POINTER_DISOWN) || check_owned_pointer_release)) {
    /* 
     *  almost copy paste code from below SWIG_POINTER_OWN setting
     */
    SV *obj = sv;
    HV *stash = SvSTASH(SvRV(obj));
    GV *gv = *(GV**)hv_fetch(stash, "OWNER", 5, TRUE);
    int owned = 0;
    if (isGV(gv)) {
      HV *hv = GvHVn(gv);
      /*
       * To set ownership (see below), a newSViv(1) entry is added. 
       * Hence, to remove ownership, we delete the entry.
       */
      if (hv_exists_ent(hv, obj, 0)) {
        owned = 1;
        if (flags & SWIG_POINTER_DISOWN) {
          hv_delete_ent(hv, obj, 0, 0);
        }
      }
    }
    if (check_owned_pointer_release && !owned) {
      return SWIG_ERROR_RELEASE_NOT_OWNED;
    }
  }

  if (tsv && (flags & SWIG_POINTER_CLEAR)) {
    SvIV_set(tsv, 0);
  }

  return SWIG_OK;
}

SWIGRUNTIME int
SWIG_Perl_ConvertPtr(SWIG_MAYBE_PERL_OBJECT SV *sv, void **ptr, swig_type_info *_t, int flags) {
  return SWIG_Perl_ConvertPtrAndOwn(sv, ptr, _t, flags, 0);
}

SWIGRUNTIME void
SWIG_Perl_MakePtr(SWIG_MAYBE_PERL_OBJECT SV *sv, void *ptr, swig_type_info *t, int flags) {
  if (ptr && (flags & (SWIG_SHADOW | SWIG_POINTER_OWN))) {
    SV *self;
    SV *obj=newSV(0);
    HV *hash=newHV();
    HV *stash;
    sv_setref_pv(obj, SWIG_Perl_TypeProxyName(t), ptr);
    stash=SvSTASH(SvRV(obj));
    if (flags & SWIG_POINTER_OWN) {
      HV *hv;
      GV *gv = *(GV**)hv_fetch(stash, "OWNER", 5, TRUE);
      if (!isGV(gv))
        gv_init(gv, stash, "OWNER", 5, FALSE);
      hv=GvHVn(gv);
      hv_store_ent(hv, obj, newSViv(1), 0);
    }
    sv_magic((SV *)hash, (SV *)obj, 'P', Nullch, 0);
    SvREFCNT_dec(obj);
    self=newRV_noinc((SV *)hash);
    sv_setsv(sv, self);
    SvREFCNT_dec((SV *)self);
    sv_bless(sv, stash);
  }
  else {
    sv_setref_pv(sv, SWIG_Perl_TypeProxyName(t), ptr);
  }
}

SWIGRUNTIMEINLINE SV *
SWIG_Perl_NewPointerObj(SWIG_MAYBE_PERL_OBJECT void *ptr, swig_type_info *t, int flags) {
  SV *result = sv_newmortal();
  SWIG_MakePtr(result, ptr, t, flags);
  return result;
}

SWIGRUNTIME void
SWIG_Perl_MakePackedObj(SWIG_MAYBE_PERL_OBJECT SV *sv, void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(SWIG_Perl_TypeProxyName(type))) > 1000) return;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  strcpy(r,SWIG_Perl_TypeProxyName(type));
  sv_setpv(sv, result);
}

SWIGRUNTIME SV *
SWIG_Perl_NewPackedObj(SWIG_MAYBE_PERL_OBJECT void *ptr, int sz, swig_type_info *type) {
  SV *result = sv_newmortal();
  SWIG_Perl_MakePackedObj(result, ptr, sz, type);
  return result;
}

/* Convert a packed pointer value */
SWIGRUNTIME int
SWIG_Perl_ConvertPacked(SWIG_MAYBE_PERL_OBJECT SV *obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c = 0;

  if ((!obj) || (!SvOK(obj))) return SWIG_ERROR;
  c = SvPV_nolen(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') return SWIG_ERROR;
  c++;
  c = SWIG_UnpackData(c,ptr,sz);
  if (ty) {
    tc = SWIG_TypeCheck(c,ty);
    if (!tc) return SWIG_ERROR;
  }
  return SWIG_OK;
}


/* Macros for low-level exception handling */
#define SWIG_croak(x)    { SWIG_Error(SWIG_RuntimeError, x); SWIG_fail; }


typedef XSPROTO(SwigPerlWrapper);
typedef SwigPerlWrapper *SwigPerlWrapperPtr;

/* Structure for command table */
typedef struct {
  const char         *name;
  SwigPerlWrapperPtr  wrapper;
} swig_command_info;

/* Information for constant table */

#define SWIG_INT     1
#define SWIG_FLOAT   2
#define SWIG_STRING  3
#define SWIG_POINTER 4
#define SWIG_BINARY  5

/* Constant information structure */
typedef struct swig_constant_info {
    int              type;
    const char      *name;
    long             lvalue;
    double           dvalue;
    void            *pvalue;
    swig_type_info **ptype;
} swig_constant_info;


/* Structure for variable table */
typedef struct {
  const char   *name;
  SwigMagicFunc   set;
  SwigMagicFunc   get;
  swig_type_info  **type;
} swig_variable_info;

/* Magic variable code */
#ifdef __cplusplus
#  define swig_create_magic(s,a,b,c) _swig_create_magic(s,const_cast<char*>(a),b,c)
#else
#  define swig_create_magic(s,a,b,c) _swig_create_magic(s,(char*)(a),b,c)
#endif
#ifndef MULTIPLICITY
SWIGRUNTIME void _swig_create_magic(SV *sv, char *name, int (*set)(SV *, MAGIC *), int (*get)(SV *,MAGIC *))
#else
SWIGRUNTIME void _swig_create_magic(SV *sv, char *name, int (*set)(struct interpreter*, SV *, MAGIC *), int (*get)(struct interpreter*, SV *,MAGIC *))
#endif
{
  MAGIC *mg;
  sv_magic(sv,sv,'U',name,(I32)strlen(name));
  mg = mg_find(sv,'U');
  mg->mg_virtual = (MGVTBL *) malloc(sizeof(MGVTBL));
  mg->mg_virtual->svt_get = (SwigMagicFunc) get;
  mg->mg_virtual->svt_set = (SwigMagicFunc) set;
  mg->mg_virtual->svt_len = 0;
  mg->mg_virtual->svt_clear = 0;
  mg->mg_virtual->svt_free = 0;
}


SWIGRUNTIME swig_module_info *
SWIG_Perl_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  SV *pointer;

  /* first check if pointer already created */
  if (!type_pointer) {
    pointer = get_sv("swig_runtime_data::type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, FALSE | GV_ADDMULTI);
    if (pointer && SvOK(pointer)) {
      type_pointer = INT2PTR(swig_type_info **, SvIV(pointer));
    }
  }

  return (swig_module_info *) type_pointer;
}

SWIGRUNTIME void
SWIG_Perl_SetModule(swig_module_info *module) {
  SV *pointer;

  /* create a new pointer */
  pointer = get_sv("swig_runtime_data::type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, TRUE | GV_ADDMULTI);
  sv_setiv(pointer, PTR2IV(module));
}

#ifdef __cplusplus
}
#endif

/* Workaround perl5 global namespace pollution. Note that undefining library
 * functions like fopen will not solve the problem on all platforms as fopen
 * might be a macro on Windows but not necessarily on other operating systems. */
#ifdef do_open
  #undef do_open
#endif
#ifdef do_close
  #undef do_close
#endif
#ifdef do_exec
  #undef do_exec
#endif
#ifdef scalar
  #undef scalar
#endif
#ifdef list
  #undef list
#endif
#ifdef apply
  #undef apply
#endif
#ifdef convert
  #undef convert
#endif
#ifdef Error
  #undef Error
#endif
#ifdef form
  #undef form
#endif
#ifdef vform
  #undef vform
#endif
#ifdef LABEL
  #undef LABEL
#endif
#ifdef METHOD
  #undef METHOD
#endif
#ifdef Move
  #undef Move
#endif
#ifdef yylex
  #undef yylex
#endif
#ifdef yyparse
  #undef yyparse
#endif
#ifdef yyerror
  #undef yyerror
#endif
#ifdef invert
  #undef invert
#endif
#ifdef ref
  #undef ref
#endif
#ifdef read
  #undef read
#endif
#ifdef write
  #undef write
#endif
#ifdef eof
  #undef eof
#endif
#ifdef close
  #undef close
#endif
#ifdef rewind
  #undef rewind
#endif
#ifdef free
  #undef free
#endif
#ifdef malloc
  #undef malloc
#endif
#ifdef calloc
  #undef calloc
#endif
#ifdef Stat
  #undef Stat
#endif
#ifdef check
  #undef check
#endif
#ifdef seekdir
  #undef seekdir
#endif
#ifdef open
  #undef open
#endif
#ifdef readdir
  #undef readdir
#endif
#ifdef bind
  #undef bind
#endif
#ifdef access
  #undef access
#endif
#ifdef stat
  #undef stat
#endif
#ifdef seed
  #undef seed
#endif

#ifdef bool
  /* Leave if macro is from C99 stdbool.h */
  #ifndef __bool_true_false_are_defined
    #undef bool
  #endif
#endif




#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_Alternative swig_types[0]
#define SWIGTYPE_p_Chksum swig_types[1]
#define SWIGTYPE_p_Dataiterator swig_types[2]
#define SWIGTYPE_p_Datamatch swig_types[3]
#define SWIGTYPE_p_Datapos swig_types[4]
#define SWIGTYPE_p_Decision swig_types[5]
#define SWIGTYPE_p_Decisionset swig_types[6]
#define SWIGTYPE_p_Dep swig_types[7]
#define SWIGTYPE_p_Job swig_types[8]
#define SWIGTYPE_p_Pool swig_types[9]
#define SWIGTYPE_p_Pool_repo_iterator swig_types[10]
#define SWIGTYPE_p_Pool_solvable_iterator swig_types[11]
#define SWIGTYPE_p_Problem swig_types[12]
#define SWIGTYPE_p_Queue swig_types[13]
#define SWIGTYPE_p_Repo swig_types[14]
#define SWIGTYPE_p_Repo_solvable_iterator swig_types[15]
#define SWIGTYPE_p_Ruleinfo swig_types[16]
#define SWIGTYPE_p_Selection swig_types[17]
#define SWIGTYPE_p_Solution swig_types[18]
#define SWIGTYPE_p_Solutionelement swig_types[19]
#define SWIGTYPE_p_SolvFp swig_types[20]
#define SWIGTYPE_p_Solver swig_types[21]
#define SWIGTYPE_p_Solvsig swig_types[22]
#define SWIGTYPE_p_Transaction swig_types[23]
#define SWIGTYPE_p_TransactionClass swig_types[24]
#define SWIGTYPE_p_XRepodata swig_types[25]
#define SWIGTYPE_p_XRule swig_types[26]
#define SWIGTYPE_p_XSolvable swig_types[27]
#define SWIGTYPE_p_char swig_types[28]
#define SWIGTYPE_p_int swig_types[29]
#define SWIGTYPE_p_p_XRule swig_types[30]
#define SWIGTYPE_p_unsigned_char swig_types[31]
#define SWIGTYPE_p_unsigned_int swig_types[32]
static swig_type_info *swig_types[34];
static swig_module_info swig_module = {swig_types, 33, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    boot_solv

#define SWIG_name   "solvc::boot_solv"
#define SWIG_prefix "solvc::"

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#ifdef __cplusplus
extern "C"
#endif
#ifndef MULTIPLICITY
SWIGEXPORT void SWIG_init (CV* cv);
#else
SWIGEXPORT void SWIG_init (pTHXo_ CV* cv);
#endif


typedef struct {
  const void *data;
  size_t len;
} BinaryBlob;



#define SWIG_PERL_ITERATOR      0x80

SWIGRUNTIMEINLINE SV *
SWIG_Perl_NewArrayObj(SWIG_MAYBE_PERL_OBJECT void *ptr, swig_type_info *t, int flags) {
  SV *result = sv_newmortal();
  if (ptr && (flags & (SWIG_SHADOW | SWIG_POINTER_OWN))) {
    SV *self;
    SV *obj=newSV(0);
    AV *array=newAV();
    HV *stash;
    sv_setref_pv(obj, (char *) SWIG_Perl_TypeProxyName(t), ptr);
    stash=SvSTASH(SvRV(obj));
    if (flags & SWIG_POINTER_OWN) {
      HV *hv;
      GV *gv=*(GV**)hv_fetch(stash, "OWNER", 5, TRUE);
      if (!isGV(gv))
        gv_init(gv, stash, "OWNER", 5, FALSE);
      hv=GvHVn(gv);
      hv_store_ent(hv, obj, newSViv(1), 0);
    }
    if (flags & SWIG_PERL_ITERATOR) {
      HV *hv;
      GV *gv=*(GV**)hv_fetch(stash, "ITERATORS", 9, TRUE);
      AV *av=newAV();
      if (!isGV(gv))
        gv_init(gv, stash, "ITERATORS", 9, FALSE);
      hv=GvHVn(gv);
      hv_store_ent(hv, obj, newRV_inc((SV *)av), 0);
    }
    sv_magic((SV *)array, (SV *)obj, 'P', Nullch, 0);
    SvREFCNT_dec(obj);
    self=newRV_noinc((SV *)array);
    sv_setsv(result, self);
    SvREFCNT_dec((SV *)self);
    sv_bless(result, stash);
  } else {
    sv_setref_pv(result, (char *) SWIG_Perl_TypeProxyName(t), ptr);
  }
  return result;
}



#include <stdbool.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

/* argh, swig undefs bool for perl */
#ifndef bool
#if !defined __STDC_VERSION__ || __STDC_VERSION__ < 202311L
typedef int bool;
#endif
#endif

#include "pool.h"
#include "poolarch.h"
#include "evr.h"
#include "solver.h"
#include "policy.h"
#include "solverdebug.h"
#include "repo_solv.h"
#include "chksum.h"
#include "selection.h"

#include "repo_write.h"
#if defined(ENABLE_RPMDB) || defined(ENABLE_RPMPKG)
#include "repo_rpmdb.h"
#endif
#ifdef ENABLE_PUBKEY
#include "repo_pubkey.h"
#endif
#ifdef ENABLE_DEBIAN
#include "repo_deb.h"
#endif
#ifdef ENABLE_RPMMD
#include "repo_rpmmd.h"
#include "repo_updateinfoxml.h"
#include "repo_deltainfoxml.h"
#include "repo_repomdxml.h"
#endif
#ifdef ENABLE_SUSEREPO
#include "repo_products.h"
#include "repo_susetags.h"
#include "repo_content.h"
#endif
#ifdef ENABLE_MDKREPO
#include "repo_mdk.h"
#endif
#ifdef ENABLE_ARCHREPO
#include "repo_arch.h"
#endif
#ifdef SUSE
#include "repo_autopattern.h"
#endif
#if defined(ENABLE_COMPLEX_DEPS) && (defined(ENABLE_SUSEREPO) || defined(ENABLE_RPMMD) || defined(ENABLE_RPMDB) || defined(ENABLE_RPMPKG))
#include "pool_parserpmrichdep.h"
#endif
#include "solv_xfopen.h"
#include "testcase.h"

/* for old ruby versions */
#ifndef RARRAY_PTR
#define RARRAY_PTR(ary) (RARRAY(ary)->ptr)
#endif
#ifndef RARRAY_LEN
#define RARRAY_LEN(ary) (RARRAY(ary)->len)
#endif

typedef void *AppObjectPtr;
typedef Id DepId;

typedef struct {
  Pool *pool;
  Id id;
} Dep;

typedef struct {
  Pool *pool;
  Id id;
} XSolvable;

typedef struct {
  Solver *solv;
  Id id;
} XRule;

typedef struct {
  Repo *repo;
  Id id;
} XRepodata;

typedef struct {
  Pool *pool;
  Id id;
} Pool_solvable_iterator;

typedef struct {
  Pool *pool;
  Id id;
} Pool_repo_iterator;

typedef struct {
  Repo *repo;
  Id id;
} Repo_solvable_iterator;

typedef struct {
  Pool *pool;
  int how;
  Id what;
} Job;

typedef struct {
  Solver *solv;
  Id id;
} Problem;

typedef struct {
  Solver *solv;
  Id problemid;
  Id id;
} Solution;

typedef struct {
  Solver *solv;
  Id problemid;
  Id solutionid;

  Id type;
  Id p;
  Id rp;
} Solutionelement;

typedef struct {
  Solver *solv;
  Id rid;
  int type;
  Id source;
  Id target;
  Id dep_id;
} Ruleinfo;

typedef struct {
  Solver *solv;
  Id type;
  Id rid;
  Id from_id;
  Id dep_id;
  Id chosen_id;
  Queue choices;
  int level;
} Alternative;

typedef struct {
  Transaction *transaction;
  int mode;
  Id type;
  int count;
  Id fromid;
  Id toid;
} TransactionClass;

typedef struct {
  Pool *pool;
  Queue q;
  int flags;
} Selection;

typedef struct {
  Solver *solv;
  Id p;
  int reason;
  Id infoid;
} Decision;

typedef struct {
  Solver *solv;
  Queue decisionlistq;
  Id p;
  int reason;
  Id infoid;
  int bits;
  int type;
  Id source;
  Id target;
  Id dep_id;
} Decisionset;

typedef struct {
  FILE *fp;
} SolvFp;

typedef Dataiterator Datamatch;

typedef int disown_helper;

struct myappdata {
  void *appdata;
  int disowned;
};

/* special internal decisionset constructor from a prepared decisionlist */
static Decisionset *decisionset_fromids(Solver *solv, Id *ids, int cnt)
{
  Decisionset *d = solv_calloc(1, sizeof(*d));
  int i;
  d->solv = solv;
  queue_init(&d->decisionlistq);
  d->p = ids[0];
  d->reason = ids[1];
  d->infoid = ids[2];
  d->bits = ids[3];
  d->type = ids[4];
  d->source = ids[5];
  d->target = ids[6];
  d->dep_id = ids[7];
  for (i = 0; i < cnt; i += 8)
    queue_insertn(&d->decisionlistq, d->decisionlistq.count, 3, ids + i);
  if (cnt > 8)
    d->infoid = 0;
  return d;
}

/* prepare a decisionlist so we can feed it to decisionset_fromids */
static void prepare_decisionset_queue(Solver *solv, Queue *q) {
  int i, cnt;
  for (i = cnt = 0; i < q->count; cnt++)
    {
      i += 1 + 8 + 8 * solver_decisionlist_merged(solv, q, i);  /* +1 as we insert one element */
      queue_insert(q, cnt, i - cnt);
    }
  if (cnt)
    queue_unshift(q, 1);        /* start of first block */
  for (i = 0; i < cnt; i++)
    q->elements[i] += cnt - i;
  q->count = cnt;   /* hack */
}



SWIGINTERN void appdata_disown_helper(void *appdata) {
  struct myappdata *myappdata = appdata;
  SV *rsv;
  if (!myappdata || !myappdata->appdata || myappdata->disowned)
    return;
  rsv = myappdata->appdata;
  if (!SvROK(rsv))
    return;
  myappdata->appdata = SvRV(rsv);
  myappdata->disowned = 1;
  SvREFCNT_dec(rsv);
}
SWIGINTERN void appdata_clr_helper(void **appdatap) {
  struct myappdata *myappdata = *(struct myappdata **)appdatap;
  if (myappdata && myappdata->appdata && !myappdata->disowned) {
    SvREFCNT_dec((SV *)myappdata->appdata);
  }
  *appdatap = solv_free(myappdata);
}
SWIGINTERN void appdata_set_helper(void **appdatap, void *appdata) {
  appdata_clr_helper(appdatap);
  if (appdata) {
    struct myappdata *myappdata = *appdatap = solv_calloc(sizeof(struct myappdata), 1);
    myappdata->appdata = appdata;
  }
}
SWIGINTERN void *appdata_get_helper(void **appdatap) {
  struct myappdata *myappdata = *appdatap;
  if (!myappdata || !myappdata->appdata)
    return 0;
  return myappdata->disowned ? newRV_noinc((SV *)myappdata->appdata) : myappdata->appdata;
}



SWIGINTERNINLINE SV *
SWIG_From_long  SWIG_PERL_DECL_ARGS_1(long value)
{
  SV *sv;
  if (IVSIZE >= sizeof(value) || (value >= IV_MIN && value <= IV_MAX))
    sv = newSViv(value);
  else
    sv = newSVpvf("%ld", value);
  return sv_2mortal(sv);
}


SWIGINTERNINLINE SV *
SWIG_From_int  SWIG_PERL_DECL_ARGS_1(int value)
{    
  return SWIG_From_long  SWIG_PERL_CALL_ARGS_1(value);
}

SWIGINTERN Selection *new_Selection(Pool *pool){
    Selection *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN void delete_Selection(Selection *self){
    queue_free(&self->q);
    solv_free(self);
  }
SWIGINTERN bool Selection_isempty(Selection *self){
    return self->q.count == 0;
  }

SWIGINTERNINLINE SV *
SWIG_From_bool  SWIG_PERL_DECL_ARGS_1(bool value)
{
  return boolSV(value);
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#include <errno.h>


#include <stdlib.h>
#ifdef _MSC_VER
# ifndef strtoull
#  define strtoull _strtoui64
# endif
# ifndef strtoll
#  define strtoll _strtoi64
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double SWIG_PERL_DECL_ARGS_2(SV *obj, double *val)
{
  if (SvNIOK(obj)) {
    if (val) *val = SvNV(obj);
    return SWIG_OK;
  } else if (SvIOK(obj)) {
    if (val) *val = (double) SvIV(obj);
    return SWIG_AddCast(SWIG_OK);
  } else {
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      double v;
      errno = 0;
      v = strtod(nptr, &endptr);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
  }
  return SWIG_TypeError;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long SWIG_PERL_DECL_ARGS_2(SV *obj, long* val)
{
  if (SvUOK(obj)) {
    UV v = SvUV(obj);
    if (UVSIZE < sizeof(*val) || v <= LONG_MAX) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (IVSIZE <= sizeof(*val) || (v >= LONG_MIN && v <= LONG_MAX)) {
      if(val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      long v;
      errno = 0;
      v = strtol(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int SWIG_PERL_DECL_ARGS_2(SV * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long SWIG_PERL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}

SWIGINTERN Selection *Selection_clone(Selection *self,int flags){
    Selection *s = new_Selection(self->pool);
    queue_init_clone(&s->q, &self->q);
    s->flags = self->flags;
    return s;
  }
SWIGINTERN void Selection_filter(Selection *self,Selection *lsel){
    if (self->pool != lsel->pool)
      queue_empty(&self->q);
    else
      selection_filter(self->pool, &self->q, &lsel->q);
  }
SWIGINTERN void Selection_add(Selection *self,Selection *lsel){
    if (self->pool == lsel->pool)
      {
        selection_add(self->pool, &self->q, &lsel->q);
        self->flags |= lsel->flags;
      }
  }
SWIGINTERN void Selection_add_raw(Selection *self,Id how,Id what){
    queue_push2(&self->q, how, what);
  }
SWIGINTERN void Selection_subtract(Selection *self,Selection *lsel){
    if (self->pool == lsel->pool)
      selection_subtract(self->pool, &self->q, &lsel->q);
  }

SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(SV *obj, char** cptr, size_t* psize, int *alloc)
{
  if (SvMAGICAL(obj)) {
     SV *tmp = sv_newmortal();
     SvSetSV(tmp, obj);
     obj = tmp;
  }
  if (SvPOK(obj)) {
    STRLEN len = 0;
    char *cstr = SvPV(obj, len); 
    size_t size = len + 1;
    if (cptr) {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      char* vptr = 0; 
      if (SWIG_ConvertPtr(obj, (void**)&vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = vptr;
	if (psize) *psize = vptr ? (strlen(vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}




SWIGINTERN void Selection_select(Selection *self,char const *name,int flags){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make(self->pool, &self->q, name, flags);
  }
SWIGINTERN void Selection_matchdeps(Selection *self,char const *name,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchdeps(self->pool, &self->q, name, flags, keyname, marker);
  }


SWIGINTERN int







SWIG_AsValDepId(void *obj, int *val) {

  static swig_type_info* desc = 0;
  void *vptr = 0;
  int ecode;
  if (!desc) desc = SWIG_TypeQuery("Dep *");





  ecode = SWIG_AsVal_int(obj, val);

  if (SWIG_IsOK(ecode))
    return ecode;



  if ((SWIG_ConvertPtr(obj, &vptr, desc, 0)) == SWIG_OK) {

    if (val)
      *val = vptr ? ((Dep *)vptr)->id : 0;
    return SWIG_OK;
  }
  return SWIG_TypeError;
}





SWIGINTERN void Selection_matchdepid(Selection *self,DepId dep,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchdepid(self->pool, &self->q, dep, flags, keyname, marker);
  }
SWIGINTERN void Selection_matchsolvable(Selection *self,XSolvable *solvable,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchsolvable(self->pool, &self->q, solvable->id, flags, keyname, marker);
  }
SWIGINTERN Queue Selection_jobs(Selection *self,int flags){
    Queue q;
    int i;
    queue_init_clone(&q, &self->q);
    for (i = 0; i < q.count; i += 2)
      q.elements[i] |= flags;
    return q;
  }
SWIGINTERN Queue Selection_solvables(Selection *self){
    Queue q;
    queue_init(&q);
    selection_solvables(self->pool, &self->q, &q);
    return q;
  }
SWIGINTERN char const *Selection___str__(Selection *self){
    return pool_selection2str(self->pool, &self->q, 0);
  }

SWIGINTERNINLINE SV *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  SV *obj = sv_newmortal();
  if (carray) {
    sv_setpvn(obj, carray, size);
  } else {
    sv_setsv(obj, &PL_sv_undef);
  }
  return obj;
}


SWIGINTERNINLINE SV * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}

SWIGINTERN char const *Selection___repr__(Selection *self){
    const char *str = pool_selection2str(self->pool, &self->q, ~0);
    return pool_tmpjoin(self->pool, "<Selection ", str, ">");
  }
SWIGINTERN Dep *new_Dep(Pool *pool,Id id){
    Dep *s;
    if (!id)
      return 0;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    s->id = id;
    return s;
  }

SWIGINTERN int
SWIG_AsVal_bool SWIG_PERL_DECL_ARGS_2(SV *obj, bool* val)
{
  if (obj == &PL_sv_yes) {
    if (val) *val = true;
    return SWIG_OK;
  } else if (obj == &PL_sv_no) { 
    if (val) *val = false;
    return SWIG_OK;
  } else {
    if (val) *val = SvTRUE(obj) ? true : false;
    return SWIG_AddCast(SWIG_OK);    
  }
}

SWIGINTERN Dep *Dep_Rel(Dep *self,int flags,DepId evrid,bool create){
    Id id = pool_rel2id(self->pool, self->id, evrid, flags, create);
    if (!id)
      return 0;
    return new_Dep(self->pool, id);
  }
SWIGINTERN Selection *Dep_Selection_name(Dep *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    if (ISRELDEP(self->id)) {
      Reldep *rd = GETRELDEP(self->pool, self->id);
      if (rd->flags == REL_EQ) {
        setflags |= self->pool->disttype == DISTTYPE_DEB || strchr(pool_id2str(self->pool, rd->evr), '-') != 0 ? SOLVER_SETEVR : SOLVER_SETEV;
        if (ISRELDEP(rd->name))
          rd = GETRELDEP(self->pool, rd->name);
      }
      if (rd->flags == REL_ARCH)
        setflags |= SOLVER_SETARCH;
    }
    queue_push2(&sel->q, SOLVER_SOLVABLE_NAME | setflags, self->id);
    return sel;
  }
SWIGINTERN Selection *Dep_Selection_provides(Dep *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    if (ISRELDEP(self->id)) {
      Reldep *rd = GETRELDEP(self->pool, self->id);
      if (rd->flags == REL_ARCH)
        setflags |= SOLVER_SETARCH;
    }
    queue_push2(&sel->q, SOLVER_SOLVABLE_PROVIDES | setflags, self->id);
    return sel;
  }
SWIGINTERN char const *Dep_str(Dep *self){
    return pool_dep2str(self->pool, self->id);
  }
SWIGINTERN bool Dep___eq__(Dep *self,Dep *s){
    return self->pool == s->pool && self->id == s->id;
  }
SWIGINTERN bool Dep___ne__(Dep *self,Dep *s){
    return !Dep___eq__(self, s);
  }
SWIGINTERN char const *Dep___repr__(Dep *self){
    char buf[20];
    sprintf(buf, "<Id #%d ", self->id);
    return solv_dupjoin(buf, pool_dep2str(self->pool, self->id), ">");
  }
SWIGINTERN Dataiterator *new_Dataiterator(Pool *pool,Repo *repo,Id p,Id key,char const *match,int flags){
    Dataiterator *di = solv_calloc(1, sizeof(*di));
    dataiterator_init(di, pool, repo, p, key, match, flags);
    return di;
  }
SWIGINTERN void delete_Dataiterator(Dataiterator *self){
    dataiterator_free(self);
    solv_free(self);
  }
SWIGINTERN Datamatch *Dataiterator___next__(Dataiterator *self){

    Dataiterator *ndi;
    if (!dataiterator_step(self)) {
      return 0;
    }
    ndi = solv_calloc(1, sizeof(*ndi));
    dataiterator_init_clone(ndi, self);
    dataiterator_strdup(ndi);
    return ndi;
  }
SWIGINTERN void Dataiterator_prepend_keyname(Dataiterator *self,Id key){
    dataiterator_prepend_keyname(self, key);
  }
SWIGINTERN void Dataiterator_skip_solvable(Dataiterator *self){
    dataiterator_skip_solvable(self);
  }
SWIGINTERN XSolvable *new_XSolvable(Pool *pool,Id id){
    XSolvable *s;
    if (!id || id >= pool->nsolvables)
      return 0;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    s->id = id;
    return s;
  }
SWIGINTERN char const *XSolvable_str(XSolvable *self){
    return pool_solvid2str(self->pool, self->id);
  }
SWIGINTERN char const *XSolvable_lookup_str(XSolvable *self,Id keyname){
    return pool_lookup_str(self->pool, self->id, keyname);
  }
SWIGINTERN Id XSolvable_lookup_id(XSolvable *self,Id keyname){
    return pool_lookup_id(self->pool, self->id, keyname);
  }

#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_DECL_ARGS_2(SV *obj, unsigned long long *val)
{
  if (SvUOK(obj)) {
    /* pretty sure this should be conditional on
     * (UVSIZE <= sizeof(*val) || v <= ULLONG_MAX) */
    if (val) *val = SvUV(obj);
    return SWIG_OK;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (v >= 0 && (IVSIZE <= sizeof(*val) || v <= ULLONG_MAX)) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      unsigned long long v;
      errno = 0;
      v = strtoull(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      const double mant_max = 1LL << DBL_MANT_DIG;
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
	if (val) *val = (unsigned long long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}
#endif

SWIGINTERN unsigned long long XSolvable_lookup_num(XSolvable *self,Id keyname,unsigned long long notfound){
    return pool_lookup_num(self->pool, self->id, keyname, notfound);
  }

#include <stdio.h>
#if (defined(_MSC_VER) && (_MSC_VER < 1900)) || defined(__BORLANDC__) || defined(_WATCOM)
# ifndef snprintf
#  define snprintf _snprintf
# endif
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE SV *
SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_DECL_ARGS_1(unsigned long long value)
{
  SV *sv;
  if (UVSIZE >= sizeof(value) || value <= UV_MAX)
    sv = newSVuv((UV)(value));
  else {
    //sv = newSVpvf("%llu", value); doesn't work in non 64bit Perl
    char temp[256];
    SWIG_snprintf(temp, sizeof(temp), "%llu", value);
    sv = newSVpv(temp, 0);
  }
  return sv_2mortal(sv);
}
#endif

SWIGINTERN bool XSolvable_lookup_void(XSolvable *self,Id keyname){
    return pool_lookup_void(self->pool, self->id, keyname);
  }
SWIGINTERN Chksum *XSolvable_lookup_checksum(XSolvable *self,Id keyname){
    Id type = 0;
    const unsigned char *b = pool_lookup_bin_checksum(self->pool, self->id, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN Queue XSolvable_lookup_idarray(XSolvable *self,Id keyname,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    Queue r;
    queue_init(&r);
    solvable_lookup_deparray(s, keyname, &r, marker);
    return r;
  }
SWIGINTERN Queue XSolvable_lookup_deparray(XSolvable *self,Id keyname,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    Queue r;
    queue_init(&r);
    solvable_lookup_deparray(s, keyname, &r, marker);
    return r;
  }

SWIGINTERNINLINE SV *
SWIG_From_unsigned_SS_long  SWIG_PERL_DECL_ARGS_1(unsigned long value)
{
  SV *sv;
  if (UVSIZE >= sizeof(value) || value <= UV_MAX)
    sv = newSVuv(value);
  else
    sv = newSVpvf("%lu", value);
  return sv_2mortal(sv);
}


SWIGINTERNINLINE SV *
SWIG_From_unsigned_SS_int  SWIG_PERL_DECL_ARGS_1(unsigned int value)
{    
  return SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1(value);
}

SWIGINTERN char const *XSolvable_lookup_location(XSolvable *self,unsigned int *OUTPUT){
    return solvable_lookup_location(self->pool->solvables + self->id, OUTPUT);
  }
SWIGINTERN char const *XSolvable_lookup_sourcepkg(XSolvable *self){
    return solvable_lookup_sourcepkg(self->pool->solvables + self->id);
  }
SWIGINTERN Dataiterator *XSolvable_Dataiterator(XSolvable *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, 0, self->id, key, match, flags);
  }
SWIGINTERN bool XSolvable_installable(XSolvable *self){
    return pool_installable(self->pool, pool_id2solvable(self->pool, self->id));
  }
SWIGINTERN bool XSolvable_isinstalled(XSolvable *self){
    Pool *pool = self->pool;
    return pool->installed && pool_id2solvable(pool, self->id)->repo == pool->installed;
  }

    SWIGINTERN void XSolvable_name_set(XSolvable *xs, const char *name) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].name = pool_str2id(pool, name, 1);
    }
    SWIGINTERN const char *XSolvable_name_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].name);
    }
  

    SWIGINTERN void XSolvable_nameid_set(XSolvable *xs, Id nameid) {
      xs->pool->solvables[xs->id].name = nameid;
    }
    SWIGINTERN Id XSolvable_nameid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].name;
    }
  

    SWIGINTERN void XSolvable_evr_set(XSolvable *xs, const char *evr) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].evr = pool_str2id(pool, evr, 1);
    }
    SWIGINTERN const char *XSolvable_evr_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].evr);
    }
  

    SWIGINTERN void XSolvable_evrid_set(XSolvable *xs, Id evrid) {
      xs->pool->solvables[xs->id].evr = evrid;
    }
    SWIGINTERN Id XSolvable_evrid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].evr;
    }
  

    SWIGINTERN void XSolvable_arch_set(XSolvable *xs, const char *arch) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].arch = pool_str2id(pool, arch, 1);
    }
    SWIGINTERN const char *XSolvable_arch_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].arch);
    }
  

    SWIGINTERN void XSolvable_archid_set(XSolvable *xs, Id archid) {
      xs->pool->solvables[xs->id].arch = archid;
    }
    SWIGINTERN Id XSolvable_archid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].arch;
    }
  

    SWIGINTERN void XSolvable_vendor_set(XSolvable *xs, const char *vendor) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].vendor = pool_str2id(pool, vendor, 1);
    }
    SWIGINTERN const char *XSolvable_vendor_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].vendor);
    }
  

    SWIGINTERN void XSolvable_vendorid_set(XSolvable *xs, Id vendorid) {
      xs->pool->solvables[xs->id].vendor = vendorid;
    }
    SWIGINTERN Id XSolvable_vendorid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].vendor;
    }
  

    SWIGINTERN Repo *XSolvable_repo_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].repo;
    }
  
SWIGINTERN void XSolvable_add_provides(XSolvable *self,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    marker = solv_depmarker(SOLVABLE_PROVIDES, marker);
    s->provides = repo_addid_dep(s->repo, s->provides, id, marker);
  }
SWIGINTERN void XSolvable_add_obsoletes(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->obsoletes = repo_addid_dep(s->repo, s->obsoletes, id, 0);
  }
SWIGINTERN void XSolvable_add_conflicts(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->conflicts = repo_addid_dep(s->repo, s->conflicts, id, 0);
  }
SWIGINTERN void XSolvable_add_requires(XSolvable *self,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    marker = solv_depmarker(SOLVABLE_REQUIRES, marker);
    s->requires = repo_addid_dep(s->repo, s->requires, id, marker);
  }
SWIGINTERN void XSolvable_add_recommends(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->recommends = repo_addid_dep(s->repo, s->recommends, id, 0);
  }
SWIGINTERN void XSolvable_add_suggests(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->suggests = repo_addid_dep(s->repo, s->suggests, id, 0);
  }
SWIGINTERN void XSolvable_add_supplements(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->supplements = repo_addid_dep(s->repo, s->supplements, id, 0);
  }
SWIGINTERN void XSolvable_add_enhances(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->enhances = repo_addid_dep(s->repo, s->enhances, id, 0);
  }
SWIGINTERN void XSolvable_unset(XSolvable *self,Id keyname){
    Solvable *s = self->pool->solvables + self->id;
    repo_unset(s->repo, self->id, keyname);
  }
SWIGINTERN void XSolvable_add_deparray(XSolvable *self,Id keyname,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    solvable_add_deparray(s, keyname, id, marker);
  }
SWIGINTERN Selection *XSolvable_Selection(XSolvable *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    queue_push2(&sel->q, SOLVER_SOLVABLE | setflags, self->id);
    return sel;
  }
SWIGINTERN bool XSolvable_identical(XSolvable *self,XSolvable *s2){
    return solvable_identical(self->pool->solvables + self->id, s2->pool->solvables + s2->id);
  }
SWIGINTERN int XSolvable_evrcmp(XSolvable *self,XSolvable *s2){
    return pool_evrcmp(self->pool, self->pool->solvables[self->id].evr, s2->pool->solvables[s2->id].evr, EVRCMP_COMPARE);
  }
SWIGINTERN bool XSolvable_matchesdep(XSolvable *self,Id keyname,DepId id,Id marker){
    return solvable_matchesdep(self->pool->solvables + self->id, keyname, id, marker);
  }
SWIGINTERN bool XSolvable___eq__(XSolvable *self,XSolvable *s){
    return self->pool == s->pool && self->id == s->id;
  }
SWIGINTERN bool XSolvable___ne__(XSolvable *self,XSolvable *s){
    return !XSolvable___eq__(self, s);
  }
SWIGINTERN char const *XSolvable___repr__(XSolvable *self){
    char buf[20];
    sprintf(buf, "<Solvable #%d ", self->id);
    return solv_dupjoin(buf, pool_solvid2str(self->pool, self->id), ">");
  }
SWIGINTERN Ruleinfo *new_Ruleinfo(Solver *solv,Id rid,Id type,Id source,Id target,Id dep_id){
    Ruleinfo *ri = solv_calloc(1, sizeof(*ri));
    ri->solv = solv;
    ri->rid = rid;
    ri->type = type;
    ri->source = source;
    ri->target = target;
    ri->dep_id = dep_id;
    return ri;
  }

    SWIGINTERN XSolvable *Ruleinfo_solvable_get(Ruleinfo *ri) {
      return new_XSolvable(ri->solv->pool, ri->source);
    }
    SWIGINTERN XSolvable *Ruleinfo_othersolvable_get(Ruleinfo *ri) {
      return new_XSolvable(ri->solv->pool, ri->target);
    }
    SWIGINTERN Dep *Ruleinfo_dep_get(Ruleinfo *ri) {
      return new_Dep(ri->solv->pool, ri->dep_id);
    }
  
SWIGINTERN char const *Ruleinfo_problemstr(Ruleinfo *self){
    return solver_problemruleinfo2str(self->solv, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN char const *Ruleinfo___str__(Ruleinfo *self){
    return solver_ruleinfo2str(self->solv, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN XRule *new_XRule(Solver *solv,Id id){
    if (!id)
      return 0;
    XRule *xr = solv_calloc(1, sizeof(*xr));
    xr->solv = solv;
    xr->id = id;
    return xr;
  }

    SWIGINTERN int XRule_type_get(XRule *xr) {
      return solver_ruleclass(xr->solv, xr->id);
    }
  
SWIGINTERN Ruleinfo *XRule_info(XRule *self){
    Id type, source, target, dep;
    type = solver_ruleinfo(self->solv, self->id, &source, &target, &dep);
    return new_Ruleinfo(self->solv, self->id, type, source, target, dep);
  }
SWIGINTERN Queue XRule_allinfos(XRule *self){
    Queue q;
    queue_init(&q);
    solver_allruleinfos(self->solv, self->id, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_learnt(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_decisionlist(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE | SOLVER_DECISIONLIST_SORTED, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_decisionsetlist(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE | SOLVER_DECISIONLIST_SORTED | SOLVER_DECISIONLIST_WITHINFO | SOLVER_DECISIONLIST_MERGEDINFO, &q);
    prepare_decisionset_queue(self->solv, &q);
    return q;
  }
SWIGINTERN bool XRule___eq__(XRule *self,XRule *xr){
    return self->solv == xr->solv && self->id == xr->id;
  }
SWIGINTERN bool XRule___ne__(XRule *self,XRule *xr){
    return !XRule___eq__(self, xr);
  }
SWIGINTERN char const *XRule___repr__(XRule *self){
    char buf[20];
    sprintf(buf, "<Rule #%d>", self->id);
    return solv_strdup(buf);
  }
SWIGINTERN XRepodata *new_XRepodata(Repo *repo,Id id){
    XRepodata *xr = solv_calloc(1, sizeof(*xr));
    xr->repo = repo;
    xr->id = id;
    return xr;
  }
SWIGINTERN Id XRepodata_new_handle(XRepodata *self){
    return repodata_new_handle(repo_id2repodata(self->repo, self->id));
  }
SWIGINTERN void XRepodata_set_id(XRepodata *self,Id solvid,Id keyname,DepId id){
    repodata_set_id(repo_id2repodata(self->repo, self->id), solvid, keyname, id);
  }
SWIGINTERN void XRepodata_set_num(XRepodata *self,Id solvid,Id keyname,unsigned long long num){
    repodata_set_num(repo_id2repodata(self->repo, self->id), solvid, keyname, num);
  }
SWIGINTERN void XRepodata_set_str(XRepodata *self,Id solvid,Id keyname,char const *str){
    repodata_set_str(repo_id2repodata(self->repo, self->id), solvid, keyname, str);
  }
SWIGINTERN void XRepodata_set_void(XRepodata *self,Id solvid,Id keyname){
    repodata_set_void(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN void XRepodata_set_poolstr(XRepodata *self,Id solvid,Id keyname,char const *str){
    repodata_set_poolstr(repo_id2repodata(self->repo, self->id), solvid, keyname, str);
  }
SWIGINTERN void XRepodata_add_idarray(XRepodata *self,Id solvid,Id keyname,DepId id){
    repodata_add_idarray(repo_id2repodata(self->repo, self->id), solvid, keyname, id);
  }
SWIGINTERN void XRepodata_add_flexarray(XRepodata *self,Id solvid,Id keyname,Id handle){
    repodata_add_flexarray(repo_id2repodata(self->repo, self->id), solvid, keyname, handle);
  }
SWIGINTERN void XRepodata_set_checksum(XRepodata *self,Id solvid,Id keyname,Chksum *chksum){
    const unsigned char *buf = solv_chksum_get(chksum, 0);
    if (buf)
      repodata_set_bin_checksum(repo_id2repodata(self->repo, self->id), solvid, keyname, solv_chksum_get_type(chksum), buf);
  }
SWIGINTERN void XRepodata_set_sourcepkg(XRepodata *self,Id solvid,char const *sourcepkg){
    repodata_set_sourcepkg(repo_id2repodata(self->repo, self->id), solvid, sourcepkg);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long SWIG_PERL_DECL_ARGS_2(SV *obj, unsigned long *val) 
{
  if (SvUOK(obj)) {
    UV v = SvUV(obj);
    if (UVSIZE <= sizeof(*val) || v <= ULONG_MAX) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (v >= 0 && (IVSIZE <= sizeof(*val) || v <= ULONG_MAX)) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      unsigned long v;
      errno = 0;
      v = strtoul(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_int SWIG_PERL_DECL_ARGS_2(SV * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}

SWIGINTERN void XRepodata_set_location(XRepodata *self,Id solvid,unsigned int mediano,char const *location){
    repodata_set_location(repo_id2repodata(self->repo, self->id), solvid, mediano, 0, location);
  }
SWIGINTERN void XRepodata_unset(XRepodata *self,Id solvid,Id keyname){
    repodata_unset(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN char const *XRepodata_lookup_str(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_str(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN Id XRepodata_lookup_id(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_id(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN unsigned long long XRepodata_lookup_num(XRepodata *self,Id solvid,Id keyname,unsigned long long notfound){
    return repodata_lookup_num(repo_id2repodata(self->repo, self->id), solvid, keyname, notfound);
  }
SWIGINTERN bool XRepodata_lookup_void(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_void(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN Queue XRepodata_lookup_idarray(XRepodata *self,Id solvid,Id keyname){
    Queue r;
    queue_init(&r);
    repodata_lookup_idarray(repo_id2repodata(self->repo, self->id), solvid, keyname, &r);
    return r;
  }
SWIGINTERN Chksum *XRepodata_lookup_checksum(XRepodata *self,Id solvid,Id keyname){
    Id type = 0;
    const unsigned char *b = repodata_lookup_bin_checksum(repo_id2repodata(self->repo, self->id), solvid, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN void XRepodata_internalize(XRepodata *self){
    repodata_internalize(repo_id2repodata(self->repo, self->id));
  }
SWIGINTERN void XRepodata_create_stubs(XRepodata *self){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    data = repodata_create_stubs(data);
    self->id = data->repodataid;
  }



SWIGINTERN int







SWIG_AsValSolvFpPtr(void *obj, FILE **val) {

  static swig_type_info* desc = 0;
  void *vptr = 0;




  if (!desc) desc = SWIG_TypeQuery("SolvFp *");



  if ((SWIG_ConvertPtr(obj, &vptr, desc, 0)) == SWIG_OK) {

    if (val)
      *val = vptr ? ((SolvFp *)vptr)->fp : 0;
    return SWIG_OK;
  }





  return SWIG_TypeError;
}






SWIGINTERN bool XRepodata_write(XRepodata *self,FILE *fp){
    return repodata_write(repo_id2repodata(self->repo, self->id), fp) == 0;
  }
SWIGINTERN Id XRepodata_str2dir(XRepodata *self,char const *dir,bool create){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    return repodata_str2dir(data, dir, create);
  }
SWIGINTERN char const *XRepodata_dir2str(XRepodata *self,Id did,char const *suf){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    return repodata_dir2str(data, did, suf);
  }
SWIGINTERN void XRepodata_add_dirstr(XRepodata *self,Id solvid,Id keyname,Id dir,char const *str){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    repodata_add_dirstr(data, solvid, keyname, dir, str);
  }
SWIGINTERN bool XRepodata_add_solv(XRepodata *self,FILE *fp,int flags){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    int r, oldstate = data->state;
    data->state = REPODATA_LOADING;
    r = repo_add_solv(data->repo, fp, flags | REPO_USE_LOADING);
    if (r || data->state == REPODATA_LOADING)
      data->state = oldstate;
    return r == 0;
  }
SWIGINTERN void XRepodata_extend_to_repo(XRepodata *self){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    repodata_extend_block(data, data->repo->start, data->repo->end - data->repo->start);
  }
SWIGINTERN bool XRepodata___eq__(XRepodata *self,XRepodata *xr){
    return self->repo == xr->repo && self->id == xr->id;
  }
SWIGINTERN bool XRepodata___ne__(XRepodata *self,XRepodata *xr){
    return !XRepodata___eq__(self, xr);
  }
SWIGINTERN char const *XRepodata___repr__(XRepodata *self){
    char buf[20];
    sprintf(buf, "<Repodata #%d>", self->id);
    return solv_strdup(buf);
  }
SWIGINTERN Pool_solvable_iterator *new_Pool_solvable_iterator(Pool *pool){
    Pool_solvable_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN XSolvable *Pool_solvable_iterator___next__(Pool_solvable_iterator *self){

    Pool *pool = self->pool;
    if (self->id >= pool->nsolvables)
      return 0;
    while (++self->id < pool->nsolvables)
      if (pool->solvables[self->id].repo)
        return new_XSolvable(pool, self->id);
    return 0;
  }
SWIGINTERN XSolvable *Pool_solvable_iterator___getitem__(Pool_solvable_iterator *self,Id key){
    Pool *pool = self->pool;
    if (key > 0 && key < pool->nsolvables && pool->solvables[key].repo)
      return new_XSolvable(pool, key);
    return 0;
  }
SWIGINTERN int Pool_solvable_iterator___len__(Pool_solvable_iterator *self){
    return self->pool->nsolvables;
  }
SWIGINTERN Pool_repo_iterator *new_Pool_repo_iterator(Pool *pool){
    Pool_repo_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN Repo *Pool_repo_iterator___next__(Pool_repo_iterator *self){

    Pool *pool = self->pool;
    if (self->id >= pool->nrepos)
      return 0;
    while (++self->id < pool->nrepos) {
      Repo *r = pool_id2repo(pool, self->id);
      if (r)
        return r;
    }
    return 0;
  }
SWIGINTERN Repo *Pool_repo_iterator___getitem__(Pool_repo_iterator *self,Id key){
    Pool *pool = self->pool;
    if (key > 0 && key < pool->nrepos)
      return pool_id2repo(pool, key);
    return 0;
  }
SWIGINTERN int Pool_repo_iterator___len__(Pool_repo_iterator *self){
    return self->pool->nrepos;
  }
SWIGINTERN Repo_solvable_iterator *new_Repo_solvable_iterator(Repo *repo){
    Repo_solvable_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->repo = repo;
    return s;
  }
SWIGINTERN XSolvable *Repo_solvable_iterator___next__(Repo_solvable_iterator *self){

    Repo *repo = self->repo;
    Pool *pool = repo->pool;
    if (repo->start > 0 && self->id < repo->start)
      self->id = repo->start - 1;
    if (self->id >= repo->end)
      return 0;
    while (++self->id < repo->end)
      if (pool->solvables[self->id].repo == repo)
        return new_XSolvable(pool, self->id);
    return 0;
  }
SWIGINTERN XSolvable *Repo_solvable_iterator___getitem__(Repo_solvable_iterator *self,Id key){
    Repo *repo = self->repo;
    Pool *pool = repo->pool;
    if (key > 0 && key < pool->nsolvables && pool->solvables[key].repo == repo)
      return new_XSolvable(pool, key);
    return 0;
  }
SWIGINTERN int Repo_solvable_iterator___len__(Repo_solvable_iterator *self){
    return self->repo->pool->nsolvables;
  }
SWIGINTERN void delete_Datamatch(Datamatch *self){
    dataiterator_free(self);
    solv_free(self);
  }

  SWIGINTERN XSolvable *Datamatch_solvable_get(Dataiterator *di) {
    return new_XSolvable(di->pool, di->solvid);
  }
  SWIGINTERN Id Datamatch_key_id_get(Dataiterator *di) {
    return di->key->name;
  }
  SWIGINTERN const char *Datamatch_key_idstr_get(Dataiterator *di) {
    return pool_id2str(di->pool, di->key->name);
  }
  SWIGINTERN Id Datamatch_type_id_get(Dataiterator *di) {
    return di->key->type;
  }
  SWIGINTERN const char *Datamatch_type_idstr_get(Dataiterator *di) {
    return pool_id2str(di->pool, di->key->type);
  }
  SWIGINTERN Id Datamatch_id_get(Dataiterator *di) {
    return di->kv.id;
  }
  SWIGINTERN const char *Datamatch_idstr_get(Dataiterator *di) {
   if (di->data && (di->key->type == REPOKEY_TYPE_DIR || di->key->type == REPOKEY_TYPE_DIRSTRARRAY || di->key->type == REPOKEY_TYPE_DIRNUMNUMARRAY))
      return repodata_dir2str(di->data,  di->kv.id, 0);
    if (di->data && di->data->localpool)
      return stringpool_id2str(&di->data->spool, di->kv.id);
    return pool_id2str(di->pool, di->kv.id);
  }
  SWIGINTERN const char * const Datamatch_str_get(Dataiterator *di) {
    return di->kv.str;
  }
  SWIGINTERN Dep *Datamatch_dep_get(Dataiterator *di) {
    if (di->key->type == REPOKEY_TYPE_DIR || di->key->type == REPOKEY_TYPE_DIRSTRARRAY || di->key->type == REPOKEY_TYPE_DIRNUMNUMARRAY)
      return 0;
    if (di->data && di->data->localpool)
      return 0;
    return new_Dep(di->pool, di->kv.id);
  }
  SWIGINTERN BinaryBlob Datamatch_binary_get(Dataiterator *di) {
    BinaryBlob bl;
    bl.data = 0;
    bl.len = 0;
    if (di->key->type == REPOKEY_TYPE_BINARY)
      {
        bl.data = di->kv.str;
        bl.len = di->kv.num;
      }
    else if ((bl.len = solv_chksum_len(di->key->type)) != 0)
      bl.data = di->kv.str;
    return bl;
  }
  SWIGINTERN unsigned long long Datamatch_num_get(Dataiterator *di) {
   if (di->key->type == REPOKEY_TYPE_NUM)
     return SOLV_KV_NUM64(&di->kv);
   return di->kv.num;
  }
  SWIGINTERN unsigned int Datamatch_num2_get(Dataiterator *di) {
    return di->kv.num2;
  }
  
SWIGINTERN Datapos *Datamatch_pos(Datamatch *self){
    Pool *pool = self->pool;
    Datapos *pos, oldpos = pool->pos;
    dataiterator_setpos(self);
    pos = solv_calloc(1, sizeof(*pos));
    *pos = pool->pos;
    pool->pos = oldpos;
    return pos;
  }
SWIGINTERN Datapos *Datamatch_parentpos(Datamatch *self){
    Pool *pool = self->pool;
    Datapos *pos, oldpos = pool->pos;
    dataiterator_setpos_parent(self);
    pos = solv_calloc(1, sizeof(*pos));
    *pos = pool->pos;
    pool->pos = oldpos;
    return pos;
  }
SWIGINTERN char const *Datamatch___str__(Datamatch *self){
    KeyValue kv = self->kv;
    const char *str = repodata_stringify(self->pool, self->data, self->key, &kv, SEARCH_FILES | SEARCH_CHECKSUMS);
    return str ? str : "";
  }
SWIGINTERN Id Datapos_lookup_id(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Id r;
    pool->pos = *self;
    r = pool_lookup_id(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN char const *Datapos_lookup_str(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *r;
    pool->pos = *self;
    r = pool_lookup_str(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN unsigned long long Datapos_lookup_num(Datapos *self,Id keyname,unsigned long long notfound){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    unsigned long long r;
    pool->pos = *self;
    r = pool_lookup_num(pool, SOLVID_POS, keyname, notfound);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN bool Datapos_lookup_void(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    int r;
    pool->pos = *self;
    r = pool_lookup_void(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN Chksum *Datapos_lookup_checksum(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Id type = 0;
    const unsigned char *b;
    pool->pos = *self;
    b = pool_lookup_bin_checksum(pool, SOLVID_POS, keyname, &type);
    pool->pos = oldpos;
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN char const *Datapos_lookup_deltaseq(Datapos *self){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *seq;
    pool->pos = *self;
    seq = pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_NAME);
    if (seq) {
      seq = pool_tmpjoin(pool, seq, "-", pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_EVR));
      seq = pool_tmpappend(pool, seq, "-", pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_NUM));
    }
    pool->pos = oldpos;
    return seq;
  }
SWIGINTERN char const *Datapos_lookup_deltalocation(Datapos *self,unsigned int *OUTPUT){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *loc;
    pool->pos = *self;
    loc = pool_lookup_deltalocation(pool, SOLVID_POS, OUTPUT);
    pool->pos = oldpos;
    return loc;
  }
SWIGINTERN Queue Datapos_lookup_idarray(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Queue r;
    queue_init(&r);
    pool->pos = *self;
    pool_lookup_idarray(pool, SOLVID_POS, keyname, &r);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN Dataiterator *Datapos_Dataiterator(Datapos *self,Id key,char const *match,int flags){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Dataiterator *di;
    pool->pos = *self;
    di = new_Dataiterator(pool, 0, SOLVID_POS, key, match, flags);
    pool->pos = oldpos;
    return di;
  }
SWIGINTERN Job *new_Job(Pool *pool,int how,Id what){
    Job *job = solv_calloc(1, sizeof(*job));
    job->pool = pool;
    job->how = how;
    job->what = what;
    return job;
  }
SWIGINTERN Queue Job_solvables(Job *self){
    Queue q;
    queue_init(&q);
    pool_job2solvables(self->pool, &q, self->how, self->what);
    return q;
  }
SWIGINTERN bool Job_isemptyupdate(Job *self){
    return pool_isemptyupdatejob(self->pool, self->how, self->what);
  }
SWIGINTERN bool Job___eq__(Job *self,Job *j){
    return self->pool == j->pool && self->how == j->how && self->what == j->what;
  }
SWIGINTERN bool Job___ne__(Job *self,Job *j){
    return !Job___eq__(self, j);
  }
SWIGINTERN char const *Job___str__(Job *self){
    return pool_job2str(self->pool, self->how, self->what, 0);
  }
SWIGINTERN char const *Job___repr__(Job *self){
    const char *str = pool_job2str(self->pool, self->how, self->what, ~0);
    return pool_tmpjoin(self->pool, "<Job ", str, ">");
  }
SWIGINTERN Pool *new_Pool(void){
    Pool *pool = pool_create();
    return pool;
  }
SWIGINTERN int Pool_setdisttype(Pool *self,int disttype){
    return pool_setdisttype(self, disttype);
  }
SWIGINTERN void Pool_set_debuglevel(Pool *self,int level){
    pool_setdebuglevel(self, level);
  }
SWIGINTERN int Pool_set_flag(Pool *self,int flag,int value){
    return pool_set_flag(self, flag, value);
  }
SWIGINTERN int Pool_get_flag(Pool *self,int flag){
    return pool_get_flag(self, flag);
  }
SWIGINTERN void Pool_set_rootdir(Pool *self,char const *rootdir){
    pool_set_rootdir(self, rootdir);
  }
SWIGINTERN char const *Pool_get_rootdir(Pool *self,int flag){
    return pool_get_rootdir(self);
  }

  SWIGINTERN int loadcallback(Pool *pool, Repodata *data, void *d) {
    int count;
    int ret = 0;
    dSP;
    XRepodata *xd = new_XRepodata(data->repo, data->repodataid);

    ENTER;
    SAVETMPS;
    PUSHMARK(SP);
    XPUSHs(SWIG_NewPointerObj(SWIG_as_voidptr(xd), SWIGTYPE_p_XRepodata, SWIG_OWNER | SWIG_SHADOW));
    PUTBACK;
    count = perl_call_sv((SV *)d, G_EVAL|G_SCALAR);
    SPAGAIN;
    if (count)
      ret = POPi;
    PUTBACK;
    FREETMPS;
    LEAVE;
    return ret;
  }

SWIGINTERN void Pool_clr_loadcallback(Pool *self){
    if (self->loadcallback == loadcallback) {
      SvREFCNT_dec(self->loadcallbackdata);
      pool_setloadcallback(self, 0, 0);
    }
  }
SWIGINTERN void Pool_set_loadcallback(Pool *self,SV *callable){
    Pool_clr_loadcallback(self);
    if (callable) {
      SvREFCNT_inc(callable);
      pool_setloadcallback(self, loadcallback, callable);
    }
  }
SWIGINTERN void delete_Pool(Pool *self){
    Pool *pool = self;
    Id repoid;
    Repo *repo;
    FOR_REPOS(repoid, repo)
      appdata_clr_helper(&repo->appdata);
    Pool_clr_loadcallback(pool);
    appdata_clr_helper(&pool->appdata);
    pool_free(pool);
  }
SWIGINTERN disown_helper Pool_free(Pool *self){
    Pool *pool = self;
    Id repoid;
    Repo *repo;
    FOR_REPOS(repoid, repo)
      appdata_clr_helper(&repo->appdata);
    Pool_clr_loadcallback(pool);
    appdata_clr_helper(&pool->appdata);
    pool_free(pool);
    return 0;
  }
SWIGINTERN disown_helper Pool_disown(Pool *self){
    return 0;
  }

  SWIGINTERN void Pool_appdata_set(Pool *pool, AppObjectPtr appdata) {
    appdata_set_helper(&pool->appdata, appdata);
  }
  SWIGINTERN AppObjectPtr Pool_appdata_get(Pool *pool) {
    return appdata_get_helper(&pool->appdata);
  }
  
SWIGINTERN void Pool_appdata_disown(Pool *self){
    appdata_disown_helper(self->appdata);
  }
SWIGINTERN Id Pool_str2id(Pool *self,char const *str,bool create){
    return pool_str2id(self, str, create);
  }
SWIGINTERN Dep *Pool_Dep(Pool *self,char const *str,bool create){
    Id id = pool_str2id(self, str, create);
    return new_Dep(self, id);
  }
SWIGINTERN Dep *Pool_parserpmrichdep(Pool *self,char const *str){
    Id id = pool_parserpmrichdep(self, str);
    return new_Dep(self, id);
  }
SWIGINTERN char const *Pool_id2str(Pool *self,Id id){
    return pool_id2str(self, id);
  }
SWIGINTERN char const *Pool_dep2str(Pool *self,Id id){
    return pool_dep2str(self, id);
  }
SWIGINTERN Id Pool_rel2id(Pool *self,Id name,Id evr,int flags,bool create){
    return pool_rel2id(self, name, evr, flags, create);
  }
SWIGINTERN Id Pool_id2langid(Pool *self,Id id,char const *lang,bool create){
    return pool_id2langid(self, id, lang, create);
  }
SWIGINTERN void Pool_setarch(Pool *self,char const *arch){
    struct utsname un;
    if (!arch) {
      if (uname(&un)) {
        perror("uname");
        return;
      }
      arch = un.machine;
    }
    pool_setarch(self, arch);
  }
SWIGINTERN Repo *Pool_add_repo(Pool *self,char const *name){
    return repo_create(self, name);
  }
SWIGINTERN char const *Pool_lookup_str(Pool *self,Id entry,Id keyname){
    return pool_lookup_str(self, entry, keyname);
  }
SWIGINTERN Id Pool_lookup_id(Pool *self,Id entry,Id keyname){
    return pool_lookup_id(self, entry, keyname);
  }
SWIGINTERN unsigned long long Pool_lookup_num(Pool *self,Id entry,Id keyname,unsigned long long notfound){
    return pool_lookup_num(self, entry, keyname, notfound);
  }
SWIGINTERN bool Pool_lookup_void(Pool *self,Id entry,Id keyname){
    return pool_lookup_void(self, entry, keyname);
  }
SWIGINTERN Chksum *Pool_lookup_checksum(Pool *self,Id entry,Id keyname){
    Id type = 0;
    const unsigned char *b = pool_lookup_bin_checksum(self, entry, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN Dataiterator *Pool_Dataiterator(Pool *self,Id key,char const *match,int flags){
    return new_Dataiterator(self, 0, 0, key, match, flags);
  }
SWIGINTERN Dataiterator *Pool_Dataiterator_solvid(Pool *self,Id p,Id key,char const *match,int flags){
    return new_Dataiterator(self, 0, p, key, match, flags);
  }
SWIGINTERN char const *Pool_solvid2str(Pool *self,Id solvid){
    return pool_solvid2str(self, solvid);
  }
SWIGINTERN char const *Pool_solvidset2str(Pool *self,Queue q){
    return pool_solvidset2str(self, &q);
  }
SWIGINTERN char const *Pool_solvableset2str(Pool *self,Queue solvables){
    return pool_solvidset2str(self, &solvables);
  }
SWIGINTERN void Pool_addfileprovides(Pool *self){
    pool_addfileprovides(self);
  }
SWIGINTERN Queue Pool_addfileprovides_queue(Pool *self){
    Queue r;
    queue_init(&r);
    pool_addfileprovides_queue(self, &r, 0);
    return r;
  }
SWIGINTERN void Pool_createwhatprovides(Pool *self){
    pool_createwhatprovides(self);
  }
SWIGINTERN XSolvable *Pool_id2solvable(Pool *self,Id id){
    return new_XSolvable(self, id);
  }

  SWIGINTERN Pool_solvable_iterator * Pool_solvables_get(Pool *pool) {
    return new_Pool_solvable_iterator(pool);
  }
  
SWIGINTERN Pool_solvable_iterator *Pool_solvables_iter(Pool *self){
    return new_Pool_solvable_iterator(self);
  }
SWIGINTERN Repo *Pool_id2repo(Pool *self,Id id){
    if (id < 1 || id >= self->nrepos)
      return 0;
    return pool_id2repo(self, id);
  }

  SWIGINTERN Pool_repo_iterator * Pool_repos_get(Pool *pool) {
    return new_Pool_repo_iterator(pool);
  }
  
SWIGINTERN Pool_repo_iterator *Pool_repos_iter(Pool *self){
    return new_Pool_repo_iterator(self);
  }

  SWIGINTERN void Pool_installed_set(Pool *pool, Repo *installed) {
    pool_set_installed(pool, installed);
  }
  SWIGINTERN Repo *Pool_installed_get(Pool *pool) {
    return pool->installed;
  }
  SWIGINTERN const char *Pool_errstr_get(Pool *pool) {
    return pool_errstr(pool);
  }
  
SWIGINTERN Queue Pool_matchprovidingids(Pool *self,char const *match,int flags){
    Pool *pool = self;
    Queue q;
    Id id;
    queue_init(&q);
    if (!flags) {
      for (id = 1; id < pool->ss.nstrings; id++)
        if (pool->whatprovides[id])
          queue_push(&q, id);
    } else {
      Datamatcher ma;
      if (!datamatcher_init(&ma, match, flags)) {
        for (id = 1; id < pool->ss.nstrings; id++)
          if (pool->whatprovides[id] && datamatcher_match(&ma, pool_id2str(pool, id)))
            queue_push(&q, id);
        datamatcher_free(&ma);
      }
    }
    return q;
  }
SWIGINTERN Job *Pool_Job(Pool *self,int how,Id what){
    return new_Job(self, how, what);
  }
SWIGINTERN Queue Pool_whatprovides(Pool *self,DepId dep){
    Pool *pool = self;
    Queue q;
    Id p, pp;
    queue_init(&q);
    FOR_PROVIDES(p, pp, dep)
      queue_push(&q, p);
    return q;
  }
SWIGINTERN Queue Pool_best_solvables(Pool *self,Queue solvables,int flags){
    Queue q;
    queue_init_clone(&q, &solvables);
    pool_best_solvables(self, &q, flags);
    return q;
  }
SWIGINTERN Id Pool_towhatprovides(Pool *self,Queue q){
    return pool_queuetowhatprovides(self, &q);
  }
SWIGINTERN void Pool_set_namespaceproviders(Pool *self,DepId ns,DepId evr,bool value){
    Id dep = pool_rel2id(self, ns, evr, REL_NAMESPACE, 1);
    pool_set_whatprovides(self, dep, value ? 2 : 1);
  }
SWIGINTERN void Pool_flush_namespaceproviders(Pool *self,DepId ns,DepId evr){
    pool_flush_namespaceproviders(self, ns, evr);
  }
SWIGINTERN Queue Pool_whatcontainsdep(Pool *self,Id keyname,DepId dep,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatcontainsdep(self, keyname, dep, &q, marker);
    return q;
  }
SWIGINTERN Queue Pool_whatmatchesdep(Pool *self,Id keyname,DepId dep,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatmatchesdep(self, keyname, dep, &q, marker);
    return q;
  }
SWIGINTERN Queue Pool_whatmatchessolvable(Pool *self,Id keyname,XSolvable *pool_solvable,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatmatchessolvable(self, keyname, pool_solvable->id, &q, marker);
    return q;
  }
SWIGINTERN bool Pool_isknownarch(Pool *self,DepId id){
    Pool *pool = self;
    if (!id || id == ID_EMPTY)
      return 0;
    if (id == ARCH_SRC || id == ARCH_NOSRC || id == ARCH_NOARCH)
      return 1;
    if (pool->id2arch && pool_arch2score(pool, id) == 0)
      return 0;
    return 1;
  }
SWIGINTERN Solver *Pool_Solver(Pool *self){
    return solver_create(self);
  }
SWIGINTERN Selection *Pool_Selection(Pool *self){
    return new_Selection(self);
  }
SWIGINTERN Selection *Pool_Selection_all(Pool *self,int setflags){
    Selection *sel = new_Selection(self);
    queue_push2(&sel->q, SOLVER_SOLVABLE_ALL | setflags, 0);
    return sel;
  }
SWIGINTERN Selection *Pool_select(Pool *self,char const *name,int flags){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make(self, &sel->q, name, flags);
    return sel;
  }
SWIGINTERN Selection *Pool_matchdeps(Pool *self,char const *name,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchdeps(self, &sel->q, name, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Selection *Pool_matchdepid(Pool *self,DepId dep,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchdepid(self, &sel->q, dep, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Selection *Pool_matchsolvable(Pool *self,XSolvable *solvable,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchsolvable(self, &sel->q, solvable->id, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Queue Pool_get_considered_list(Pool *self){
    Queue q;
    queue_init(&q);
    int i;
    for (i = 2; i < self->nsolvables; i++) {
      if (self->solvables[i].repo && (!self->considered || MAPTST(self->considered, i)))
        queue_push(&q, i);
    }
    return q;
  }
SWIGINTERN Queue Pool_get_disabled_list(Pool *self){
    Queue q;
    queue_init(&q);
    int i;
    for (i = 2; i < self->nsolvables; i++) {
      if (self->solvables[i].repo && (self->considered && !MAPTST(self->considered, i)))
        queue_push(&q, i);
    }
    return q;
  }
SWIGINTERN void Pool_set_considered_list(Pool *self,Queue q){
    int i;
    Id p;
    if (!self->considered) {
      self->considered = solv_calloc(1, sizeof(Map));
      map_init(self->considered, self->nsolvables);
    }
    map_empty(self->considered);
    MAPSET(self->considered, 1);
    for (i = 0; i < q.count; i++) {
      p = q.elements[i];
      if (p > 0 && p < self->nsolvables)
        MAPSET(self->considered, p);
    }
  }
SWIGINTERN void Pool_set_disabled_list(Pool *self,Queue q){
    int i;
    Id p;
    if (!q.count) {
      if (self->considered) {
        map_free(self->considered);
        self->considered = solv_free(self->considered);
      }
      return;
    }
    if (!self->considered) {
      self->considered = solv_calloc(1, sizeof(Map));
      map_init(self->considered, self->nsolvables);
    }
    map_setall(self->considered);
    for (i = 0; i < q.count; i++) {
      p = q.elements[i];
      if (p > 0 && p < self->nsolvables)
        MAPCLR(self->considered, p);
    }
  }
SWIGINTERN void Pool_setpooljobs(Pool *self,Queue solvejobs){
    queue_free(&self->pooljobs);
    queue_init_clone(&self->pooljobs, &solvejobs);
  }
SWIGINTERN Queue Pool_getpooljobs(Pool *self){
    Queue q;
    queue_init_clone(&q, &self->pooljobs);
    return q;
  }
SWIGINTERN void Repo_free(Repo *self,bool reuseids){
    appdata_clr_helper(&self->appdata);
    repo_free(self, reuseids);
  }
SWIGINTERN void Repo_empty(Repo *self,bool reuseids){
    repo_empty(self, reuseids);
  }
SWIGINTERN bool Repo_isempty(Repo *self){
    return !self->nsolvables;
  }

  SWIGINTERN void Repo_appdata_set(Repo *repo, AppObjectPtr appdata) {
    appdata_set_helper(&repo->appdata, appdata);
  }
  SWIGINTERN AppObjectPtr Repo_appdata_get(Repo *repo) {
    return appdata_get_helper(&repo->appdata);
  }
  
SWIGINTERN bool Repo_add_solv__SWIG_0(Repo *self,char const *name,int flags){
    FILE *fp = fopen(name, "r");
    int r;
    if (!fp)
      return 0;
    r = repo_add_solv(self, fp, flags);
    fclose(fp);
    return r == 0;
  }
SWIGINTERN bool Repo_add_solv__SWIG_1(Repo *self,FILE *fp,int flags){
    return repo_add_solv(self, fp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_solvable(Repo *self){
    Id solvid = repo_add_solvable(self);
    return new_XSolvable(self->pool, solvid);
  }
SWIGINTERN bool Repo_add_rpmdb(Repo *self,int flags){
    return repo_add_rpmdb(self, 0, flags) == 0;
  }
SWIGINTERN bool Repo_add_rpmdb_reffp(Repo *self,FILE *reffp,int flags){
    return repo_add_rpmdb_reffp(self, reffp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_rpm(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_rpm(self, name, flags));
  }
SWIGINTERN bool Repo_add_rpmdb_pubkeys(Repo *self,int flags){
    return repo_add_rpmdb_pubkeys(self, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_pubkey(Repo *self,char const *keyfile,int flags){
    return new_XSolvable(self->pool, repo_add_pubkey(self, keyfile, flags));
  }
SWIGINTERN bool Repo_add_keyring(Repo *self,FILE *fp,int flags){
    return repo_add_keyring(self, fp, flags);
  }
SWIGINTERN bool Repo_add_keydir(Repo *self,char const *keydir,char const *suffix,int flags){
    return repo_add_keydir(self, keydir, suffix, flags);
  }
SWIGINTERN bool Repo_add_rpmmd(Repo *self,FILE *fp,char const *language,int flags){
    return repo_add_rpmmd(self, fp, language, flags) == 0;
  }
SWIGINTERN bool Repo_add_repomdxml(Repo *self,FILE *fp,int flags){
    return repo_add_repomdxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_updateinfoxml(Repo *self,FILE *fp,int flags){
    return repo_add_updateinfoxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_deltainfoxml(Repo *self,FILE *fp,int flags){
    return repo_add_deltainfoxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_debdb(Repo *self,int flags){
    return repo_add_debdb(self, flags) == 0;
  }
SWIGINTERN bool Repo_add_debpackages(Repo *self,FILE *fp,int flags){
    return repo_add_debpackages(self, fp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_deb(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_deb(self, name, flags));
  }
SWIGINTERN bool Repo_add_susetags(Repo *self,FILE *fp,Id defvendor,char const *language,int flags){
    return repo_add_susetags(self, fp, defvendor, language, flags) == 0;
  }
SWIGINTERN bool Repo_add_content(Repo *self,FILE *fp,int flags){
    return repo_add_content(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_products(Repo *self,char const *proddir,int flags){
    return repo_add_products(self, proddir, flags) == 0;
  }
SWIGINTERN bool Repo_add_mdk(Repo *self,FILE *fp,int flags){
    return repo_add_mdk(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_mdk_info(Repo *self,FILE *fp,int flags){
    return repo_add_mdk_info(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_arch_repo(Repo *self,FILE *fp,int flags){
    return repo_add_arch_repo(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_arch_local(Repo *self,char const *dir,int flags){
    return repo_add_arch_local(self, dir, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_arch_pkg(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_arch_pkg(self, name, flags));
  }
SWIGINTERN void Repo_internalize(Repo *self){
    repo_internalize(self);
  }
SWIGINTERN bool Repo_write(Repo *self,FILE *fp){
    return repo_write(self, fp) == 0;
  }
SWIGINTERN bool Repo_write_first_repodata(Repo *self,FILE *fp){
    int oldnrepodata = self->nrepodata;
    int res;
    self->nrepodata = oldnrepodata > 2 ? 2 : oldnrepodata;
    res = repo_write(self, fp);
    self->nrepodata = oldnrepodata;
    return res == 0;
  }
SWIGINTERN Dataiterator *Repo_Dataiterator(Repo *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, self, 0, key, match, flags);
  }
SWIGINTERN Dataiterator *Repo_Dataiterator_meta(Repo *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, self, SOLVID_META, key, match, flags);
  }

  SWIGINTERN Id Repo_id_get(Repo *repo) {
    return repo->repoid;
  }
  

  SWIGINTERN Repo_solvable_iterator * Repo_solvables_get(Repo *repo) {
    return new_Repo_solvable_iterator(repo);
  }
  

  SWIGINTERN Datapos * Repo_meta_get(Repo *repo) {
    Datapos *pos = solv_calloc(1, sizeof(*pos));
    pos->solvid = SOLVID_META;
    pos->repo = repo;
    return pos;
  }
  
SWIGINTERN Repo_solvable_iterator *Repo_solvables_iter(Repo *self){
    return new_Repo_solvable_iterator(self);
  }
SWIGINTERN XRepodata *Repo_add_repodata(Repo *self,int flags){
    Repodata *rd = repo_add_repodata(self, flags);
    return new_XRepodata(self, rd->repodataid);
  }
SWIGINTERN void Repo_create_stubs(Repo *self){
    Repodata *data;
    if (!self->nrepodata)
      return;
    data = repo_id2repodata(self, self->nrepodata - 1);
    if (data->state != REPODATA_STUB)
      (void)repodata_create_stubs(data);
  }
SWIGINTERN bool Repo_iscontiguous(Repo *self){
    int i;
    for (i = self->start; i < self->end; i++)
      if (self->pool->solvables[i].repo != self)
        return 0;
    return 1;
  }
SWIGINTERN XRepodata *Repo_first_repodata(Repo *self){
    Repodata *data;
    int i;
    if (self->nrepodata < 2)
      return 0;
    /* make sure all repodatas but the first are extensions */
    data = repo_id2repodata(self, 1);
    if (data->loadcallback)
       return 0;
    for (i = 2; i < self->nrepodata; i++)
      {
        data = repo_id2repodata(self, i);
        if (!data->loadcallback)
          return 0;       /* oops, not an extension */
      }
    return new_XRepodata(self, 1);
  }
SWIGINTERN Selection *Repo_Selection(Repo *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    setflags |= SOLVER_SETREPO;
    queue_push2(&sel->q, SOLVER_SOLVABLE_REPO | setflags, self->repoid);
    return sel;
  }
SWIGINTERN XSolvable *Repo_find_pubkey(Repo *self,char const *keyid){
    return new_XSolvable(self->pool, repo_find_pubkey(self, keyid));
  }
SWIGINTERN Repo *Repo_createshadow(Repo *self,char const *name){
    Repo *repo = repo_create(self->pool, name);
    if (self->idarraysize) {
      repo_reserve_ids(repo, 0, self->idarraysize);
      memcpy(repo->idarraydata, self->idarraydata, sizeof(Id) * self->idarraysize);
      repo->idarraysize = self->idarraysize;
    }
    repo->start = self->start;
    repo->end = self->end;
    repo->nsolvables = self->nsolvables;
    return repo;
  }
SWIGINTERN void Repo_moveshadow(Repo *self,Queue q){
    Pool *pool = self->pool;
    int i;
    for (i = 0; i < q.count; i++) {
      Solvable *s;
      Id p = q.elements[i];
      if (p < self->start || p >= self->end)
        continue;
      s = pool->solvables + p;
      if (self->idarraysize != s->repo->idarraysize)
        continue;
      s->repo = self;
    }
  }
SWIGINTERN bool Repo___eq__(Repo *self,Repo *repo){
    return self == repo;
  }
SWIGINTERN bool Repo___ne__(Repo *self,Repo *repo){
    return self != repo;
  }
SWIGINTERN char const *Repo___str__(Repo *self){
    char buf[20];
    if (self->name)
      return solv_strdup(self->name);
    sprintf(buf, "Repo#%d", self->repoid);
    return solv_strdup(buf);
  }
SWIGINTERN char const *Repo___repr__(Repo *self){
    char buf[20];
    if (self->name)
      {
        sprintf(buf, "<Repo #%d ", self->repoid);
        return solv_dupjoin(buf, self->name, ">");
      }
    sprintf(buf, "<Repo #%d>", self->repoid);
    return solv_strdup(buf);
  }
SWIGINTERN Decision *new_Decision(Solver *solv,Id p,int reason,Id infoid){
    Decision *d = solv_calloc(1, sizeof(*d));
    d->solv = solv;
    d->p = p;
    d->reason = reason;
    d->infoid = infoid;
    return d;
  }

    SWIGINTERN XRule *Decision_rule_get(Decision *d) {
      return d->reason == SOLVER_REASON_WEAKDEP || d->infoid <= 0 ? 0 : new_XRule(d->solv, d->infoid);
    }
    SWIGINTERN XSolvable *Decision_solvable_get(Decision *d) {
      return new_XSolvable(d->solv->pool, d->p >= 0 ? d->p : -d->p);
    }
  
SWIGINTERN Ruleinfo *Decision_info(Decision *self){
    Id type, source, target, dep;
    if (self->reason == SOLVER_REASON_WEAKDEP) {
      type = solver_weakdepinfo(self->solv, self->p, &source, &target, &dep);
    } else if (self->infoid) {
      type = solver_ruleinfo(self->solv, self->infoid, &source, &target, &dep);
    } else {
      return 0;
    }
    return new_Ruleinfo(self->solv, self->infoid, type, source, target, dep);
  }
SWIGINTERN Queue Decision_allinfos(Decision *self){
    Queue q;
    queue_init(&q);
    if (self->reason == SOLVER_REASON_WEAKDEP) {
      solver_allweakdepinfos(self->solv, self->p, &q);
    } else if (self->infoid) {
      solver_allruleinfos(self->solv, self->infoid, &q);
    }
    return q;
  }
SWIGINTERN char const *Decision_reasonstr(Decision *self,bool noinfo){
    if (noinfo)
      return solver_reason2str(self->solv, self->reason);
    return solver_decisionreason2str(self->solv, self->p, self->reason, self->infoid);
  }
SWIGINTERN char const *Decision___str__(Decision *self){
    Pool *pool = self->solv->pool;
    if (self->p == 0 && self->reason == SOLVER_REASON_UNSOLVABLE)
      return "unsolvable";
    if (self->p >= 0)
      return pool_tmpjoin(pool, "install ", pool_solvid2str(pool, self->p), 0);
    else
      return pool_tmpjoin(pool, "conflict ", pool_solvid2str(pool, -self->p), 0);
  }
SWIGINTERN Decisionset *new_Decisionset(Solver *solv){
    Decisionset *d = solv_calloc(1, sizeof(*d));
    d->solv = solv;
    queue_init(&d->decisionlistq);
    return d;
  }
SWIGINTERN void delete_Decisionset(Decisionset *self){
    queue_free(&self->decisionlistq);
    solv_free(self);
  }
SWIGINTERN Ruleinfo *Decisionset_info(Decisionset *self){
    return new_Ruleinfo(self->solv, self->infoid, self->type, self->source, self->target, self->dep_id);
  }

    SWIGINTERN Dep *Decisionset_dep_get(Decisionset *d) {
      return new_Dep(d->solv->pool, d->dep_id);
    }
  
SWIGINTERN Queue Decisionset_solvables(Decisionset *self){
    Queue q;
    int i;
    queue_init(&q);
    for (i = 0; i < self->decisionlistq.count; i += 3)
      if (self->decisionlistq.elements[i] != 0)
        queue_push(&q, self->decisionlistq.elements[i] > 0 ? self->decisionlistq.elements[i] : -self->decisionlistq.elements[i]);
    return q;
  }
SWIGINTERN Queue Decisionset_decisions(Decisionset *self){
    Queue q;
    queue_init_clone(&q, &self->decisionlistq);
    return q;
  }
SWIGINTERN char const *Decisionset_reasonstr(Decisionset *self,bool noinfo){
    if (noinfo || !self->type)
      return solver_reason2str(self->solv, self->reason);
    return solver_decisioninfo2str(self->solv, self->bits, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN char const *Decisionset___str__(Decisionset *self){
    Pool *pool = self->solv->pool;
    Queue q;
    int i;
    const char *s;
    if (!self->decisionlistq.elements)
      return "";
    if (self->p == 0 && self->reason == SOLVER_REASON_UNSOLVABLE)
      return "unsolvable";
    queue_init(&q);
    for (i = 0; i < self->decisionlistq.count; i += 3)
      if (self->decisionlistq.elements[i] != 0)
        queue_push(&q, self->decisionlistq.elements[i] > 0 ? self->decisionlistq.elements[i] : -self->decisionlistq.elements[i]);
    s = pool_solvidset2str(pool, &q);
    queue_free(&q);
    return pool_tmpjoin(pool, self->p >= 0 ? "install " : "conflict ", s, 0);
  }
SWIGINTERN void delete_Solver(Solver *self){
    solver_free(self);
  }
SWIGINTERN int Solver_set_flag(Solver *self,int flag,int value){
    return solver_set_flag(self, flag, value);
  }
SWIGINTERN int Solver_get_flag(Solver *self,int flag){
    return solver_get_flag(self, flag);
  }
SWIGINTERN Queue Solver_solve(Solver *self,Queue solvejobs){
    Queue q;
    int i, cnt;
    queue_init(&q);
    solver_solve(self, &solvejobs);
    cnt = solver_problem_count(self);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN Transaction *Solver_transaction(Solver *self){
    return solver_create_transaction(self);
  }
SWIGINTERN int Solver_describe_decision(Solver *self,XSolvable *s,XRule **OUTPUT){
    Id ruleid;
    int reason = solver_describe_decision(self, s->id, &ruleid);
    *OUTPUT = new_XRule(self, ruleid);
    return reason;
  }
SWIGINTERN Queue Solver_describe_weakdep_decision_raw(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_describe_weakdep_decision(self, s->id, &q);
    return q;
  }
SWIGINTERN int Solver_alternatives_count(Solver *self){
    return solver_alternatives_count(self);
  }
SWIGINTERN Alternative *Solver_get_alternative(Solver *self,Id aid){
    Alternative *a = solv_calloc(1, sizeof(*a));
    a->solv = self;
    queue_init(&a->choices);
    a->type = solver_get_alternative(self, aid, &a->dep_id, &a->from_id, &a->chosen_id, &a->choices, &a->level);
    if (!a->type) {
      queue_free(&a->choices);
      solv_free(a);
      return 0;
    }
    if (a->type == SOLVER_ALTERNATIVE_TYPE_RULE) {
      a->rid = a->dep_id;
      a->dep_id = 0;
    }
    return a;
  }
SWIGINTERN Queue Solver_alternatives(Solver *self){
    Queue q;
    int i, cnt;
    queue_init(&q);
    cnt = solver_alternatives_count(self);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN bool Solver_write_testcase(Solver *self,char const *dir){
    return testcase_write(self, dir, TESTCASE_RESULT_TRANSACTION | TESTCASE_RESULT_PROBLEMS, 0, 0);
  }
SWIGINTERN Queue Solver_raw_decisions(Solver *self,int filter){
    Queue q;
    queue_init(&q);
    solver_get_decisionqueue(self, &q);
    if (filter) {
      int i, j;
      for (i = j = 0; i < q.count; i++)
        if ((filter > 0 && q.elements[i] > 1) ||
            (filter < 0 && q.elements[i] < 0))
          q.elements[j++] = q.elements[i];
      queue_truncate(&q, j);
    }
    return q;
  }
SWIGINTERN Queue Solver_all_decisions(Solver *self,int filter){
    int i, j, cnt;
    Queue q;
    queue_init(&q);
    solver_get_decisionqueue(self, &q);
    if (filter) {
      for (i = j = 0; i < q.count; i++)
        if ((filter > 0 && q.elements[i] > 1) ||
            (filter < 0 && q.elements[i] < 0))
          q.elements[j++] = q.elements[i];
      queue_truncate(&q, j);
    }
    cnt = q.count;
    for (i = 0; i < cnt; i++) {
      Id ruleid, p = q.elements[i];
      int reason;
      if (p == 0 || p == 1)
        continue;       /* ignore system solvable */
      reason = solver_describe_decision(self, p > 0 ? p : -p, &ruleid);
      queue_push(&q, p);
      queue_push2(&q, reason, ruleid);
    }
    queue_deleten(&q, 0, cnt);
    return q;
  }
SWIGINTERN Decision *Solver_get_decision(Solver *self,XSolvable *s){
    Id info;
    int lvl = solver_get_decisionlevel(self, s->id);
    Id p = lvl > 0 ? s->id : -s->id;
    int reason = solver_describe_decision(self, p, &info);
    return new_Decision(self, p, reason, info);
  }
SWIGINTERN Queue Solver_get_learnt(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self, s->id, SOLVER_DECISIONLIST_SOLVABLE, &q);
    return q;
  }
SWIGINTERN Queue Solver_get_decisionlist(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self, s->id, SOLVER_DECISIONLIST_SOLVABLE, &q);
    return q;
  }
SWIGINTERN Queue Solver_get_recommended(Solver *self,bool noselected){
    Queue q;
    queue_init(&q);
    solver_get_recommendations(self, &q, NULL, noselected);
    return q;
  }
SWIGINTERN Queue Solver_get_suggested(Solver *self,bool noselected){
    Queue q;
    queue_init(&q);
    solver_get_recommendations(self, NULL, &q, noselected);
    return q;
  }
SWIGINTERN Chksum *new_Chksum__SWIG_0(Id type){
    return solv_chksum_create(type);
  }
SWIGINTERN Chksum *new_Chksum__SWIG_1(Id type,char const *hex){
    unsigned char buf[64];
    int l = solv_chksum_len(type);
    if (!l)
      return 0;
    if (solv_hex2bin(&hex, buf, sizeof(buf)) != l || hex[0])
      return 0;
    return solv_chksum_create_from_bin(type, buf);
  }

SWIGINTERNINLINE int
SWIG_AsVal_size_t SWIG_PERL_DECL_ARGS_2(SV * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}

SWIGINTERN Chksum *Chksum_from_bin(Id type,unsigned char const *str,size_t len){
    return len == solv_chksum_len(type) ? solv_chksum_create_from_bin(type, str) : 0;
  }
SWIGINTERN void delete_Chksum(Chksum *self){
    solv_chksum_free(self, 0);
  }

  SWIGINTERN Id Chksum_type_get(Chksum *chk) {
    return solv_chksum_get_type(chk);
  }
  
SWIGINTERN void Chksum_add(Chksum *self,unsigned char const *str,size_t len){
    solv_chksum_add(self, str, (int)len);
  }
SWIGINTERN void Chksum_add_fp(Chksum *self,FILE *fp){
    char buf[4096];
    int l;
    while ((l = fread(buf, 1, sizeof(buf), fp)) > 0)
      solv_chksum_add(self, buf, l);
    rewind(fp);         /* convenience */
  }
SWIGINTERN void Chksum_add_fd(Chksum *self,int fd){
    char buf[4096];
    int l;
    while ((l = read(fd, buf, sizeof(buf))) > 0)
      solv_chksum_add(self, buf, l);
    lseek(fd, 0, 0);    /* convenience */
  }
SWIGINTERN void Chksum_add_stat(Chksum *self,char const *filename){
    struct stat stb;
    if (stat(filename, &stb))
      memset(&stb, 0, sizeof(stb));
    solv_chksum_add(self, &stb.st_dev, sizeof(stb.st_dev));
    solv_chksum_add(self, &stb.st_ino, sizeof(stb.st_ino));
    solv_chksum_add(self, &stb.st_size, sizeof(stb.st_size));
    solv_chksum_add(self, &stb.st_mtime, sizeof(stb.st_mtime));
  }
SWIGINTERN void Chksum_add_fstat(Chksum *self,int fd){
    struct stat stb;
    if (fstat(fd, &stb))
      memset(&stb, 0, sizeof(stb));
    solv_chksum_add(self, &stb.st_dev, sizeof(stb.st_dev));
    solv_chksum_add(self, &stb.st_ino, sizeof(stb.st_ino));
    solv_chksum_add(self, &stb.st_size, sizeof(stb.st_size));
    solv_chksum_add(self, &stb.st_mtime, sizeof(stb.st_mtime));
  }
SWIGINTERN BinaryBlob Chksum_raw(Chksum *self){
    BinaryBlob bl;
    int l;
    const unsigned char *b;
    b = solv_chksum_get(self, &l);
    bl.data = b;
    bl.len = l;
    return bl;
  }
SWIGINTERN char *Chksum_hex(Chksum *self){
    int l;
    const unsigned char *b;
    char *ret;

    b = solv_chksum_get(self, &l);
    ret = solv_malloc(2 * l + 1);
    solv_bin2hex(b, l, ret);
    return ret;
  }
SWIGINTERN char const *Chksum_typestr(Chksum *self){
    return solv_chksum_type2str(solv_chksum_get_type(self));
  }
SWIGINTERN bool Chksum___eq__(Chksum *self,Chksum *chk){
    return solv_chksum_cmp(self, chk);
  }
SWIGINTERN bool Chksum___ne__(Chksum *self,Chksum *chk){
    return !solv_chksum_cmp(self, chk);
  }
SWIGINTERN char const *Chksum___str__(Chksum *self){
    const char *str;
    const char *h = 0;
    if (solv_chksum_isfinished(self))
      h = Chksum_hex(self);
    str = solv_dupjoin(solv_chksum_type2str(solv_chksum_get_type(self)), ":", h ? h : "unfinished");
    solv_free((void *)h);
    return str;
  }
SWIGINTERN char const *Chksum___repr__(Chksum *self){
    const char *h = Chksum___str__(self);
    const char *str = solv_dupjoin("<Chksum ", h, ">");
    solv_free((void *)h);
    return str;
  }
SWIGINTERN Solvsig *new_Solvsig(FILE *fp){
    return solvsig_create(fp);
  }
SWIGINTERN void delete_Solvsig(Solvsig *self){
    solvsig_free(self);
  }
SWIGINTERN Chksum *Solvsig_Chksum(Solvsig *self){
    return self->htype ? (Chksum *)solv_chksum_create(self->htype) : 0;
  }
SWIGINTERN XSolvable *Solvsig_verify(Solvsig *self,Repo *repo,Chksum *chksum){
    Id p = solvsig_verify(self, repo, chksum);
    return new_XSolvable(repo->pool, p);
  }
SWIGINTERN void delete_SolvFp(SolvFp *self){
    if (self->fp)
      fclose(self->fp);
    free(self);
  }
SWIGINTERN int SolvFp_fileno(SolvFp *self){
    return self->fp ? fileno(self->fp) : -1;
  }
SWIGINTERN int SolvFp_dup(SolvFp *self){
    return self->fp ? dup(fileno(self->fp)) : -1;
  }
SWIGINTERN bool SolvFp_write(SolvFp *self,unsigned char const *str,size_t len){
    return fwrite(str, len, 1, self->fp) == 1;
  }
SWIGINTERN bool SolvFp_flush(SolvFp *self){
    if (!self->fp)
      return 1;
    return fflush(self->fp) == 0;
  }
SWIGINTERN bool SolvFp_close(SolvFp *self){
    bool ret;
    if (!self->fp)
      return 1;
    ret = fclose(self->fp) == 0;
    self->fp = 0;
    return ret;
  }
SWIGINTERN void SolvFp_cloexec(SolvFp *self,bool state){
    if (!self->fp || fileno(self->fp) == -1)
      return;
    solv_setcloexec(fileno(self->fp), state);
  }

  SWIGINTERN SolvFp *solvfp_xfopen_fd(const char *fn, int fd, const char *mode) {
    SolvFp *sfp;
    FILE *fp;
    fd = dup(fd);
    if (fd == -1)
      return 0;
    solv_setcloexec(fd, 1);
    fp = solv_xfopen_fd(fn, fd, mode);
    if (!fp) {
      close(fd);
      return 0;
    }
    sfp = solv_calloc(1, sizeof(SolvFp));
    sfp->fp = fp;
    return sfp;
  }
  SWIGINTERN SolvFp *solvfp_xfopen(const char *fn, const char *mode) {
    SolvFp *sfp;
    FILE *fp;
    fp = solv_xfopen(fn, mode);
    if (!fp)
      return 0;
    if (fileno(fp) != -1)
      solv_setcloexec(fileno(fp), 1);
    sfp = solv_calloc(1, sizeof(SolvFp));
    sfp->fp = fp;
    return sfp;
  }

SWIGINTERN Problem *new_Problem(Solver *solv,Id id){
    Problem *p;
    p = solv_calloc(1, sizeof(*p));
    p->solv = solv;
    p->id = id;
    return p;
  }
SWIGINTERN XRule *Problem_findproblemrule(Problem *self){
    Id r = solver_findproblemrule(self->solv, self->id);
    return new_XRule(self->solv, r);
  }
SWIGINTERN Queue Problem_findallproblemrules(Problem *self,int unfiltered){
    Solver *solv = self->solv;
    Id probr;
    int i, j;
    Queue q;
    queue_init(&q);
    solver_findallproblemrules(solv, self->id, &q);
    if (!unfiltered)
      {
        for (i = j = 0; i < q.count; i++)
          {
            SolverRuleinfo rclass;
            probr = q.elements[i];
            rclass = solver_ruleclass(solv, probr);
            if (rclass == SOLVER_RULE_UPDATE || rclass == SOLVER_RULE_JOB)
              continue;
            q.elements[j++] = probr;
          }
        if (j)
          queue_truncate(&q, j);
      }
    return q;
  }
SWIGINTERN int Problem_solution_count(Problem *self){
    return solver_solution_count(self->solv, self->id);
  }
SWIGINTERN Queue Problem_solutions(Problem *self){
    Queue q;
    int i, cnt;
    queue_init(&q);
    cnt = solver_solution_count(self->solv, self->id);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN Queue Problem_get_learnt(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM, &q);
    return q;
  }
SWIGINTERN Queue Problem_get_decisionlist(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM | SOLVER_DECISIONLIST_SORTED, &q);
    return q;
  }
SWIGINTERN Queue Problem_get_decisionsetlist(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM | SOLVER_DECISIONLIST_SORTED | SOLVER_DECISIONLIST_WITHINFO | SOLVER_DECISIONLIST_MERGEDINFO, &q);
    prepare_decisionset_queue(self->solv, &q);
    return q;
  }
SWIGINTERN char const *Problem___str__(Problem *self){
    return solver_problem2str(self->solv, self->id);
  }
SWIGINTERN Solution *new_Solution(Problem *p,Id id){
    Solution *s;
    s = solv_calloc(1, sizeof(*s));
    s->solv = p->solv;
    s->problemid = p->id;
    s->id = id;
    return s;
  }
SWIGINTERN int Solution_element_count(Solution *self){
    return solver_solutionelement_count(self->solv, self->problemid, self->id);
  }
SWIGINTERN Queue Solution_elements(Solution *self,bool expandreplaces){
    Queue q;
    queue_init(&q);
    solver_all_solutionelements(self->solv, self->problemid, self->id, expandreplaces, &q);
    return q;
  }
SWIGINTERN Solutionelement *new_Solutionelement(Solver *solv,Id problemid,Id solutionid,Id type,Id p,Id rp){
    Solutionelement *e;
    e = solv_calloc(1, sizeof(*e));
    e->solv = solv;
    e->problemid = problemid;
    e->solutionid = solutionid;
    e->type = type;
    e->p = p;
    e->rp = rp;
    return e;
  }
SWIGINTERN char const *Solutionelement_str(Solutionelement *self){
    return solver_solutionelementtype2str(self->solv, self->type, self->p, self->rp);
  }
SWIGINTERN Queue Solutionelement_replaceelements(Solutionelement *self){
    Queue q;
    int illegal;

    queue_init(&q);
    if (self->type != SOLVER_SOLUTION_REPLACE || self->p <= 0 || self->rp <= 0)
      illegal = 0;
    else
      illegal = policy_is_illegal(self->solv, self->solv->pool->solvables + self->p, self->solv->pool->solvables + self->rp, 0);
    if ((illegal & POLICY_ILLEGAL_DOWNGRADE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_DOWNGRADE);
    if ((illegal & POLICY_ILLEGAL_ARCHCHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_ARCHCHANGE);
    if ((illegal & POLICY_ILLEGAL_VENDORCHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_VENDORCHANGE);
    if ((illegal & POLICY_ILLEGAL_NAMECHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_NAMECHANGE);
    if (!q.count)
      queue_push(&q, self->type);
    return q;
  }
SWIGINTERN int Solutionelement_illegalreplace(Solutionelement *self){
    if (self->type != SOLVER_SOLUTION_REPLACE || self->p <= 0 || self->rp <= 0)
      return 0;
    return policy_is_illegal(self->solv, self->solv->pool->solvables + self->p, self->solv->pool->solvables + self->rp, 0);
  }

    SWIGINTERN XSolvable *Solutionelement_solvable_get(Solutionelement *e) {
      return new_XSolvable(e->solv->pool, e->p);
    }
    SWIGINTERN XSolvable *Solutionelement_replacement_get(Solutionelement *e) {
      return new_XSolvable(e->solv->pool, e->rp);
    }
    SWIGINTERN int Solutionelement_jobidx_get(Solutionelement *e) {
      if (e->type != SOLVER_SOLUTION_JOB && e->type != SOLVER_SOLUTION_POOLJOB)
        return -1;
      return (e->p - 1) / 2;
    }
  
SWIGINTERN Job *Solutionelement_Job(Solutionelement *self){
    Id extraflags = solver_solutionelement_extrajobflags(self->solv, self->problemid, self->solutionid);
    if (self->type == SOLVER_SOLUTION_JOB || self->type == SOLVER_SOLUTION_POOLJOB)
      return new_Job(self->solv->pool, SOLVER_NOOP, 0);
    if (self->type == SOLVER_SOLUTION_INFARCH || self->type == SOLVER_SOLUTION_DISTUPGRADE || self->type == SOLVER_SOLUTION_BEST)
      return new_Job(self->solv->pool, SOLVER_INSTALL|SOLVER_SOLVABLE|SOLVER_NOTBYUSER|extraflags, self->p);
    if (self->type == SOLVER_SOLUTION_REPLACE || self->type == SOLVER_SOLUTION_REPLACE_DOWNGRADE || self->type == SOLVER_SOLUTION_REPLACE_ARCHCHANGE || self->type == SOLVER_SOLUTION_REPLACE_VENDORCHANGE || self->type == SOLVER_SOLUTION_REPLACE_NAMECHANGE)
      return new_Job(self->solv->pool, SOLVER_INSTALL|SOLVER_SOLVABLE|SOLVER_NOTBYUSER|extraflags, self->rp);
    if (self->type == SOLVER_SOLUTION_ERASE)
      return new_Job(self->solv->pool, SOLVER_ERASE|SOLVER_SOLVABLE|extraflags, self->p);
    return 0;
  }
SWIGINTERN void delete_Alternative(Alternative *self){
    queue_free(&self->choices);
    solv_free(self);
  }

    SWIGINTERN XSolvable *Alternative_chosen_get(Alternative *a) {
      return new_XSolvable(a->solv->pool, a->chosen_id);
    }
    SWIGINTERN XRule *Alternative_rule_get(Alternative *a) {
      return new_XRule(a->solv, a->rid);
    }
    SWIGINTERN XSolvable *Alternative_depsolvable_get(Alternative *a) {
      return new_XSolvable(a->solv->pool, a->from_id);
    }
    SWIGINTERN Dep *Alternative_dep_get(Alternative *a) {
      return new_Dep(a->solv->pool, a->dep_id);
    }
  
SWIGINTERN Queue Alternative_choices_raw(Alternative *self){
    Queue q;
    queue_init_clone(&q, &self->choices);
    return q;
  }
SWIGINTERN Queue Alternative_choices(Alternative *self){
    int i;
    Queue q;
    queue_init_clone(&q, &self->choices);
    for (i = 0; i < q.count; i++)
      if (q.elements[i] < 0)
        q.elements[i] = -q.elements[i];
    return q;
  }
SWIGINTERN char const *Alternative___str__(Alternative *self){
    return solver_alternative2str(self->solv, self->type, self->type == SOLVER_ALTERNATIVE_TYPE_RULE ? self->rid : self->dep_id, self->from_id);
  }
SWIGINTERN void delete_Transaction(Transaction *self){
    transaction_free(self);
  }
SWIGINTERN bool Transaction_isempty(Transaction *self){
    return self->steps.count == 0;
  }
SWIGINTERN XSolvable *Transaction_othersolvable(Transaction *self,XSolvable *s){
    Id op = transaction_obs_pkg(self, s->id);
    return new_XSolvable(self->pool, op);
  }
SWIGINTERN Queue Transaction_allothersolvables(Transaction *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    transaction_all_obs_pkgs(self, s->id, &q);
    return q;
  }
SWIGINTERN Queue Transaction_classify(Transaction *self,int mode){
    Queue q;
    queue_init(&q);
    transaction_classify(self, mode, &q);
    return q;
  }
SWIGINTERN Queue Transaction_newpackages(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    queue_truncate(&q, cut);
    return q;
  }
SWIGINTERN Queue Transaction_keptpackages(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    if (cut)
      queue_deleten(&q, 0, cut);
    return q;
  }
SWIGINTERN Queue Transaction_newsolvables(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    queue_truncate(&q, cut);
    return q;
  }
SWIGINTERN Queue Transaction_keptsolvables(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    if (cut)
      queue_deleten(&q, 0, cut);
    return q;
  }
SWIGINTERN Queue Transaction_steps(Transaction *self){
    Queue q;
    queue_init_clone(&q, &self->steps);
    return q;
  }
SWIGINTERN int Transaction_steptype(Transaction *self,XSolvable *s,int mode){
    return transaction_type(self, s->id, mode);
  }
SWIGINTERN long long Transaction_calc_installsizechange(Transaction *self){
    return transaction_calc_installsizechange(self);
  }

#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE SV *
SWIG_From_long_SS_long  SWIG_PERL_DECL_ARGS_1(long long value)
{
  SV *sv;
  if (IVSIZE >= sizeof(value) || (value >= IV_MIN && value <= IV_MAX))
    sv = newSViv((IV)(value));
  else {
    //sv = newSVpvf("%lld", value); doesn't work in non 64bit Perl
    char temp[256];
    SWIG_snprintf(temp, sizeof(temp), "%lld", value);
    sv = newSVpv(temp, 0);
  }
  return sv_2mortal(sv);
}
#endif

SWIGINTERN void Transaction_order(Transaction *self,int flags){
    transaction_order(self, flags);
  }
SWIGINTERN TransactionClass *new_TransactionClass(Transaction *trans,int mode,Id type,int count,Id fromid,Id toid){
    TransactionClass *cl = solv_calloc(1, sizeof(*cl));
    cl->transaction = trans;
    cl->mode = mode;
    cl->type = type;
    cl->count = count;
    cl->fromid = fromid;
    cl->toid = toid;
    return cl;
  }
SWIGINTERN Queue TransactionClass_solvables(TransactionClass *self){
    Queue q;
    queue_init(&q);
    transaction_classify_pkgs(self->transaction, self->mode, self->type, self->fromid, self->toid, &q);
    return q;
  }

    SWIGINTERN const char *TransactionClass_fromstr_get(TransactionClass *cl) {
      return pool_id2str(cl->transaction->pool, cl->fromid);
    }
    SWIGINTERN const char *TransactionClass_tostr_get(TransactionClass *cl) {
      return pool_id2str(cl->transaction->pool, cl->toid);
    }
  
#ifdef __cplusplus
extern "C" {
#endif

#define MAGIC_CLASS
SWIGCLASS_STATIC int swig_magic_readonly(pTHX_ SV *SWIGUNUSEDPARM(sv), MAGIC *SWIGUNUSEDPARM(mg)) {
    MAGIC_PPERL
    croak("Value is read-only.");
    return 0;
}

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif
XS(_wrap_Selection_pool_get) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_pool_get" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_flags_get) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_flags_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_flags_get" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = (int)(int) ((arg1)->flags);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Selection) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Selection(pool);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Selection" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Selection *)new_Selection(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Selection) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Selection(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Selection" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    delete_Selection(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_isempty) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_isempty(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_isempty" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = (bool)Selection_isempty(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_clone) {
  {
    Selection *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Selection_clone(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_clone" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_clone" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)Selection_clone(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_filter) {
  {
    Selection *arg1 = 0 ;
    Selection *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Selection_filter(self,lsel);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_filter" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_filter" "', argument " "2"" of type '" "Selection *""'"); 
    }
    arg2 = (Selection *)(argp2);
    Selection_filter(arg1,arg2);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_add) {
  {
    Selection *arg1 = 0 ;
    Selection *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Selection_add(self,lsel);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_add" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_add" "', argument " "2"" of type '" "Selection *""'"); 
    }
    arg2 = (Selection *)(argp2);
    Selection_add(arg1,arg2);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_add_raw) {
  {
    Selection *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Selection_add_raw(self,how,what);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_add_raw" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_add_raw" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_add_raw" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    Selection_add_raw(arg1,arg2,arg3);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_subtract) {
  {
    Selection *arg1 = 0 ;
    Selection *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Selection_subtract(self,lsel);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_subtract" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_subtract" "', argument " "2"" of type '" "Selection *""'"); 
    }
    arg2 = (Selection *)(argp2);
    Selection_subtract(arg1,arg2);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_select) {
  {
    Selection *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Selection_select(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_select" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_select" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_select" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    Selection_select(arg1,(char const *)arg2,arg3);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_matchdeps) {
  {
    Selection *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Selection_matchdeps(self,name,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchdeps" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_matchdeps" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchdeps" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchdeps" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchdeps" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    Selection_matchdeps(arg1,(char const *)arg2,arg3,arg4,arg5);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_matchdepid) {
  {
    Selection *arg1 = 0 ;
    DepId arg2 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Selection_matchdepid(self,dep,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchdepid" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_matchdepid" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchdepid" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchdepid" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchdepid" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    Selection_matchdepid(arg1,arg2,arg3,arg4,arg5);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_matchsolvable) {
  {
    Selection *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Selection_matchsolvable(self,solvable,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchsolvable" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_matchsolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchsolvable" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchsolvable" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchsolvable" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    Selection_matchsolvable(arg1,arg2,arg3,arg4,arg5);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_jobs) {
  {
    Selection *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Selection_jobs(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_jobs" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_jobs" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = Selection_jobs(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 2;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 2)
      {
        Id id = *idp;
#define result resultx
        Job * result = new_Job(arg1->pool, id, idp[1]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_solvables) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_solvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_solvables" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = Selection_solvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_str) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_str" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = (char *)Selection___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Selection_repr) {
  {
    Selection *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Selection_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Selection, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_repr" "', argument " "1"" of type '" "Selection *""'"); 
    }
    arg1 = (Selection *)(argp1);
    result = (char *)Selection___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_pool_get) {
  {
    Dep *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dep_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_pool_get" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_id_get) {
  {
    Dep *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dep_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_id_get" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Dep) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_Dep(pool,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Dep" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Dep" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Dep *)new_Dep(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_Rel) {
  {
    Dep *arg1 = 0 ;
    int arg2 ;
    DepId arg3 ;
    bool arg4 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    bool val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Dep_Rel(self,flags,evrid,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Rel" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Rel" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Dep_Rel" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Dep_Rel" "', argument " "4"" of type '" "bool""'");
      } 
      arg4 = (bool)(val4);
    }
    result = (Dep *)Dep_Rel(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_Selection_name) {
  {
    Dep *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Dep_Selection_name(self,setflags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Selection_name" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Selection_name" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)Dep_Selection_name(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_Selection_provides) {
  {
    Dep *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Dep_Selection_provides(self,setflags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Selection_provides" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Selection_provides" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)Dep_Selection_provides(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_str) {
  {
    Dep *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dep_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_str" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    result = (char *)Dep_str(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep___eq__) {
  {
    Dep *arg1 = 0 ;
    Dep *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Dep___eq__(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___eq__" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Dep___eq__" "', argument " "2"" of type '" "Dep *""'"); 
    }
    arg2 = (Dep *)(argp2);
    result = (bool)Dep___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep___ne__) {
  {
    Dep *arg1 = 0 ;
    Dep *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Dep___ne__(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___ne__" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Dep___ne__" "', argument " "2"" of type '" "Dep *""'"); 
    }
    arg2 = (Dep *)(argp2);
    result = (bool)Dep___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dep_repr) {
  {
    Dep *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dep_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_repr" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    result = (char *)Dep___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Dep) {
  {
    Dep *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Dep(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dep, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Dep" "', argument " "1"" of type '" "Dep *""'"); 
    }
    arg1 = (Dep *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Dataiterator) {
  {
    Pool *arg1 = 0 ;
    Repo *arg2 = 0 ;
    Id arg3 ;
    Id arg4 ;
    char *arg5 = 0 ;
    int arg6 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int val6 ;
    int ecode6 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 6) || (items > 6)) {
      SWIG_croak("Usage: new_Dataiterator(pool,repo,p,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Dataiterator" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Dataiterator" "', argument " "2"" of type '" "Repo *""'"); 
    }
    arg2 = (Repo *)(argp2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Dataiterator" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Dataiterator" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_Dataiterator" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    ecode6 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(5), &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Dataiterator" "', argument " "6"" of type '" "int""'");
    } 
    arg6 = (int)(val6);
    result = (Dataiterator *)new_Dataiterator(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    
    
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Dataiterator) {
  {
    Dataiterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Dataiterator(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dataiterator, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Dataiterator" "', argument " "1"" of type '" "Dataiterator *""'"); 
    }
    arg1 = (Dataiterator *)(argp1);
    delete_Dataiterator(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dataiterator___next__) {
  {
    Dataiterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Datamatch *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dataiterator___next__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator___next__" "', argument " "1"" of type '" "Dataiterator *""'"); 
    }
    arg1 = (Dataiterator *)(argp1);
    result = (Datamatch *)Dataiterator___next__(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datamatch, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dataiterator_prepend_keyname) {
  {
    Dataiterator *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Dataiterator_prepend_keyname(self,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator_prepend_keyname" "', argument " "1"" of type '" "Dataiterator *""'"); 
    }
    arg1 = (Dataiterator *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dataiterator_prepend_keyname" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    Dataiterator_prepend_keyname(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Dataiterator_skip_solvable) {
  {
    Dataiterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Dataiterator_skip_solvable(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator_skip_solvable" "', argument " "1"" of type '" "Dataiterator *""'"); 
    }
    arg1 = (Dataiterator *)(argp1);
    Dataiterator_skip_solvable(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_pool_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_pool_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_id_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_id_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_XSolvable) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_XSolvable(pool,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XSolvable" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XSolvable" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XSolvable *)new_XSolvable(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_str) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_str" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_str(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_str) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_lookup_str(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_str" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (char *)XSolvable_lookup_str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_id) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_lookup_id(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_id" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Id)XSolvable_lookup_id(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_num) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    unsigned long long arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    unsigned long long val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    unsigned long long result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XSolvable_lookup_num(self,keyname,notfound);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_num" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_num" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_num" "', argument " "3"" of type '" "unsigned long long""'");
      } 
      arg3 = (unsigned long long)(val3);
    }
    result = (unsigned long long)XSolvable_lookup_num(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long long)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_void) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_lookup_void(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_void" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_void" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (bool)XSolvable_lookup_void(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_checksum) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_lookup_checksum(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_checksum" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_checksum" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Chksum *)XSolvable_lookup_checksum(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_idarray) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    Id arg3 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XSolvable_lookup_idarray(self,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_idarray" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_idarray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_idarray" "', argument " "3"" of type '" "Id""'");
      } 
      arg3 = (Id)(val3);
    }
    result = XSolvable_lookup_idarray(arg1,arg2,arg3);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_deparray) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    Id arg3 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XSolvable_lookup_deparray(self,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_deparray" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_deparray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_deparray" "', argument " "3"" of type '" "Id""'");
      } 
      arg3 = (Id)(val3);
    }
    result = XSolvable_lookup_deparray(arg1,arg2,arg3);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Dep * result = new_Dep(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_location) {
  {
    XSolvable *arg1 = 0 ;
    unsigned int *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int temp2 ;
    int res2 = SWIG_TMPOBJ ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    arg2 = &temp2;
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_lookup_location(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_location" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_lookup_location(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    if (SWIG_IsTmpObj(res2)) {
      /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,59,%append_output@*/
      if (argvi >= items) EXTEND(sp, argvi+1);
      ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((*arg2)); argvi++  /*@SWIG@*/;
    } else {
      int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN | 0) : 0;
      /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,59,%append_output@*/
      if (argvi >= items) EXTEND(sp, argvi+1);
      ST(argvi) = SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags); argvi++  /*@SWIG@*/;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_lookup_sourcepkg) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_lookup_sourcepkg(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_sourcepkg" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_lookup_sourcepkg(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_Dataiterator) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: XSolvable_Dataiterator(self,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_Dataiterator" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_Dataiterator" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XSolvable_Dataiterator" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_Dataiterator" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (Dataiterator *)XSolvable_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_installable) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_installable(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_installable" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (bool)XSolvable_installable(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_isinstalled) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_isinstalled(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_isinstalled" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (bool)XSolvable_isinstalled(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_name_set) {
  {
    XSolvable *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_name_set(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_name_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_name_set" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    XSolvable_name_set(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_name_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_name_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_name_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_name_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_nameid_set) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_nameid_set(self,nameid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_nameid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_nameid_set" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    XSolvable_nameid_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_nameid_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_nameid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_nameid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Id)XSolvable_nameid_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_evr_set) {
  {
    XSolvable *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_evr_set(self,evr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evr_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_evr_set" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    XSolvable_evr_set(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_evr_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_evr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evr_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_evr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_evrid_set) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_evrid_set(self,evrid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_evrid_set" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    XSolvable_evrid_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_evrid_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_evrid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Id)XSolvable_evrid_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_arch_set) {
  {
    XSolvable *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_arch_set(self,arch);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_arch_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_arch_set" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    XSolvable_arch_set(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_arch_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_arch_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_arch_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_arch_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_archid_set) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_archid_set(self,archid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_archid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_archid_set" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    XSolvable_archid_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_archid_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_archid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_archid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Id)XSolvable_archid_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_vendor_set) {
  {
    XSolvable *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_vendor_set(self,vendor);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendor_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_vendor_set" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    XSolvable_vendor_set(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_vendor_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_vendor_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendor_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable_vendor_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_vendorid_set) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_vendorid_set(self,vendorid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendorid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_vendorid_set" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    XSolvable_vendorid_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_vendorid_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_vendorid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendorid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Id)XSolvable_vendorid_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_repo_get) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_repo_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_repo_get" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (Repo *)XSolvable_repo_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_provides) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    Id arg3 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XSolvable_add_provides(self,id,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_provides" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_provides" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_provides" "', argument " "3"" of type '" "Id""'");
      } 
      arg3 = (Id)(val3);
    }
    XSolvable_add_provides(arg1,arg2,arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_obsoletes) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_obsoletes(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_obsoletes" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_obsoletes" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_obsoletes(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_conflicts) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_conflicts(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_conflicts" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_conflicts" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_conflicts(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_requires) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    Id arg3 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XSolvable_add_requires(self,id,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_requires" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_requires" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_requires" "', argument " "3"" of type '" "Id""'");
      } 
      arg3 = (Id)(val3);
    }
    XSolvable_add_requires(arg1,arg2,arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_recommends) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_recommends(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_recommends" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_recommends" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_recommends(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_suggests) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_suggests(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_suggests" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_suggests" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_suggests(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_supplements) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_supplements(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_supplements" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_supplements" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_supplements(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_enhances) {
  {
    XSolvable *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_add_enhances(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_enhances" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_enhances" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    XSolvable_add_enhances(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_unset) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_unset(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_unset" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_unset" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    XSolvable_unset(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_add_deparray) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    DepId arg3 ;
    Id arg4 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XSolvable_add_deparray(self,keyname,id,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_deparray" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_deparray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_deparray" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_add_deparray" "', argument " "4"" of type '" "Id""'");
      } 
      arg4 = (Id)(val4);
    }
    XSolvable_add_deparray(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_Selection) {
  {
    XSolvable *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_Selection(self,setflags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_Selection" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_Selection" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)XSolvable_Selection(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_identical) {
  {
    XSolvable *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_identical(self,s2);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_identical" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_identical" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (bool)XSolvable_identical(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_evrcmp) {
  {
    XSolvable *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable_evrcmp(self,s2);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrcmp" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_evrcmp" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (int)XSolvable_evrcmp(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_matchesdep) {
  {
    XSolvable *arg1 = 0 ;
    Id arg2 ;
    DepId arg3 ;
    Id arg4 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XSolvable_matchesdep(self,keyname,id,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_matchesdep" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_matchesdep" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_matchesdep" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_matchesdep" "', argument " "4"" of type '" "Id""'");
      } 
      arg4 = (Id)(val4);
    }
    result = (bool)XSolvable_matchesdep(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable___eq__) {
  {
    XSolvable *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable___eq__(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___eq__" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable___eq__" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (bool)XSolvable___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable___ne__) {
  {
    XSolvable *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XSolvable___ne__(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___ne__" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable___ne__" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (bool)XSolvable___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XSolvable_repr) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XSolvable_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_repr" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    result = (char *)XSolvable___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_XSolvable) {
  {
    XSolvable *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_XSolvable(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XSolvable, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XSolvable" "', argument " "1"" of type '" "XSolvable *""'"); 
    }
    arg1 = (XSolvable *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_solv_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_solv_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_type_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_type_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (int)(int) ((arg1)->type);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_dep_id_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_dep_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_dep_id_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (Id)(Id) ((arg1)->dep_id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Ruleinfo) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    Id arg4 ;
    Id arg5 ;
    Id arg6 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int val6 ;
    int ecode6 = 0 ;
    int argvi = 0;
    Ruleinfo *result = 0 ;
    dXSARGS;
    
    if ((items < 6) || (items > 6)) {
      SWIG_croak("Usage: new_Ruleinfo(solv,rid,type,source,target,dep_id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Ruleinfo" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Ruleinfo" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Ruleinfo" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Ruleinfo" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Ruleinfo" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
    ecode6 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(5), &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Ruleinfo" "', argument " "6"" of type '" "Id""'");
    } 
    arg6 = (Id)(val6);
    result = (Ruleinfo *)new_Ruleinfo(arg1,arg2,arg3,arg4,arg5,arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_solvable_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_solvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_solvable_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (XSolvable *)Ruleinfo_solvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_othersolvable_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_othersolvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_othersolvable_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (XSolvable *)Ruleinfo_othersolvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_dep_get) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_dep_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_dep_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (Dep *)Ruleinfo_dep_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_problemstr) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_problemstr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_problemstr" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (char *)Ruleinfo_problemstr(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Ruleinfo_str) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Ruleinfo_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_str" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    result = (char *)Ruleinfo___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Ruleinfo) {
  {
    Ruleinfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Ruleinfo(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Ruleinfo, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Ruleinfo" "', argument " "1"" of type '" "Ruleinfo *""'"); 
    }
    arg1 = (Ruleinfo *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_solv_get) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_solv_get" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_id_get) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_id_get" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_XRule) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XRule *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_XRule(solv,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XRule" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XRule" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XRule *)new_XRule(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_type_get) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_type_get" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = (int)XRule_type_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_info) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Ruleinfo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_info(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_info" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = (Ruleinfo *)XRule_info(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_allinfos) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_allinfos(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_allinfos" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = XRule_allinfos(arg1);
    {
      int i;
      int cnt = (&result)->count / 4;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 4)
      {
        Id id = *idp;
#define result resultx
        Ruleinfo * result = new_Ruleinfo(arg1->solv, arg1->id, id, idp[1], idp[2], idp[3]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_get_learnt) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_get_learnt(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_learnt" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = XRule_get_learnt(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XRule * result = new_XRule(arg1->solv, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_get_decisionlist) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_get_decisionlist(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_decisionlist" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = XRule_get_decisionlist(arg1);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_get_decisionsetlist) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_get_decisionsetlist(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_decisionsetlist" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = XRule_get_decisionsetlist(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Decisionset * result = decisionset_fromids(arg1->solv, idp + id, idp[1] - id + 1);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule___eq__) {
  {
    XRule *arg1 = 0 ;
    XRule *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XRule___eq__(self,xr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___eq__" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRule___eq__" "', argument " "2"" of type '" "XRule *""'"); 
    }
    arg2 = (XRule *)(argp2);
    result = (bool)XRule___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule___ne__) {
  {
    XRule *arg1 = 0 ;
    XRule *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XRule___ne__(self,xr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___ne__" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRule___ne__" "', argument " "2"" of type '" "XRule *""'"); 
    }
    arg2 = (XRule *)(argp2);
    result = (bool)XRule___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRule_repr) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRule_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_repr" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    result = (char *)XRule___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_XRule) {
  {
    XRule *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_XRule(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRule, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XRule" "', argument " "1"" of type '" "XRule *""'"); 
    }
    arg1 = (XRule *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_repo_get) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_repo_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_repo_get" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    result = (Repo *)(Repo *) ((arg1)->repo);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_id_get) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_id_get" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_XRepodata) {
  {
    Repo *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XRepodata *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_XRepodata(repo,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XRepodata" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XRepodata" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XRepodata *)new_XRepodata(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_new_handle) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_new_handle(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_new_handle" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    result = (Id)XRepodata_new_handle(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_id) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    DepId arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    DepId val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_id(self,solvid,keyname,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_id" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_id" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsValDepId(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_set_id" "', argument " "4"" of type '" "DepId""'");
    } 
    arg4 = (DepId)(val4);
    XRepodata_set_id(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_num) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    unsigned long long arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    unsigned long long val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_num(self,solvid,keyname,num);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_num" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_num" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_num" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_set_num" "', argument " "4"" of type '" "unsigned long long""'");
    } 
    arg4 = (unsigned long long)(val4);
    XRepodata_set_num(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_str) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_str(self,solvid,keyname,str);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_str" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_str" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_str" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    XRepodata_set_str(arg1,arg2,arg3,(char const *)arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_void) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_set_void(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_void" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_void" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_void" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    XRepodata_set_void(arg1,arg2,arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_poolstr) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_poolstr(self,solvid,keyname,str);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_poolstr" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_poolstr" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_poolstr" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_poolstr" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    XRepodata_set_poolstr(arg1,arg2,arg3,(char const *)arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_add_idarray) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    DepId arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    DepId val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_add_idarray(self,solvid,keyname,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_idarray" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_idarray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_idarray" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsValDepId(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_idarray" "', argument " "4"" of type '" "DepId""'");
    } 
    arg4 = (DepId)(val4);
    XRepodata_add_idarray(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_add_flexarray) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    Id arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_add_flexarray(self,solvid,keyname,handle);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_flexarray" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_flexarray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_flexarray" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_flexarray" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    XRepodata_add_flexarray(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_checksum) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    Chksum *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    void *argp4 = 0 ;
    int res4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_checksum(self,solvid,keyname,chksum);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_checksum" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_checksum" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_checksum" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    res4 = SWIG_ConvertPtr(ST(3), &argp4,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_checksum" "', argument " "4"" of type '" "Chksum *""'"); 
    }
    arg4 = (Chksum *)(argp4);
    XRepodata_set_checksum(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_sourcepkg) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_set_sourcepkg(self,solvid,sourcepkg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_sourcepkg" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_sourcepkg" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XRepodata_set_sourcepkg" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    XRepodata_set_sourcepkg(arg1,arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_set_location) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    unsigned int arg3 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    unsigned int val3 ;
    int ecode3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_set_location(self,solvid,mediano,location);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_location" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_location" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_location" "', argument " "3"" of type '" "unsigned int""'");
    } 
    arg3 = (unsigned int)(val3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_location" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    XRepodata_set_location(arg1,arg2,arg3,(char const *)arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_unset) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_unset(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_unset" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_unset" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_unset" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    XRepodata_unset(arg1,arg2,arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_str) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_lookup_str(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_str" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_str" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (char *)XRepodata_lookup_str(arg1,arg2,arg3);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_id) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_lookup_id(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_id" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_id" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Id)XRepodata_lookup_id(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_num) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    unsigned long long arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    unsigned long long val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    unsigned long long result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XRepodata_lookup_num(self,solvid,keyname,notfound);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_num" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_num" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_num" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_lookup_num" "', argument " "4"" of type '" "unsigned long long""'");
      } 
      arg4 = (unsigned long long)(val4);
    }
    result = (unsigned long long)XRepodata_lookup_num(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long long)(result)); argvi++ ;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_void) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_lookup_void(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_void" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_void" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_void" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (bool)XRepodata_lookup_void(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_idarray) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_lookup_idarray(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_idarray" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_idarray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_idarray" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = XRepodata_lookup_idarray(arg1,arg2,arg3);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_lookup_checksum) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_lookup_checksum(self,solvid,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_checksum" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_checksum" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_checksum" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Chksum *)XRepodata_lookup_checksum(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_internalize) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_internalize(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_internalize" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    XRepodata_internalize(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_create_stubs) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_create_stubs(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_create_stubs" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    XRepodata_create_stubs(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_write) {
  {
    XRepodata *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XRepodata_write(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_write" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_write" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    result = (bool)XRepodata_write(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_str2dir) {
  {
    XRepodata *arg1 = 0 ;
    char *arg2 = 0 ;
    bool arg3 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    bool val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_str2dir(self,dir,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_str2dir" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata_str2dir" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_str2dir" "', argument " "3"" of type '" "bool""'");
      } 
      arg3 = (bool)(val3);
    }
    result = (Id)XRepodata_str2dir(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_dir2str) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_dir2str(self,did,suf);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_dir2str" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_dir2str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XRepodata_dir2str" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (char *)XRepodata_dir2str(arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_add_dirstr) {
  {
    XRepodata *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    Id arg4 ;
    char *arg5 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 5) || (items > 5)) {
      SWIG_croak("Usage: XRepodata_add_dirstr(self,solvid,keyname,dir,str);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_dirstr" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_dirstr" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_dirstr" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_dirstr" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "XRepodata_add_dirstr" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    XRepodata_add_dirstr(arg1,arg2,arg3,arg4,(char const *)arg5);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    XSRETURN(argvi);
    fail:;
    
    
    
    
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_add_solv) {
  {
    XRepodata *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XRepodata_add_solv(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_solv" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_solv" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_solv" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)XRepodata_add_solv(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_extend_to_repo) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_extend_to_repo(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_extend_to_repo" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    XRepodata_extend_to_repo(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata___eq__) {
  {
    XRepodata *arg1 = 0 ;
    XRepodata *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XRepodata___eq__(self,xr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___eq__" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata___eq__" "', argument " "2"" of type '" "XRepodata *""'"); 
    }
    arg2 = (XRepodata *)(argp2);
    result = (bool)XRepodata___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata___ne__) {
  {
    XRepodata *arg1 = 0 ;
    XRepodata *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XRepodata___ne__(self,xr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___ne__" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata___ne__" "', argument " "2"" of type '" "XRepodata *""'"); 
    }
    arg2 = (XRepodata *)(argp2);
    result = (bool)XRepodata___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XRepodata_repr) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XRepodata_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_repr" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    result = (char *)XRepodata___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_XRepodata) {
  {
    XRepodata *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_XRepodata(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_XRepodata, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XRepodata" "', argument " "1"" of type '" "XRepodata *""'"); 
    }
    arg1 = (XRepodata *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Pool_solvable_iterator) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Pool_solvable_iterator(pool);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Pool_solvable_iterator" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_solvable_iterator *)new_Pool_solvable_iterator(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvable_iterator___next__) {
  {
    Pool_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_solvable_iterator___next__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___next__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
    }
    arg1 = (Pool_solvable_iterator *)(argp1);
    result = (XSolvable *)Pool_solvable_iterator___next__(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvable_iterator___getitem__) {
  {
    Pool_solvable_iterator *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_solvable_iterator___getitem__(self,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___getitem__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
    }
    arg1 = (Pool_solvable_iterator *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_solvable_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XSolvable *)Pool_solvable_iterator___getitem__(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvable_iterator___len__) {
  {
    Pool_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_solvable_iterator___len__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___len__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
    }
    arg1 = (Pool_solvable_iterator *)(argp1);
    result = (int)Pool_solvable_iterator___len__(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Pool_solvable_iterator) {
  {
    Pool_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Pool_solvable_iterator(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_solvable_iterator, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool_solvable_iterator" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
    }
    arg1 = (Pool_solvable_iterator *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Pool_repo_iterator) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_repo_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Pool_repo_iterator(pool);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Pool_repo_iterator" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_repo_iterator *)new_Pool_repo_iterator(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_repo_iterator___next__) {
  {
    Pool_repo_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_repo_iterator___next__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___next__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
    }
    arg1 = (Pool_repo_iterator *)(argp1);
    result = (Repo *)Pool_repo_iterator___next__(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_repo_iterator___getitem__) {
  {
    Pool_repo_iterator *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_repo_iterator___getitem__(self,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___getitem__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
    }
    arg1 = (Pool_repo_iterator *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_repo_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Repo *)Pool_repo_iterator___getitem__(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_repo_iterator___len__) {
  {
    Pool_repo_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_repo_iterator___len__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___len__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
    }
    arg1 = (Pool_repo_iterator *)(argp1);
    result = (int)Pool_repo_iterator___len__(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Pool_repo_iterator) {
  {
    Pool_repo_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Pool_repo_iterator(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool_repo_iterator, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool_repo_iterator" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
    }
    arg1 = (Pool_repo_iterator *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Repo_solvable_iterator) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Repo_solvable_iterator(repo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Repo_solvable_iterator" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Repo_solvable_iterator *)new_Repo_solvable_iterator(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_solvable_iterator___next__) {
  {
    Repo_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_solvable_iterator___next__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___next__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
    }
    arg1 = (Repo_solvable_iterator *)(argp1);
    result = (XSolvable *)Repo_solvable_iterator___next__(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_solvable_iterator___getitem__) {
  {
    Repo_solvable_iterator *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_solvable_iterator___getitem__(self,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___getitem__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
    }
    arg1 = (Repo_solvable_iterator *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_solvable_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XSolvable *)Repo_solvable_iterator___getitem__(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_solvable_iterator___len__) {
  {
    Repo_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_solvable_iterator___len__(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___len__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
    }
    arg1 = (Repo_solvable_iterator *)(argp1);
    result = (int)Repo_solvable_iterator___len__(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Repo_solvable_iterator) {
  {
    Repo_solvable_iterator *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Repo_solvable_iterator(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo_solvable_iterator, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Repo_solvable_iterator" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
    }
    arg1 = (Repo_solvable_iterator *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_pool_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_pool_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_repo_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_repo_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_repo_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Repo *)(Repo *) ((arg1)->repo);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_solvid_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_solvid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_solvid_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Id)(Id) ((arg1)->solvid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Datamatch) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Datamatch(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Datamatch" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    delete_Datamatch(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_solvable_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_solvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_solvable_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (XSolvable *)Datamatch_solvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_key_id_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_key_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_key_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Id)Datamatch_key_id_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_key_idstr_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_key_idstr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_key_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (char *)Datamatch_key_idstr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_type_id_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_type_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_type_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Id)Datamatch_type_id_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_type_idstr_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_type_idstr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_type_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (char *)Datamatch_type_idstr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_id_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Id)Datamatch_id_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_idstr_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_idstr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (char *)Datamatch_idstr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_str_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_str_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_str_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (char *)Datamatch_str_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_dep_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_dep_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_dep_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Dep *)Datamatch_dep_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_binary_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    BinaryBlob result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_binary_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_binary_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = Datamatch_binary_get(arg1);
    ST(argvi) = SWIG_FromCharPtrAndSize((&result)->data, (&result)->len);
    
    argvi++;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_num_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_num_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_num_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (unsigned long long)Datamatch_num_get(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_num2_get) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_num2_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_num2_get" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (unsigned int)Datamatch_num2_get(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_pos) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Datapos *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_pos(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_pos" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Datapos *)Datamatch_pos(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_parentpos) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Datapos *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_parentpos(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_parentpos" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (Datapos *)Datamatch_parentpos(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datamatch_stringify) {
  {
    Datamatch *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datamatch_stringify(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_stringify" "', argument " "1"" of type '" "Datamatch *""'"); 
    }
    arg1 = (Datamatch *)(argp1);
    result = (char *)Datamatch___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_repo_get) {
  {
    Datapos *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datapos_repo_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_repo_get" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    result = (Repo *)(Repo *) ((arg1)->repo);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_id) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Datapos_lookup_id(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_id" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Id)Datapos_lookup_id(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_str) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Datapos_lookup_str(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_str" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (char *)Datapos_lookup_str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_num) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    unsigned long long arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    unsigned long long val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    unsigned long long result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Datapos_lookup_num(self,keyname,notfound);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_num" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_num" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Datapos_lookup_num" "', argument " "3"" of type '" "unsigned long long""'");
      } 
      arg3 = (unsigned long long)(val3);
    }
    result = (unsigned long long)Datapos_lookup_num(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long long)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_void) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Datapos_lookup_void(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_void" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_void" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (bool)Datapos_lookup_void(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_checksum) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Datapos_lookup_checksum(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_checksum" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_checksum" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Chksum *)Datapos_lookup_checksum(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_deltaseq) {
  {
    Datapos *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datapos_lookup_deltaseq(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_deltaseq" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    result = (char *)Datapos_lookup_deltaseq(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_deltalocation) {
  {
    Datapos *arg1 = 0 ;
    unsigned int *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int temp2 ;
    int res2 = SWIG_TMPOBJ ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    arg2 = &temp2;
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Datapos_lookup_deltalocation(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_deltalocation" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    result = (char *)Datapos_lookup_deltalocation(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    if (SWIG_IsTmpObj(res2)) {
      /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,59,%append_output@*/
      if (argvi >= items) EXTEND(sp, argvi+1);
      ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((*arg2)); argvi++  /*@SWIG@*/;
    } else {
      int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN | 0) : 0;
      /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,59,%append_output@*/
      if (argvi >= items) EXTEND(sp, argvi+1);
      ST(argvi) = SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags); argvi++  /*@SWIG@*/;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_lookup_idarray) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Datapos_lookup_idarray(self,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_idarray" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_idarray" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = Datapos_lookup_idarray(arg1,arg2);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Datapos_Dataiterator) {
  {
    Datapos *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: Datapos_Dataiterator(self,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_Dataiterator" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_Dataiterator" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Datapos_Dataiterator" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Datapos_Dataiterator" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (Dataiterator *)Datapos_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Datapos) {
  {
    Datapos *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Datapos(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Datapos, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Datapos" "', argument " "1"" of type '" "Datapos *""'"); 
    }
    arg1 = (Datapos *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_pool_get) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_pool_get" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_how_set) {
  {
    Job *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Job_how_set(self,how);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_how_set" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Job_how_set" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    if (arg1) (arg1)->how = arg2;
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_how_get) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_how_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_how_get" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (int) ((arg1)->how);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_what_set) {
  {
    Job *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Job_what_set(self,what);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_what_set" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Job_what_set" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (arg1) (arg1)->what = arg2;
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_what_get) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_what_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_what_get" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (Id) ((arg1)->what);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Job) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Job *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: new_Job(pool,how,what);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Job" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Job" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Job" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Job *)new_Job(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_solvables) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_solvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_solvables" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = Job_solvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_isemptyupdate) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_isemptyupdate(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_isemptyupdate" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (bool)Job_isemptyupdate(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job___eq__) {
  {
    Job *arg1 = 0 ;
    Job *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Job___eq__(self,j);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___eq__" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Job___eq__" "', argument " "2"" of type '" "Job *""'"); 
    }
    arg2 = (Job *)(argp2);
    result = (bool)Job___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job___ne__) {
  {
    Job *arg1 = 0 ;
    Job *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Job___ne__(self,j);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___ne__" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Job___ne__" "', argument " "2"" of type '" "Job *""'"); 
    }
    arg2 = (Job *)(argp2);
    result = (bool)Job___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_str) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_str" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (char *)Job___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Job_repr) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Job_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_repr" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    result = (char *)Job___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Job) {
  {
    Job *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Job(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Job, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Job" "', argument " "1"" of type '" "Job *""'"); 
    }
    arg1 = (Job *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Pool) {
  {
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 0) || (items > 0)) {
      SWIG_croak("Usage: new_Pool();");
    }
    result = (Pool *)new_Pool();
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_setdisttype) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_setdisttype(self,disttype);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setdisttype" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_setdisttype" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (int)Pool_setdisttype(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_debuglevel) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_set_debuglevel(self,level);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_debuglevel" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_debuglevel" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    Pool_set_debuglevel(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_flag) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_set_flag(self,flag,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_flag" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_flag" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_set_flag" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    result = (int)Pool_set_flag(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_get_flag) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_get_flag(self,flag);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_flag" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_get_flag" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (int)Pool_get_flag(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_rootdir) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_set_rootdir(self,rootdir);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_rootdir" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_set_rootdir" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    Pool_set_rootdir(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_get_rootdir) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_get_rootdir(self,flag);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_rootdir" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_get_rootdir" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (char *)Pool_get_rootdir(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_clr_loadcallback) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_clr_loadcallback(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_clr_loadcallback" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    Pool_clr_loadcallback(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_loadcallback) {
  {
    Pool *arg1 = 0 ;
    SV *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_set_loadcallback(self,callable);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_loadcallback" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    arg2 = ST(1);
    Pool_set_loadcallback(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Pool) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Pool(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    delete_Pool(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_free) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    disown_helper result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_free(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_free" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_free(arg1);
    {
      SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      ST(argvi) = SWIG_From_int((int)(0));
      
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_disown) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    disown_helper result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_disown(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_disown" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_disown(arg1);
    {
      SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      ST(argvi) = SWIG_From_int((int)(0));
      
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_appdata_set) {
  {
    Pool *arg1 = 0 ;
    AppObjectPtr arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_appdata_set(self,appdata);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_set" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      if (ST(1)) {
        arg2 = newSV(0);
        sv_setsv((SV *)arg2, ST(1));
      } else
      arg2 = (void *)0;
    }
    Pool_appdata_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_appdata_get) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    AppObjectPtr result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_appdata_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_get" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (AppObjectPtr)Pool_appdata_get(arg1);
    {
      ST(argvi) = sv_2mortal(result ? SvREFCNT_inc(result) : newSV(0));
      argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_appdata_disown) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_appdata_disown(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_disown" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    Pool_appdata_disown(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_str2id) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    bool arg3 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    bool val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Pool_str2id(self,str,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_str2id" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_str2id" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_str2id" "', argument " "3"" of type '" "bool""'");
      } 
      arg3 = (bool)(val3);
    }
    result = (Id)Pool_str2id(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Dep) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    bool arg3 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    bool val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Pool_Dep(self,str,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dep" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_Dep" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Dep" "', argument " "3"" of type '" "bool""'");
      } 
      arg3 = (bool)(val3);
    }
    result = (Dep *)Pool_Dep(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_parserpmrichdep) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_parserpmrichdep(self,str);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_parserpmrichdep" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_parserpmrichdep" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (Dep *)Pool_parserpmrichdep(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_id2str) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_id2str(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (char *)Pool_id2str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_dep2str) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_dep2str(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_dep2str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_dep2str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (char *)Pool_dep2str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_rel2id) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    int arg4 ;
    bool arg5 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    bool val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Pool_rel2id(self,name,evr,flags,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_rel2id" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_rel2id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_rel2id" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_rel2id" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_rel2id" "', argument " "5"" of type '" "bool""'");
      } 
      arg5 = (bool)(val5);
    }
    result = (Id)Pool_rel2id(arg1,arg2,arg3,arg4,arg5);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_id2langid) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    bool arg4 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    bool val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_id2langid(self,id,lang,create);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2langid" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2langid" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_id2langid" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_id2langid" "', argument " "4"" of type '" "bool""'");
      } 
      arg4 = (bool)(val4);
    }
    result = (Id)Pool_id2langid(arg1,arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_setarch) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Pool_setarch(self,arch);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setarch" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_setarch" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    Pool_setarch(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_add_repo) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_add_repo(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_add_repo" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_add_repo" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (Repo *)Pool_add_repo(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_lookup_str) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_lookup_str(self,entry,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_str" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (char *)Pool_lookup_str(arg1,arg2,arg3);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_lookup_id) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_lookup_id(self,entry,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_id" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_id" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_id" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Id)Pool_lookup_id(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_lookup_num) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    unsigned long long arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    unsigned long long val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    unsigned long long result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_lookup_num(self,entry,keyname,notfound);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_num" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_num" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_num" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_lookup_num" "', argument " "4"" of type '" "unsigned long long""'");
      } 
      arg4 = (unsigned long long)(val4);
    }
    result = (unsigned long long)Pool_lookup_num(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_From_unsigned_SS_long_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long long)(result)); argvi++ ;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_lookup_void) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_lookup_void(self,entry,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_void" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_void" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_void" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (bool)Pool_lookup_void(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_lookup_checksum) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_lookup_checksum(self,entry,keyname);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_checksum" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_checksum" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_checksum" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Chksum *)Pool_lookup_checksum(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Dataiterator) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: Pool_Dataiterator(self,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dataiterator" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Dataiterator" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_Dataiterator" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_Dataiterator" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (Dataiterator *)Pool_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Dataiterator_solvid) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    char *arg4 = 0 ;
    int arg5 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 5)) {
      SWIG_croak("Usage: Pool_Dataiterator_solvid(self,p,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dataiterator_solvid" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Dataiterator_solvid" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Dataiterator_solvid" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    if (items > 3) {
      res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
      if (!SWIG_IsOK(res4)) {
        SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Pool_Dataiterator_solvid" "', argument " "4"" of type '" "char const *""'");
      }
      arg4 = (char *)(buf4);
    }
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_Dataiterator_solvid" "', argument " "5"" of type '" "int""'");
      } 
      arg5 = (int)(val5);
    }
    result = (Dataiterator *)Pool_Dataiterator_solvid(arg1,arg2,arg3,(char const *)arg4,arg5);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvid2str) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_solvid2str(self,solvid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvid2str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_solvid2str" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (char *)Pool_solvid2str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvidset2str) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_solvidset2str(self,q);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvidset2str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        int v;
        int e = SWIG_AsVal_int(*sv, &v);
        if (!SWIG_IsOK(e))
        SWIG_croak("array in argument 2 must contain only " "integers");
        queue_push(&arg2, v);
      }
    }
    
    result = (char *)Pool_solvidset2str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvableset2str) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_solvableset2str(self,solvables);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvableset2str" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        XSolvable * obj;
        int e = SWIG_ConvertPtr(*sv, (void **)&obj, SWIGTYPE_p_XSolvable, 0 | 0);
        if (!SWIG_IsOK(e))
        SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""XSolvable *");
        queue_push(&arg2, obj->id);
      }
    }
    
    result = (char *)Pool_solvableset2str(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_addfileprovides) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_addfileprovides(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_addfileprovides" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    Pool_addfileprovides(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_addfileprovides_queue) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_addfileprovides_queue(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_addfileprovides_queue" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_addfileprovides_queue(arg1);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_createwhatprovides) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_createwhatprovides(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_createwhatprovides" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    Pool_createwhatprovides(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_id2solvable) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_id2solvable(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2solvable" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2solvable" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (XSolvable *)Pool_id2solvable(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvables_get) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_solvables_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvables_get" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_solvable_iterator *)Pool_solvables_get(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_solvables_iter) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_solvables_iter(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvables_iter" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_solvable_iterator *)Pool_solvables_iter(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_id2repo) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_id2repo(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2repo" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2repo" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Repo *)Pool_id2repo(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_repos_get) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_repo_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_repos_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repos_get" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_repo_iterator *)Pool_repos_get(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_repos_iter) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool_repo_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_repos_iter(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repos_iter" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Pool_repo_iterator *)Pool_repos_iter(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_installed_set) {
  {
    Pool *arg1 = 0 ;
    Repo *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_installed_set(self,installed);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_installed_set" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Repo, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_installed_set" "', argument " "2"" of type '" "Repo *""'"); 
    }
    arg2 = (Repo *)(argp2);
    Pool_installed_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_installed_get) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_installed_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_installed_get" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Repo *)Pool_installed_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_errstr_get) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_errstr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_errstr_get" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (char *)Pool_errstr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_matchprovidingids) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_matchprovidingids(self,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchprovidingids" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchprovidingids" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchprovidingids" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    result = Pool_matchprovidingids(arg1,(char const *)arg2,arg3);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Job) {
  {
    Pool *arg1 = 0 ;
    int arg2 ;
    Id arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Job *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_Job(self,how,what);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Job" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Job" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Job" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    result = (Job *)Pool_Job(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_whatprovides) {
  {
    Pool *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_whatprovides(self,dep);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatprovides" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatprovides" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    result = Pool_whatprovides(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_best_solvables) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Pool_best_solvables(self,solvables,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_best_solvables" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        XSolvable * obj;
        int e = SWIG_ConvertPtr(*sv, (void **)&obj, SWIGTYPE_p_XSolvable, 0 | 0);
        if (!SWIG_IsOK(e))
        SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""XSolvable *");
        queue_push(&arg2, obj->id);
      }
    }
    
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_best_solvables" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = Pool_best_solvables(arg1,arg2,arg3);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    {
      queue_free(&arg2);
    }
    
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_towhatprovides) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_towhatprovides(self,q);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_towhatprovides" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        int v;
        int e = SWIG_AsVal_int(*sv, &v);
        if (!SWIG_IsOK(e))
        SWIG_croak("array in argument 2 must contain only " "integers");
        queue_push(&arg2, v);
      }
    }
    
    result = (Id)Pool_towhatprovides(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_namespaceproviders) {
  {
    Pool *arg1 = 0 ;
    DepId arg2 ;
    DepId arg3 ;
    bool arg4 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    bool val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_set_namespaceproviders(self,ns,evr,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_namespaceproviders" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_namespaceproviders" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_set_namespaceproviders" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_set_namespaceproviders" "', argument " "4"" of type '" "bool""'");
      } 
      arg4 = (bool)(val4);
    }
    Pool_set_namespaceproviders(arg1,arg2,arg3,arg4);
    ST(argvi) = &PL_sv_undef;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_flush_namespaceproviders) {
  {
    Pool *arg1 = 0 ;
    DepId arg2 ;
    DepId arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_flush_namespaceproviders(self,ns,evr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_flush_namespaceproviders" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_flush_namespaceproviders" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_flush_namespaceproviders" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    Pool_flush_namespaceproviders(arg1,arg2,arg3);
    ST(argvi) = &PL_sv_undef;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_whatcontainsdep) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    DepId arg3 ;
    Id arg4 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_whatcontainsdep(self,keyname,dep,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatcontainsdep" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatcontainsdep" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_whatcontainsdep" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatcontainsdep" "', argument " "4"" of type '" "Id""'");
      } 
      arg4 = (Id)(val4);
    }
    result = Pool_whatcontainsdep(arg1,arg2,arg3,arg4);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_whatmatchesdep) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    DepId arg3 ;
    Id arg4 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    DepId val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_whatmatchesdep(self,keyname,dep,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatmatchesdep" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatmatchesdep" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsValDepId(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_whatmatchesdep" "', argument " "3"" of type '" "DepId""'");
    } 
    arg3 = (DepId)(val3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatmatchesdep" "', argument " "4"" of type '" "Id""'");
      } 
      arg4 = (Id)(val4);
    }
    result = Pool_whatmatchesdep(arg1,arg2,arg3,arg4);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_whatmatchessolvable) {
  {
    Pool *arg1 = 0 ;
    Id arg2 ;
    XSolvable *arg3 = 0 ;
    Id arg4 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Pool_whatmatchessolvable(self,keyname,pool_solvable,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatmatchessolvable" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatmatchessolvable" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_whatmatchessolvable" "', argument " "3"" of type '" "XSolvable *""'"); 
    }
    arg3 = (XSolvable *)(argp3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatmatchessolvable" "', argument " "4"" of type '" "Id""'");
      } 
      arg4 = (Id)(val4);
    }
    result = Pool_whatmatchessolvable(arg1,arg2,arg3,arg4);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_isknownarch) {
  {
    Pool *arg1 = 0 ;
    DepId arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_isknownarch(self,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_isknownarch" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_isknownarch" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    result = (bool)Pool_isknownarch(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Solver) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_Solver(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Solver" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Solver *)Pool_Solver(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Selection) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_Selection(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Selection" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = (Selection *)Pool_Selection(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_Selection_all) {
  {
    Pool *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Pool_Selection_all(self,setflags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Selection_all" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Selection_all" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)Pool_Selection_all(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_select) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Pool_select(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_select" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_select" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_select" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    result = (Selection *)Pool_select(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_matchdeps) {
  {
    Pool *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Pool_matchdeps(self,name,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchdeps" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchdeps" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchdeps" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchdeps" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchdeps" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    result = (Selection *)Pool_matchdeps(arg1,(char const *)arg2,arg3,arg4,arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_matchdepid) {
  {
    Pool *arg1 = 0 ;
    DepId arg2 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    DepId val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Pool_matchdepid(self,dep,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchdepid" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    ecode2 = SWIG_AsValDepId(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_matchdepid" "', argument " "2"" of type '" "DepId""'");
    } 
    arg2 = (DepId)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchdepid" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchdepid" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchdepid" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    result = (Selection *)Pool_matchdepid(arg1,arg2,arg3,arg4,arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_matchsolvable) {
  {
    Pool *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    int arg3 ;
    Id arg4 ;
    Id arg5 = -1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Pool_matchsolvable(self,solvable,flags,keyname,marker);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchsolvable" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchsolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchsolvable" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchsolvable" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchsolvable" "', argument " "5"" of type '" "Id""'");
      } 
      arg5 = (Id)(val5);
    }
    result = (Selection *)Pool_matchsolvable(arg1,arg2,arg3,arg4,arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_get_considered_list) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_get_considered_list(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_considered_list" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_get_considered_list(arg1);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_get_disabled_list) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_get_disabled_list(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_disabled_list" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_get_disabled_list(arg1);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_considered_list) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_set_considered_list(self,q);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_considered_list" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        int v;
        int e = SWIG_AsVal_int(*sv, &v);
        if (!SWIG_IsOK(e))
        SWIG_croak("array in argument 2 must contain only " "integers");
        queue_push(&arg2, v);
      }
    }
    
    Pool_set_considered_list(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_set_disabled_list) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_set_disabled_list(self,q);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_disabled_list" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        int v;
        int e = SWIG_AsVal_int(*sv, &v);
        if (!SWIG_IsOK(e))
        SWIG_croak("array in argument 2 must contain only " "integers");
        queue_push(&arg2, v);
      }
    }
    
    Pool_set_disabled_list(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_setpooljobs) {
  {
    Pool *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Pool_setpooljobs(self,solvejobs);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setpooljobs" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        Job * obj;
        int e = SWIG_ConvertPtr(*sv, (void **)&obj, SWIGTYPE_p_Job, 0 | 0);
        if (!SWIG_IsOK(e))
        SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""Job *");
        queue_push2(&arg2, obj->how, obj->what);
      }
    }
    
    Pool_setpooljobs(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Pool_getpooljobs) {
  {
    Pool *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Pool_getpooljobs(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Pool, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_getpooljobs" "', argument " "1"" of type '" "Pool *""'"); 
    }
    arg1 = (Pool *)(argp1);
    result = Pool_getpooljobs(arg1);
    {
      int i;
      int cnt = (&result)->count / 2;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 2)
      {
        Id id = *idp;
#define result resultx
        Job * result = new_Job(arg1, id, idp[1]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_pool_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_pool_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_name_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_name_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_name_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (char *)(char *) ((arg1)->name);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_priority_set) {
  {
    Repo *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_priority_set(self,priority);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_priority_set" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_priority_set" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    if (arg1) (arg1)->priority = arg2;
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_priority_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_priority_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_priority_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (int) ((arg1)->priority);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_subpriority_set) {
  {
    Repo *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_subpriority_set(self,subpriority);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_subpriority_set" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_subpriority_set" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    if (arg1) (arg1)->subpriority = arg2;
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_subpriority_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_subpriority_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_subpriority_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (int) ((arg1)->subpriority);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_nsolvables_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_nsolvables_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_nsolvables_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (int)(int) ((arg1)->nsolvables);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_free) {
  {
    Repo *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_free(self,reuseids);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_free" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_free" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    Repo_free(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_empty) {
  {
    Repo *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_empty(self,reuseids);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_empty" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_empty" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    Repo_empty(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_isempty) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_isempty(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_isempty" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (bool)Repo_isempty(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_appdata_set) {
  {
    Repo *arg1 = 0 ;
    AppObjectPtr arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_appdata_set(self,appdata);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_appdata_set" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    {
      if (ST(1)) {
        arg2 = newSV(0);
        sv_setsv((SV *)arg2, ST(1));
      } else
      arg2 = (void *)0;
    }
    Repo_appdata_set(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_appdata_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    AppObjectPtr result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_appdata_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_appdata_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (AppObjectPtr)Repo_appdata_get(arg1);
    {
      ST(argvi) = sv_2mortal(result ? SvREFCNT_inc(result) : newSV(0));
      argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_solv__SWIG_0) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_solv(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solv" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_solv" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_solv" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_solv__SWIG_0(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_solv__SWIG_1) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_solv(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solv" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_solv" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_solv" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_solv__SWIG_1(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_solv) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if ((items >= 2) && (items <= 3)) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p_Repo, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        {
          int res = SWIG_AsValSolvFpPtr(ST(1), NULL);
          _v = SWIG_CheckState(res);
        }
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (items > 2) {
        {
          {
            int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), NULL);
            _v = SWIG_CheckState(res);
          }
        }
        if (!_v) goto check_1;
        _ranki += _v*_pi;
        _rankm += _pi;
        _pi *= SWIG_MAXCASTRANK;
      }
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if ((items >= 2) && (items <= 3)) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p_Repo, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (items > 2) {
        {
          {
            int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), NULL);
            _v = SWIG_CheckState(res);
          }
        }
        if (!_v) goto check_2;
        _ranki += _v*_pi;
        _rankm += _pi;
        _pi *= SWIG_MAXCASTRANK;
      }
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_Repo_add_solv__SWIG_1); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_Repo_add_solv__SWIG_0); return;
    }
  }
  
  croak("No matching function for overloaded 'Repo_add_solv'");
  XSRETURN(0);
}


XS(_wrap_Repo_add_solvable) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_add_solvable(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solvable" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (XSolvable *)Repo_add_solvable(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_rpmdb) {
  {
    Repo *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_add_rpmdb(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (bool)Repo_add_rpmdb(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_rpmdb_reffp) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_rpmdb_reffp(self,reffp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb_reffp" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb_reffp" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_rpmdb_reffp" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_rpmdb_reffp(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_rpm) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_rpm(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpm" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_rpm" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_rpm" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (XSolvable *)Repo_add_rpm(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_rpmdb_pubkeys) {
  {
    Repo *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_add_rpmdb_pubkeys(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb_pubkeys" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb_pubkeys" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (bool)Repo_add_rpmdb_pubkeys(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_pubkey) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_pubkey(self,keyfile,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_pubkey" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_pubkey" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_pubkey" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (XSolvable *)Repo_add_pubkey(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_keyring) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_keyring(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_keyring" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_keyring" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_keyring" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_keyring(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_keydir) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Repo_add_keydir(self,keydir,suffix,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_keydir" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_keydir" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_add_keydir" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_add_keydir" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (bool)Repo_add_keydir(arg1,(char const *)arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_rpmmd) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: Repo_add_rpmmd(self,fp,language,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmmd" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmmd" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_add_rpmmd" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_add_rpmmd" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (bool)Repo_add_rpmmd(arg1,arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_repomdxml) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_repomdxml(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_repomdxml" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_repomdxml" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_repomdxml" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_repomdxml(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_updateinfoxml) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_updateinfoxml(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_updateinfoxml" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_updateinfoxml" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_updateinfoxml" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_updateinfoxml(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_deltainfoxml) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_deltainfoxml(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_deltainfoxml" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_deltainfoxml" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_deltainfoxml" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_deltainfoxml(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_debdb) {
  {
    Repo *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_add_debdb(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_debdb" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_debdb" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (bool)Repo_add_debdb(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_debpackages) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_debpackages(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_debpackages" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_debpackages" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_debpackages" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_debpackages(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_deb) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_deb(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_deb" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_deb" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_deb" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (XSolvable *)Repo_add_deb(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_susetags) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    Id arg3 ;
    char *arg4 = 0 ;
    int arg5 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Repo_add_susetags(self,fp,defvendor,language,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_susetags" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_susetags" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_susetags" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Repo_add_susetags" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    if (items > 4) {
      ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
      if (!SWIG_IsOK(ecode5)) {
        SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Repo_add_susetags" "', argument " "5"" of type '" "int""'");
      } 
      arg5 = (int)(val5);
    }
    result = (bool)Repo_add_susetags(arg1,arg2,arg3,(char const *)arg4,arg5);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_content) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_content(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_content" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_content" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_content" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_content(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_products) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_products(self,proddir,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_products" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_products" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_products" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_products(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_mdk) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_mdk(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_mdk" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_mdk" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_mdk" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_mdk(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_mdk_info) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_mdk_info(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_mdk_info" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_mdk_info" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_mdk_info" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_mdk_info(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_arch_repo) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_arch_repo(self,fp,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_repo" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_arch_repo" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_repo" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_arch_repo(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_arch_local) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_arch_local(self,dir,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_local" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_arch_local" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_local" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (bool)Repo_add_arch_local(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_arch_pkg) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    int arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: Repo_add_arch_pkg(self,name,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_pkg" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_arch_pkg" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
      if (!SWIG_IsOK(ecode3)) {
        SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_pkg" "', argument " "3"" of type '" "int""'");
      } 
      arg3 = (int)(val3);
    }
    result = (XSolvable *)Repo_add_arch_pkg(arg1,(char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_internalize) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_internalize(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_internalize" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    Repo_internalize(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_write) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_write(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_write" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_write" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    result = (bool)Repo_write(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_write_first_repodata) {
  {
    Repo *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_write_first_repodata(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_write_first_repodata" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_write_first_repodata" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    result = (bool)Repo_write_first_repodata(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_Dataiterator) {
  {
    Repo *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: Repo_Dataiterator(self,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Dataiterator" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Dataiterator" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_Dataiterator" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_Dataiterator" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (Dataiterator *)Repo_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_Dataiterator_meta) {
  {
    Repo *arg1 = 0 ;
    Id arg2 ;
    char *arg3 = 0 ;
    int arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Dataiterator *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: Repo_Dataiterator_meta(self,key,match,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Dataiterator_meta" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Dataiterator_meta" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_Dataiterator_meta" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_Dataiterator_meta" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (Dataiterator *)Repo_Dataiterator_meta(arg1,arg2,(char const *)arg3,arg4);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_id_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_id_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Id)Repo_id_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_solvables_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_solvables_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvables_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Repo_solvable_iterator *)Repo_solvables_get(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_OWNER | SWIG_SHADOW); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_meta_get) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Datapos *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_meta_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_meta_get" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Datapos *)Repo_meta_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_solvables_iter) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Repo_solvable_iterator *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_solvables_iter(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvables_iter" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (Repo_solvable_iterator *)Repo_solvables_iter(arg1);
    {
      ST(argvi) = SWIG_Perl_NewArrayObj(SWIG_PERL_OBJECT_CALL SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_OWNER | SWIG_SHADOW | SWIG_PERL_ITERATOR); argvi++;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_add_repodata) {
  {
    Repo *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    XRepodata *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_add_repodata(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_repodata" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_repodata" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (XRepodata *)Repo_add_repodata(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_create_stubs) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_create_stubs(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_create_stubs" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    Repo_create_stubs(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_iscontiguous) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_iscontiguous(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_iscontiguous" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (bool)Repo_iscontiguous(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_first_repodata) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XRepodata *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_first_repodata(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_first_repodata" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (XRepodata *)Repo_first_repodata(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_Selection) {
  {
    Repo *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Selection *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Repo_Selection(self,setflags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Selection" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Selection" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (Selection *)Repo_Selection(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_find_pubkey) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_find_pubkey(self,keyid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_find_pubkey" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_find_pubkey" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (XSolvable *)Repo_find_pubkey(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_createshadow) {
  {
    Repo *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    Repo *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_createshadow(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_createshadow" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_createshadow" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (Repo *)Repo_createshadow(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_moveshadow) {
  {
    Repo *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo_moveshadow(self,q);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_moveshadow" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        int v;
        int e = SWIG_AsVal_int(*sv, &v);
        if (!SWIG_IsOK(e))
        SWIG_croak("array in argument 2 must contain only " "integers");
        queue_push(&arg2, v);
      }
    }
    
    Repo_moveshadow(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Repo___eq__) {
  {
    Repo *arg1 = 0 ;
    Repo *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo___eq__(self,repo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___eq__" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo___eq__" "', argument " "2"" of type '" "Repo *""'"); 
    }
    arg2 = (Repo *)(argp2);
    result = (bool)Repo___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo___ne__) {
  {
    Repo *arg1 = 0 ;
    Repo *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Repo___ne__(self,repo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___ne__" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo___ne__" "', argument " "2"" of type '" "Repo *""'"); 
    }
    arg2 = (Repo *)(argp2);
    result = (bool)Repo___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_str) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_str" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (char *)Repo___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Repo_repr) {
  {
    Repo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Repo_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_repr" "', argument " "1"" of type '" "Repo *""'"); 
    }
    arg1 = (Repo *)(argp1);
    result = (char *)Repo___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_solv_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_solv_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_p_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_p_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_p_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (Id)(Id) ((arg1)->p);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_reason_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_reason_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_reason_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (int)(int) ((arg1)->reason);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_infoid_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_infoid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_infoid_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (Id)(Id) ((arg1)->infoid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Decision) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    int arg3 ;
    Id arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    Decision *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: new_Decision(solv,p,reason,infoid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Decision" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Decision" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Decision" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Decision" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    result = (Decision *)new_Decision(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_rule_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XRule *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_rule_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_rule_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (XRule *)Decision_rule_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_solvable_get) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_solvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_solvable_get" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (XSolvable *)Decision_solvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_info) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Ruleinfo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_info(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_info" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (Ruleinfo *)Decision_info(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_allinfos) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_allinfos(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_allinfos" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = Decision_allinfos(arg1);
    {
      int i;
      int cnt = (&result)->count / 4;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 4)
      {
        Id id = *idp;
#define result resultx
        Ruleinfo * result = new_Ruleinfo(arg1->solv, arg1->infoid, id, idp[1], idp[2], idp[3]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_reasonstr) {
  {
    Decision *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Decision_reasonstr(self,noinfo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_reasonstr" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Decision_reasonstr" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    result = (char *)Decision_reasonstr(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decision_str) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decision_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_str" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    result = (char *)Decision___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Decision) {
  {
    Decision *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Decision(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decision, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Decision" "', argument " "1"" of type '" "Decision *""'"); 
    }
    arg1 = (Decision *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_solv_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_solv_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_p_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_p_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_p_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Id)(Id) ((arg1)->p);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_reason_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_reason_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_reason_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (int)(int) ((arg1)->reason);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_infoid_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_infoid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_infoid_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Id)(Id) ((arg1)->infoid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_bits_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_bits_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_bits_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (int)(int) ((arg1)->bits);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_type_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_type_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (int)(int) ((arg1)->type);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_dep_id_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_dep_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_dep_id_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Id)(Id) ((arg1)->dep_id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Decisionset) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Decisionset *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Decisionset(solv);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Decisionset" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    result = (Decisionset *)new_Decisionset(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Decisionset) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Decisionset(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Decisionset" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    delete_Decisionset(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_info) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Ruleinfo *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_info(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_info" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Ruleinfo *)Decisionset_info(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_dep_get) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_dep_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_dep_get" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (Dep *)Decisionset_dep_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_solvables) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_solvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_solvables" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = Decisionset_solvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->solv->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_decisions) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_decisions(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_decisions" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = Decisionset_decisions(arg1);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_reasonstr) {
  {
    Decisionset *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Decisionset_reasonstr(self,noinfo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_reasonstr" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Decisionset_reasonstr" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    result = (char *)Decisionset_reasonstr(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Decisionset_str) {
  {
    Decisionset *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Decisionset_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_str" "', argument " "1"" of type '" "Decisionset *""'"); 
    }
    arg1 = (Decisionset *)(argp1);
    result = (char *)Decisionset___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_pool_get) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solver_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_pool_get" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Solver) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Solver(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solver" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    delete_Solver(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_set_flag) {
  {
    Solver *arg1 = 0 ;
    int arg2 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Solver_set_flag(self,flag,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_set_flag" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_set_flag" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Solver_set_flag" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    result = (int)Solver_set_flag(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_flag) {
  {
    Solver *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_flag(self,flag);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_flag" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_flag" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (int)Solver_get_flag(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_solve) {
  {
    Solver *arg1 = 0 ;
    Queue arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    {
      queue_init(&arg2);
    }
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_solve(self,solvejobs);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_solve" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    {
      AV *av;
      int i, size;
      if (!SvROK(ST(1)) || SvTYPE(SvRV(ST(1))) != SVt_PVAV)
      SWIG_croak("argument 2 is not an array reference.");
      av = (AV*)SvRV(ST(1));
      size = av_len(av);
      for (i = 0; i <= size; i++) {
        SV **sv = av_fetch(av, i, 0);
        Job * obj;
        int e = SWIG_ConvertPtr(*sv, (void **)&obj, SWIGTYPE_p_Job, 0 | 0);
        if (!SWIG_IsOK(e))
        SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""Job *");
        queue_push2(&arg2, obj->how, obj->what);
      }
    }
    
    result = Solver_solve(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Problem * result = new_Problem(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Problem, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    {
      queue_free(&arg2);
    }
    XSRETURN(argvi);
    fail:;
    
    {
      queue_free(&arg2);
    }
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_transaction) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Transaction *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solver_transaction(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_transaction" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    result = (Transaction *)Solver_transaction(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Transaction, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_describe_decision) {
  {
    Solver *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    XRule **arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    XRule *temp3 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    arg3 = &temp3;
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_describe_decision(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_describe_decision" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_describe_decision" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (int)Solver_describe_decision(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,59,%append_output@*/
    if (argvi >= items) EXTEND(sp, argvi+1);
    ST(argvi) = SWIG_NewPointerObj((void*)(*arg3), SWIGTYPE_p_XRule, SWIG_POINTER_OWN | 0); argvi++  /*@SWIG@*/;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_describe_weakdep_decision_raw) {
  {
    Solver *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_describe_weakdep_decision_raw(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_describe_weakdep_decision_raw" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_describe_weakdep_decision_raw" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = Solver_describe_weakdep_decision_raw(arg1,arg2);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_alternatives_count) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solver_alternatives_count(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_alternatives_count" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    result = (int)Solver_alternatives_count(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_alternative) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Alternative *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_alternative(self,aid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_alternative" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_alternative" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Alternative *)Solver_get_alternative(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Alternative, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_alternatives) {
  {
    Solver *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solver_alternatives(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_alternatives" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    result = Solver_alternatives(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Alternative * result = Solver_get_alternative(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Alternative, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_write_testcase) {
  {
    Solver *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_write_testcase(self,dir);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_write_testcase" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_write_testcase" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (bool)Solver_write_testcase(arg1,(char const *)arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_raw_decisions) {
  {
    Solver *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Solver_raw_decisions(self,filter);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_raw_decisions" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_raw_decisions" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = Solver_raw_decisions(arg1,arg2);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_all_decisions) {
  {
    Solver *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Solver_all_decisions(self,filter);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_all_decisions" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_all_decisions" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = Solver_all_decisions(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Decision * result = new_Decision(arg1, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_decision) {
  {
    Solver *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    Decision *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_decision(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_decision" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_decision" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (Decision *)Solver_get_decision(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_learnt) {
  {
    Solver *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_learnt(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_learnt" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_learnt" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = Solver_get_learnt(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XRule * result = new_XRule(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_decisionlist) {
  {
    Solver *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_decisionlist(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_decisionlist" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_decisionlist" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = Solver_get_decisionlist(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Decision * result = new_Decision(arg1, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_recommended) {
  {
    Solver *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_recommended(self,noselected);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_recommended" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_recommended" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    result = Solver_get_recommended(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solver_get_suggested) {
  {
    Solver *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Solver_get_suggested(self,noselected);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_suggested" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_suggested" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    result = Solver_get_suggested(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Chksum__SWIG_0) {
  {
    Id arg1 ;
    int val1 ;
    int ecode1 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Chksum(type);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Chksum" "', argument " "1"" of type '" "Id""'");
    } 
    arg1 = (Id)(val1);
    result = (Chksum *)new_Chksum__SWIG_0(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Chksum__SWIG_1) {
  {
    Id arg1 ;
    char *arg2 = 0 ;
    int val1 ;
    int ecode1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_Chksum(type,hex);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Chksum" "', argument " "1"" of type '" "Id""'");
    } 
    arg1 = (Id)(val1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Chksum" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (Chksum *)new_Chksum__SWIG_1(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_new_Chksum) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if (items == 1) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        {
          int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), NULL);
          _v = SWIG_CheckState(res);
        }
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if (items == 2) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        {
          int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), NULL);
          _v = SWIG_CheckState(res);
        }
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_Chksum__SWIG_0); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_Chksum__SWIG_1); return;
    }
  }
  
  croak("No matching function for overloaded 'new_Chksum'");
  XSRETURN(0);
}


XS(_wrap_Chksum_from_bin) {
  {
    Id arg1 ;
    unsigned char *arg2 = 0 ;
    size_t arg3 ;
    int val1 ;
    int ecode1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    size_t size2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_from_bin(type,str,len);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Chksum_from_bin" "', argument " "1"" of type '" "Id""'");
    } 
    arg1 = (Id)(val1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, &size2, &alloc2);
    if (buf2 && size2)
    size2--;
    
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_from_bin" "', argument " "2"" of type '" "const char *""'");
      
    }
    arg2 = (unsigned char *)buf2;
    arg3 = size2;
    result = (Chksum *)Chksum_from_bin(arg1,(unsigned char const *)arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Chksum) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Chksum(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Chksum" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    delete_Chksum(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_type_get) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_type_get" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = (Id)Chksum_type_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_add) {
  {
    Chksum *arg1 = 0 ;
    unsigned char *arg2 = 0 ;
    size_t arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    size_t size2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_add(self,str,len);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, &size2, &alloc2);
    if (buf2 && size2)
    size2--;
    
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_add" "', argument " "2"" of type '" "const char *""'");
      
    }
    arg2 = (unsigned char *)buf2;
    arg3 = size2;
    Chksum_add(arg1,(unsigned char const *)arg2,arg3);
    {
      ST(argvi) = sv_2mortal(SvREFCNT_inc(ST(0)));argvi++;
    }
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_add_fp) {
  {
    Chksum *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    FILE *val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_add_fp(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fp" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    ecode2 = SWIG_AsValSolvFpPtr(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fp" "', argument " "2"" of type '" "FILE *""'");
    } 
    arg2 = (FILE *)(val2);
    Chksum_add_fp(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_add_fd) {
  {
    Chksum *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_add_fd(self,fd);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fd" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fd" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    Chksum_add_fd(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_add_stat) {
  {
    Chksum *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_add_stat(self,filename);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_stat" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_add_stat" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    Chksum_add_stat(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_add_fstat) {
  {
    Chksum *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum_add_fstat(self,fd);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fstat" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fstat" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    Chksum_add_fstat(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_raw) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    BinaryBlob result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_raw(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_raw" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = Chksum_raw(arg1);
    ST(argvi) = SWIG_FromCharPtrAndSize((&result)->data, (&result)->len);
    
    argvi++;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_hex) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_hex(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_hex" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = (char *)Chksum_hex(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_typestr) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_typestr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_typestr" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = (char *)Chksum_typestr(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum___eq__) {
  {
    Chksum *arg1 = 0 ;
    Chksum *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum___eq__(self,chk);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___eq__" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum___eq__" "', argument " "2"" of type '" "Chksum *""'"); 
    }
    arg2 = (Chksum *)(argp2);
    result = (bool)Chksum___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum___ne__) {
  {
    Chksum *arg1 = 0 ;
    Chksum *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Chksum___ne__(self,chk);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___ne__" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum___ne__" "', argument " "2"" of type '" "Chksum *""'"); 
    }
    arg2 = (Chksum *)(argp2);
    result = (bool)Chksum___ne__(arg1,arg2);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_str) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_str" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = (char *)Chksum___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Chksum_repr) {
  {
    Chksum *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Chksum_repr(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_repr" "', argument " "1"" of type '" "Chksum *""'"); 
    }
    arg1 = (Chksum *)(argp1);
    result = (char *)Chksum___repr__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free((char*)result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_htype_get) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solvsig_htype_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_htype_get" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    result = (Id)(Id) ((arg1)->htype);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_created_get) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solvsig_created_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_created_get" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    result = (unsigned int)(unsigned int) ((arg1)->created);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_expires_get) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solvsig_expires_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_expires_get" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    result = (unsigned int)(unsigned int) ((arg1)->expires);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_keyid_get) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solvsig_keyid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_keyid_get" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    result = (char *)(char *) ((arg1)->keyid);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Solvsig) {
  {
    FILE *arg1 = 0 ;
    FILE *val1 ;
    int ecode1 = 0 ;
    int argvi = 0;
    Solvsig *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Solvsig(fp);");
    }
    ecode1 = SWIG_AsValSolvFpPtr(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Solvsig" "', argument " "1"" of type '" "FILE *""'");
    } 
    arg1 = (FILE *)(val1);
    result = (Solvsig *)new_Solvsig(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solvsig, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Solvsig) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Solvsig(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solvsig" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    delete_Solvsig(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_Chksum) {
  {
    Solvsig *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Chksum *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solvsig_Chksum(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_Chksum" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    result = (Chksum *)Solvsig_Chksum(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solvsig_verify) {
  {
    Solvsig *arg1 = 0 ;
    Repo *arg2 = 0 ;
    Chksum *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Solvsig_verify(self,repo,chksum);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_verify" "', argument " "1"" of type '" "Solvsig *""'"); 
    }
    arg1 = (Solvsig *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_Repo, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solvsig_verify" "', argument " "2"" of type '" "Repo *""'"); 
    }
    arg2 = (Repo *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_Chksum, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Solvsig_verify" "', argument " "3"" of type '" "Chksum *""'"); 
    }
    arg3 = (Chksum *)(argp3);
    result = (XSolvable *)Solvsig_verify(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_SolvFp) {
  {
    SolvFp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_SolvFp(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SolvFp" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    delete_SolvFp(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_fileno) {
  {
    SolvFp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SolvFp_fileno(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_fileno" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    result = (int)SolvFp_fileno(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_dup) {
  {
    SolvFp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SolvFp_dup(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_dup" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    result = (int)SolvFp_dup(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_write) {
  {
    SolvFp *arg1 = 0 ;
    unsigned char *arg2 = 0 ;
    size_t arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    size_t size2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: SolvFp_write(self,str,len);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_write" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, &size2, &alloc2);
    if (buf2 && size2)
    size2--;
    
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SolvFp_write" "', argument " "2"" of type '" "const char *""'");
      
    }
    arg2 = (unsigned char *)buf2;
    arg3 = size2;
    result = (bool)SolvFp_write(arg1,(unsigned char const *)arg2,arg3);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_flush) {
  {
    SolvFp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SolvFp_flush(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_flush" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    result = (bool)SolvFp_flush(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_close) {
  {
    SolvFp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SolvFp_close(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_close" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    result = (bool)SolvFp_close(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SolvFp_cloexec) {
  {
    SolvFp *arg1 = 0 ;
    bool arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: SolvFp_cloexec(self,state);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_cloexec" "', argument " "1"" of type '" "SolvFp *""'"); 
    }
    arg1 = (SolvFp *)(argp1);
    ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SolvFp_cloexec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
    SolvFp_cloexec(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_xfopen) {
  {
    char *arg1 = 0 ;
    char *arg2 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    SolvFp *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: xfopen(fn,mode);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "xfopen" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "xfopen" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (SolvFp *)solvfp_xfopen((char const *)arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_SolvFp, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_xfopen_fd) {
  {
    char *arg1 = 0 ;
    int arg2 ;
    char *arg3 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    SolvFp *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: xfopen_fd(fn,fd,mode);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "xfopen_fd" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "xfopen_fd" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "xfopen_fd" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (SolvFp *)solvfp_xfopen_fd((char const *)arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_SolvFp, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_solv_get) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solv_get" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_id_get) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_id_get" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Problem) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Problem *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_Problem(solv,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Problem" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Problem" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Problem *)new_Problem(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Problem, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_findproblemrule) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XRule *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_findproblemrule(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_findproblemrule" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = (XRule *)Problem_findproblemrule(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_findallproblemrules) {
  {
    Problem *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Problem_findallproblemrules(self,unfiltered);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_findallproblemrules" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Problem_findallproblemrules" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = Problem_findallproblemrules(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XRule * result = new_XRule(arg1->solv, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_solution_count) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_solution_count(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solution_count" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = (int)Problem_solution_count(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_solutions) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_solutions(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solutions" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = Problem_solutions(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Solution * result = new_Solution(arg1, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solution, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_get_learnt) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_get_learnt(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_learnt" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = Problem_get_learnt(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XRule * result = new_XRule(arg1->solv, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_get_decisionlist) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_get_decisionlist(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_decisionlist" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = Problem_get_decisionlist(arg1);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_get_decisionsetlist) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_get_decisionsetlist(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_decisionsetlist" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = Problem_get_decisionsetlist(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Decisionset * result = decisionset_fromids(arg1->solv, idp + id, idp[1] - id + 1);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Problem_str) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Problem_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_str" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    result = (char *)Problem___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Problem) {
  {
    Problem *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Problem(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Problem" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solution_solv_get) {
  {
    Solution *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solution_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_solv_get" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solution_problemid_get) {
  {
    Solution *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solution_problemid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_problemid_get" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    result = (Id)(Id) ((arg1)->problemid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solution_id_get) {
  {
    Solution *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solution_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_id_get" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    result = (Id)(Id) ((arg1)->id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Solution) {
  {
    Problem *arg1 = 0 ;
    Id arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Solution *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_Solution(p,id);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Problem, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Solution" "', argument " "1"" of type '" "Problem *""'"); 
    }
    arg1 = (Problem *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Solution" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    result = (Solution *)new_Solution(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solution, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solution_element_count) {
  {
    Solution *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solution_element_count(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_element_count" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    result = (int)Solution_element_count(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solution_elements) {
  {
    Solution *arg1 = 0 ;
    bool arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    bool val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Solution_elements(self,expandreplaces);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_elements" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_bool SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solution_elements" "', argument " "2"" of type '" "bool""'");
      } 
      arg2 = (bool)(val2);
    }
    result = Solution_elements(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 3;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 3)
      {
        Id id = *idp;
#define result resultx
        Solutionelement * result = new_Solutionelement(arg1->solv, arg1->problemid, arg1->id, id, idp[1], idp[2]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Solution) {
  {
    Solution *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Solution(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solution, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solution" "', argument " "1"" of type '" "Solution *""'"); 
    }
    arg1 = (Solution *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_solv_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solv_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_problemid_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_problemid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_problemid_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (Id)(Id) ((arg1)->problemid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_solutionid_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_solutionid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solutionid_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (Id)(Id) ((arg1)->solutionid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_type_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_type_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (Id)(Id) ((arg1)->type);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Solutionelement) {
  {
    Solver *arg1 = 0 ;
    Id arg2 ;
    Id arg3 ;
    Id arg4 ;
    Id arg5 ;
    Id arg6 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int val6 ;
    int ecode6 = 0 ;
    int argvi = 0;
    Solutionelement *result = 0 ;
    dXSARGS;
    
    if ((items < 6) || (items > 6)) {
      SWIG_croak("Usage: new_Solutionelement(solv,problemid,solutionid,type,p,rp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solver, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Solutionelement" "', argument " "1"" of type '" "Solver *""'"); 
    }
    arg1 = (Solver *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Solutionelement" "', argument " "2"" of type '" "Id""'");
    } 
    arg2 = (Id)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Solutionelement" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Solutionelement" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
    ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Solutionelement" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
    ecode6 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(5), &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Solutionelement" "', argument " "6"" of type '" "Id""'");
    } 
    arg6 = (Id)(val6);
    result = (Solutionelement *)new_Solutionelement(arg1,arg2,arg3,arg4,arg5,arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_str) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_str" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (char *)Solutionelement_str(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_replaceelements) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_replaceelements(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_replaceelements" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = Solutionelement_replaceelements(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        Solutionelement * result = new_Solutionelement(arg1->solv, arg1->problemid, arg1->solutionid, id, arg1->p, arg1->rp);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_illegalreplace) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_illegalreplace(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_illegalreplace" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (int)Solutionelement_illegalreplace(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_solvable_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_solvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solvable_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (XSolvable *)Solutionelement_solvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_replacement_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_replacement_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_replacement_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (XSolvable *)Solutionelement_replacement_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_jobidx_get) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_jobidx_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_jobidx_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (int)Solutionelement_jobidx_get(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Solutionelement_Job) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Job *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Solutionelement_Job(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_Job" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    result = (Job *)Solutionelement_Job(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Solutionelement) {
  {
    Solutionelement *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Solutionelement(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Solutionelement, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solutionelement" "', argument " "1"" of type '" "Solutionelement *""'"); 
    }
    arg1 = (Solutionelement *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_solv_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Solver *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_solv_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_solv_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (Solver *)(Solver *) ((arg1)->solv);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_type_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_type_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (Id)(Id) ((arg1)->type);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_dep_id_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_dep_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_dep_id_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (Id)(Id) ((arg1)->dep_id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_chosen_id_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_chosen_id_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_chosen_id_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (Id)(Id) ((arg1)->chosen_id);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_level_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_level_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_level_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (int)(int) ((arg1)->level);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Alternative) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Alternative(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Alternative" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    delete_Alternative(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_chosen_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_chosen_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_chosen_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (XSolvable *)Alternative_chosen_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_rule_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XRule *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_rule_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_rule_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (XRule *)Alternative_rule_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_depsolvable_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_depsolvable_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_depsolvable_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (XSolvable *)Alternative_depsolvable_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_dep_get) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Dep *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_dep_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_dep_get" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (Dep *)Alternative_dep_get(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_choices_raw) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_choices_raw(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_choices_raw" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = Alternative_choices_raw(arg1);
    {
      int i;
      if (argvi + (&result)->count + 1 >= items) {
        EXTEND(sp, (argvi + (&result)->count + 1) - items + 1);
      }
      for (i = 0; i < (&result)->count; i++)
      ST(argvi++) = SvREFCNT_inc(SWIG_From_int((&result)->elements[i]));
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_choices) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_choices(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_choices" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = Alternative_choices(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->solv->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, 0 | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Alternative_str) {
  {
    Alternative *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Alternative_str(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_str" "', argument " "1"" of type '" "Alternative *""'"); 
    }
    arg1 = (Alternative *)(argp1);
    result = (char *)Alternative___str__(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_pool_get) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Pool *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_pool_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_pool_get" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = (Pool *)(Pool *) ((arg1)->pool);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Transaction) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Transaction(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Transaction" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    delete_Transaction(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_isempty) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    bool result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_isempty(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_isempty" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = (bool)Transaction_isempty(arg1);
    ST(argvi) = SWIG_From_bool  SWIG_PERL_CALL_ARGS_1((bool)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_othersolvable) {
  {
    Transaction *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    XSolvable *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transaction_othersolvable(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_othersolvable" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_othersolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = (XSolvable *)Transaction_othersolvable(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_allothersolvables) {
  {
    Transaction *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transaction_allothersolvables(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_allothersolvables" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_allothersolvables" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    result = Transaction_allothersolvables(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_classify) {
  {
    Transaction *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Transaction_classify(self,mode);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_classify" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transaction_classify" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = Transaction_classify(arg1,arg2);
    {
      int i;
      int cnt = (&result)->count / 4;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 4)
      {
        Id id = *idp;
#define result resultx
        TransactionClass * result = new_TransactionClass(arg1, arg2, id, idp[1], idp[2], idp[3]);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TransactionClass, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_newpackages) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_newpackages(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_newpackages" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = Transaction_newpackages(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_keptpackages) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_keptpackages(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_keptpackages" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = Transaction_keptpackages(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_newsolvables) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_newsolvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_newsolvables" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = Transaction_newsolvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_keptsolvables) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_keptsolvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_keptsolvables" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = Transaction_keptsolvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_steps) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_steps(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_steps" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = Transaction_steps(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_steptype) {
  {
    Transaction *arg1 = 0 ;
    XSolvable *arg2 = 0 ;
    int arg3 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Transaction_steptype(self,s,mode);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_steptype" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_steptype" "', argument " "2"" of type '" "XSolvable *""'"); 
    }
    arg2 = (XSolvable *)(argp2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Transaction_steptype" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
    result = (int)Transaction_steptype(arg1,arg2,arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_calc_installsizechange) {
  {
    Transaction *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    long long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transaction_calc_installsizechange(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_calc_installsizechange" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    result = (long long)Transaction_calc_installsizechange(arg1);
    ST(argvi) = SWIG_From_long_SS_long  SWIG_PERL_CALL_ARGS_1((long long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transaction_order) {
  {
    Transaction *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Transaction_order(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_order" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transaction_order" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    Transaction_order(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_transaction_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Transaction *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_transaction_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_transaction_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (Transaction *)(Transaction *) ((arg1)->transaction);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Transaction, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_type_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_type_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_type_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (Id)(Id) ((arg1)->type);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_fromid_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_fromid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_fromid_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (Id)(Id) ((arg1)->fromid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_toid_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Id result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_toid_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_toid_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (Id)(Id) ((arg1)->toid);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_count_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_count_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_count_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (int)(int) ((arg1)->count);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_TransactionClass) {
  {
    Transaction *arg1 = 0 ;
    int arg2 ;
    Id arg3 ;
    int arg4 ;
    Id arg5 ;
    Id arg6 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int val3 ;
    int ecode3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int val6 ;
    int ecode6 = 0 ;
    int argvi = 0;
    TransactionClass *result = 0 ;
    dXSARGS;
    
    if ((items < 6) || (items > 6)) {
      SWIG_croak("Usage: new_TransactionClass(trans,mode,type,count,fromid,toid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TransactionClass" "', argument " "1"" of type '" "Transaction *""'"); 
    }
    arg1 = (Transaction *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_TransactionClass" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    ecode3 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(2), &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_TransactionClass" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_TransactionClass" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
    ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_TransactionClass" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
    ecode6 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(5), &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_TransactionClass" "', argument " "6"" of type '" "Id""'");
    } 
    arg6 = (Id)(val6);
    result = (TransactionClass *)new_TransactionClass(arg1,arg2,arg3,arg4,arg5,arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TransactionClass, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_solvables) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    Queue result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_solvables(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_solvables" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = TransactionClass_solvables(arg1);
    {
      int i;
      int cnt = (&result)->count / 1;
      Id *idp = (&result)->elements;
      if (argvi + cnt + 1 >= items) {
        EXTEND(sp, (argvi + cnt + 1) - items + 1);
      }
      for (i = 0; i < cnt; i++, idp += 1)
      {
        Id id = *idp;
#define result resultx
        XSolvable * result = new_XSolvable(arg1->transaction->pool, id);
        ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_OWNER | 0); argvi++ ;
        SvREFCNT_inc(ST(argvi - 1));
#undef result
      }
      queue_free(&result);
      ST(argvi) = 0;
    }
    
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_fromstr_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_fromstr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_fromstr_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (char *)TransactionClass_fromstr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_TransactionClass_tostr_get) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: TransactionClass_tostr_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_tostr_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    result = (char *)TransactionClass_tostr_get(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_TransactionClass) {
  {
    TransactionClass *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_TransactionClass(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_TransactionClass, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TransactionClass" "', argument " "1"" of type '" "TransactionClass *""'"); 
    }
    arg1 = (TransactionClass *)(argp1);
    free((char *) arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}



/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_Alternative = {"_p_Alternative", "Alternative *", 0, 0, (void*)"solv::Alternative", 0};
static swig_type_info _swigt__p_Chksum = {"_p_Chksum", "Chksum *", 0, 0, (void*)"solv::Chksum", 0};
static swig_type_info _swigt__p_Dataiterator = {"_p_Dataiterator", "Dataiterator *", 0, 0, (void*)"solv::Dataiterator", 0};
static swig_type_info _swigt__p_Datamatch = {"_p_Datamatch", "Datamatch *", 0, 0, (void*)"solv::Datamatch", 0};
static swig_type_info _swigt__p_Datapos = {"_p_Datapos", "Datapos *", 0, 0, (void*)"solv::Datapos", 0};
static swig_type_info _swigt__p_Decision = {"_p_Decision", "Decision *", 0, 0, (void*)"solv::Decision", 0};
static swig_type_info _swigt__p_Decisionset = {"_p_Decisionset", "Decisionset *", 0, 0, (void*)"solv::Decisionset", 0};
static swig_type_info _swigt__p_Dep = {"_p_Dep", "Dep *", 0, 0, (void*)"solv::Dep", 0};
static swig_type_info _swigt__p_Job = {"_p_Job", "Job *", 0, 0, (void*)"solv::Job", 0};
static swig_type_info _swigt__p_Pool = {"_p_Pool", "Pool *", 0, 0, (void*)"solv::Pool", 0};
static swig_type_info _swigt__p_Pool_repo_iterator = {"_p_Pool_repo_iterator", "Pool_repo_iterator *", 0, 0, (void*)"solv::Pool_repo_iterator", 0};
static swig_type_info _swigt__p_Pool_solvable_iterator = {"_p_Pool_solvable_iterator", "Pool_solvable_iterator *", 0, 0, (void*)"solv::Pool_solvable_iterator", 0};
static swig_type_info _swigt__p_Problem = {"_p_Problem", "Problem *", 0, 0, (void*)"solv::Problem", 0};
static swig_type_info _swigt__p_Queue = {"_p_Queue", "Queue *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Repo = {"_p_Repo", "Repo *", 0, 0, (void*)"solv::Repo", 0};
static swig_type_info _swigt__p_Repo_solvable_iterator = {"_p_Repo_solvable_iterator", "Repo_solvable_iterator *", 0, 0, (void*)"solv::Repo_solvable_iterator", 0};
static swig_type_info _swigt__p_Ruleinfo = {"_p_Ruleinfo", "Ruleinfo *", 0, 0, (void*)"solv::Ruleinfo", 0};
static swig_type_info _swigt__p_Selection = {"_p_Selection", "Selection *", 0, 0, (void*)"solv::Selection", 0};
static swig_type_info _swigt__p_Solution = {"_p_Solution", "Solution *", 0, 0, (void*)"solv::Solution", 0};
static swig_type_info _swigt__p_Solutionelement = {"_p_Solutionelement", "Solutionelement *", 0, 0, (void*)"solv::Solutionelement", 0};
static swig_type_info _swigt__p_SolvFp = {"_p_SolvFp", "SolvFp *", 0, 0, (void*)"solv::SolvFp", 0};
static swig_type_info _swigt__p_Solver = {"_p_Solver", "Solver *", 0, 0, (void*)"solv::Solver", 0};
static swig_type_info _swigt__p_Solvsig = {"_p_Solvsig", "Solvsig *", 0, 0, (void*)"solv::Solvsig", 0};
static swig_type_info _swigt__p_Transaction = {"_p_Transaction", "Transaction *", 0, 0, (void*)"solv::Transaction", 0};
static swig_type_info _swigt__p_TransactionClass = {"_p_TransactionClass", "TransactionClass *", 0, 0, (void*)"solv::TransactionClass", 0};
static swig_type_info _swigt__p_XRepodata = {"_p_XRepodata", "XRepodata *", 0, 0, (void*)"solv::XRepodata", 0};
static swig_type_info _swigt__p_XRule = {"_p_XRule", "XRule *", 0, 0, (void*)"solv::XRule", 0};
static swig_type_info _swigt__p_XSolvable = {"_p_XSolvable", "XSolvable *", 0, 0, (void*)"solv::XSolvable", 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "Id *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_XRule = {"_p_p_XRule", "XRule **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_Alternative,
  &_swigt__p_Chksum,
  &_swigt__p_Dataiterator,
  &_swigt__p_Datamatch,
  &_swigt__p_Datapos,
  &_swigt__p_Decision,
  &_swigt__p_Decisionset,
  &_swigt__p_Dep,
  &_swigt__p_Job,
  &_swigt__p_Pool,
  &_swigt__p_Pool_repo_iterator,
  &_swigt__p_Pool_solvable_iterator,
  &_swigt__p_Problem,
  &_swigt__p_Queue,
  &_swigt__p_Repo,
  &_swigt__p_Repo_solvable_iterator,
  &_swigt__p_Ruleinfo,
  &_swigt__p_Selection,
  &_swigt__p_Solution,
  &_swigt__p_Solutionelement,
  &_swigt__p_SolvFp,
  &_swigt__p_Solver,
  &_swigt__p_Solvsig,
  &_swigt__p_Transaction,
  &_swigt__p_TransactionClass,
  &_swigt__p_XRepodata,
  &_swigt__p_XRule,
  &_swigt__p_XSolvable,
  &_swigt__p_char,
  &_swigt__p_int,
  &_swigt__p_p_XRule,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
};

static swig_cast_info _swigc__p_Alternative[] = {  {&_swigt__p_Alternative, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Chksum[] = {  {&_swigt__p_Chksum, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Dataiterator[] = {  {&_swigt__p_Dataiterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Datamatch[] = {  {&_swigt__p_Datamatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Datapos[] = {  {&_swigt__p_Datapos, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Decision[] = {  {&_swigt__p_Decision, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Decisionset[] = {  {&_swigt__p_Decisionset, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Dep[] = {  {&_swigt__p_Dep, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Job[] = {  {&_swigt__p_Job, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool[] = {  {&_swigt__p_Pool, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool_repo_iterator[] = {  {&_swigt__p_Pool_repo_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool_solvable_iterator[] = {  {&_swigt__p_Pool_solvable_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Problem[] = {  {&_swigt__p_Problem, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Queue[] = {  {&_swigt__p_Queue, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Repo[] = {  {&_swigt__p_Repo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Repo_solvable_iterator[] = {  {&_swigt__p_Repo_solvable_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Ruleinfo[] = {  {&_swigt__p_Ruleinfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Selection[] = {  {&_swigt__p_Selection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solution[] = {  {&_swigt__p_Solution, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solutionelement[] = {  {&_swigt__p_Solutionelement, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SolvFp[] = {  {&_swigt__p_SolvFp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solver[] = {  {&_swigt__p_Solver, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solvsig[] = {  {&_swigt__p_Solvsig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Transaction[] = {  {&_swigt__p_Transaction, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TransactionClass[] = {  {&_swigt__p_TransactionClass, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XRepodata[] = {  {&_swigt__p_XRepodata, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XRule[] = {  {&_swigt__p_XRule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XSolvable[] = {  {&_swigt__p_XSolvable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_XRule[] = {  {&_swigt__p_p_XRule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_Alternative,
  _swigc__p_Chksum,
  _swigc__p_Dataiterator,
  _swigc__p_Datamatch,
  _swigc__p_Datapos,
  _swigc__p_Decision,
  _swigc__p_Decisionset,
  _swigc__p_Dep,
  _swigc__p_Job,
  _swigc__p_Pool,
  _swigc__p_Pool_repo_iterator,
  _swigc__p_Pool_solvable_iterator,
  _swigc__p_Problem,
  _swigc__p_Queue,
  _swigc__p_Repo,
  _swigc__p_Repo_solvable_iterator,
  _swigc__p_Ruleinfo,
  _swigc__p_Selection,
  _swigc__p_Solution,
  _swigc__p_Solutionelement,
  _swigc__p_SolvFp,
  _swigc__p_Solver,
  _swigc__p_Solvsig,
  _swigc__p_Transaction,
  _swigc__p_TransactionClass,
  _swigc__p_XRepodata,
  _swigc__p_XRule,
  _swigc__p_XSolvable,
  _swigc__p_char,
  _swigc__p_int,
  _swigc__p_p_XRule,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_constant_info swig_constants[] = {
{0,0,0,0,0,0}
};
#ifdef __cplusplus
}
#endif
static swig_variable_info swig_variables[] = {
{0,0,0,0}
};
static swig_command_info swig_commands[] = {
{"solvc::Selection_pool_get", _wrap_Selection_pool_get},
{"solvc::Selection_flags_get", _wrap_Selection_flags_get},
{"solvc::new_Selection", _wrap_new_Selection},
{"solvc::delete_Selection", _wrap_delete_Selection},
{"solvc::Selection_isempty", _wrap_Selection_isempty},
{"solvc::Selection_clone", _wrap_Selection_clone},
{"solvc::Selection_filter", _wrap_Selection_filter},
{"solvc::Selection_add", _wrap_Selection_add},
{"solvc::Selection_add_raw", _wrap_Selection_add_raw},
{"solvc::Selection_subtract", _wrap_Selection_subtract},
{"solvc::Selection_select", _wrap_Selection_select},
{"solvc::Selection_matchdeps", _wrap_Selection_matchdeps},
{"solvc::Selection_matchdepid", _wrap_Selection_matchdepid},
{"solvc::Selection_matchsolvable", _wrap_Selection_matchsolvable},
{"solvc::Selection_jobs", _wrap_Selection_jobs},
{"solvc::Selection_solvables", _wrap_Selection_solvables},
{"solvc::Selection_str", _wrap_Selection_str},
{"solvc::Selection_repr", _wrap_Selection_repr},
{"solvc::Dep_pool_get", _wrap_Dep_pool_get},
{"solvc::Dep_id_get", _wrap_Dep_id_get},
{"solvc::new_Dep", _wrap_new_Dep},
{"solvc::Dep_Rel", _wrap_Dep_Rel},
{"solvc::Dep_Selection_name", _wrap_Dep_Selection_name},
{"solvc::Dep_Selection_provides", _wrap_Dep_Selection_provides},
{"solvc::Dep_str", _wrap_Dep_str},
{"solvc::Dep___eq__", _wrap_Dep___eq__},
{"solvc::Dep___ne__", _wrap_Dep___ne__},
{"solvc::Dep_repr", _wrap_Dep_repr},
{"solvc::delete_Dep", _wrap_delete_Dep},
{"solvc::new_Dataiterator", _wrap_new_Dataiterator},
{"solvc::delete_Dataiterator", _wrap_delete_Dataiterator},
{"solvc::Dataiterator___next__", _wrap_Dataiterator___next__},
{"solvc::Dataiterator_prepend_keyname", _wrap_Dataiterator_prepend_keyname},
{"solvc::Dataiterator_skip_solvable", _wrap_Dataiterator_skip_solvable},
{"solvc::XSolvable_pool_get", _wrap_XSolvable_pool_get},
{"solvc::XSolvable_id_get", _wrap_XSolvable_id_get},
{"solvc::new_XSolvable", _wrap_new_XSolvable},
{"solvc::XSolvable_str", _wrap_XSolvable_str},
{"solvc::XSolvable_lookup_str", _wrap_XSolvable_lookup_str},
{"solvc::XSolvable_lookup_id", _wrap_XSolvable_lookup_id},
{"solvc::XSolvable_lookup_num", _wrap_XSolvable_lookup_num},
{"solvc::XSolvable_lookup_void", _wrap_XSolvable_lookup_void},
{"solvc::XSolvable_lookup_checksum", _wrap_XSolvable_lookup_checksum},
{"solvc::XSolvable_lookup_idarray", _wrap_XSolvable_lookup_idarray},
{"solvc::XSolvable_lookup_deparray", _wrap_XSolvable_lookup_deparray},
{"solvc::XSolvable_lookup_location", _wrap_XSolvable_lookup_location},
{"solvc::XSolvable_lookup_sourcepkg", _wrap_XSolvable_lookup_sourcepkg},
{"solvc::XSolvable_Dataiterator", _wrap_XSolvable_Dataiterator},
{"solvc::XSolvable_installable", _wrap_XSolvable_installable},
{"solvc::XSolvable_isinstalled", _wrap_XSolvable_isinstalled},
{"solvc::XSolvable_name_set", _wrap_XSolvable_name_set},
{"solvc::XSolvable_name_get", _wrap_XSolvable_name_get},
{"solvc::XSolvable_nameid_set", _wrap_XSolvable_nameid_set},
{"solvc::XSolvable_nameid_get", _wrap_XSolvable_nameid_get},
{"solvc::XSolvable_evr_set", _wrap_XSolvable_evr_set},
{"solvc::XSolvable_evr_get", _wrap_XSolvable_evr_get},
{"solvc::XSolvable_evrid_set", _wrap_XSolvable_evrid_set},
{"solvc::XSolvable_evrid_get", _wrap_XSolvable_evrid_get},
{"solvc::XSolvable_arch_set", _wrap_XSolvable_arch_set},
{"solvc::XSolvable_arch_get", _wrap_XSolvable_arch_get},
{"solvc::XSolvable_archid_set", _wrap_XSolvable_archid_set},
{"solvc::XSolvable_archid_get", _wrap_XSolvable_archid_get},
{"solvc::XSolvable_vendor_set", _wrap_XSolvable_vendor_set},
{"solvc::XSolvable_vendor_get", _wrap_XSolvable_vendor_get},
{"solvc::XSolvable_vendorid_set", _wrap_XSolvable_vendorid_set},
{"solvc::XSolvable_vendorid_get", _wrap_XSolvable_vendorid_get},
{"solvc::XSolvable_repo_get", _wrap_XSolvable_repo_get},
{"solvc::XSolvable_add_provides", _wrap_XSolvable_add_provides},
{"solvc::XSolvable_add_obsoletes", _wrap_XSolvable_add_obsoletes},
{"solvc::XSolvable_add_conflicts", _wrap_XSolvable_add_conflicts},
{"solvc::XSolvable_add_requires", _wrap_XSolvable_add_requires},
{"solvc::XSolvable_add_recommends", _wrap_XSolvable_add_recommends},
{"solvc::XSolvable_add_suggests", _wrap_XSolvable_add_suggests},
{"solvc::XSolvable_add_supplements", _wrap_XSolvable_add_supplements},
{"solvc::XSolvable_add_enhances", _wrap_XSolvable_add_enhances},
{"solvc::XSolvable_unset", _wrap_XSolvable_unset},
{"solvc::XSolvable_add_deparray", _wrap_XSolvable_add_deparray},
{"solvc::XSolvable_Selection", _wrap_XSolvable_Selection},
{"solvc::XSolvable_identical", _wrap_XSolvable_identical},
{"solvc::XSolvable_evrcmp", _wrap_XSolvable_evrcmp},
{"solvc::XSolvable_matchesdep", _wrap_XSolvable_matchesdep},
{"solvc::XSolvable___eq__", _wrap_XSolvable___eq__},
{"solvc::XSolvable___ne__", _wrap_XSolvable___ne__},
{"solvc::XSolvable_repr", _wrap_XSolvable_repr},
{"solvc::delete_XSolvable", _wrap_delete_XSolvable},
{"solvc::Ruleinfo_solv_get", _wrap_Ruleinfo_solv_get},
{"solvc::Ruleinfo_type_get", _wrap_Ruleinfo_type_get},
{"solvc::Ruleinfo_dep_id_get", _wrap_Ruleinfo_dep_id_get},
{"solvc::new_Ruleinfo", _wrap_new_Ruleinfo},
{"solvc::Ruleinfo_solvable_get", _wrap_Ruleinfo_solvable_get},
{"solvc::Ruleinfo_othersolvable_get", _wrap_Ruleinfo_othersolvable_get},
{"solvc::Ruleinfo_dep_get", _wrap_Ruleinfo_dep_get},
{"solvc::Ruleinfo_problemstr", _wrap_Ruleinfo_problemstr},
{"solvc::Ruleinfo_str", _wrap_Ruleinfo_str},
{"solvc::delete_Ruleinfo", _wrap_delete_Ruleinfo},
{"solvc::XRule_solv_get", _wrap_XRule_solv_get},
{"solvc::XRule_id_get", _wrap_XRule_id_get},
{"solvc::new_XRule", _wrap_new_XRule},
{"solvc::XRule_type_get", _wrap_XRule_type_get},
{"solvc::XRule_info", _wrap_XRule_info},
{"solvc::XRule_allinfos", _wrap_XRule_allinfos},
{"solvc::XRule_get_learnt", _wrap_XRule_get_learnt},
{"solvc::XRule_get_decisionlist", _wrap_XRule_get_decisionlist},
{"solvc::XRule_get_decisionsetlist", _wrap_XRule_get_decisionsetlist},
{"solvc::XRule___eq__", _wrap_XRule___eq__},
{"solvc::XRule___ne__", _wrap_XRule___ne__},
{"solvc::XRule_repr", _wrap_XRule_repr},
{"solvc::delete_XRule", _wrap_delete_XRule},
{"solvc::XRepodata_repo_get", _wrap_XRepodata_repo_get},
{"solvc::XRepodata_id_get", _wrap_XRepodata_id_get},
{"solvc::new_XRepodata", _wrap_new_XRepodata},
{"solvc::XRepodata_new_handle", _wrap_XRepodata_new_handle},
{"solvc::XRepodata_set_id", _wrap_XRepodata_set_id},
{"solvc::XRepodata_set_num", _wrap_XRepodata_set_num},
{"solvc::XRepodata_set_str", _wrap_XRepodata_set_str},
{"solvc::XRepodata_set_void", _wrap_XRepodata_set_void},
{"solvc::XRepodata_set_poolstr", _wrap_XRepodata_set_poolstr},
{"solvc::XRepodata_add_idarray", _wrap_XRepodata_add_idarray},
{"solvc::XRepodata_add_flexarray", _wrap_XRepodata_add_flexarray},
{"solvc::XRepodata_set_checksum", _wrap_XRepodata_set_checksum},
{"solvc::XRepodata_set_sourcepkg", _wrap_XRepodata_set_sourcepkg},
{"solvc::XRepodata_set_location", _wrap_XRepodata_set_location},
{"solvc::XRepodata_unset", _wrap_XRepodata_unset},
{"solvc::XRepodata_lookup_str", _wrap_XRepodata_lookup_str},
{"solvc::XRepodata_lookup_id", _wrap_XRepodata_lookup_id},
{"solvc::XRepodata_lookup_num", _wrap_XRepodata_lookup_num},
{"solvc::XRepodata_lookup_void", _wrap_XRepodata_lookup_void},
{"solvc::XRepodata_lookup_idarray", _wrap_XRepodata_lookup_idarray},
{"solvc::XRepodata_lookup_checksum", _wrap_XRepodata_lookup_checksum},
{"solvc::XRepodata_internalize", _wrap_XRepodata_internalize},
{"solvc::XRepodata_create_stubs", _wrap_XRepodata_create_stubs},
{"solvc::XRepodata_write", _wrap_XRepodata_write},
{"solvc::XRepodata_str2dir", _wrap_XRepodata_str2dir},
{"solvc::XRepodata_dir2str", _wrap_XRepodata_dir2str},
{"solvc::XRepodata_add_dirstr", _wrap_XRepodata_add_dirstr},
{"solvc::XRepodata_add_solv", _wrap_XRepodata_add_solv},
{"solvc::XRepodata_extend_to_repo", _wrap_XRepodata_extend_to_repo},
{"solvc::XRepodata___eq__", _wrap_XRepodata___eq__},
{"solvc::XRepodata___ne__", _wrap_XRepodata___ne__},
{"solvc::XRepodata_repr", _wrap_XRepodata_repr},
{"solvc::delete_XRepodata", _wrap_delete_XRepodata},
{"solvc::new_Pool_solvable_iterator", _wrap_new_Pool_solvable_iterator},
{"solvc::Pool_solvable_iterator___next__", _wrap_Pool_solvable_iterator___next__},
{"solvc::Pool_solvable_iterator___getitem__", _wrap_Pool_solvable_iterator___getitem__},
{"solvc::Pool_solvable_iterator___len__", _wrap_Pool_solvable_iterator___len__},
{"solvc::delete_Pool_solvable_iterator", _wrap_delete_Pool_solvable_iterator},
{"solvc::new_Pool_repo_iterator", _wrap_new_Pool_repo_iterator},
{"solvc::Pool_repo_iterator___next__", _wrap_Pool_repo_iterator___next__},
{"solvc::Pool_repo_iterator___getitem__", _wrap_Pool_repo_iterator___getitem__},
{"solvc::Pool_repo_iterator___len__", _wrap_Pool_repo_iterator___len__},
{"solvc::delete_Pool_repo_iterator", _wrap_delete_Pool_repo_iterator},
{"solvc::new_Repo_solvable_iterator", _wrap_new_Repo_solvable_iterator},
{"solvc::Repo_solvable_iterator___next__", _wrap_Repo_solvable_iterator___next__},
{"solvc::Repo_solvable_iterator___getitem__", _wrap_Repo_solvable_iterator___getitem__},
{"solvc::Repo_solvable_iterator___len__", _wrap_Repo_solvable_iterator___len__},
{"solvc::delete_Repo_solvable_iterator", _wrap_delete_Repo_solvable_iterator},
{"solvc::Datamatch_pool_get", _wrap_Datamatch_pool_get},
{"solvc::Datamatch_repo_get", _wrap_Datamatch_repo_get},
{"solvc::Datamatch_solvid_get", _wrap_Datamatch_solvid_get},
{"solvc::delete_Datamatch", _wrap_delete_Datamatch},
{"solvc::Datamatch_solvable_get", _wrap_Datamatch_solvable_get},
{"solvc::Datamatch_key_id_get", _wrap_Datamatch_key_id_get},
{"solvc::Datamatch_key_idstr_get", _wrap_Datamatch_key_idstr_get},
{"solvc::Datamatch_type_id_get", _wrap_Datamatch_type_id_get},
{"solvc::Datamatch_type_idstr_get", _wrap_Datamatch_type_idstr_get},
{"solvc::Datamatch_id_get", _wrap_Datamatch_id_get},
{"solvc::Datamatch_idstr_get", _wrap_Datamatch_idstr_get},
{"solvc::Datamatch_str_get", _wrap_Datamatch_str_get},
{"solvc::Datamatch_dep_get", _wrap_Datamatch_dep_get},
{"solvc::Datamatch_binary_get", _wrap_Datamatch_binary_get},
{"solvc::Datamatch_num_get", _wrap_Datamatch_num_get},
{"solvc::Datamatch_num2_get", _wrap_Datamatch_num2_get},
{"solvc::Datamatch_pos", _wrap_Datamatch_pos},
{"solvc::Datamatch_parentpos", _wrap_Datamatch_parentpos},
{"solvc::Datamatch_stringify", _wrap_Datamatch_stringify},
{"solvc::Datapos_repo_get", _wrap_Datapos_repo_get},
{"solvc::Datapos_lookup_id", _wrap_Datapos_lookup_id},
{"solvc::Datapos_lookup_str", _wrap_Datapos_lookup_str},
{"solvc::Datapos_lookup_num", _wrap_Datapos_lookup_num},
{"solvc::Datapos_lookup_void", _wrap_Datapos_lookup_void},
{"solvc::Datapos_lookup_checksum", _wrap_Datapos_lookup_checksum},
{"solvc::Datapos_lookup_deltaseq", _wrap_Datapos_lookup_deltaseq},
{"solvc::Datapos_lookup_deltalocation", _wrap_Datapos_lookup_deltalocation},
{"solvc::Datapos_lookup_idarray", _wrap_Datapos_lookup_idarray},
{"solvc::Datapos_Dataiterator", _wrap_Datapos_Dataiterator},
{"solvc::delete_Datapos", _wrap_delete_Datapos},
{"solvc::Job_pool_get", _wrap_Job_pool_get},
{"solvc::Job_how_set", _wrap_Job_how_set},
{"solvc::Job_how_get", _wrap_Job_how_get},
{"solvc::Job_what_set", _wrap_Job_what_set},
{"solvc::Job_what_get", _wrap_Job_what_get},
{"solvc::new_Job", _wrap_new_Job},
{"solvc::Job_solvables", _wrap_Job_solvables},
{"solvc::Job_isemptyupdate", _wrap_Job_isemptyupdate},
{"solvc::Job___eq__", _wrap_Job___eq__},
{"solvc::Job___ne__", _wrap_Job___ne__},
{"solvc::Job_str", _wrap_Job_str},
{"solvc::Job_repr", _wrap_Job_repr},
{"solvc::delete_Job", _wrap_delete_Job},
{"solvc::new_Pool", _wrap_new_Pool},
{"solvc::Pool_setdisttype", _wrap_Pool_setdisttype},
{"solvc::Pool_set_debuglevel", _wrap_Pool_set_debuglevel},
{"solvc::Pool_set_flag", _wrap_Pool_set_flag},
{"solvc::Pool_get_flag", _wrap_Pool_get_flag},
{"solvc::Pool_set_rootdir", _wrap_Pool_set_rootdir},
{"solvc::Pool_get_rootdir", _wrap_Pool_get_rootdir},
{"solvc::Pool_clr_loadcallback", _wrap_Pool_clr_loadcallback},
{"solvc::Pool_set_loadcallback", _wrap_Pool_set_loadcallback},
{"solvc::delete_Pool", _wrap_delete_Pool},
{"solvc::Pool_free", _wrap_Pool_free},
{"solvc::Pool_disown", _wrap_Pool_disown},
{"solvc::Pool_appdata_set", _wrap_Pool_appdata_set},
{"solvc::Pool_appdata_get", _wrap_Pool_appdata_get},
{"solvc::Pool_appdata_disown", _wrap_Pool_appdata_disown},
{"solvc::Pool_str2id", _wrap_Pool_str2id},
{"solvc::Pool_Dep", _wrap_Pool_Dep},
{"solvc::Pool_parserpmrichdep", _wrap_Pool_parserpmrichdep},
{"solvc::Pool_id2str", _wrap_Pool_id2str},
{"solvc::Pool_dep2str", _wrap_Pool_dep2str},
{"solvc::Pool_rel2id", _wrap_Pool_rel2id},
{"solvc::Pool_id2langid", _wrap_Pool_id2langid},
{"solvc::Pool_setarch", _wrap_Pool_setarch},
{"solvc::Pool_add_repo", _wrap_Pool_add_repo},
{"solvc::Pool_lookup_str", _wrap_Pool_lookup_str},
{"solvc::Pool_lookup_id", _wrap_Pool_lookup_id},
{"solvc::Pool_lookup_num", _wrap_Pool_lookup_num},
{"solvc::Pool_lookup_void", _wrap_Pool_lookup_void},
{"solvc::Pool_lookup_checksum", _wrap_Pool_lookup_checksum},
{"solvc::Pool_Dataiterator", _wrap_Pool_Dataiterator},
{"solvc::Pool_Dataiterator_solvid", _wrap_Pool_Dataiterator_solvid},
{"solvc::Pool_solvid2str", _wrap_Pool_solvid2str},
{"solvc::Pool_solvidset2str", _wrap_Pool_solvidset2str},
{"solvc::Pool_solvableset2str", _wrap_Pool_solvableset2str},
{"solvc::Pool_addfileprovides", _wrap_Pool_addfileprovides},
{"solvc::Pool_addfileprovides_queue", _wrap_Pool_addfileprovides_queue},
{"solvc::Pool_createwhatprovides", _wrap_Pool_createwhatprovides},
{"solvc::Pool_id2solvable", _wrap_Pool_id2solvable},
{"solvc::Pool_solvables_get", _wrap_Pool_solvables_get},
{"solvc::Pool_solvables_iter", _wrap_Pool_solvables_iter},
{"solvc::Pool_id2repo", _wrap_Pool_id2repo},
{"solvc::Pool_repos_get", _wrap_Pool_repos_get},
{"solvc::Pool_repos_iter", _wrap_Pool_repos_iter},
{"solvc::Pool_installed_set", _wrap_Pool_installed_set},
{"solvc::Pool_installed_get", _wrap_Pool_installed_get},
{"solvc::Pool_errstr_get", _wrap_Pool_errstr_get},
{"solvc::Pool_matchprovidingids", _wrap_Pool_matchprovidingids},
{"solvc::Pool_Job", _wrap_Pool_Job},
{"solvc::Pool_whatprovides", _wrap_Pool_whatprovides},
{"solvc::Pool_best_solvables", _wrap_Pool_best_solvables},
{"solvc::Pool_towhatprovides", _wrap_Pool_towhatprovides},
{"solvc::Pool_set_namespaceproviders", _wrap_Pool_set_namespaceproviders},
{"solvc::Pool_flush_namespaceproviders", _wrap_Pool_flush_namespaceproviders},
{"solvc::Pool_whatcontainsdep", _wrap_Pool_whatcontainsdep},
{"solvc::Pool_whatmatchesdep", _wrap_Pool_whatmatchesdep},
{"solvc::Pool_whatmatchessolvable", _wrap_Pool_whatmatchessolvable},
{"solvc::Pool_isknownarch", _wrap_Pool_isknownarch},
{"solvc::Pool_Solver", _wrap_Pool_Solver},
{"solvc::Pool_Selection", _wrap_Pool_Selection},
{"solvc::Pool_Selection_all", _wrap_Pool_Selection_all},
{"solvc::Pool_select", _wrap_Pool_select},
{"solvc::Pool_matchdeps", _wrap_Pool_matchdeps},
{"solvc::Pool_matchdepid", _wrap_Pool_matchdepid},
{"solvc::Pool_matchsolvable", _wrap_Pool_matchsolvable},
{"solvc::Pool_get_considered_list", _wrap_Pool_get_considered_list},
{"solvc::Pool_get_disabled_list", _wrap_Pool_get_disabled_list},
{"solvc::Pool_set_considered_list", _wrap_Pool_set_considered_list},
{"solvc::Pool_set_disabled_list", _wrap_Pool_set_disabled_list},
{"solvc::Pool_setpooljobs", _wrap_Pool_setpooljobs},
{"solvc::Pool_getpooljobs", _wrap_Pool_getpooljobs},
{"solvc::Repo_pool_get", _wrap_Repo_pool_get},
{"solvc::Repo_name_get", _wrap_Repo_name_get},
{"solvc::Repo_priority_set", _wrap_Repo_priority_set},
{"solvc::Repo_priority_get", _wrap_Repo_priority_get},
{"solvc::Repo_subpriority_set", _wrap_Repo_subpriority_set},
{"solvc::Repo_subpriority_get", _wrap_Repo_subpriority_get},
{"solvc::Repo_nsolvables_get", _wrap_Repo_nsolvables_get},
{"solvc::Repo_free", _wrap_Repo_free},
{"solvc::Repo_empty", _wrap_Repo_empty},
{"solvc::Repo_isempty", _wrap_Repo_isempty},
{"solvc::Repo_appdata_set", _wrap_Repo_appdata_set},
{"solvc::Repo_appdata_get", _wrap_Repo_appdata_get},
{"solvc::Repo_add_solv", _wrap_Repo_add_solv},
{"solvc::Repo_add_solvable", _wrap_Repo_add_solvable},
{"solvc::Repo_add_rpmdb", _wrap_Repo_add_rpmdb},
{"solvc::Repo_add_rpmdb_reffp", _wrap_Repo_add_rpmdb_reffp},
{"solvc::Repo_add_rpm", _wrap_Repo_add_rpm},
{"solvc::Repo_add_rpmdb_pubkeys", _wrap_Repo_add_rpmdb_pubkeys},
{"solvc::Repo_add_pubkey", _wrap_Repo_add_pubkey},
{"solvc::Repo_add_keyring", _wrap_Repo_add_keyring},
{"solvc::Repo_add_keydir", _wrap_Repo_add_keydir},
{"solvc::Repo_add_rpmmd", _wrap_Repo_add_rpmmd},
{"solvc::Repo_add_repomdxml", _wrap_Repo_add_repomdxml},
{"solvc::Repo_add_updateinfoxml", _wrap_Repo_add_updateinfoxml},
{"solvc::Repo_add_deltainfoxml", _wrap_Repo_add_deltainfoxml},
{"solvc::Repo_add_debdb", _wrap_Repo_add_debdb},
{"solvc::Repo_add_debpackages", _wrap_Repo_add_debpackages},
{"solvc::Repo_add_deb", _wrap_Repo_add_deb},
{"solvc::Repo_add_susetags", _wrap_Repo_add_susetags},
{"solvc::Repo_add_content", _wrap_Repo_add_content},
{"solvc::Repo_add_products", _wrap_Repo_add_products},
{"solvc::Repo_add_mdk", _wrap_Repo_add_mdk},
{"solvc::Repo_add_mdk_info", _wrap_Repo_add_mdk_info},
{"solvc::Repo_add_arch_repo", _wrap_Repo_add_arch_repo},
{"solvc::Repo_add_arch_local", _wrap_Repo_add_arch_local},
{"solvc::Repo_add_arch_pkg", _wrap_Repo_add_arch_pkg},
{"solvc::Repo_internalize", _wrap_Repo_internalize},
{"solvc::Repo_write", _wrap_Repo_write},
{"solvc::Repo_write_first_repodata", _wrap_Repo_write_first_repodata},
{"solvc::Repo_Dataiterator", _wrap_Repo_Dataiterator},
{"solvc::Repo_Dataiterator_meta", _wrap_Repo_Dataiterator_meta},
{"solvc::Repo_id_get", _wrap_Repo_id_get},
{"solvc::Repo_solvables_get", _wrap_Repo_solvables_get},
{"solvc::Repo_meta_get", _wrap_Repo_meta_get},
{"solvc::Repo_solvables_iter", _wrap_Repo_solvables_iter},
{"solvc::Repo_add_repodata", _wrap_Repo_add_repodata},
{"solvc::Repo_create_stubs", _wrap_Repo_create_stubs},
{"solvc::Repo_iscontiguous", _wrap_Repo_iscontiguous},
{"solvc::Repo_first_repodata", _wrap_Repo_first_repodata},
{"solvc::Repo_Selection", _wrap_Repo_Selection},
{"solvc::Repo_find_pubkey", _wrap_Repo_find_pubkey},
{"solvc::Repo_createshadow", _wrap_Repo_createshadow},
{"solvc::Repo_moveshadow", _wrap_Repo_moveshadow},
{"solvc::Repo___eq__", _wrap_Repo___eq__},
{"solvc::Repo___ne__", _wrap_Repo___ne__},
{"solvc::Repo_str", _wrap_Repo_str},
{"solvc::Repo_repr", _wrap_Repo_repr},
{"solvc::Decision_solv_get", _wrap_Decision_solv_get},
{"solvc::Decision_p_get", _wrap_Decision_p_get},
{"solvc::Decision_reason_get", _wrap_Decision_reason_get},
{"solvc::Decision_infoid_get", _wrap_Decision_infoid_get},
{"solvc::new_Decision", _wrap_new_Decision},
{"solvc::Decision_rule_get", _wrap_Decision_rule_get},
{"solvc::Decision_solvable_get", _wrap_Decision_solvable_get},
{"solvc::Decision_info", _wrap_Decision_info},
{"solvc::Decision_allinfos", _wrap_Decision_allinfos},
{"solvc::Decision_reasonstr", _wrap_Decision_reasonstr},
{"solvc::Decision_str", _wrap_Decision_str},
{"solvc::delete_Decision", _wrap_delete_Decision},
{"solvc::Decisionset_solv_get", _wrap_Decisionset_solv_get},
{"solvc::Decisionset_p_get", _wrap_Decisionset_p_get},
{"solvc::Decisionset_reason_get", _wrap_Decisionset_reason_get},
{"solvc::Decisionset_infoid_get", _wrap_Decisionset_infoid_get},
{"solvc::Decisionset_bits_get", _wrap_Decisionset_bits_get},
{"solvc::Decisionset_type_get", _wrap_Decisionset_type_get},
{"solvc::Decisionset_dep_id_get", _wrap_Decisionset_dep_id_get},
{"solvc::new_Decisionset", _wrap_new_Decisionset},
{"solvc::delete_Decisionset", _wrap_delete_Decisionset},
{"solvc::Decisionset_info", _wrap_Decisionset_info},
{"solvc::Decisionset_dep_get", _wrap_Decisionset_dep_get},
{"solvc::Decisionset_solvables", _wrap_Decisionset_solvables},
{"solvc::Decisionset_decisions", _wrap_Decisionset_decisions},
{"solvc::Decisionset_reasonstr", _wrap_Decisionset_reasonstr},
{"solvc::Decisionset_str", _wrap_Decisionset_str},
{"solvc::Solver_pool_get", _wrap_Solver_pool_get},
{"solvc::delete_Solver", _wrap_delete_Solver},
{"solvc::Solver_set_flag", _wrap_Solver_set_flag},
{"solvc::Solver_get_flag", _wrap_Solver_get_flag},
{"solvc::Solver_solve", _wrap_Solver_solve},
{"solvc::Solver_transaction", _wrap_Solver_transaction},
{"solvc::Solver_describe_decision", _wrap_Solver_describe_decision},
{"solvc::Solver_describe_weakdep_decision_raw", _wrap_Solver_describe_weakdep_decision_raw},
{"solvc::Solver_alternatives_count", _wrap_Solver_alternatives_count},
{"solvc::Solver_get_alternative", _wrap_Solver_get_alternative},
{"solvc::Solver_alternatives", _wrap_Solver_alternatives},
{"solvc::Solver_write_testcase", _wrap_Solver_write_testcase},
{"solvc::Solver_raw_decisions", _wrap_Solver_raw_decisions},
{"solvc::Solver_all_decisions", _wrap_Solver_all_decisions},
{"solvc::Solver_get_decision", _wrap_Solver_get_decision},
{"solvc::Solver_get_learnt", _wrap_Solver_get_learnt},
{"solvc::Solver_get_decisionlist", _wrap_Solver_get_decisionlist},
{"solvc::Solver_get_recommended", _wrap_Solver_get_recommended},
{"solvc::Solver_get_suggested", _wrap_Solver_get_suggested},
{"solvc::new_Chksum", _wrap_new_Chksum},
{"solvc::Chksum_from_bin", _wrap_Chksum_from_bin},
{"solvc::delete_Chksum", _wrap_delete_Chksum},
{"solvc::Chksum_type_get", _wrap_Chksum_type_get},
{"solvc::Chksum_add", _wrap_Chksum_add},
{"solvc::Chksum_add_fp", _wrap_Chksum_add_fp},
{"solvc::Chksum_add_fd", _wrap_Chksum_add_fd},
{"solvc::Chksum_add_stat", _wrap_Chksum_add_stat},
{"solvc::Chksum_add_fstat", _wrap_Chksum_add_fstat},
{"solvc::Chksum_raw", _wrap_Chksum_raw},
{"solvc::Chksum_hex", _wrap_Chksum_hex},
{"solvc::Chksum_typestr", _wrap_Chksum_typestr},
{"solvc::Chksum___eq__", _wrap_Chksum___eq__},
{"solvc::Chksum___ne__", _wrap_Chksum___ne__},
{"solvc::Chksum_str", _wrap_Chksum_str},
{"solvc::Chksum_repr", _wrap_Chksum_repr},
{"solvc::Solvsig_htype_get", _wrap_Solvsig_htype_get},
{"solvc::Solvsig_created_get", _wrap_Solvsig_created_get},
{"solvc::Solvsig_expires_get", _wrap_Solvsig_expires_get},
{"solvc::Solvsig_keyid_get", _wrap_Solvsig_keyid_get},
{"solvc::new_Solvsig", _wrap_new_Solvsig},
{"solvc::delete_Solvsig", _wrap_delete_Solvsig},
{"solvc::Solvsig_Chksum", _wrap_Solvsig_Chksum},
{"solvc::Solvsig_verify", _wrap_Solvsig_verify},
{"solvc::delete_SolvFp", _wrap_delete_SolvFp},
{"solvc::SolvFp_fileno", _wrap_SolvFp_fileno},
{"solvc::SolvFp_dup", _wrap_SolvFp_dup},
{"solvc::SolvFp_write", _wrap_SolvFp_write},
{"solvc::SolvFp_flush", _wrap_SolvFp_flush},
{"solvc::SolvFp_close", _wrap_SolvFp_close},
{"solvc::SolvFp_cloexec", _wrap_SolvFp_cloexec},
{"solvc::xfopen", _wrap_xfopen},
{"solvc::xfopen_fd", _wrap_xfopen_fd},
{"solvc::Problem_solv_get", _wrap_Problem_solv_get},
{"solvc::Problem_id_get", _wrap_Problem_id_get},
{"solvc::new_Problem", _wrap_new_Problem},
{"solvc::Problem_findproblemrule", _wrap_Problem_findproblemrule},
{"solvc::Problem_findallproblemrules", _wrap_Problem_findallproblemrules},
{"solvc::Problem_solution_count", _wrap_Problem_solution_count},
{"solvc::Problem_solutions", _wrap_Problem_solutions},
{"solvc::Problem_get_learnt", _wrap_Problem_get_learnt},
{"solvc::Problem_get_decisionlist", _wrap_Problem_get_decisionlist},
{"solvc::Problem_get_decisionsetlist", _wrap_Problem_get_decisionsetlist},
{"solvc::Problem_str", _wrap_Problem_str},
{"solvc::delete_Problem", _wrap_delete_Problem},
{"solvc::Solution_solv_get", _wrap_Solution_solv_get},
{"solvc::Solution_problemid_get", _wrap_Solution_problemid_get},
{"solvc::Solution_id_get", _wrap_Solution_id_get},
{"solvc::new_Solution", _wrap_new_Solution},
{"solvc::Solution_element_count", _wrap_Solution_element_count},
{"solvc::Solution_elements", _wrap_Solution_elements},
{"solvc::delete_Solution", _wrap_delete_Solution},
{"solvc::Solutionelement_solv_get", _wrap_Solutionelement_solv_get},
{"solvc::Solutionelement_problemid_get", _wrap_Solutionelement_problemid_get},
{"solvc::Solutionelement_solutionid_get", _wrap_Solutionelement_solutionid_get},
{"solvc::Solutionelement_type_get", _wrap_Solutionelement_type_get},
{"solvc::new_Solutionelement", _wrap_new_Solutionelement},
{"solvc::Solutionelement_str", _wrap_Solutionelement_str},
{"solvc::Solutionelement_replaceelements", _wrap_Solutionelement_replaceelements},
{"solvc::Solutionelement_illegalreplace", _wrap_Solutionelement_illegalreplace},
{"solvc::Solutionelement_solvable_get", _wrap_Solutionelement_solvable_get},
{"solvc::Solutionelement_replacement_get", _wrap_Solutionelement_replacement_get},
{"solvc::Solutionelement_jobidx_get", _wrap_Solutionelement_jobidx_get},
{"solvc::Solutionelement_Job", _wrap_Solutionelement_Job},
{"solvc::delete_Solutionelement", _wrap_delete_Solutionelement},
{"solvc::Alternative_solv_get", _wrap_Alternative_solv_get},
{"solvc::Alternative_type_get", _wrap_Alternative_type_get},
{"solvc::Alternative_dep_id_get", _wrap_Alternative_dep_id_get},
{"solvc::Alternative_chosen_id_get", _wrap_Alternative_chosen_id_get},
{"solvc::Alternative_level_get", _wrap_Alternative_level_get},
{"solvc::delete_Alternative", _wrap_delete_Alternative},
{"solvc::Alternative_chosen_get", _wrap_Alternative_chosen_get},
{"solvc::Alternative_rule_get", _wrap_Alternative_rule_get},
{"solvc::Alternative_depsolvable_get", _wrap_Alternative_depsolvable_get},
{"solvc::Alternative_dep_get", _wrap_Alternative_dep_get},
{"solvc::Alternative_choices_raw", _wrap_Alternative_choices_raw},
{"solvc::Alternative_choices", _wrap_Alternative_choices},
{"solvc::Alternative_str", _wrap_Alternative_str},
{"solvc::Transaction_pool_get", _wrap_Transaction_pool_get},
{"solvc::delete_Transaction", _wrap_delete_Transaction},
{"solvc::Transaction_isempty", _wrap_Transaction_isempty},
{"solvc::Transaction_othersolvable", _wrap_Transaction_othersolvable},
{"solvc::Transaction_allothersolvables", _wrap_Transaction_allothersolvables},
{"solvc::Transaction_classify", _wrap_Transaction_classify},
{"solvc::Transaction_newpackages", _wrap_Transaction_newpackages},
{"solvc::Transaction_keptpackages", _wrap_Transaction_keptpackages},
{"solvc::Transaction_newsolvables", _wrap_Transaction_newsolvables},
{"solvc::Transaction_keptsolvables", _wrap_Transaction_keptsolvables},
{"solvc::Transaction_steps", _wrap_Transaction_steps},
{"solvc::Transaction_steptype", _wrap_Transaction_steptype},
{"solvc::Transaction_calc_installsizechange", _wrap_Transaction_calc_installsizechange},
{"solvc::Transaction_order", _wrap_Transaction_order},
{"solvc::TransactionClass_transaction_get", _wrap_TransactionClass_transaction_get},
{"solvc::TransactionClass_type_get", _wrap_TransactionClass_type_get},
{"solvc::TransactionClass_fromid_get", _wrap_TransactionClass_fromid_get},
{"solvc::TransactionClass_toid_get", _wrap_TransactionClass_toid_get},
{"solvc::TransactionClass_count_get", _wrap_TransactionClass_count_get},
{"solvc::new_TransactionClass", _wrap_new_TransactionClass},
{"solvc::TransactionClass_solvables", _wrap_TransactionClass_solvables},
{"solvc::TransactionClass_fromstr_get", _wrap_TransactionClass_fromstr_get},
{"solvc::TransactionClass_tostr_get", _wrap_TransactionClass_tostr_get},
{"solvc::delete_TransactionClass", _wrap_delete_TransactionClass},
{0,0}
};
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#if defined(__cplusplus) && ! defined(XSPROTO)
extern "C"
#endif

XS(SWIG_init) {
  dXSARGS;
  int i;
  (void)items;
  
  SWIG_InitializeModule(0);
  
  /* Install commands */
  for (i = 0; swig_commands[i].name; i++) {
    /* Casts only needed for Perl < 5.10. */
#ifdef __cplusplus
    newXS(const_cast<char*>(swig_commands[i].name), swig_commands[i].wrapper, const_cast<char*>(__FILE__));
#else
    newXS((char*)swig_commands[i].name, swig_commands[i].wrapper, (char*)__FILE__);
#endif
  }
  
  /* Install variables */
  for (i = 0; swig_variables[i].name; i++) {
    SV *sv;
    sv = get_sv(swig_variables[i].name, TRUE | 0x2 | GV_ADDMULTI);
    if (swig_variables[i].type) {
      SWIG_MakePtr(sv,(void *)1, *swig_variables[i].type,0);
    } else {
      sv_setiv(sv,(IV) 0);
    }
    swig_create_magic(sv, swig_variables[i].name, swig_variables[i].set, swig_variables[i].get); 
  }
  
  /* Install constant */
  for (i = 0; swig_constants[i].type; i++) {
    SV *sv;
    sv = get_sv(swig_constants[i].name, TRUE | 0x2 | GV_ADDMULTI);
    switch(swig_constants[i].type) {
    case SWIG_INT:
      sv_setiv(sv, (IV) swig_constants[i].lvalue);
      break;
    case SWIG_FLOAT:
      sv_setnv(sv, (double) swig_constants[i].dvalue);
      break;
    case SWIG_STRING:
      sv_setpv(sv, (const char *) swig_constants[i].pvalue);
      break;
    case SWIG_POINTER:
      SWIG_MakePtr(sv, swig_constants[i].pvalue, *(swig_constants[i].ptype),0);
      break;
    case SWIG_BINARY:
      SWIG_MakePackedObj(sv, swig_constants[i].pvalue, swig_constants[i].lvalue, *(swig_constants[i].ptype));
      break;
    default:
      break;
    }
    SvREADONLY_on(sv);
  }
  
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ID_NULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ID_NULL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ID_EMPTY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ID_EMPTY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_VENDOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_VENDOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CONFLICTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CONFLICTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_REQUIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_REQUIRES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_RECOMMENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_RECOMMENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SUGGESTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SUGGESTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SUPPLEMENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SUPPLEMENTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ENHANCES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ENHANCES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "RPM_RPMDBID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(RPM_RPMDBID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PREREQMARKER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PREREQMARKER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_FILEMARKER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_FILEMARKER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_INSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_INSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_MODALIAS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_MODALIAS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_SPLITPROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_SPLITPROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_LANGUAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_LANGUAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_FILESYSTEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_FILESYSTEM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NAMESPACE_OTHERPROVIDERS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(NAMESPACE_OTHERPROVIDERS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SYSTEM_SYSTEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SYSTEM_SYSTEM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ARCH_SRC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ARCH_SRC)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ARCH_NOSRC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ARCH_NOSRC)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ARCH_NOARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ARCH_NOARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ARCH_ALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ARCH_ALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ARCH_ANY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ARCH_ANY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_SOLVABLES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_SOLVABLES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_DELTAINFO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_DELTAINFO)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_EXTERNAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_EXTERNAL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_KEYS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_KEYS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_LOCATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_LOCATION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_VOID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_VOID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_CONSTANT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_CONSTANT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_CONSTANTID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_CONSTANTID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_ID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_ID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_NUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_NUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_DIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_DIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_STR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_BINARY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_BINARY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_IDARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_IDARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_REL_IDARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_REL_IDARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_DIRSTRARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_DIRSTRARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_DIRNUMNUMARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_DIRNUMNUMARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_MD5", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_MD5)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_SHA1", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_SHA1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_SHA224", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_SHA224)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_SHA256", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_SHA256)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_SHA384", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_SHA384)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_SHA512", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_SHA512)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_FIXARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_FIXARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_FLEXARRAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_FLEXARRAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOKEY_TYPE_DELETED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOKEY_TYPE_DELETED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SUMMARY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SUMMARY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_DESCRIPTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_DESCRIPTION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_DISTRIBUTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_DISTRIBUTION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_AUTHORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_AUTHORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PACKAGER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PACKAGER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_GROUP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_GROUP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_URL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_URL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_KEYWORDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_KEYWORDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_LICENSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_LICENSE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_BUILDTIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_BUILDTIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_BUILDHOST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_BUILDHOST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_EULA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_EULA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CPEID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CPEID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MESSAGEINS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MESSAGEINS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MESSAGEDEL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MESSAGEDEL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_INSTALLSIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_INSTALLSIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_DISKUSAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_DISKUSAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_FILELIST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_FILELIST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_INSTALLTIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_INSTALLTIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MEDIADIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MEDIADIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MEDIAFILE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MEDIAFILE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MEDIANR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MEDIANR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MEDIABASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MEDIABASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_DOWNLOADSIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_DOWNLOADSIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SOURCEARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SOURCEARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SOURCENAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SOURCENAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SOURCEEVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SOURCEEVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ISVISIBLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ISVISIBLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_TRIGGERS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_TRIGGERS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CHECKSUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CHECKSUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PKGID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PKGID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_HDRID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_HDRID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_LEADSIGID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_LEADSIGID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PATCHCATEGORY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PATCHCATEGORY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_HEADEREND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_HEADEREND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CHANGELOG", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CHANGELOG)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CHANGELOG_AUTHOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CHANGELOG_AUTHOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CHANGELOG_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CHANGELOG_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CHANGELOG_TEXT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CHANGELOG_TEXT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_INSTALLSTATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_INSTALLSTATUS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_PREREQ_IGNOREINST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_PREREQ_IGNOREINST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CATEGORY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CATEGORY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_INCLUDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_INCLUDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_EXTENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_EXTENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ICON", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ICON)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ORDER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ORDER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REBOOT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REBOOT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_RESTART", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_RESTART)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_RELOGIN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_RELOGIN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MESSAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MESSAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_SEVERITY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_SEVERITY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_RIGHTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_RIGHTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION_FILENAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION_FILENAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTION_FLAGS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTION_FLAGS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REFERENCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REFERENCE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REFERENCE_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REFERENCE_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REFERENCE_HREF", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REFERENCE_HREF)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REFERENCE_ID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REFERENCE_ID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_REFERENCE_TITLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_REFERENCE_TITLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_REFERENCEFILE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_REFERENCEFILE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_SHORTLABEL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_SHORTLABEL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_DISTPRODUCT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_DISTPRODUCT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_DISTVERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_DISTVERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_URL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_URL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_URL_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_URL_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_FLAGS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_FLAGS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_PRODUCTLINE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_PRODUCTLINE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_REGISTER_TARGET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_REGISTER_TARGET)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_REGISTER_FLAVOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_REGISTER_FLAVOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_REGISTER_RELEASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_REGISTER_RELEASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_UPDATES_REPOID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_UPDATES_REPOID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_UPDATES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_UPDATES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PRODUCT_ENDOFLIFE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PRODUCT_ENDOFLIFE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_DATADIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_DATADIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_DESCRDIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_DESCRDIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_DEFAULTVENDOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_DEFAULTVENDOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_FILE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_FILE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_FILE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_FILE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_FILE_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_FILE_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_FILE_CHECKSUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_FILE_CHECKSUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_SHARE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_SHARE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_SHARE_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_SHARE_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SUSETAGS_SHARE_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_SHARE_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_ADDEDFILEPROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_ADDEDFILEPROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_RPMDBCOOKIE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_RPMDBCOOKIE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_FILTEREDFILELIST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_FILTEREDFILELIST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_TIMESTAMP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_TIMESTAMP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_EXPIRE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_EXPIRE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_UPDATES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_UPDATES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_DISTROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_DISTROS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_PRODUCT_LABEL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_PRODUCT_LABEL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_PRODUCT_CPEID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_PRODUCT_CPEID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_KEYWORDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_KEYWORDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REVISION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REVISION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_TOOLVERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_TOOLVERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_PACKAGE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_PACKAGE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_PACKAGE_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_PACKAGE_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_PACKAGE_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_PACKAGE_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_LOCATION_DIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_LOCATION_DIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_LOCATION_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_LOCATION_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_LOCATION_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_LOCATION_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_LOCATION_SUFFIX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_LOCATION_SUFFIX)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_DOWNLOADSIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_DOWNLOADSIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_CHECKSUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_CHECKSUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_BASE_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_BASE_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_SEQ_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_SEQ_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_SEQ_EVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_SEQ_EVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_SEQ_NUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_SEQ_NUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DELTA_LOCATION_BASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DELTA_LOCATION_BASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_LOCATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_LOCATION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_TIMESTAMP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_TIMESTAMP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_CHECKSUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_CHECKSUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_OPENCHECKSUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_OPENCHECKSUM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REPOSITORY_REPOMD_SIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPOSITORY_REPOMD_SIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_KEYID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_KEYID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_FINGERPRINT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_FINGERPRINT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_EXPIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_EXPIRES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_SIGNATURES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_SIGNATURES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_DATA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_DATA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PUBKEY_SUBKEYOF", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(PUBKEY_SUBKEYOF)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SIGNATURE_ISSUER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SIGNATURE_ISSUER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SIGNATURE_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SIGNATURE_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SIGNATURE_EXPIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SIGNATURE_EXPIRES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SIGNATURE_DATA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SIGNATURE_DATA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE_STREAM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE_STREAM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE_VERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE_VERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE_CONTEXT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE_CONTEXT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_MODULE_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_MODULE_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_BUILDVERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_BUILDVERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_BUILDFLAVOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_BUILDFLAVOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_STATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_STATUS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "LIBSOLV_SELF_DESTRUCT_PKG", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(LIBSOLV_SELF_DESTRUCT_PKG)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_CONSTRAINS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_CONSTRAINS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_TRACK_FEATURES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_TRACK_FEATURES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ISDEFAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ISDEFAULT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_LANGONLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_LANGONLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "UPDATE_COLLECTIONLIST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(UPDATE_COLLECTIONLIST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_MULTIARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_MULTIARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_SIGNATUREDATA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_SIGNATUREDATA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVABLE_ORDERWITHREQUIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVABLE_ORDERWITHREQUIRES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ID_NUM_INTERNAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(ID_NUM_INTERNAL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVID_META", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVID_META)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOLVID_POS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVID_POS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_EQ", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_EQ)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_GT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_GT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_LT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_LT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_AND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_AND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_OR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_OR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_WITH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_WITH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_NAMESPACE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_NAMESPACE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_FILECONFLICT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_FILECONFLICT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_COND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_COND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_COMPAT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_COMPAT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_KIND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_KIND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_MULTIARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_MULTIARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_ELSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_ELSE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_WITHOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_WITHOUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_UNLESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_UNLESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "REL_CONDA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REL_CONDA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_PROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_PROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_FILELIST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_FILELIST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_CANON", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_CANON)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_DOTARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_DOTARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_REL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_REL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_INSTALLED_ONLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_INSTALLED_ONLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_GLOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_GLOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_FLAT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_FLAT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_NOCASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_NOCASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_SKIP_KIND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_SKIP_KIND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_MATCH_DEPSTR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_MATCH_DEPSTR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_SOURCE_ONLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_SOURCE_ONLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_WITH_SOURCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_WITH_SOURCE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_WITH_DISABLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_WITH_DISABLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_WITH_BADARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_WITH_BADARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_WITH_ALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_WITH_ALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_ADD", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_ADD)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_SUBTRACT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_SUBTRACT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_FILTER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_FILTER_KEEP_IFEMPTY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_FILTER_KEEP_IFEMPTY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Selection_SELECTION_FILTER_SWAPPED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SELECTION_FILTER_SWAPPED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Selection, (void*) "solv::Selection");
  SWIG_TypeClientData(SWIGTYPE_p_Dep, (void*) "solv::Dep");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_STRING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_STRING)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_STRINGSTART", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_STRINGSTART)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_STRINGEND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_STRINGEND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_SUBSTRING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_SUBSTRING)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_GLOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_GLOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_REGEX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_REGEX)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_NOCASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_NOCASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_FILES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_FILES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_COMPLETE_FILELIST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_COMPLETE_FILELIST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Dataiterator_SEARCH_CHECKSUMS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SEARCH_CHECKSUMS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Dataiterator, (void*) "solv::Dataiterator");
  SWIG_TypeClientData(SWIGTYPE_p_XSolvable, (void*) "solv::XSolvable");
  SWIG_TypeClientData(SWIGTYPE_p_Ruleinfo, (void*) "solv::Ruleinfo");
  SWIG_TypeClientData(SWIGTYPE_p_XRule, (void*) "solv::XRule");
  SWIG_TypeClientData(SWIGTYPE_p_XRepodata, (void*) "solv::XRepodata");
  SWIG_TypeClientData(SWIGTYPE_p_Pool_solvable_iterator, (void*) "solv::Pool_solvable_iterator");
  SWIG_TypeClientData(SWIGTYPE_p_Pool_repo_iterator, (void*) "solv::Pool_repo_iterator");
  SWIG_TypeClientData(SWIGTYPE_p_Repo_solvable_iterator, (void*) "solv::Repo_solvable_iterator");
  SWIG_TypeClientData(SWIGTYPE_p_Datamatch, (void*) "solv::Datamatch");
  SWIG_TypeClientData(SWIGTYPE_p_Datapos, (void*) "solv::Datapos");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE_PROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE_PROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE_ONE_OF", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE_ONE_OF)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE_REPO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE_REPO)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SOLVABLE_ALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLVABLE_ALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SELECTMASK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SELECTMASK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_NOOP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_NOOP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_INSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_INSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_ERASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ERASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_UPDATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_UPDATE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_WEAKENDEPS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_WEAKENDEPS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_MULTIVERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_MULTIVERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_LOCK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_LOCK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_DISTUPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_DISTUPGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_VERIFY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_VERIFY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_DROP_ORPHANED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_DROP_ORPHANED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_USERINSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_USERINSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_ALLOWUNINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ALLOWUNINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_FAVOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FAVOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_DISFAVOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_DISFAVOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_EXCLUDEFROMWEAK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_EXCLUDEFROMWEAK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_JOBMASK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_JOBMASK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_WEAK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_WEAK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_ESSENTIAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ESSENTIAL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_CLEANDEPS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_CLEANDEPS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_FORCEBEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FORCEBEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_TARGETED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TARGETED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_NOTBYUSER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_NOTBYUSER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETEV", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETEV)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETEVR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETEVR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETVENDOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETVENDOR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETREPO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETREPO)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETNAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETNAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_NOAUTOSET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_NOAUTOSET)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Job_SOLVER_SETMASK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SETMASK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Job, (void*) "solv::Job");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_PROMOTEEPOCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_PROMOTEEPOCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_FORBIDSELFCONFLICTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_FORBIDSELFCONFLICTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_OBSOLETEUSESPROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_OBSOLETEUSESPROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_IMPLICITOBSOLETEUSESPROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_IMPLICITOBSOLETEUSESPROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_OBSOLETEUSESCOLORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_OBSOLETEUSESCOLORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_IMPLICITOBSOLETEUSESCOLORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_IMPLICITOBSOLETEUSESCOLORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_NOINSTALLEDOBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_NOINSTALLEDOBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_HAVEDISTEPOCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_HAVEDISTEPOCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_NOOBSOLETESMULTIVERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_NOOBSOLETESMULTIVERSION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_ADDFILEPROVIDESFILTERED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_ADDFILEPROVIDESFILTERED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_NOWHATPROVIDESAUX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_NOWHATPROVIDESAUX)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_POOL_FLAG_WHATPROVIDESWITHDISABLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POOL_FLAG_WHATPROVIDESWITHDISABLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_RPM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_RPM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_DEB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_DEB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_ARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_ARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_HAIKU", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_HAIKU)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_CONDA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_CONDA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Pool_DISTTYPE_APK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(DISTTYPE_APK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Pool, (void*) "solv::Pool");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_REUSE_REPODATA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_REUSE_REPODATA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_NO_INTERNALIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_NO_INTERNALIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_LOCALPOOL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_LOCALPOOL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_USE_LOADING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_USE_LOADING)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_EXTEND_SOLVABLES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_EXTEND_SOLVABLES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_USE_ROOTDIR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_USE_ROOTDIR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_REPO_NO_LOCATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(REPO_NO_LOCATION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_SOLV_ADD_NO_STUBS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLV_ADD_NO_STUBS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Repo_SUSETAGS_RECORD_SHARES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SUSETAGS_RECORD_SHARES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Repo, (void*) "solv::Repo");
  SWIG_TypeClientData(SWIGTYPE_p_Decision, (void*) "solv::Decision");
  SWIG_TypeClientData(SWIGTYPE_p_Decisionset, (void*) "solv::Decisionset");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_UNKNOWN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_UNKNOWN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_NOT_INSTALLABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_NOT_INSTALLABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_NOTHING_PROVIDES_DEP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_NOTHING_PROVIDES_DEP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_REQUIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_REQUIRES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_SELF_CONFLICT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_SELF_CONFLICT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_CONFLICTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_CONFLICTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_SAME_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_SAME_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_IMPLICIT_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_IMPLICIT_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_INSTALLED_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_INSTALLED_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_RECOMMENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_RECOMMENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_CONSTRAINS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_CONSTRAINS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_PKG_SUPPLEMENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_PKG_SUPPLEMENTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_UPDATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_UPDATE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_FEATURE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_FEATURE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_JOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_JOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_JOB_NOTHING_PROVIDES_DEP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_JOB_NOTHING_PROVIDES_DEP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_JOB_PROVIDED_BY_SYSTEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_JOB_PROVIDED_BY_SYSTEM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_JOB_UNKNOWN_PACKAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_JOB_UNKNOWN_PACKAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_JOB_UNSUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_JOB_UNSUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_DISTUPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_DISTUPGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_INFARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_INFARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_CHOICE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_CHOICE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_LEARNT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_LEARNT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_BEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_BEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_YUMOBS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_YUMOBS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_RECOMMENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_RECOMMENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_BLACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_BLACK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_STRICT_REPO_PRIORITY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_STRICT_REPO_PRIORITY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_JOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_JOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_POOLJOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_POOLJOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_INFARCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_INFARCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_DISTUPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_DISTUPGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_BEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_BEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_ERASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_ERASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_REPLACE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_REPLACE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_REPLACE_DOWNGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_REPLACE_DOWNGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_REPLACE_ARCHCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_REPLACE_ARCHCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_REPLACE_VENDORCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_REPLACE_VENDORCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_SOLUTION_REPLACE_NAMECHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_SOLUTION_REPLACE_NAMECHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_POLICY_ILLEGAL_DOWNGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POLICY_ILLEGAL_DOWNGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_POLICY_ILLEGAL_ARCHCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POLICY_ILLEGAL_ARCHCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_POLICY_ILLEGAL_VENDORCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POLICY_ILLEGAL_VENDORCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_POLICY_ILLEGAL_NAMECHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(POLICY_ILLEGAL_NAMECHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ALLOW_DOWNGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ALLOW_DOWNGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ALLOW_ARCHCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ALLOW_ARCHCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ALLOW_VENDORCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ALLOW_VENDORCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ALLOW_NAMECHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ALLOW_NAMECHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ALLOW_UNINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ALLOW_UNINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_NO_UPDATEPROVIDE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_NO_UPDATEPROVIDE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_SPLITPROVIDES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_SPLITPROVIDES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_IGNORE_RECOMMENDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_IGNORE_RECOMMENDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ADD_ALREADY_RECOMMENDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ADD_ALREADY_RECOMMENDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_NO_INFARCHCHECK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_NO_INFARCHCHECK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_BEST_OBEY_POLICY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_BEST_OBEY_POLICY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_NO_AUTOTARGET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_NO_AUTOTARGET)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_DUP_ALLOW_DOWNGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_DUP_ALLOW_DOWNGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_DUP_ALLOW_ARCHCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_DUP_ALLOW_ARCHCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_DUP_ALLOW_VENDORCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_DUP_ALLOW_VENDORCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_DUP_ALLOW_NAMECHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_DUP_ALLOW_NAMECHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_KEEP_ORPHANS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_KEEP_ORPHANS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_BREAK_ORPHANS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_BREAK_ORPHANS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_FOCUS_INSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_FOCUS_INSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_YUM_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_YUM_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_NEED_UPDATEPROVIDE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_NEED_UPDATEPROVIDE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_FOCUS_BEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_FOCUS_BEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_STRONG_RECOMMENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_STRONG_RECOMMENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_INSTALL_ALSO_UPDATES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_INSTALL_ALSO_UPDATES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_ONLY_NAMESPACE_RECOMMENDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_ONLY_NAMESPACE_RECOMMENDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_STRICT_REPO_PRIORITY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_STRICT_REPO_PRIORITY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_FLAG_FOCUS_NEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_FLAG_FOCUS_NEW)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_UNRELATED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_UNRELATED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_UNIT_RULE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_UNIT_RULE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_KEEP_INSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_KEEP_INSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_RESOLVE_JOB", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_RESOLVE_JOB)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_UPDATE_INSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_UPDATE_INSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_CLEANDEPS_ERASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_CLEANDEPS_ERASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_RESOLVE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_RESOLVE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_WEAKDEP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_WEAKDEP)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_RESOLVE_ORPHAN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_RESOLVE_ORPHAN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_RECOMMENDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_RECOMMENDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_SUPPLEMENTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_SUPPLEMENTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_UNSOLVABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_UNSOLVABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_REASON_PREMISE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_REASON_PREMISE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Solver_SOLVER_RULE_RPM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_RULE_RPM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Solver, (void*) "solv::Solver");
  SWIG_TypeClientData(SWIGTYPE_p_Chksum, (void*) "solv::Chksum");
  SWIG_TypeClientData(SWIGTYPE_p_Solvsig, (void*) "solv::Solvsig");
  SWIG_TypeClientData(SWIGTYPE_p_SolvFp, (void*) "solv::SolvFp");
  SWIG_TypeClientData(SWIGTYPE_p_Problem, (void*) "solv::Problem");
  SWIG_TypeClientData(SWIGTYPE_p_Solution, (void*) "solv::Solution");
  SWIG_TypeClientData(SWIGTYPE_p_Solutionelement, (void*) "solv::Solutionelement");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Alternative_SOLVER_ALTERNATIVE_TYPE_RULE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ALTERNATIVE_TYPE_RULE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Alternative_SOLVER_ALTERNATIVE_TYPE_RECOMMENDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ALTERNATIVE_TYPE_RECOMMENDS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Alternative_SOLVER_ALTERNATIVE_TYPE_SUGGESTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_ALTERNATIVE_TYPE_SUGGESTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Alternative, (void*) "solv::Alternative");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_IGNORE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_IGNORE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_ERASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_ERASE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_REINSTALLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_REINSTALLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_DOWNGRADED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_DOWNGRADED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_CHANGED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_CHANGED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_UPGRADED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_UPGRADED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_OBSOLETED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_OBSOLETED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_INSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_INSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_REINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_REINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_DOWNGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_DOWNGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_CHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_CHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_UPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_UPGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_MULTIINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_MULTIINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_MULTIREINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_MULTIREINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_MAXTYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_MAXTYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_SHOW_ACTIVE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_SHOW_ACTIVE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_SHOW_ALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_SHOW_ALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_SHOW_OBSOLETES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_SHOW_OBSOLETES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_SHOW_MULTIINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_SHOW_MULTIINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_CHANGE_IS_REINSTALL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_CHANGE_IS_REINSTALL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_OBSOLETE_IS_UPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_OBSOLETE_IS_UPGRADE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_MERGE_VENDORCHANGES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_MERGE_VENDORCHANGES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_MERGE_ARCHCHANGES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_MERGE_ARCHCHANGES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_RPM_ONLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_RPM_ONLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_ARCHCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_ARCHCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_VENDORCHANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_VENDORCHANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "Transaction_SOLVER_TRANSACTION_KEEP_ORDERDATA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOLVER_TRANSACTION_KEEP_ORDERDATA)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_Transaction, (void*) "solv::Transaction");
  SWIG_TypeClientData(SWIGTYPE_p_TransactionClass, (void*) "solv::TransactionClass");
  ST(0) = &PL_sv_yes;
  XSRETURN(1);
}

