/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

#define PY_SSIZE_T_CLEAN



#define SWIG_VERSION 0x040401
#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

#define SWIG_name    "_solv"
/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if !defined(PY_SSIZE_T_CLEAN) && !defined(SWIG_NO_PY_SSIZE_T_CLEAN)
#define PY_SSIZE_T_CLEAN
#endif

#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus) && __cplusplus >=201703L
#pragma GCC diagnostic ignored "-Wregister" /* For python-2.7 headers that use register */
#endif
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */

#if defined(_MSC_VER) && _MSC_VER >= 1929
/* Workaround compilation errors when redefining _DEBUG in MSVC 2019 version 16.10 and later
 * See https://github.com/swig/swig/issues/2090 */
# include <corecrt.h>
#endif

# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif


#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

#include <stdio.h>
#include <stdlib.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if PY_VERSION_HEX >= 0x03030000 && !defined(SWIG_NO_HEAPTYPES)
#if !defined(SWIG_HEAPTYPES)
#define SWIG_HEAPTYPES
#endif
#endif

/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif

#if defined(SWIG_HEAPTYPES)
#if PY_VERSION_HEX < 0x030c0000
#include <structmember.h>
#define Py_READONLY READONLY
#define Py_T_PYSSIZET T_PYSSIZET
#endif
#endif

#include <stddef.h> /* For offsetof */


/* Wrapper around PyUnicode_AsUTF8AndSize - call Py_XDECREF on the returned pbytes when finished with the returned string */
SWIGINTERN const char *
SWIG_PyUnicode_AsUTF8AndSize(PyObject *str, Py_ssize_t *psize, PyObject **pbytes)
{
#if PY_VERSION_HEX >= 0x03030000
# if !defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030A0000
  *pbytes = NULL;
  return PyUnicode_AsUTF8AndSize(str, psize);
# else
  const char *chars;
  *pbytes = PyUnicode_AsUTF8String(str);
  chars = *pbytes ? PyBytes_AsString(*pbytes) : NULL;
  if (chars && psize)
    *psize = PyBytes_Size(*pbytes);
  return chars;
# endif
#else
  char *chars = NULL;
  *pbytes = NULL;
  PyString_AsStringAndSize(str, &chars, psize);
  return chars;
#endif
}

SWIGINTERN PyObject *
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#define SWIG_RUNTIME_MODULE "swig_runtime_data" SWIG_RUNTIME_VERSION

/* SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user interface files check for it. */
# define SWIGPY_USE_CAPSULE
#ifdef SWIGPYTHON_BUILTIN
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule_builtin" SWIG_TYPE_TABLE_NAME
#else
# define SWIGPY_CAPSULE_ATTR_NAME "type_pointer_capsule" SWIG_TYPE_TABLE_NAME
#endif
#define SWIGPY_CAPSULE_NAME SWIG_RUNTIME_MODULE "." SWIGPY_CAPSULE_ATTR_NAME

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

#if defined(Py_LIMITED_API)
# define PyTuple_GET_ITEM PyTuple_GetItem
/* Note that PyTuple_SetItem() has different semantics from PyTuple_SET_ITEM as it decref's the original tuple item, so in general they cannot be used
  interchangeably. However in SWIG-generated code PyTuple_SET_ITEM is only used with newly initialized tuples without any items and for them this does work. */
# define PyTuple_SET_ITEM PyTuple_SetItem
# define PyTuple_GET_SIZE PyTuple_Size
# define PyCFunction_GET_FLAGS PyCFunction_GetFlags
# define PyCFunction_GET_FUNCTION PyCFunction_GetFunction
# define PyCFunction_GET_SELF PyCFunction_GetSelf
# define PyList_GET_ITEM PyList_GetItem
# define PyList_SET_ITEM PyList_SetItem
# define PySliceObject PyObject
#endif

/* Increment and Decrement wrappers - for portability when using the stable abi and for performance otherwise */
#ifdef Py_LIMITED_API
# define SWIG_Py_INCREF Py_IncRef
# define SWIG_Py_XINCREF Py_IncRef
# define SWIG_Py_DECREF Py_DecRef
# define SWIG_Py_XDECREF Py_DecRef
#else
# define SWIG_Py_INCREF Py_INCREF
# define SWIG_Py_XINCREF Py_XINCREF
# define SWIG_Py_DECREF Py_DECREF
# define SWIG_Py_XDECREF Py_XDECREF
#endif

#if PY_VERSION_HEX >= 0x03000000
#if (PY_VERSION_HEX >= 0x030d00a6) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a6)
# define SWIG_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
SWIGINTERN PyObject *
SWIG_PyType_GetFullyQualifiedName(PyTypeObject *type) {
  PyObject *result = NULL;
  PyObject *qualname = PyObject_GetAttrString((PyObject *)type, "__qualname__");
  if (qualname) {
    PyObject *mod = PyObject_GetAttrString((PyObject *)type, "__module__");
    if (mod) {
      if (PyUnicode_Check(mod) && PyUnicode_CompareWithASCIIString(mod, "builtins") && PyUnicode_CompareWithASCIIString(mod, "__main__")) {
        result = PyUnicode_FromFormat("%U%c%U", mod, '.', qualname);
        SWIG_Py_DECREF(qualname);
      } else {
        result = qualname;
      }
      SWIG_Py_DECREF(mod);
    } else {
      result = qualname;
    }
  }

  return result;
}
#endif
#endif

/* gh-114329 added PyList_GetItemRef() to Python 3.13.0a4 */
#if (PY_VERSION_HEX >= 0x030d00a4) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a4)
# define SWIG_PyList_GetItemRef PyList_GetItemRef
#else
SWIGINTERN PyObject *
SWIG_PyList_GetItemRef(PyObject *op, Py_ssize_t index) {
  PyObject *item = PyList_GetItem(op, index);
  Py_XINCREF(item);
  return item;
}
#endif

/* gh-106004 added PyDict_GetItemRef() and PyDict_GetItemStringRef() to Python 3.13.0a1
   functions are renamed here for compatibility with abi3audit */
#if (PY_VERSION_HEX >= 0x030d00a1) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d00a1)
# define SWIG_PyDict_GetItemRef PyDict_GetItemRef
# define SWIG_PyDict_GetItemStringRef PyDict_GetItemStringRef
#else
SWIGINTERN int
SWIG_PyDict_GetItemRef(PyObject *mp, PyObject *key, PyObject **result) {
#if PY_VERSION_HEX >= 0x03000000
  PyObject *item = PyDict_GetItemWithError(mp, key);
#else
  PyObject *item = _PyDict_GetItemWithError(mp, key);
#endif
  if (item != NULL) {
    *result = (PyObject *)(item);
    SWIG_Py_INCREF(*result);
    return 1;
  }
  if (!PyErr_Occurred()) {
    *result = NULL;
    return 0;
  }
  *result = NULL;
  return -1;
}

SWIGINTERN int
SWIG_PyDict_GetItemStringRef(PyObject *mp, const char *key, PyObject **result) {
  int res;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *key_obj = PyUnicode_FromString(key);
#else
  PyObject *key_obj = PyString_FromString(key);
#endif
  if (key_obj == NULL) {
    *result = NULL;
    return -1;
  }
  res = SWIG_PyDict_GetItemRef(mp, key_obj, result);
  Py_DECREF(key_obj);
  return res;
}
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  case SWIG_NullReferenceError:
    type = PyExc_TypeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyObject *bytes = NULL;
    const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
    PyErr_Clear();
    SWIG_Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_DECREF(old_str);
    SWIG_Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    if (newvalue) {
      SWIG_Py_XDECREF(value);
      PyErr_Restore(type, newvalue, traceback);
    } else {
      PyErr_Restore(type, value, traceback);
    }
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      if PY_VERSION_HEX < 0x03070000
#        define SWIG_PYTHON_INITIALIZE_THREADS PyEval_InitThreads()
#      else
#        define SWIG_PYTHON_INITIALIZE_THREADS
#      endif
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { status = false; PyEval_RestoreThread(save); }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03050000
# error "This version of SWIG only supports Python 3 >= 3.5"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   

/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  SWIG_Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  SWIG_Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  SWIG_Py_DECREF(obj);
}

#endif

/* SWIG runtime data Python module */
static PyObject *Swig_runtime_data_module_global = NULL;

/* Create/obtain the single swig_runtime_data module which is used across different SWIG generated modules */
SWIGINTERN PyObject *
SWIG_runtime_data_module(void) {
  if (!Swig_runtime_data_module_global) {
#if (PY_VERSION_HEX >= 0x030d0000) && (!defined(Py_LIMITED_API) || Py_LIMITED_API+0 >= 0x030d0000)
    /* free-threading note: the GIL is always enabled when this function is first called
       by SWIG_init, so there's no risk of race conditions */
    Swig_runtime_data_module_global = PyImport_AddModuleRef(SWIG_RUNTIME_MODULE);
#elif PY_VERSION_HEX >= 0x03000000
    Swig_runtime_data_module_global = PyImport_AddModule(SWIG_RUNTIME_MODULE);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#else
    static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
    Swig_runtime_data_module_global = Py_InitModule(SWIG_RUNTIME_MODULE, swig_empty_runtime_method_table);
    SWIG_Py_XINCREF(Swig_runtime_data_module_global);
#endif
  }
  assert(Swig_runtime_data_module_global);
  return Swig_runtime_data_module_global;
}

/* Append a value to the result obj */
SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj, int is_void) {
  if (!result) {
    result = obj;
  } else if (result == Py_None && is_void) {
    SWIG_Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      if (result) {
        PyList_SET_ITEM(result, 0, o2);
      } else {
        SWIG_Py_DECREF(obj);
        return o2;
      }
    }
    PyList_Append(result,obj);
    SWIG_Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
 
/* -----------------------------------------------------------------------------
 * global variable support code.
 * ----------------------------------------------------------------------------- */
 
typedef struct swig_globalvar {   
  char       *name;                  /* Name of global variable */
  PyObject *(*get_attr)(void);       /* Return the current value */
  int       (*set_attr)(PyObject *); /* Set the value */
  struct swig_globalvar *next;
} swig_globalvar;

typedef struct swig_varlinkobject {
  PyObject_HEAD
  swig_globalvar *vars;
} swig_varlinkobject;

SWIGINTERN PyObject *
SwigVarLink_repr(PyObject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_InternFromString("<Swig global variables>");
#else
  return PyString_FromString("<Swig global variables>");
#endif
}

SWIGINTERN PyObject *
SwigVarLink_str(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
#if PY_VERSION_HEX >= 0x03000000
  PyObject *str = PyUnicode_InternFromString("(");
  PyObject *tail;
  PyObject *joined;
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    tail = PyUnicode_FromString(var->name);
    joined = PyUnicode_Concat(str, tail);
    SWIG_Py_DECREF(str);
    SWIG_Py_DECREF(tail);
    str = joined;
    if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        SWIG_Py_DECREF(str);
        SWIG_Py_DECREF(tail);
        str = joined;
    }
  }
  tail = PyUnicode_InternFromString(")");
  joined = PyUnicode_Concat(str, tail);
  SWIG_Py_DECREF(str);
  SWIG_Py_DECREF(tail);
  str = joined;
#else
  PyObject *str = PyString_FromString("(");
  swig_globalvar *var;
  for (var = v->vars; var; var=var->next) {
    PyString_ConcatAndDel(&str,PyString_FromString(var->name));
    if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
  }
  PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
  return str;
}

SWIGINTERN void
SwigVarLink_dealloc(PyObject *o) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  swig_globalvar *var = v->vars;
  while (var) {
    swig_globalvar *n = var->next;
    free(var->name);
    free(var);
    var = n;
  }
}

SWIGINTERN PyObject *
SwigVarLink_getattr(PyObject *o, char *n) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  PyObject *res = NULL;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->get_attr)();
      break;
    }
    var = var->next;
  }
  if (res == NULL && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN int
SwigVarLink_setattr(PyObject *o, char *n, PyObject *p) {
  swig_varlinkobject *v = (swig_varlinkobject *) o;
  int res = 1;
  swig_globalvar *var = v->vars;
  while (var) {
    if (strcmp(var->name,n) == 0) {
      res = (*var->set_attr)(p);
      break;
    }
    var = var->next;
  }
  if (res == 1 && !PyErr_Occurred()) {
    PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
  }
  return res;
}

SWIGINTERN PyTypeObject*
SwigVarLink_TypeOnce(void) {
  static char SwigVarLink_doc[] = "Swig variable link object";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject varlink_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                  /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigVarLink", /* tp_name */
      sizeof(swig_varlinkobject),         /* tp_basicsize */
      0,                                  /* tp_itemsize */
      (destructor) SwigVarLink_dealloc,   /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                       /* tp_print */
#else
      (Py_ssize_t)0,                      /* tp_vectorcall_offset */
#endif
      (getattrfunc) SwigVarLink_getattr,  /* tp_getattr */
      (setattrfunc) SwigVarLink_setattr,  /* tp_setattr */
      0,                                  /* tp_compare */
      (reprfunc) SwigVarLink_repr,        /* tp_repr */
      0,                                  /* tp_as_number */
      0,                                  /* tp_as_sequence */
      0,                                  /* tp_as_mapping */
      0,                                  /* tp_hash */
      0,                                  /* tp_call */
      (reprfunc) SwigVarLink_str,         /* tp_str */
      0,                                  /* tp_getattro */
      0,                                  /* tp_setattro */
      0,                                  /* tp_as_buffer */
      0,                                  /* tp_flags */
      SwigVarLink_doc,                    /* tp_doc */
      0,                                  /* tp_traverse */
      0,                                  /* tp_clear */
      0,                                  /* tp_richcompare */
      0,                                  /* tp_weaklistoffset */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
      0,                                  /* tp_del */
      0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                  /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                  /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                  /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                  /* tp_allocs */
      0,                                  /* tp_frees */
      0,                                  /* tp_maxalloc */
      0,                                  /* tp_prev */
      0                                   /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    varlink_type = tmp;
    type_init = 1;
    if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigVarLink", (PyObject *)&varlink_type) == 0)
      SWIG_Py_INCREF((PyObject *)&varlink_type);
  }
  return &varlink_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigVarLink_dealloc },
    { Py_tp_repr, (void *)SwigVarLink_repr },
    { Py_tp_getattr, (void *)SwigVarLink_getattr },
    { Py_tp_setattr, (void *)SwigVarLink_setattr },
    { Py_tp_str, (void *)SwigVarLink_str },
    { Py_tp_doc, (void *)SwigVarLink_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigVarLink",
    sizeof(swig_varlinkobject),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigVarLink", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyTypeObject*
SwigVarLink_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigVarLink_TypeOnce();
  return type;
}

/* Create a variable linking object for use later */
SWIGINTERN PyObject *
SWIG_Python_newvarlink(void) {
  swig_varlinkobject *result = PyObject_New(swig_varlinkobject, SwigVarLink_Type());
  if (result) {
    result->vars = 0;
  }
  return ((PyObject*) result);
}

SWIGINTERN void 
SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
  swig_varlinkobject *v = (swig_varlinkobject *) p;
  swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
  if (gv) {
    size_t size = strlen(name)+1;
    gv->name = (char *)malloc(size);
    if (gv->name) {
      memcpy(gv->name, name, size);
      gv->get_attr = get_attr;
      gv->set_attr = set_attr;
      gv->next = v->vars;
    }
  }
  v->vars = gv;
}


static PyObject *Swig_Globals_global = NULL;
  
SWIGINTERN PyObject *
SWIG_globals(void) {
  if (Swig_Globals_global == NULL) {
    Swig_Globals_global = SWIG_newvarlink();
  }
  return Swig_Globals_global;
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  SWIG_Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    SWIG_Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      SWIG_Py_INCREF(obj);
      data->newargs = obj;
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
        data->newargs = PyTuple_New(1);
        if (data->newargs) {
          SWIG_Py_INCREF(obj);
          PyTuple_SET_ITEM(data->newargs, 0, obj);
        } else {
          SWIG_Py_DECREF(data->newraw);
          SWIG_Py_DECREF(data->klass);
          free(data);
          return 0;
        }
      } else {
        SWIG_Py_INCREF(obj);
        data->newargs = obj;
      }
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      data->delargs = !(PyCFunction_GET_FLAGS(data->destroy) & METH_O);
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data)
{
  SWIG_Py_XDECREF(data->klass);
  SWIG_Py_XDECREF(data->newraw);
  SWIG_Py_XDECREF(data->newargs);
  SWIG_Py_XDECREF(data->destroy);
  free(data);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
  PyObject *swigdict;
  PyObject *weakreflist;
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->swigdict)
    sobj->swigdict = PyDict_New();

  SWIG_Py_XINCREF(sobj->swigdict);
  return sobj->swigdict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    PyObject *val = SwigPyObject_long(v);
    if (val) {
      PyObject *ofmt;
      PyTuple_SET_ITEM(args, 0, val);
      ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
        res = PyUnicode_Format(ofmt,args);
#else
        res = PyString_Format(ofmt,args);
#endif
        SWIG_Py_DECREF(ofmt);
      }
    }
    SWIG_Py_DECREF(args);
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (repr && v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
    if (nrep) {
# if PY_VERSION_HEX >= 0x03000000
      PyObject *joined = PyUnicode_Concat(repr, nrep);
      SWIG_Py_DECREF(repr);
      SWIG_Py_DECREF(nrep);
      repr = joined;
# else
      PyString_ConcatAndDel(&repr,nrep);
# endif
    } else {
      SWIG_Py_DECREF(repr);
      repr = NULL;
    }
  }
  return repr;
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in SwigPyObject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(PyObject *v, PyObject *w)
{
  /* tp_compare is only called when both objects have the same type, so
   * the casts are guaranteed to be ok. */
  void *i = ((SwigPyObject *)v)->ptr;
  void *j = ((SwigPyObject *)w)->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

SWIGRUNTIMEINLINE int SwigPyObject_Check(PyObject *);

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(PyObject *v, PyObject *w, int op)
{
  PyObject* res = NULL;
  if (!PyErr_Occurred()) {
    /* Per https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_richcompare
     * the first argument is guaranteed to be an instance of SwigPyObject, but the
     * second is not, so we typecheck that one. */
    if ((op != Py_EQ && op != Py_NE) || !SwigPyObject_Check(w)) {
      SWIG_Py_INCREF(Py_NotImplemented);
      return Py_NotImplemented;
    }
    res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  }
  return res;
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyObject_Type();
  PyTypeObject *op_type = Py_TYPE(op);
#ifdef SWIGPYTHON_BUILTIN
  /* Only builtin types have SwigPyObject as a base type */
  return PyType_IsSubtype(op_type, target_tp);
#else
  /* Check for an exact match to SwigPyObject */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyObject");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyObject") == 0;
#endif
  }
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

static PyObject* Swig_Capsule_global = NULL;

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        if (tmp) {
          res = SWIG_Python_CallFunctor(destroy, tmp);
        } else {
          res = 0;
        }
        SWIG_Py_XDECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      SWIG_Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
    SWIG_Py_XDECREF(Swig_Capsule_global);
  }
  SWIG_Py_XDECREF(next);
  SWIG_Py_XDECREF(sobj->swigdict);
  PyObject_Free(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  ((SwigPyObject *)next)->next = sobj->next;
  sobj->next = next;
  SWIG_Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    SWIG_Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SWIG_Py_DECREF(SwigPyObject_acquire(v,args));
      } else {
        SWIG_Py_DECREF(SwigPyObject_disown(v,args));
      }
    } 
    return obj;
  }
}

static PyMethodDef
SwigPyObject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char SwigPyObject_doc[] = "Swig object holding a C/C++ pointer";
#ifndef SWIG_HEAPTYPES
  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyObject",  /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyObject_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      offsetof(SwigPyObject, weakreflist),  /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      SwigPyObject_methods,                 /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyObject", (PyObject *)&swigpyobject_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpyobject_type);
  }
  return &swigpyobject_type;
#else
  static PyMemberDef SwigPyObject_members[] = {
      { (char *)"__dictoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, swigdict), Py_READONLY, NULL },
      { (char *)"__weaklistoffset__", Py_T_PYSSIZET, offsetof(SwigPyObject, weakreflist), Py_READONLY, NULL },
      { NULL, 0, 0, 0, NULL }
  };
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyObject_dealloc },
    { Py_tp_repr, (void *)SwigPyObject_repr },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyObject_doc },
    { Py_tp_richcompare, (void *)SwigPyObject_richcompare },
    { Py_tp_methods, (void *)SwigPyObject_methods },
    { Py_nb_int, (void *)SwigPyObject_long },
    { Py_tp_members, (void *)SwigPyObject_members },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyObject",
    sizeof(SwigPyObject),
    0,
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
#if !defined(Py_LIMITED_API)
/* While this __dictoffset__ is only used with the builtin wrappers, SwigPyObject ought to be
   identical when created for use by proxy class wrappers in case it is shared across multiple modules. */
#if PY_VERSION_HEX < 0x03090000
    /* Workaround as __dictoffset__ and __weaklistoffset__ above are only supported from python-3.9 */
  if (pytype) {
    ((PyTypeObject *)pytype)->tp_dictoffset = offsetof(SwigPyObject, swigdict);
    ((PyTypeObject *)pytype)->tp_weaklistoffset = offsetof(SwigPyObject, weakreflist);
  }
#endif
#endif
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyObject", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_New(SwigPyObject, SwigPyObject_Type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
    sobj->swigdict = 0;
    sobj->weakreflist = 0;
    if (own == SWIG_POINTER_OWN) {
      /* Obtain a reference to the Python capsule wrapping the module information, so that the
       * module information is correctly destroyed after all SWIG python objects have been freed
       * by the GC (and corresponding destructors invoked) */
      SWIG_Py_XINCREF(Swig_Capsule_global);
    }
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_Type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  PyTypeObject *target_tp = SwigPyPacked_Type();
  PyTypeObject *op_type = Py_TYPE(op);
  /* Check for an exact match to SwigPyPacked */
  if (op_type == target_tp) {
    return 1;
  } else {
    /* Fallback for multiple modules */
#if PY_VERSION_HEX >= 0x03000000
    int cmp;
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(op_type);
    if (!tpname)
      return 0;
    cmp = PyUnicode_CompareWithASCIIString(tpname, SWIG_RUNTIME_MODULE ".SwigPyPacked");
    SWIG_Py_DECREF(tpname);
    return cmp == 0;
#else
    return strcmp(op_type->tp_name, SWIG_RUNTIME_MODULE ".SwigPyPacked") == 0;
#endif
  }
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_Free(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char SwigPyPacked_doc[] = "Swig object holding a C/C++ function pointer";
#ifndef SWIG_HEAPTYPES
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      SWIG_RUNTIME_MODULE ".SwigPyPacked",  /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
#if PY_VERSION_HEX < 0x030800b4
      (printfunc)0,                         /* tp_print */
#else
      (Py_ssize_t)0,                        /* tp_vectorcall_offset */
#endif
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      SwigPyPacked_doc,                     /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#if PY_VERSION_HEX >= 0x030c0000
      0,                                    /* tp_watched */
#endif
#if PY_VERSION_HEX >= 0x030d00a4
      0,                                    /* tp_versions_used */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    PyObject *runtime_data_module = SWIG_runtime_data_module();
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) != 0)
      return NULL;
    if (PyModule_AddObject(runtime_data_module, "SwigPyPacked", (PyObject *)&swigpypacked_type) == 0)
      SWIG_Py_INCREF((PyObject *)&swigpypacked_type);
  }
  return &swigpypacked_type;
#else
  PyType_Slot slots[] = {
    { Py_tp_dealloc, (void *)SwigPyPacked_dealloc },
    { Py_tp_repr, (void *)SwigPyPacked_repr },
    { Py_tp_str, (void *)SwigPyPacked_str },
    { Py_tp_getattro, (void *)PyObject_GenericGetAttr },
    { Py_tp_doc, (void *)SwigPyPacked_doc },
    { 0, NULL }
  };
  PyType_Spec spec = {
    SWIG_RUNTIME_MODULE ".SwigPyPacked",
    sizeof(SwigPyPacked),
    0,
    Py_TPFLAGS_DEFAULT,
    slots
  };
  PyObject *pytype = PyType_FromSpec(&spec);
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  if (pytype && PyModule_AddObject(runtime_data_module, "SwigPyPacked", pytype) == 0)
    SWIG_Py_INCREF(pytype);
  return (PyTypeObject *)pytype;
#endif
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_New(SwigPyPacked, SwigPyPacked_Type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_Free((PyObject *)sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX >= 0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

/* Returns a borrowed reference to the 'this' object */
SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
  if (PyWeakref_CheckProxy(pyobj)) {
#if PY_VERSION_HEX >= 0x030d0000
    if (PyWeakref_GetRef(pyobj, &pyobj) > 0)
      Py_DECREF(pyobj);
    else
      pyobj = NULL;
#else
    pyobj = PyWeakref_GetObject(pyobj);
#endif
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
      if (PyWeakref_CheckProxy(pyobj)) {
        PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
      obj = PyObject_GetAttr(pyobj, SWIG_This());
      if (obj) {
	SWIG_Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj, SWIG_This());
  if (obj) {
    SWIG_Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE) && !sobj->own) {
      res = SWIG_ERROR_RELEASE_NOT_OWNED;
    } else {
      if (own)
        *own = *own | sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
        sobj->own = 0;
      }
      if (flags & SWIG_POINTER_CLEAR) {
        sobj->ptr = 0;
      }
      res = SWIG_OK;
    }
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            SWIG_Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
#ifndef Py_LIMITED_API
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
#else
    PyObject* pystr_doc = PyObject_GetAttrString(obj, "__doc__");
    PyObject *bytes = NULL;
    const char *doc = pystr_doc ? SWIG_PyUnicode_AsUTF8AndSize(pystr_doc, NULL, &bytes) : 0;
#endif
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
#ifdef Py_LIMITED_API
    SWIG_Py_XDECREF(bytes);
    SWIG_Py_XDECREF(pystr_doc);
#endif
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
        PyObject *dict = *dictptr;
        if (dict == NULL) {
          dict = PyDict_New();
          *dictptr = dict;
        }
        if (dict) {
          PyDict_SetItem(dict, SWIG_This(), swig_this);
        } else{
          SWIG_Py_DECREF(inst);
          inst = 0;
        }
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        SWIG_Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
#ifndef Py_LIMITED_API
        newfunc newfn = ((PyTypeObject *)data->newargs)->tp_new;
#else
        newfunc newfn = (newfunc)PyType_GetSlot((PyTypeObject *)data->newargs, Py_tp_new);
#endif
        inst = newfn((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        SWIG_Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            SWIG_Py_DECREF(inst);
            inst = 0;
          } else {
            PyType_Modified(Py_TYPE(inst));
          }
        }
      }
      SWIG_Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      SWIG_Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    if (dict) {
      return PyDict_SetItem(dict, SWIG_This(), swig_this);
    } else{
      return -1;
    }
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SWIG_Py_DECREF(SwigPyObject_append((PyObject*) sthis, obj[1]));
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
#ifndef Py_LIMITED_API
        allocfunc alloc = clientdata->pytype->tp_alloc;
#else
        allocfunc alloc = (allocfunc)PyType_GetSlot(clientdata->pytype, Py_tp_alloc);
#endif
        PyObject *next_self = alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
      if (newobj) {
        newobj->swigdict = 0;
        newobj->weakreflist = 0;
      }
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    SWIG_Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

static PyObject *Swig_TypeCache_global = NULL;

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  if (Swig_TypeCache_global == NULL) {
    Swig_TypeCache_global = PyDict_New();
  }
  return Swig_TypeCache_global;
}

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
#ifdef SWIG_LINK_RUNTIME
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
  }
#else
  void *type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
  if (PyErr_Occurred()) {
    PyErr_Clear();
    type_pointer = (void *)0;
  }
#endif
  return (swig_module_info *) type_pointer;
}


#if defined(SWIG_REFCNT_DEBUG)
#define SWIG_PYOBJ_REFCNT(OBJ) fprintf(stdout, "" #OBJ " count %ld\n", (OBJ ? Py_REFCNT(OBJ) : 0))
#else
#define SWIG_PYOBJ_REFCNT(OBJ)
#endif

static int interpreter_counter = 0; /* how many (sub-)interpreters are using swig_module's types */

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  if (--interpreter_counter != 0) /* another sub-interpreter may still be using the swig_module's types */
    return;
  for (i = 0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      ty->clientdata = 0;
      if (data) SwigPyClientData_Del(data);
    }
  }
  SWIG_PYOBJ_REFCNT(Swig_This_global);
  SWIG_Py_XDECREF(Swig_This_global);
  Swig_This_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Globals_global);
  SWIG_Py_XDECREF(Swig_Globals_global);
  Swig_Globals_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_TypeCache_global);
  SWIG_Py_XDECREF(Swig_TypeCache_global);
  Swig_TypeCache_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_Capsule_global);
  Swig_Capsule_global = NULL;

  SWIG_PYOBJ_REFCNT(Swig_runtime_data_module_global);
  SWIG_Py_XDECREF(Swig_runtime_data_module_global);
  Swig_runtime_data_module_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
  PyObject *runtime_data_module = SWIG_runtime_data_module();
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && runtime_data_module) {
    if (PyModule_AddObject(runtime_data_module, SWIGPY_CAPSULE_ATTR_NAME, pointer) == 0) {
      ++interpreter_counter;
      Swig_Capsule_global = pointer;
    } else {
      SWIG_Py_DECREF(pointer);
    }
  } else {
    SWIG_Py_XDECREF(pointer);
  }
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  swig_type_info *descriptor;
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *obj;
  SWIG_PyDict_GetItemStringRef(cache, type, &obj);
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      if (obj) PyDict_SetItemString(cache, type, obj);
    }
  }
  SWIG_Py_XDECREF(obj);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      PyObject *bytes = NULL;
      const char *tmp = SWIG_PyUnicode_AsUTF8AndSize(old_str, NULL, &bytes);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      SWIG_Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Py_XDECREF(bytes);
      SWIG_Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = Py_TYPE(obj);
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(Py_TYPE(name));
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%S'", tpname);
    SWIG_Py_DECREF(tpname);
#else
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%s'", Py_TYPE(name)->tp_name);
#endif
    return -1;
  } else {
    SWIG_Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) != 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = Py_TYPE(descr)->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      SWIG_Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        goto done;
    }
#if PY_VERSION_HEX >= 0x03000000
    {
      PyObject *tpname = SWIG_PyType_GetFullyQualifiedName(tp);
      PyErr_Format(PyExc_AttributeError, "'%S' object has no attribute '%s'", tpname, PyString_AsString(encoded_name));
      SWIG_Py_DECREF(tpname);
    }
#else
    PyErr_Format(PyExc_AttributeError, "'%s' object has no attribute '%s'", tp->tp_name, PyString_AsString(encoded_name));
#endif
    SWIG_Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  SWIG_Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_Alternative swig_types[0]
#define SWIGTYPE_p_Chksum swig_types[1]
#define SWIGTYPE_p_Dataiterator swig_types[2]
#define SWIGTYPE_p_Datamatch swig_types[3]
#define SWIGTYPE_p_Datapos swig_types[4]
#define SWIGTYPE_p_Decision swig_types[5]
#define SWIGTYPE_p_Decisionset swig_types[6]
#define SWIGTYPE_p_Dep swig_types[7]
#define SWIGTYPE_p_FILE swig_types[8]
#define SWIGTYPE_p_Job swig_types[9]
#define SWIGTYPE_p_Pool swig_types[10]
#define SWIGTYPE_p_Pool_repo_iterator swig_types[11]
#define SWIGTYPE_p_Pool_solvable_iterator swig_types[12]
#define SWIGTYPE_p_Problem swig_types[13]
#define SWIGTYPE_p_Queue swig_types[14]
#define SWIGTYPE_p_Repo swig_types[15]
#define SWIGTYPE_p_Repo_solvable_iterator swig_types[16]
#define SWIGTYPE_p_Ruleinfo swig_types[17]
#define SWIGTYPE_p_Selection swig_types[18]
#define SWIGTYPE_p_Solution swig_types[19]
#define SWIGTYPE_p_Solutionelement swig_types[20]
#define SWIGTYPE_p_SolvFp swig_types[21]
#define SWIGTYPE_p_Solver swig_types[22]
#define SWIGTYPE_p_Solvsig swig_types[23]
#define SWIGTYPE_p_Transaction swig_types[24]
#define SWIGTYPE_p_TransactionClass swig_types[25]
#define SWIGTYPE_p_XRepodata swig_types[26]
#define SWIGTYPE_p_XRule swig_types[27]
#define SWIGTYPE_p_XSolvable swig_types[28]
#define SWIGTYPE_p_char swig_types[29]
#define SWIGTYPE_p_int swig_types[30]
#define SWIGTYPE_p_p_XRule swig_types[31]
#define SWIGTYPE_p_unsigned_char swig_types[32]
#define SWIGTYPE_p_unsigned_int swig_types[33]
static swig_type_info *swig_types[35];
static swig_module_info swig_module = {swig_types, 34, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _solv.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__solv

#else
#  define SWIG_init    init_solv

#endif

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


typedef struct {
  const void *data;
  size_t len;
} BinaryBlob;


#include <stdbool.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

/* argh, swig undefs bool for perl */
#ifndef bool
#if !defined __STDC_VERSION__ || __STDC_VERSION__ < 202311L
typedef int bool;
#endif
#endif

#include "pool.h"
#include "poolarch.h"
#include "evr.h"
#include "solver.h"
#include "policy.h"
#include "solverdebug.h"
#include "repo_solv.h"
#include "chksum.h"
#include "selection.h"

#include "repo_write.h"
#if defined(ENABLE_RPMDB) || defined(ENABLE_RPMPKG)
#include "repo_rpmdb.h"
#endif
#ifdef ENABLE_PUBKEY
#include "repo_pubkey.h"
#endif
#ifdef ENABLE_DEBIAN
#include "repo_deb.h"
#endif
#ifdef ENABLE_RPMMD
#include "repo_rpmmd.h"
#include "repo_updateinfoxml.h"
#include "repo_deltainfoxml.h"
#include "repo_repomdxml.h"
#endif
#ifdef ENABLE_SUSEREPO
#include "repo_products.h"
#include "repo_susetags.h"
#include "repo_content.h"
#endif
#ifdef ENABLE_MDKREPO
#include "repo_mdk.h"
#endif
#ifdef ENABLE_ARCHREPO
#include "repo_arch.h"
#endif
#ifdef SUSE
#include "repo_autopattern.h"
#endif
#if defined(ENABLE_COMPLEX_DEPS) && (defined(ENABLE_SUSEREPO) || defined(ENABLE_RPMMD) || defined(ENABLE_RPMDB) || defined(ENABLE_RPMPKG))
#include "pool_parserpmrichdep.h"
#endif
#include "solv_xfopen.h"
#include "testcase.h"

/* for old ruby versions */
#ifndef RARRAY_PTR
#define RARRAY_PTR(ary) (RARRAY(ary)->ptr)
#endif
#ifndef RARRAY_LEN
#define RARRAY_LEN(ary) (RARRAY(ary)->len)
#endif

typedef void *AppObjectPtr;
typedef Id DepId;

typedef struct {
  Pool *pool;
  Id id;
} Dep;

typedef struct {
  Pool *pool;
  Id id;
} XSolvable;

typedef struct {
  Solver *solv;
  Id id;
} XRule;

typedef struct {
  Repo *repo;
  Id id;
} XRepodata;

typedef struct {
  Pool *pool;
  Id id;
} Pool_solvable_iterator;

typedef struct {
  Pool *pool;
  Id id;
} Pool_repo_iterator;

typedef struct {
  Repo *repo;
  Id id;
} Repo_solvable_iterator;

typedef struct {
  Pool *pool;
  int how;
  Id what;
} Job;

typedef struct {
  Solver *solv;
  Id id;
} Problem;

typedef struct {
  Solver *solv;
  Id problemid;
  Id id;
} Solution;

typedef struct {
  Solver *solv;
  Id problemid;
  Id solutionid;

  Id type;
  Id p;
  Id rp;
} Solutionelement;

typedef struct {
  Solver *solv;
  Id rid;
  int type;
  Id source;
  Id target;
  Id dep_id;
} Ruleinfo;

typedef struct {
  Solver *solv;
  Id type;
  Id rid;
  Id from_id;
  Id dep_id;
  Id chosen_id;
  Queue choices;
  int level;
} Alternative;

typedef struct {
  Transaction *transaction;
  int mode;
  Id type;
  int count;
  Id fromid;
  Id toid;
} TransactionClass;

typedef struct {
  Pool *pool;
  Queue q;
  int flags;
} Selection;

typedef struct {
  Solver *solv;
  Id p;
  int reason;
  Id infoid;
} Decision;

typedef struct {
  Solver *solv;
  Queue decisionlistq;
  Id p;
  int reason;
  Id infoid;
  int bits;
  int type;
  Id source;
  Id target;
  Id dep_id;
} Decisionset;

typedef struct {
  FILE *fp;
} SolvFp;

typedef Dataiterator Datamatch;

typedef int disown_helper;

struct myappdata {
  void *appdata;
  int disowned;
};

/* special internal decisionset constructor from a prepared decisionlist */
static Decisionset *decisionset_fromids(Solver *solv, Id *ids, int cnt)
{
  Decisionset *d = solv_calloc(1, sizeof(*d));
  int i;
  d->solv = solv;
  queue_init(&d->decisionlistq);
  d->p = ids[0];
  d->reason = ids[1];
  d->infoid = ids[2];
  d->bits = ids[3];
  d->type = ids[4];
  d->source = ids[5];
  d->target = ids[6];
  d->dep_id = ids[7];
  for (i = 0; i < cnt; i += 8)
    queue_insertn(&d->decisionlistq, d->decisionlistq.count, 3, ids + i);
  if (cnt > 8)
    d->infoid = 0;
  return d;
}

/* prepare a decisionlist so we can feed it to decisionset_fromids */
static void prepare_decisionset_queue(Solver *solv, Queue *q) {
  int i, cnt;
  for (i = cnt = 0; i < q->count; cnt++)
    {
      i += 1 + 8 + 8 * solver_decisionlist_merged(solv, q, i);  /* +1 as we insert one element */
      queue_insert(q, cnt, i - cnt);
    }
  if (cnt)
    queue_unshift(q, 1);        /* start of first block */
  for (i = 0; i < cnt; i++)
    q->elements[i] += cnt - i;
  q->count = cnt;   /* hack */
}



SWIGINTERN void appdata_disown_helper(void *appdata) {
  struct myappdata *myappdata = appdata;
  if (!myappdata || !myappdata->appdata || myappdata->disowned)
    return;
  myappdata->disowned = 1;
  Py_DECREF((PyObject *)myappdata->appdata);
}
SWIGINTERN void appdata_clr_helper(void **appdatap) {
  struct myappdata *myappdata = *(struct myappdata **)appdatap;
  if (myappdata && myappdata->appdata && !myappdata->disowned) {
    Py_DECREF((PyObject *)myappdata->appdata);
  }
  *appdatap = solv_free(myappdata);
}
SWIGINTERN void appdata_set_helper(void **appdatap, void *appdata) {
  appdata_clr_helper(appdatap);
  if (appdata) {
    struct myappdata *myappdata = *appdatap = solv_calloc(sizeof(struct myappdata), 1);
    myappdata->appdata = appdata;
  }
}
SWIGINTERN void *appdata_get_helper(void **appdatap) {
  struct myappdata *myappdata = *(struct myappdata **)appdatap;
  return myappdata ? myappdata->appdata : 0;
}



SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}

SWIGINTERN Selection *new_Selection(Pool *pool){
    Selection *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN void delete_Selection(Selection *self){
    queue_free(&self->q);
    solv_free(self);
  }
SWIGINTERN bool Selection_isempty(Selection *self){
    return self->q.count == 0;
  }

SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <errno.h>


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than LONG_MAX (not portably calculated easily)
      // Note that double(LONG_MAX) is stored in a double rounded up by one (for 64-bit long)
      // 0x7ffffffffffffc00LL == (int64_t)std::nextafter(double(__uint128_t(LONG_MAX)+1), double(0))
      const double long_max = sizeof(long) == 8 ? 0x7ffffffffffffc00LL : LONG_MAX;
      // No equivalent needed for 64-bit double(LONG_MIN) is exactly LONG_MIN
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, long_max)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}

SWIGINTERN Selection *Selection_clone(Selection *self,int flags){
    Selection *s = new_Selection(self->pool);
    queue_init_clone(&s->q, &self->q);
    s->flags = self->flags;
    return s;
  }
SWIGINTERN void Selection_filter(Selection *self,Selection *lsel){
    if (self->pool != lsel->pool)
      queue_empty(&self->q);
    else
      selection_filter(self->pool, &self->q, &lsel->q);
  }
SWIGINTERN void Selection_add(Selection *self,Selection *lsel){
    if (self->pool == lsel->pool)
      {
        selection_add(self->pool, &self->q, &lsel->q);
        self->flags |= lsel->flags;
      }
  }
SWIGINTERN void Selection_add_raw(Selection *self,Id how,Id what){
    queue_push2(&self->q, how, what);
  }
SWIGINTERN void Selection_subtract(Selection *self,Selection *lsel){
    if (self->pool == lsel->pool)
      selection_subtract(self->pool, &self->q, &lsel->q);
  }

SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


/* Return string from Python obj. NOTE: obj must remain in scope in order
   to use the returned cptr (but only when alloc is set to SWIG_OLDOBJ) */
SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char **cptr, size_t *psize, int *alloc)
{
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    PyObject *bytes = NULL;
    int ret = SWIG_OK;
    if (alloc)
      *alloc = SWIG_OLDOBJ;
#if PY_VERSION_HEX >= 0x03000000 && defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    cstr = (char *)SWIG_PyUnicode_AsUTF8AndSize(obj, &len, &bytes);
    if (!cstr)
      return SWIG_TypeError;
    /* The returned string is only duplicated if the char * returned is not owned and memory managed by obj */
    if (bytes && cptr) {
      if (alloc) {
        cstr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        *alloc = SWIG_NEWOBJ;
      } else {
        /* alloc must be set in order to clean up allocated memory */
        return SWIG_RuntimeError;
      }
    }
#endif
    if (cptr) *cptr = cstr;
    if (psize) *psize = len + 1;
    SWIG_Py_XDECREF(bytes);
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX < 0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = (char *)memcpy(malloc((len + 1)*sizeof(char)), cstr, sizeof(char)*(len + 1));
        }
        if (psize) *psize = len + 1;

        SWIG_Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        SWIG_Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}




SWIGINTERN void Selection_select(Selection *self,char const *name,int flags){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make(self->pool, &self->q, name, flags);
  }
SWIGINTERN void Selection_matchdeps(Selection *self,char const *name,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchdeps(self->pool, &self->q, name, flags, keyname, marker);
  }


SWIGINTERN int







SWIG_AsValDepId(void *obj, int *val) {

  static swig_type_info* desc = 0;
  void *vptr = 0;
  int ecode;
  if (!desc) desc = SWIG_TypeQuery("Dep *");





  ecode = SWIG_AsVal_int(obj, val);

  if (SWIG_IsOK(ecode))
    return ecode;



  if ((SWIG_ConvertPtr(obj, &vptr, desc, 0)) == SWIG_OK) {

    if (val)
      *val = vptr ? ((Dep *)vptr)->id : 0;
    return SWIG_OK;
  }
  return SWIG_TypeError;
}





SWIGINTERN void Selection_matchdepid(Selection *self,DepId dep,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchdepid(self->pool, &self->q, dep, flags, keyname, marker);
  }
SWIGINTERN void Selection_matchsolvable(Selection *self,XSolvable *solvable,int flags,Id keyname,Id marker){
    if ((flags & SELECTION_MODEBITS) == 0)
      flags |= SELECTION_FILTER | SELECTION_WITH_ALL;
    self->flags = selection_make_matchsolvable(self->pool, &self->q, solvable->id, flags, keyname, marker);
  }
SWIGINTERN Queue Selection_jobs(Selection *self,int flags){
    Queue q;
    int i;
    queue_init_clone(&q, &self->q);
    for (i = 0; i < q.count; i += 2)
      q.elements[i] |= flags;
    return q;
  }
SWIGINTERN Queue Selection_solvables(Selection *self){
    Queue q;
    queue_init(&q);
    selection_solvables(self->pool, &self->q, &q);
    return q;
  }
SWIGINTERN char const *Selection___str__(Selection *self){
    return pool_selection2str(self->pool, &self->q, 0);
  }

SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > (size_t)PY_SSIZE_T_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj((char *)(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, (Py_ssize_t)(size));
#else
      return PyUnicode_DecodeUTF8(carray, (Py_ssize_t)(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, (Py_ssize_t)(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}

SWIGINTERN char const *Selection___repr__(Selection *self){
    const char *str = pool_selection2str(self->pool, &self->q, ~0);
    return pool_tmpjoin(self->pool, "<Selection ", str, ">");
  }
SWIGINTERN Dep *new_Dep(Pool *pool,Id id){
    Dep *s;
    if (!id)
      return 0;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    s->id = id;
    return s;
  }

SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  int r = PyObject_IsTrue(obj);
  if (r == -1)
    return SWIG_ERROR;
  if (val) *val = r ? true : false;
  return SWIG_OK;
}

SWIGINTERN Dep *Dep_Rel(Dep *self,int flags,DepId evrid,bool create){
    Id id = pool_rel2id(self->pool, self->id, evrid, flags, create);
    if (!id)
      return 0;
    return new_Dep(self->pool, id);
  }
SWIGINTERN Selection *Dep_Selection_name(Dep *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    if (ISRELDEP(self->id)) {
      Reldep *rd = GETRELDEP(self->pool, self->id);
      if (rd->flags == REL_EQ) {
        setflags |= self->pool->disttype == DISTTYPE_DEB || strchr(pool_id2str(self->pool, rd->evr), '-') != 0 ? SOLVER_SETEVR : SOLVER_SETEV;
        if (ISRELDEP(rd->name))
          rd = GETRELDEP(self->pool, rd->name);
      }
      if (rd->flags == REL_ARCH)
        setflags |= SOLVER_SETARCH;
    }
    queue_push2(&sel->q, SOLVER_SOLVABLE_NAME | setflags, self->id);
    return sel;
  }
SWIGINTERN Selection *Dep_Selection_provides(Dep *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    if (ISRELDEP(self->id)) {
      Reldep *rd = GETRELDEP(self->pool, self->id);
      if (rd->flags == REL_ARCH)
        setflags |= SOLVER_SETARCH;
    }
    queue_push2(&sel->q, SOLVER_SOLVABLE_PROVIDES | setflags, self->id);
    return sel;
  }
SWIGINTERN char const *Dep_str(Dep *self){
    return pool_dep2str(self->pool, self->id);
  }
SWIGINTERN bool Dep___eq__(Dep *self,Dep *s){
    return self->pool == s->pool && self->id == s->id;
  }
SWIGINTERN bool Dep___ne__(Dep *self,Dep *s){
    return !Dep___eq__(self, s);
  }
SWIGINTERN int Dep___hash__(Dep *self){
    return self->id;
  }
SWIGINTERN char const *Dep___str__(Dep *self){
    return pool_dep2str(self->pool, self->id);
  }
SWIGINTERN char const *Dep___repr__(Dep *self){
    char buf[20];
    sprintf(buf, "<Id #%d ", self->id);
    return solv_dupjoin(buf, pool_dep2str(self->pool, self->id), ">");
  }
SWIGINTERN Dataiterator *new_Dataiterator(Pool *pool,Repo *repo,Id p,Id key,char const *match,int flags){
    Dataiterator *di = solv_calloc(1, sizeof(*di));
    dataiterator_init(di, pool, repo, p, key, match, flags);
    return di;
  }
SWIGINTERN void delete_Dataiterator(Dataiterator *self){
    dataiterator_free(self);
    solv_free(self);
  }
SWIGINTERN Datamatch *Dataiterator___next__(Dataiterator *self){

    Dataiterator *ndi;
    if (!dataiterator_step(self)) {
      return 0;
    }
    ndi = solv_calloc(1, sizeof(*ndi));
    dataiterator_init_clone(ndi, self);
    dataiterator_strdup(ndi);
    return ndi;
  }
SWIGINTERN void Dataiterator_prepend_keyname(Dataiterator *self,Id key){
    dataiterator_prepend_keyname(self, key);
  }
SWIGINTERN void Dataiterator_skip_solvable(Dataiterator *self){
    dataiterator_skip_solvable(self);
  }
SWIGINTERN XSolvable *new_XSolvable(Pool *pool,Id id){
    XSolvable *s;
    if (!id || id >= pool->nsolvables)
      return 0;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    s->id = id;
    return s;
  }
SWIGINTERN char const *XSolvable_str(XSolvable *self){
    return pool_solvid2str(self->pool, self->id);
  }
SWIGINTERN char const *XSolvable_lookup_str(XSolvable *self,Id keyname){
    return pool_lookup_str(self->pool, self->id, keyname);
  }
SWIGINTERN Id XSolvable_lookup_id(XSolvable *self,Id keyname){
    return pool_lookup_id(self->pool, self->id, keyname);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      // Largest double not larger than ULONG_MAX (not portably calculated easily)
      // Note that double(ULONG_MAX) is stored in a double rounded up by one (for 64-bit unsigned long)
      // 0xfffffffffffff800ULL == (uint64_t)std::nextafter(double(__uint128_t(ULONG_MAX)+1), double(0))
      const double ulong_max = sizeof(unsigned long) == 8 ? 0xfffffffffffff800ULL : ULONG_MAX;
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ulong_max)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, 0, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif

SWIGINTERN unsigned long long XSolvable_lookup_num(XSolvable *self,Id keyname,unsigned long long notfound){
    return pool_lookup_num(self->pool, self->id, keyname, notfound);
  }

#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyInt_FromLong((long)(value));
}
#endif

SWIGINTERN bool XSolvable_lookup_void(XSolvable *self,Id keyname){
    return pool_lookup_void(self->pool, self->id, keyname);
  }
SWIGINTERN Chksum *XSolvable_lookup_checksum(XSolvable *self,Id keyname){
    Id type = 0;
    const unsigned char *b = pool_lookup_bin_checksum(self->pool, self->id, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN Queue XSolvable_lookup_idarray(XSolvable *self,Id keyname,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    Queue r;
    queue_init(&r);
    solvable_lookup_deparray(s, keyname, &r, marker);
    return r;
  }
SWIGINTERN Queue XSolvable_lookup_deparray(XSolvable *self,Id keyname,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    Queue r;
    queue_init(&r);
    solvable_lookup_deparray(s, keyname, &r, marker);
    return r;
  }

SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}

SWIGINTERN char const *XSolvable_lookup_location(XSolvable *self,unsigned int *OUTPUT){
    return solvable_lookup_location(self->pool->solvables + self->id, OUTPUT);
  }
SWIGINTERN char const *XSolvable_lookup_sourcepkg(XSolvable *self){
    return solvable_lookup_sourcepkg(self->pool->solvables + self->id);
  }
SWIGINTERN Dataiterator *XSolvable_Dataiterator(XSolvable *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, 0, self->id, key, match, flags);
  }
SWIGINTERN bool XSolvable_installable(XSolvable *self){
    return pool_installable(self->pool, pool_id2solvable(self->pool, self->id));
  }
SWIGINTERN bool XSolvable_isinstalled(XSolvable *self){
    Pool *pool = self->pool;
    return pool->installed && pool_id2solvable(pool, self->id)->repo == pool->installed;
  }

    SWIGINTERN void XSolvable_name_set(XSolvable *xs, const char *name) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].name = pool_str2id(pool, name, 1);
    }
    SWIGINTERN const char *XSolvable_name_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].name);
    }
  

    SWIGINTERN void XSolvable_nameid_set(XSolvable *xs, Id nameid) {
      xs->pool->solvables[xs->id].name = nameid;
    }
    SWIGINTERN Id XSolvable_nameid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].name;
    }
  

    SWIGINTERN void XSolvable_evr_set(XSolvable *xs, const char *evr) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].evr = pool_str2id(pool, evr, 1);
    }
    SWIGINTERN const char *XSolvable_evr_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].evr);
    }
  

    SWIGINTERN void XSolvable_evrid_set(XSolvable *xs, Id evrid) {
      xs->pool->solvables[xs->id].evr = evrid;
    }
    SWIGINTERN Id XSolvable_evrid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].evr;
    }
  

    SWIGINTERN void XSolvable_arch_set(XSolvable *xs, const char *arch) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].arch = pool_str2id(pool, arch, 1);
    }
    SWIGINTERN const char *XSolvable_arch_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].arch);
    }
  

    SWIGINTERN void XSolvable_archid_set(XSolvable *xs, Id archid) {
      xs->pool->solvables[xs->id].arch = archid;
    }
    SWIGINTERN Id XSolvable_archid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].arch;
    }
  

    SWIGINTERN void XSolvable_vendor_set(XSolvable *xs, const char *vendor) {
      Pool *pool = xs->pool;
      pool->solvables[xs->id].vendor = pool_str2id(pool, vendor, 1);
    }
    SWIGINTERN const char *XSolvable_vendor_get(XSolvable *xs) {
      Pool *pool = xs->pool;
      return pool_id2str(pool, pool->solvables[xs->id].vendor);
    }
  

    SWIGINTERN void XSolvable_vendorid_set(XSolvable *xs, Id vendorid) {
      xs->pool->solvables[xs->id].vendor = vendorid;
    }
    SWIGINTERN Id XSolvable_vendorid_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].vendor;
    }
  

    SWIGINTERN Repo *XSolvable_repo_get(XSolvable *xs) {
      return xs->pool->solvables[xs->id].repo;
    }
  
SWIGINTERN void XSolvable_add_provides(XSolvable *self,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    marker = solv_depmarker(SOLVABLE_PROVIDES, marker);
    s->provides = repo_addid_dep(s->repo, s->provides, id, marker);
  }
SWIGINTERN void XSolvable_add_obsoletes(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->obsoletes = repo_addid_dep(s->repo, s->obsoletes, id, 0);
  }
SWIGINTERN void XSolvable_add_conflicts(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->conflicts = repo_addid_dep(s->repo, s->conflicts, id, 0);
  }
SWIGINTERN void XSolvable_add_requires(XSolvable *self,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    marker = solv_depmarker(SOLVABLE_REQUIRES, marker);
    s->requires = repo_addid_dep(s->repo, s->requires, id, marker);
  }
SWIGINTERN void XSolvable_add_recommends(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->recommends = repo_addid_dep(s->repo, s->recommends, id, 0);
  }
SWIGINTERN void XSolvable_add_suggests(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->suggests = repo_addid_dep(s->repo, s->suggests, id, 0);
  }
SWIGINTERN void XSolvable_add_supplements(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->supplements = repo_addid_dep(s->repo, s->supplements, id, 0);
  }
SWIGINTERN void XSolvable_add_enhances(XSolvable *self,DepId id){
    Solvable *s = self->pool->solvables + self->id;
    s->enhances = repo_addid_dep(s->repo, s->enhances, id, 0);
  }
SWIGINTERN void XSolvable_unset(XSolvable *self,Id keyname){
    Solvable *s = self->pool->solvables + self->id;
    repo_unset(s->repo, self->id, keyname);
  }
SWIGINTERN void XSolvable_add_deparray(XSolvable *self,Id keyname,DepId id,Id marker){
    Solvable *s = self->pool->solvables + self->id;
    solvable_add_deparray(s, keyname, id, marker);
  }
SWIGINTERN Selection *XSolvable_Selection(XSolvable *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    queue_push2(&sel->q, SOLVER_SOLVABLE | setflags, self->id);
    return sel;
  }
SWIGINTERN bool XSolvable_identical(XSolvable *self,XSolvable *s2){
    return solvable_identical(self->pool->solvables + self->id, s2->pool->solvables + s2->id);
  }
SWIGINTERN int XSolvable_evrcmp(XSolvable *self,XSolvable *s2){
    return pool_evrcmp(self->pool, self->pool->solvables[self->id].evr, s2->pool->solvables[s2->id].evr, EVRCMP_COMPARE);
  }
SWIGINTERN bool XSolvable_matchesdep(XSolvable *self,Id keyname,DepId id,Id marker){
    return solvable_matchesdep(self->pool->solvables + self->id, keyname, id, marker);
  }
SWIGINTERN bool XSolvable___eq__(XSolvable *self,XSolvable *s){
    return self->pool == s->pool && self->id == s->id;
  }
SWIGINTERN bool XSolvable___ne__(XSolvable *self,XSolvable *s){
    return !XSolvable___eq__(self, s);
  }
SWIGINTERN int XSolvable___hash__(XSolvable *self){
    return self->id;
  }
SWIGINTERN char const *XSolvable___str__(XSolvable *self){
    return pool_solvid2str(self->pool, self->id);
  }
SWIGINTERN char const *XSolvable___repr__(XSolvable *self){
    char buf[20];
    sprintf(buf, "<Solvable #%d ", self->id);
    return solv_dupjoin(buf, pool_solvid2str(self->pool, self->id), ">");
  }
SWIGINTERN Ruleinfo *new_Ruleinfo(Solver *solv,Id rid,Id type,Id source,Id target,Id dep_id){
    Ruleinfo *ri = solv_calloc(1, sizeof(*ri));
    ri->solv = solv;
    ri->rid = rid;
    ri->type = type;
    ri->source = source;
    ri->target = target;
    ri->dep_id = dep_id;
    return ri;
  }

    SWIGINTERN XSolvable *Ruleinfo_solvable_get(Ruleinfo *ri) {
      return new_XSolvable(ri->solv->pool, ri->source);
    }
    SWIGINTERN XSolvable *Ruleinfo_othersolvable_get(Ruleinfo *ri) {
      return new_XSolvable(ri->solv->pool, ri->target);
    }
    SWIGINTERN Dep *Ruleinfo_dep_get(Ruleinfo *ri) {
      return new_Dep(ri->solv->pool, ri->dep_id);
    }
  
SWIGINTERN char const *Ruleinfo_problemstr(Ruleinfo *self){
    return solver_problemruleinfo2str(self->solv, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN char const *Ruleinfo___str__(Ruleinfo *self){
    return solver_ruleinfo2str(self->solv, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN XRule *new_XRule(Solver *solv,Id id){
    if (!id)
      return 0;
    XRule *xr = solv_calloc(1, sizeof(*xr));
    xr->solv = solv;
    xr->id = id;
    return xr;
  }

    SWIGINTERN int XRule_type_get(XRule *xr) {
      return solver_ruleclass(xr->solv, xr->id);
    }
  
SWIGINTERN Ruleinfo *XRule_info(XRule *self){
    Id type, source, target, dep;
    type = solver_ruleinfo(self->solv, self->id, &source, &target, &dep);
    return new_Ruleinfo(self->solv, self->id, type, source, target, dep);
  }
SWIGINTERN Queue XRule_allinfos(XRule *self){
    Queue q;
    queue_init(&q);
    solver_allruleinfos(self->solv, self->id, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_learnt(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_decisionlist(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE | SOLVER_DECISIONLIST_SORTED, &q);
    return q;
  }
SWIGINTERN Queue XRule_get_decisionsetlist(XRule *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_LEARNTRULE | SOLVER_DECISIONLIST_SORTED | SOLVER_DECISIONLIST_WITHINFO | SOLVER_DECISIONLIST_MERGEDINFO, &q);
    prepare_decisionset_queue(self->solv, &q);
    return q;
  }
SWIGINTERN bool XRule___eq__(XRule *self,XRule *xr){
    return self->solv == xr->solv && self->id == xr->id;
  }
SWIGINTERN bool XRule___ne__(XRule *self,XRule *xr){
    return !XRule___eq__(self, xr);
  }
SWIGINTERN int XRule___hash__(XRule *self){
    return self->id;
  }
SWIGINTERN char const *XRule___repr__(XRule *self){
    char buf[20];
    sprintf(buf, "<Rule #%d>", self->id);
    return solv_strdup(buf);
  }
SWIGINTERN XRepodata *new_XRepodata(Repo *repo,Id id){
    XRepodata *xr = solv_calloc(1, sizeof(*xr));
    xr->repo = repo;
    xr->id = id;
    return xr;
  }
SWIGINTERN Id XRepodata_new_handle(XRepodata *self){
    return repodata_new_handle(repo_id2repodata(self->repo, self->id));
  }
SWIGINTERN void XRepodata_set_id(XRepodata *self,Id solvid,Id keyname,DepId id){
    repodata_set_id(repo_id2repodata(self->repo, self->id), solvid, keyname, id);
  }
SWIGINTERN void XRepodata_set_num(XRepodata *self,Id solvid,Id keyname,unsigned long long num){
    repodata_set_num(repo_id2repodata(self->repo, self->id), solvid, keyname, num);
  }
SWIGINTERN void XRepodata_set_str(XRepodata *self,Id solvid,Id keyname,char const *str){
    repodata_set_str(repo_id2repodata(self->repo, self->id), solvid, keyname, str);
  }
SWIGINTERN void XRepodata_set_void(XRepodata *self,Id solvid,Id keyname){
    repodata_set_void(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN void XRepodata_set_poolstr(XRepodata *self,Id solvid,Id keyname,char const *str){
    repodata_set_poolstr(repo_id2repodata(self->repo, self->id), solvid, keyname, str);
  }
SWIGINTERN void XRepodata_add_idarray(XRepodata *self,Id solvid,Id keyname,DepId id){
    repodata_add_idarray(repo_id2repodata(self->repo, self->id), solvid, keyname, id);
  }
SWIGINTERN void XRepodata_add_flexarray(XRepodata *self,Id solvid,Id keyname,Id handle){
    repodata_add_flexarray(repo_id2repodata(self->repo, self->id), solvid, keyname, handle);
  }
SWIGINTERN void XRepodata_set_checksum(XRepodata *self,Id solvid,Id keyname,Chksum *chksum){
    const unsigned char *buf = solv_chksum_get(chksum, 0);
    if (buf)
      repodata_set_bin_checksum(repo_id2repodata(self->repo, self->id), solvid, keyname, solv_chksum_get_type(chksum), buf);
  }
SWIGINTERN void XRepodata_set_sourcepkg(XRepodata *self,Id solvid,char const *sourcepkg){
    repodata_set_sourcepkg(repo_id2repodata(self->repo, self->id), solvid, sourcepkg);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}

SWIGINTERN void XRepodata_set_location(XRepodata *self,Id solvid,unsigned int mediano,char const *location){
    repodata_set_location(repo_id2repodata(self->repo, self->id), solvid, mediano, 0, location);
  }
SWIGINTERN void XRepodata_unset(XRepodata *self,Id solvid,Id keyname){
    repodata_unset(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN char const *XRepodata_lookup_str(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_str(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN Id XRepodata_lookup_id(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_id(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN unsigned long long XRepodata_lookup_num(XRepodata *self,Id solvid,Id keyname,unsigned long long notfound){
    return repodata_lookup_num(repo_id2repodata(self->repo, self->id), solvid, keyname, notfound);
  }
SWIGINTERN bool XRepodata_lookup_void(XRepodata *self,Id solvid,Id keyname){
    return repodata_lookup_void(repo_id2repodata(self->repo, self->id), solvid, keyname);
  }
SWIGINTERN Queue XRepodata_lookup_idarray(XRepodata *self,Id solvid,Id keyname){
    Queue r;
    queue_init(&r);
    repodata_lookup_idarray(repo_id2repodata(self->repo, self->id), solvid, keyname, &r);
    return r;
  }
SWIGINTERN Chksum *XRepodata_lookup_checksum(XRepodata *self,Id solvid,Id keyname){
    Id type = 0;
    const unsigned char *b = repodata_lookup_bin_checksum(repo_id2repodata(self->repo, self->id), solvid, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN void XRepodata_internalize(XRepodata *self){
    repodata_internalize(repo_id2repodata(self->repo, self->id));
  }
SWIGINTERN void XRepodata_create_stubs(XRepodata *self){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    data = repodata_create_stubs(data);
    self->id = data->repodataid;
  }

SWIGINTERN int
SWIG_AsValFilePtr(PyObject *obj, FILE **val) {
  static swig_type_info* desc = 0;
  void *vptr = 0;
  if (!desc) desc = SWIG_TypeQuery("FILE *");
  if ((SWIG_ConvertPtr(obj, &vptr, desc, 0)) == SWIG_OK) {
    if (val) *val = (FILE *)vptr;
    return SWIG_OK;
  }
#if PY_VERSION_HEX < 0x03000000
  if (PyFile_Check(obj)) {
    if (val) *val =  PyFile_AsFile(obj);
    return SWIG_OK;
  }
#endif
  return SWIG_TypeError;
}



SWIGINTERN int







SWIG_AsValSolvFpPtr(void *obj, FILE **val) {

  static swig_type_info* desc = 0;
  void *vptr = 0;

  int ecode;


  if (!desc) desc = SWIG_TypeQuery("SolvFp *");



  if ((SWIG_ConvertPtr(obj, &vptr, desc, 0)) == SWIG_OK) {

    if (val)
      *val = vptr ? ((SolvFp *)vptr)->fp : 0;
    return SWIG_OK;
  }

  ecode = SWIG_AsValFilePtr(obj, val);
  if (ecode == SWIG_OK)
    return ecode;

  return SWIG_TypeError;
}






SWIGINTERN bool XRepodata_write(XRepodata *self,FILE *fp){
    return repodata_write(repo_id2repodata(self->repo, self->id), fp) == 0;
  }
SWIGINTERN Id XRepodata_str2dir(XRepodata *self,char const *dir,bool create){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    return repodata_str2dir(data, dir, create);
  }
SWIGINTERN char const *XRepodata_dir2str(XRepodata *self,Id did,char const *suf){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    return repodata_dir2str(data, did, suf);
  }
SWIGINTERN void XRepodata_add_dirstr(XRepodata *self,Id solvid,Id keyname,Id dir,char const *str){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    repodata_add_dirstr(data, solvid, keyname, dir, str);
  }
SWIGINTERN bool XRepodata_add_solv(XRepodata *self,FILE *fp,int flags){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    int r, oldstate = data->state;
    data->state = REPODATA_LOADING;
    r = repo_add_solv(data->repo, fp, flags | REPO_USE_LOADING);
    if (r || data->state == REPODATA_LOADING)
      data->state = oldstate;
    return r == 0;
  }
SWIGINTERN void XRepodata_extend_to_repo(XRepodata *self){
    Repodata *data = repo_id2repodata(self->repo, self->id);
    repodata_extend_block(data, data->repo->start, data->repo->end - data->repo->start);
  }
SWIGINTERN bool XRepodata___eq__(XRepodata *self,XRepodata *xr){
    return self->repo == xr->repo && self->id == xr->id;
  }
SWIGINTERN bool XRepodata___ne__(XRepodata *self,XRepodata *xr){
    return !XRepodata___eq__(self, xr);
  }
SWIGINTERN int XRepodata___hash__(XRepodata *self){
    return self->id;
  }
SWIGINTERN char const *XRepodata___repr__(XRepodata *self){
    char buf[20];
    sprintf(buf, "<Repodata #%d>", self->id);
    return solv_strdup(buf);
  }
SWIGINTERN Pool_solvable_iterator *new_Pool_solvable_iterator(Pool *pool){
    Pool_solvable_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN XSolvable *Pool_solvable_iterator___next__(Pool_solvable_iterator *self){

    Pool *pool = self->pool;
    if (self->id >= pool->nsolvables)
      return 0;
    while (++self->id < pool->nsolvables)
      if (pool->solvables[self->id].repo)
        return new_XSolvable(pool, self->id);
    return 0;
  }
SWIGINTERN XSolvable *Pool_solvable_iterator___getitem__(Pool_solvable_iterator *self,Id key){
    Pool *pool = self->pool;
    if (key > 0 && key < pool->nsolvables && pool->solvables[key].repo)
      return new_XSolvable(pool, key);
    return 0;
  }
SWIGINTERN int Pool_solvable_iterator___len__(Pool_solvable_iterator *self){
    return self->pool->nsolvables;
  }
SWIGINTERN Pool_repo_iterator *new_Pool_repo_iterator(Pool *pool){
    Pool_repo_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->pool = pool;
    return s;
  }
SWIGINTERN Repo *Pool_repo_iterator___next__(Pool_repo_iterator *self){

    Pool *pool = self->pool;
    if (self->id >= pool->nrepos)
      return 0;
    while (++self->id < pool->nrepos) {
      Repo *r = pool_id2repo(pool, self->id);
      if (r)
        return r;
    }
    return 0;
  }
SWIGINTERN Repo *Pool_repo_iterator___getitem__(Pool_repo_iterator *self,Id key){
    Pool *pool = self->pool;
    if (key > 0 && key < pool->nrepos)
      return pool_id2repo(pool, key);
    return 0;
  }
SWIGINTERN int Pool_repo_iterator___len__(Pool_repo_iterator *self){
    return self->pool->nrepos;
  }
SWIGINTERN Repo_solvable_iterator *new_Repo_solvable_iterator(Repo *repo){
    Repo_solvable_iterator *s;
    s = solv_calloc(1, sizeof(*s));
    s->repo = repo;
    return s;
  }
SWIGINTERN XSolvable *Repo_solvable_iterator___next__(Repo_solvable_iterator *self){

    Repo *repo = self->repo;
    Pool *pool = repo->pool;
    if (repo->start > 0 && self->id < repo->start)
      self->id = repo->start - 1;
    if (self->id >= repo->end)
      return 0;
    while (++self->id < repo->end)
      if (pool->solvables[self->id].repo == repo)
        return new_XSolvable(pool, self->id);
    return 0;
  }
SWIGINTERN XSolvable *Repo_solvable_iterator___getitem__(Repo_solvable_iterator *self,Id key){
    Repo *repo = self->repo;
    Pool *pool = repo->pool;
    if (key > 0 && key < pool->nsolvables && pool->solvables[key].repo == repo)
      return new_XSolvable(pool, key);
    return 0;
  }
SWIGINTERN int Repo_solvable_iterator___len__(Repo_solvable_iterator *self){
    return self->repo->pool->nsolvables;
  }
SWIGINTERN void delete_Datamatch(Datamatch *self){
    dataiterator_free(self);
    solv_free(self);
  }

  SWIGINTERN XSolvable *Datamatch_solvable_get(Dataiterator *di) {
    return new_XSolvable(di->pool, di->solvid);
  }
  SWIGINTERN Id Datamatch_key_id_get(Dataiterator *di) {
    return di->key->name;
  }
  SWIGINTERN const char *Datamatch_key_idstr_get(Dataiterator *di) {
    return pool_id2str(di->pool, di->key->name);
  }
  SWIGINTERN Id Datamatch_type_id_get(Dataiterator *di) {
    return di->key->type;
  }
  SWIGINTERN const char *Datamatch_type_idstr_get(Dataiterator *di) {
    return pool_id2str(di->pool, di->key->type);
  }
  SWIGINTERN Id Datamatch_id_get(Dataiterator *di) {
    return di->kv.id;
  }
  SWIGINTERN const char *Datamatch_idstr_get(Dataiterator *di) {
   if (di->data && (di->key->type == REPOKEY_TYPE_DIR || di->key->type == REPOKEY_TYPE_DIRSTRARRAY || di->key->type == REPOKEY_TYPE_DIRNUMNUMARRAY))
      return repodata_dir2str(di->data,  di->kv.id, 0);
    if (di->data && di->data->localpool)
      return stringpool_id2str(&di->data->spool, di->kv.id);
    return pool_id2str(di->pool, di->kv.id);
  }
  SWIGINTERN const char * const Datamatch_str_get(Dataiterator *di) {
    return di->kv.str;
  }
  SWIGINTERN Dep *Datamatch_dep_get(Dataiterator *di) {
    if (di->key->type == REPOKEY_TYPE_DIR || di->key->type == REPOKEY_TYPE_DIRSTRARRAY || di->key->type == REPOKEY_TYPE_DIRNUMNUMARRAY)
      return 0;
    if (di->data && di->data->localpool)
      return 0;
    return new_Dep(di->pool, di->kv.id);
  }
  SWIGINTERN BinaryBlob Datamatch_binary_get(Dataiterator *di) {
    BinaryBlob bl;
    bl.data = 0;
    bl.len = 0;
    if (di->key->type == REPOKEY_TYPE_BINARY)
      {
        bl.data = di->kv.str;
        bl.len = di->kv.num;
      }
    else if ((bl.len = solv_chksum_len(di->key->type)) != 0)
      bl.data = di->kv.str;
    return bl;
  }
  SWIGINTERN unsigned long long Datamatch_num_get(Dataiterator *di) {
   if (di->key->type == REPOKEY_TYPE_NUM)
     return SOLV_KV_NUM64(&di->kv);
   return di->kv.num;
  }
  SWIGINTERN unsigned int Datamatch_num2_get(Dataiterator *di) {
    return di->kv.num2;
  }
  
SWIGINTERN Datapos *Datamatch_pos(Datamatch *self){
    Pool *pool = self->pool;
    Datapos *pos, oldpos = pool->pos;
    dataiterator_setpos(self);
    pos = solv_calloc(1, sizeof(*pos));
    *pos = pool->pos;
    pool->pos = oldpos;
    return pos;
  }
SWIGINTERN Datapos *Datamatch_parentpos(Datamatch *self){
    Pool *pool = self->pool;
    Datapos *pos, oldpos = pool->pos;
    dataiterator_setpos_parent(self);
    pos = solv_calloc(1, sizeof(*pos));
    *pos = pool->pos;
    pool->pos = oldpos;
    return pos;
  }
SWIGINTERN char const *Datamatch___str__(Datamatch *self){
    KeyValue kv = self->kv;
    const char *str = repodata_stringify(self->pool, self->data, self->key, &kv, SEARCH_FILES | SEARCH_CHECKSUMS);
    return str ? str : "";
  }
SWIGINTERN Id Datapos_lookup_id(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Id r;
    pool->pos = *self;
    r = pool_lookup_id(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN char const *Datapos_lookup_str(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *r;
    pool->pos = *self;
    r = pool_lookup_str(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN unsigned long long Datapos_lookup_num(Datapos *self,Id keyname,unsigned long long notfound){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    unsigned long long r;
    pool->pos = *self;
    r = pool_lookup_num(pool, SOLVID_POS, keyname, notfound);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN bool Datapos_lookup_void(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    int r;
    pool->pos = *self;
    r = pool_lookup_void(pool, SOLVID_POS, keyname);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN Chksum *Datapos_lookup_checksum(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Id type = 0;
    const unsigned char *b;
    pool->pos = *self;
    b = pool_lookup_bin_checksum(pool, SOLVID_POS, keyname, &type);
    pool->pos = oldpos;
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN char const *Datapos_lookup_deltaseq(Datapos *self){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *seq;
    pool->pos = *self;
    seq = pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_NAME);
    if (seq) {
      seq = pool_tmpjoin(pool, seq, "-", pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_EVR));
      seq = pool_tmpappend(pool, seq, "-", pool_lookup_str(pool, SOLVID_POS, DELTA_SEQ_NUM));
    }
    pool->pos = oldpos;
    return seq;
  }
SWIGINTERN char const *Datapos_lookup_deltalocation(Datapos *self,unsigned int *OUTPUT){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    const char *loc;
    pool->pos = *self;
    loc = pool_lookup_deltalocation(pool, SOLVID_POS, OUTPUT);
    pool->pos = oldpos;
    return loc;
  }
SWIGINTERN Queue Datapos_lookup_idarray(Datapos *self,Id keyname){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Queue r;
    queue_init(&r);
    pool->pos = *self;
    pool_lookup_idarray(pool, SOLVID_POS, keyname, &r);
    pool->pos = oldpos;
    return r;
  }
SWIGINTERN Dataiterator *Datapos_Dataiterator(Datapos *self,Id key,char const *match,int flags){
    Pool *pool = self->repo->pool;
    Datapos oldpos = pool->pos;
    Dataiterator *di;
    pool->pos = *self;
    di = new_Dataiterator(pool, 0, SOLVID_POS, key, match, flags);
    pool->pos = oldpos;
    return di;
  }
SWIGINTERN Job *new_Job(Pool *pool,int how,Id what){
    Job *job = solv_calloc(1, sizeof(*job));
    job->pool = pool;
    job->how = how;
    job->what = what;
    return job;
  }
SWIGINTERN Queue Job_solvables(Job *self){
    Queue q;
    queue_init(&q);
    pool_job2solvables(self->pool, &q, self->how, self->what);
    return q;
  }
SWIGINTERN bool Job_isemptyupdate(Job *self){
    return pool_isemptyupdatejob(self->pool, self->how, self->what);
  }
SWIGINTERN bool Job___eq__(Job *self,Job *j){
    return self->pool == j->pool && self->how == j->how && self->what == j->what;
  }
SWIGINTERN bool Job___ne__(Job *self,Job *j){
    return !Job___eq__(self, j);
  }
SWIGINTERN char const *Job___str__(Job *self){
    return pool_job2str(self->pool, self->how, self->what, 0);
  }
SWIGINTERN char const *Job___repr__(Job *self){
    const char *str = pool_job2str(self->pool, self->how, self->what, ~0);
    return pool_tmpjoin(self->pool, "<Job ", str, ">");
  }
SWIGINTERN Pool *new_Pool(void){
    Pool *pool = pool_create();
    return pool;
  }
SWIGINTERN int Pool_setdisttype(Pool *self,int disttype){
    return pool_setdisttype(self, disttype);
  }
SWIGINTERN void Pool_set_debuglevel(Pool *self,int level){
    pool_setdebuglevel(self, level);
  }
SWIGINTERN int Pool_set_flag(Pool *self,int flag,int value){
    return pool_set_flag(self, flag, value);
  }
SWIGINTERN int Pool_get_flag(Pool *self,int flag){
    return pool_get_flag(self, flag);
  }
SWIGINTERN void Pool_set_rootdir(Pool *self,char const *rootdir){
    pool_set_rootdir(self, rootdir);
  }
SWIGINTERN char const *Pool_get_rootdir(Pool *self,int flag){
    return pool_get_rootdir(self);
  }

  SWIGINTERN int loadcallback(Pool *pool, Repodata *data, void *d) {
    XRepodata *xd = new_XRepodata(data->repo, data->repodataid);
    PyObject *args = Py_BuildValue("(O)", SWIG_NewPointerObj(SWIG_as_voidptr(xd), SWIGTYPE_p_XRepodata, SWIG_POINTER_OWN | 0));
    PyObject *result = PyObject_Call((PyObject *)d, args, NULL);

    int ecode = 0;
    int vresult = 0;
    Py_DECREF(args);
    if (!result)
      return 0; /* exception */
    ecode = SWIG_AsVal_int(result, &vresult);
    Py_DECREF(result);
    return SWIG_IsOK(ecode) ? vresult : 0;
  }
  
SWIGINTERN void Pool_clr_loadcallback(Pool *self){
    if (self->loadcallback == loadcallback) {
      PyObject *obj = self->loadcallbackdata;
      Py_DECREF(obj);
      pool_setloadcallback(self, 0, 0);
    }
  }
SWIGINTERN void Pool_set_loadcallback(Pool *self,PyObject *callable){
    Pool_clr_loadcallback(self);
    if (callable) {
      Py_INCREF(callable);
      pool_setloadcallback(self, loadcallback, callable);
    }
  }
SWIGINTERN void delete_Pool(Pool *self){
    Pool *pool = self;
    Id repoid;
    Repo *repo;
    FOR_REPOS(repoid, repo)
      appdata_clr_helper(&repo->appdata);
    Pool_clr_loadcallback(pool);
    appdata_clr_helper(&pool->appdata);
    pool_free(pool);
  }
SWIGINTERN disown_helper Pool_free(Pool *self){
    Pool *pool = self;
    Id repoid;
    Repo *repo;
    FOR_REPOS(repoid, repo)
      appdata_clr_helper(&repo->appdata);
    Pool_clr_loadcallback(pool);
    appdata_clr_helper(&pool->appdata);
    pool_free(pool);
    return 0;
  }
SWIGINTERN disown_helper Pool_disown(Pool *self){
    return 0;
  }

  SWIGINTERN void Pool_appdata_set(Pool *pool, AppObjectPtr appdata) {
    appdata_set_helper(&pool->appdata, appdata);
  }
  SWIGINTERN AppObjectPtr Pool_appdata_get(Pool *pool) {
    return appdata_get_helper(&pool->appdata);
  }
  
SWIGINTERN void Pool_appdata_disown(Pool *self){
    appdata_disown_helper(self->appdata);
  }
SWIGINTERN Id Pool_str2id(Pool *self,char const *str,bool create){
    return pool_str2id(self, str, create);
  }
SWIGINTERN Dep *Pool_Dep(Pool *self,char const *str,bool create){
    Id id = pool_str2id(self, str, create);
    return new_Dep(self, id);
  }
SWIGINTERN Dep *Pool_parserpmrichdep(Pool *self,char const *str){
    Id id = pool_parserpmrichdep(self, str);
    return new_Dep(self, id);
  }
SWIGINTERN char const *Pool_id2str(Pool *self,Id id){
    return pool_id2str(self, id);
  }
SWIGINTERN char const *Pool_dep2str(Pool *self,Id id){
    return pool_dep2str(self, id);
  }
SWIGINTERN Id Pool_rel2id(Pool *self,Id name,Id evr,int flags,bool create){
    return pool_rel2id(self, name, evr, flags, create);
  }
SWIGINTERN Id Pool_id2langid(Pool *self,Id id,char const *lang,bool create){
    return pool_id2langid(self, id, lang, create);
  }
SWIGINTERN void Pool_setarch(Pool *self,char const *arch){
    struct utsname un;
    if (!arch) {
      if (uname(&un)) {
        perror("uname");
        return;
      }
      arch = un.machine;
    }
    pool_setarch(self, arch);
  }
SWIGINTERN Repo *Pool_add_repo(Pool *self,char const *name){
    return repo_create(self, name);
  }
SWIGINTERN char const *Pool_lookup_str(Pool *self,Id entry,Id keyname){
    return pool_lookup_str(self, entry, keyname);
  }
SWIGINTERN Id Pool_lookup_id(Pool *self,Id entry,Id keyname){
    return pool_lookup_id(self, entry, keyname);
  }
SWIGINTERN unsigned long long Pool_lookup_num(Pool *self,Id entry,Id keyname,unsigned long long notfound){
    return pool_lookup_num(self, entry, keyname, notfound);
  }
SWIGINTERN bool Pool_lookup_void(Pool *self,Id entry,Id keyname){
    return pool_lookup_void(self, entry, keyname);
  }
SWIGINTERN Chksum *Pool_lookup_checksum(Pool *self,Id entry,Id keyname){
    Id type = 0;
    const unsigned char *b = pool_lookup_bin_checksum(self, entry, keyname, &type);
    return solv_chksum_create_from_bin(type, b);
  }
SWIGINTERN Dataiterator *Pool_Dataiterator(Pool *self,Id key,char const *match,int flags){
    return new_Dataiterator(self, 0, 0, key, match, flags);
  }
SWIGINTERN Dataiterator *Pool_Dataiterator_solvid(Pool *self,Id p,Id key,char const *match,int flags){
    return new_Dataiterator(self, 0, p, key, match, flags);
  }
SWIGINTERN char const *Pool_solvid2str(Pool *self,Id solvid){
    return pool_solvid2str(self, solvid);
  }
SWIGINTERN char const *Pool_solvidset2str(Pool *self,Queue q){
    return pool_solvidset2str(self, &q);
  }
SWIGINTERN char const *Pool_solvableset2str(Pool *self,Queue solvables){
    return pool_solvidset2str(self, &solvables);
  }
SWIGINTERN void Pool_addfileprovides(Pool *self){
    pool_addfileprovides(self);
  }
SWIGINTERN Queue Pool_addfileprovides_queue(Pool *self){
    Queue r;
    queue_init(&r);
    pool_addfileprovides_queue(self, &r, 0);
    return r;
  }
SWIGINTERN void Pool_createwhatprovides(Pool *self){
    pool_createwhatprovides(self);
  }
SWIGINTERN XSolvable *Pool_id2solvable(Pool *self,Id id){
    return new_XSolvable(self, id);
  }

  SWIGINTERN Pool_solvable_iterator * Pool_solvables_get(Pool *pool) {
    return new_Pool_solvable_iterator(pool);
  }
  
SWIGINTERN Pool_solvable_iterator *Pool_solvables_iter(Pool *self){
    return new_Pool_solvable_iterator(self);
  }
SWIGINTERN Repo *Pool_id2repo(Pool *self,Id id){
    if (id < 1 || id >= self->nrepos)
      return 0;
    return pool_id2repo(self, id);
  }

  SWIGINTERN Pool_repo_iterator * Pool_repos_get(Pool *pool) {
    return new_Pool_repo_iterator(pool);
  }
  
SWIGINTERN Pool_repo_iterator *Pool_repos_iter(Pool *self){
    return new_Pool_repo_iterator(self);
  }

  SWIGINTERN void Pool_installed_set(Pool *pool, Repo *installed) {
    pool_set_installed(pool, installed);
  }
  SWIGINTERN Repo *Pool_installed_get(Pool *pool) {
    return pool->installed;
  }
  SWIGINTERN const char *Pool_errstr_get(Pool *pool) {
    return pool_errstr(pool);
  }
  
SWIGINTERN Queue Pool_matchprovidingids(Pool *self,char const *match,int flags){
    Pool *pool = self;
    Queue q;
    Id id;
    queue_init(&q);
    if (!flags) {
      for (id = 1; id < pool->ss.nstrings; id++)
        if (pool->whatprovides[id])
          queue_push(&q, id);
    } else {
      Datamatcher ma;
      if (!datamatcher_init(&ma, match, flags)) {
        for (id = 1; id < pool->ss.nstrings; id++)
          if (pool->whatprovides[id] && datamatcher_match(&ma, pool_id2str(pool, id)))
            queue_push(&q, id);
        datamatcher_free(&ma);
      }
    }
    return q;
  }
SWIGINTERN Job *Pool_Job(Pool *self,int how,Id what){
    return new_Job(self, how, what);
  }
SWIGINTERN Queue Pool_whatprovides(Pool *self,DepId dep){
    Pool *pool = self;
    Queue q;
    Id p, pp;
    queue_init(&q);
    FOR_PROVIDES(p, pp, dep)
      queue_push(&q, p);
    return q;
  }
SWIGINTERN Queue Pool_best_solvables(Pool *self,Queue solvables,int flags){
    Queue q;
    queue_init_clone(&q, &solvables);
    pool_best_solvables(self, &q, flags);
    return q;
  }
SWIGINTERN Id Pool_towhatprovides(Pool *self,Queue q){
    return pool_queuetowhatprovides(self, &q);
  }
SWIGINTERN void Pool_set_namespaceproviders(Pool *self,DepId ns,DepId evr,bool value){
    Id dep = pool_rel2id(self, ns, evr, REL_NAMESPACE, 1);
    pool_set_whatprovides(self, dep, value ? 2 : 1);
  }
SWIGINTERN void Pool_flush_namespaceproviders(Pool *self,DepId ns,DepId evr){
    pool_flush_namespaceproviders(self, ns, evr);
  }
SWIGINTERN Queue Pool_whatcontainsdep(Pool *self,Id keyname,DepId dep,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatcontainsdep(self, keyname, dep, &q, marker);
    return q;
  }
SWIGINTERN Queue Pool_whatmatchesdep(Pool *self,Id keyname,DepId dep,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatmatchesdep(self, keyname, dep, &q, marker);
    return q;
  }
SWIGINTERN Queue Pool_whatmatchessolvable(Pool *self,Id keyname,XSolvable *pool_solvable,Id marker){
    Queue q;
    queue_init(&q);
    pool_whatmatchessolvable(self, keyname, pool_solvable->id, &q, marker);
    return q;
  }
SWIGINTERN bool Pool_isknownarch(Pool *self,DepId id){
    Pool *pool = self;
    if (!id || id == ID_EMPTY)
      return 0;
    if (id == ARCH_SRC || id == ARCH_NOSRC || id == ARCH_NOARCH)
      return 1;
    if (pool->id2arch && pool_arch2score(pool, id) == 0)
      return 0;
    return 1;
  }
SWIGINTERN Solver *Pool_Solver(Pool *self){
    return solver_create(self);
  }
SWIGINTERN Selection *Pool_Selection(Pool *self){
    return new_Selection(self);
  }
SWIGINTERN Selection *Pool_Selection_all(Pool *self,int setflags){
    Selection *sel = new_Selection(self);
    queue_push2(&sel->q, SOLVER_SOLVABLE_ALL | setflags, 0);
    return sel;
  }
SWIGINTERN Selection *Pool_select(Pool *self,char const *name,int flags){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make(self, &sel->q, name, flags);
    return sel;
  }
SWIGINTERN Selection *Pool_matchdeps(Pool *self,char const *name,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchdeps(self, &sel->q, name, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Selection *Pool_matchdepid(Pool *self,DepId dep,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchdepid(self, &sel->q, dep, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Selection *Pool_matchsolvable(Pool *self,XSolvable *solvable,int flags,Id keyname,Id marker){
    Selection *sel = new_Selection(self);
    sel->flags = selection_make_matchsolvable(self, &sel->q, solvable->id, flags, keyname, marker);
    return sel;
  }
SWIGINTERN Queue Pool_get_considered_list(Pool *self){
    Queue q;
    queue_init(&q);
    int i;
    for (i = 2; i < self->nsolvables; i++) {
      if (self->solvables[i].repo && (!self->considered || MAPTST(self->considered, i)))
        queue_push(&q, i);
    }
    return q;
  }
SWIGINTERN Queue Pool_get_disabled_list(Pool *self){
    Queue q;
    queue_init(&q);
    int i;
    for (i = 2; i < self->nsolvables; i++) {
      if (self->solvables[i].repo && (self->considered && !MAPTST(self->considered, i)))
        queue_push(&q, i);
    }
    return q;
  }
SWIGINTERN void Pool_set_considered_list(Pool *self,Queue q){
    int i;
    Id p;
    if (!self->considered) {
      self->considered = solv_calloc(1, sizeof(Map));
      map_init(self->considered, self->nsolvables);
    }
    map_empty(self->considered);
    MAPSET(self->considered, 1);
    for (i = 0; i < q.count; i++) {
      p = q.elements[i];
      if (p > 0 && p < self->nsolvables)
        MAPSET(self->considered, p);
    }
  }
SWIGINTERN void Pool_set_disabled_list(Pool *self,Queue q){
    int i;
    Id p;
    if (!q.count) {
      if (self->considered) {
        map_free(self->considered);
        self->considered = solv_free(self->considered);
      }
      return;
    }
    if (!self->considered) {
      self->considered = solv_calloc(1, sizeof(Map));
      map_init(self->considered, self->nsolvables);
    }
    map_setall(self->considered);
    for (i = 0; i < q.count; i++) {
      p = q.elements[i];
      if (p > 0 && p < self->nsolvables)
        MAPCLR(self->considered, p);
    }
  }
SWIGINTERN void Pool_setpooljobs(Pool *self,Queue solvejobs){
    queue_free(&self->pooljobs);
    queue_init_clone(&self->pooljobs, &solvejobs);
  }
SWIGINTERN Queue Pool_getpooljobs(Pool *self){
    Queue q;
    queue_init_clone(&q, &self->pooljobs);
    return q;
  }
SWIGINTERN void Repo_free(Repo *self,bool reuseids){
    appdata_clr_helper(&self->appdata);
    repo_free(self, reuseids);
  }
SWIGINTERN void Repo_empty(Repo *self,bool reuseids){
    repo_empty(self, reuseids);
  }
SWIGINTERN bool Repo_isempty(Repo *self){
    return !self->nsolvables;
  }

  SWIGINTERN void Repo_appdata_set(Repo *repo, AppObjectPtr appdata) {
    appdata_set_helper(&repo->appdata, appdata);
  }
  SWIGINTERN AppObjectPtr Repo_appdata_get(Repo *repo) {
    return appdata_get_helper(&repo->appdata);
  }
  
SWIGINTERN bool Repo_add_solv__SWIG_0(Repo *self,char const *name,int flags){
    FILE *fp = fopen(name, "r");
    int r;
    if (!fp)
      return 0;
    r = repo_add_solv(self, fp, flags);
    fclose(fp);
    return r == 0;
  }
SWIGINTERN bool Repo_add_solv__SWIG_1(Repo *self,FILE *fp,int flags){
    return repo_add_solv(self, fp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_solvable(Repo *self){
    Id solvid = repo_add_solvable(self);
    return new_XSolvable(self->pool, solvid);
  }
SWIGINTERN bool Repo_add_rpmdb(Repo *self,int flags){
    return repo_add_rpmdb(self, 0, flags) == 0;
  }
SWIGINTERN bool Repo_add_rpmdb_reffp(Repo *self,FILE *reffp,int flags){
    return repo_add_rpmdb_reffp(self, reffp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_rpm(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_rpm(self, name, flags));
  }
SWIGINTERN bool Repo_add_rpmdb_pubkeys(Repo *self,int flags){
    return repo_add_rpmdb_pubkeys(self, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_pubkey(Repo *self,char const *keyfile,int flags){
    return new_XSolvable(self->pool, repo_add_pubkey(self, keyfile, flags));
  }
SWIGINTERN bool Repo_add_keyring(Repo *self,FILE *fp,int flags){
    return repo_add_keyring(self, fp, flags);
  }
SWIGINTERN bool Repo_add_keydir(Repo *self,char const *keydir,char const *suffix,int flags){
    return repo_add_keydir(self, keydir, suffix, flags);
  }
SWIGINTERN bool Repo_add_rpmmd(Repo *self,FILE *fp,char const *language,int flags){
    return repo_add_rpmmd(self, fp, language, flags) == 0;
  }
SWIGINTERN bool Repo_add_repomdxml(Repo *self,FILE *fp,int flags){
    return repo_add_repomdxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_updateinfoxml(Repo *self,FILE *fp,int flags){
    return repo_add_updateinfoxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_deltainfoxml(Repo *self,FILE *fp,int flags){
    return repo_add_deltainfoxml(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_debdb(Repo *self,int flags){
    return repo_add_debdb(self, flags) == 0;
  }
SWIGINTERN bool Repo_add_debpackages(Repo *self,FILE *fp,int flags){
    return repo_add_debpackages(self, fp, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_deb(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_deb(self, name, flags));
  }
SWIGINTERN bool Repo_add_susetags(Repo *self,FILE *fp,Id defvendor,char const *language,int flags){
    return repo_add_susetags(self, fp, defvendor, language, flags) == 0;
  }
SWIGINTERN bool Repo_add_content(Repo *self,FILE *fp,int flags){
    return repo_add_content(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_products(Repo *self,char const *proddir,int flags){
    return repo_add_products(self, proddir, flags) == 0;
  }
SWIGINTERN bool Repo_add_mdk(Repo *self,FILE *fp,int flags){
    return repo_add_mdk(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_mdk_info(Repo *self,FILE *fp,int flags){
    return repo_add_mdk_info(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_arch_repo(Repo *self,FILE *fp,int flags){
    return repo_add_arch_repo(self, fp, flags) == 0;
  }
SWIGINTERN bool Repo_add_arch_local(Repo *self,char const *dir,int flags){
    return repo_add_arch_local(self, dir, flags) == 0;
  }
SWIGINTERN XSolvable *Repo_add_arch_pkg(Repo *self,char const *name,int flags){
    return new_XSolvable(self->pool, repo_add_arch_pkg(self, name, flags));
  }
SWIGINTERN void Repo_internalize(Repo *self){
    repo_internalize(self);
  }
SWIGINTERN bool Repo_write(Repo *self,FILE *fp){
    return repo_write(self, fp) == 0;
  }
SWIGINTERN bool Repo_write_first_repodata(Repo *self,FILE *fp){
    int oldnrepodata = self->nrepodata;
    int res;
    self->nrepodata = oldnrepodata > 2 ? 2 : oldnrepodata;
    res = repo_write(self, fp);
    self->nrepodata = oldnrepodata;
    return res == 0;
  }
SWIGINTERN Dataiterator *Repo_Dataiterator(Repo *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, self, 0, key, match, flags);
  }
SWIGINTERN Dataiterator *Repo_Dataiterator_meta(Repo *self,Id key,char const *match,int flags){
    return new_Dataiterator(self->pool, self, SOLVID_META, key, match, flags);
  }

  SWIGINTERN Id Repo_id_get(Repo *repo) {
    return repo->repoid;
  }
  

  SWIGINTERN Repo_solvable_iterator * Repo_solvables_get(Repo *repo) {
    return new_Repo_solvable_iterator(repo);
  }
  

  SWIGINTERN Datapos * Repo_meta_get(Repo *repo) {
    Datapos *pos = solv_calloc(1, sizeof(*pos));
    pos->solvid = SOLVID_META;
    pos->repo = repo;
    return pos;
  }
  
SWIGINTERN Repo_solvable_iterator *Repo_solvables_iter(Repo *self){
    return new_Repo_solvable_iterator(self);
  }
SWIGINTERN XRepodata *Repo_add_repodata(Repo *self,int flags){
    Repodata *rd = repo_add_repodata(self, flags);
    return new_XRepodata(self, rd->repodataid);
  }
SWIGINTERN void Repo_create_stubs(Repo *self){
    Repodata *data;
    if (!self->nrepodata)
      return;
    data = repo_id2repodata(self, self->nrepodata - 1);
    if (data->state != REPODATA_STUB)
      (void)repodata_create_stubs(data);
  }
SWIGINTERN bool Repo_iscontiguous(Repo *self){
    int i;
    for (i = self->start; i < self->end; i++)
      if (self->pool->solvables[i].repo != self)
        return 0;
    return 1;
  }
SWIGINTERN XRepodata *Repo_first_repodata(Repo *self){
    Repodata *data;
    int i;
    if (self->nrepodata < 2)
      return 0;
    /* make sure all repodatas but the first are extensions */
    data = repo_id2repodata(self, 1);
    if (data->loadcallback)
       return 0;
    for (i = 2; i < self->nrepodata; i++)
      {
        data = repo_id2repodata(self, i);
        if (!data->loadcallback)
          return 0;       /* oops, not an extension */
      }
    return new_XRepodata(self, 1);
  }
SWIGINTERN Selection *Repo_Selection(Repo *self,int setflags){
    Selection *sel = new_Selection(self->pool);
    setflags |= SOLVER_SETREPO;
    queue_push2(&sel->q, SOLVER_SOLVABLE_REPO | setflags, self->repoid);
    return sel;
  }
SWIGINTERN XSolvable *Repo_find_pubkey(Repo *self,char const *keyid){
    return new_XSolvable(self->pool, repo_find_pubkey(self, keyid));
  }
SWIGINTERN Repo *Repo_createshadow(Repo *self,char const *name){
    Repo *repo = repo_create(self->pool, name);
    if (self->idarraysize) {
      repo_reserve_ids(repo, 0, self->idarraysize);
      memcpy(repo->idarraydata, self->idarraydata, sizeof(Id) * self->idarraysize);
      repo->idarraysize = self->idarraysize;
    }
    repo->start = self->start;
    repo->end = self->end;
    repo->nsolvables = self->nsolvables;
    return repo;
  }
SWIGINTERN void Repo_moveshadow(Repo *self,Queue q){
    Pool *pool = self->pool;
    int i;
    for (i = 0; i < q.count; i++) {
      Solvable *s;
      Id p = q.elements[i];
      if (p < self->start || p >= self->end)
        continue;
      s = pool->solvables + p;
      if (self->idarraysize != s->repo->idarraysize)
        continue;
      s->repo = self;
    }
  }
SWIGINTERN bool Repo___eq__(Repo *self,Repo *repo){
    return self == repo;
  }
SWIGINTERN bool Repo___ne__(Repo *self,Repo *repo){
    return self != repo;
  }
SWIGINTERN int Repo___hash__(Repo *self){
    return self->repoid;
  }
SWIGINTERN char const *Repo___str__(Repo *self){
    char buf[20];
    if (self->name)
      return solv_strdup(self->name);
    sprintf(buf, "Repo#%d", self->repoid);
    return solv_strdup(buf);
  }
SWIGINTERN char const *Repo___repr__(Repo *self){
    char buf[20];
    if (self->name)
      {
        sprintf(buf, "<Repo #%d ", self->repoid);
        return solv_dupjoin(buf, self->name, ">");
      }
    sprintf(buf, "<Repo #%d>", self->repoid);
    return solv_strdup(buf);
  }
SWIGINTERN Decision *new_Decision(Solver *solv,Id p,int reason,Id infoid){
    Decision *d = solv_calloc(1, sizeof(*d));
    d->solv = solv;
    d->p = p;
    d->reason = reason;
    d->infoid = infoid;
    return d;
  }

    SWIGINTERN XRule *Decision_rule_get(Decision *d) {
      return d->reason == SOLVER_REASON_WEAKDEP || d->infoid <= 0 ? 0 : new_XRule(d->solv, d->infoid);
    }
    SWIGINTERN XSolvable *Decision_solvable_get(Decision *d) {
      return new_XSolvable(d->solv->pool, d->p >= 0 ? d->p : -d->p);
    }
  
SWIGINTERN Ruleinfo *Decision_info(Decision *self){
    Id type, source, target, dep;
    if (self->reason == SOLVER_REASON_WEAKDEP) {
      type = solver_weakdepinfo(self->solv, self->p, &source, &target, &dep);
    } else if (self->infoid) {
      type = solver_ruleinfo(self->solv, self->infoid, &source, &target, &dep);
    } else {
      return 0;
    }
    return new_Ruleinfo(self->solv, self->infoid, type, source, target, dep);
  }
SWIGINTERN Queue Decision_allinfos(Decision *self){
    Queue q;
    queue_init(&q);
    if (self->reason == SOLVER_REASON_WEAKDEP) {
      solver_allweakdepinfos(self->solv, self->p, &q);
    } else if (self->infoid) {
      solver_allruleinfos(self->solv, self->infoid, &q);
    }
    return q;
  }
SWIGINTERN char const *Decision_reasonstr(Decision *self,bool noinfo){
    if (noinfo)
      return solver_reason2str(self->solv, self->reason);
    return solver_decisionreason2str(self->solv, self->p, self->reason, self->infoid);
  }
SWIGINTERN char const *Decision___str__(Decision *self){
    Pool *pool = self->solv->pool;
    if (self->p == 0 && self->reason == SOLVER_REASON_UNSOLVABLE)
      return "unsolvable";
    if (self->p >= 0)
      return pool_tmpjoin(pool, "install ", pool_solvid2str(pool, self->p), 0);
    else
      return pool_tmpjoin(pool, "conflict ", pool_solvid2str(pool, -self->p), 0);
  }
SWIGINTERN Decisionset *new_Decisionset(Solver *solv){
    Decisionset *d = solv_calloc(1, sizeof(*d));
    d->solv = solv;
    queue_init(&d->decisionlistq);
    return d;
  }
SWIGINTERN void delete_Decisionset(Decisionset *self){
    queue_free(&self->decisionlistq);
    solv_free(self);
  }
SWIGINTERN Ruleinfo *Decisionset_info(Decisionset *self){
    return new_Ruleinfo(self->solv, self->infoid, self->type, self->source, self->target, self->dep_id);
  }

    SWIGINTERN Dep *Decisionset_dep_get(Decisionset *d) {
      return new_Dep(d->solv->pool, d->dep_id);
    }
  
SWIGINTERN Queue Decisionset_solvables(Decisionset *self){
    Queue q;
    int i;
    queue_init(&q);
    for (i = 0; i < self->decisionlistq.count; i += 3)
      if (self->decisionlistq.elements[i] != 0)
        queue_push(&q, self->decisionlistq.elements[i] > 0 ? self->decisionlistq.elements[i] : -self->decisionlistq.elements[i]);
    return q;
  }
SWIGINTERN Queue Decisionset_decisions(Decisionset *self){
    Queue q;
    queue_init_clone(&q, &self->decisionlistq);
    return q;
  }
SWIGINTERN char const *Decisionset_reasonstr(Decisionset *self,bool noinfo){
    if (noinfo || !self->type)
      return solver_reason2str(self->solv, self->reason);
    return solver_decisioninfo2str(self->solv, self->bits, self->type, self->source, self->target, self->dep_id);
  }
SWIGINTERN char const *Decisionset___str__(Decisionset *self){
    Pool *pool = self->solv->pool;
    Queue q;
    int i;
    const char *s;
    if (!self->decisionlistq.elements)
      return "";
    if (self->p == 0 && self->reason == SOLVER_REASON_UNSOLVABLE)
      return "unsolvable";
    queue_init(&q);
    for (i = 0; i < self->decisionlistq.count; i += 3)
      if (self->decisionlistq.elements[i] != 0)
        queue_push(&q, self->decisionlistq.elements[i] > 0 ? self->decisionlistq.elements[i] : -self->decisionlistq.elements[i]);
    s = pool_solvidset2str(pool, &q);
    queue_free(&q);
    return pool_tmpjoin(pool, self->p >= 0 ? "install " : "conflict ", s, 0);
  }
SWIGINTERN void delete_Solver(Solver *self){
    solver_free(self);
  }
SWIGINTERN int Solver_set_flag(Solver *self,int flag,int value){
    return solver_set_flag(self, flag, value);
  }
SWIGINTERN int Solver_get_flag(Solver *self,int flag){
    return solver_get_flag(self, flag);
  }
SWIGINTERN Queue Solver_solve(Solver *self,Queue solvejobs){
    Queue q;
    int i, cnt;
    queue_init(&q);
    solver_solve(self, &solvejobs);
    cnt = solver_problem_count(self);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN Transaction *Solver_transaction(Solver *self){
    return solver_create_transaction(self);
  }
SWIGINTERN int Solver_describe_decision(Solver *self,XSolvable *s,XRule **OUTPUT){
    Id ruleid;
    int reason = solver_describe_decision(self, s->id, &ruleid);
    *OUTPUT = new_XRule(self, ruleid);
    return reason;
  }
SWIGINTERN Queue Solver_describe_weakdep_decision_raw(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_describe_weakdep_decision(self, s->id, &q);
    return q;
  }
SWIGINTERN int Solver_alternatives_count(Solver *self){
    return solver_alternatives_count(self);
  }
SWIGINTERN Alternative *Solver_get_alternative(Solver *self,Id aid){
    Alternative *a = solv_calloc(1, sizeof(*a));
    a->solv = self;
    queue_init(&a->choices);
    a->type = solver_get_alternative(self, aid, &a->dep_id, &a->from_id, &a->chosen_id, &a->choices, &a->level);
    if (!a->type) {
      queue_free(&a->choices);
      solv_free(a);
      return 0;
    }
    if (a->type == SOLVER_ALTERNATIVE_TYPE_RULE) {
      a->rid = a->dep_id;
      a->dep_id = 0;
    }
    return a;
  }
SWIGINTERN Queue Solver_alternatives(Solver *self){
    Queue q;
    int i, cnt;
    queue_init(&q);
    cnt = solver_alternatives_count(self);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN bool Solver_write_testcase(Solver *self,char const *dir){
    return testcase_write(self, dir, TESTCASE_RESULT_TRANSACTION | TESTCASE_RESULT_PROBLEMS, 0, 0);
  }
SWIGINTERN Queue Solver_raw_decisions(Solver *self,int filter){
    Queue q;
    queue_init(&q);
    solver_get_decisionqueue(self, &q);
    if (filter) {
      int i, j;
      for (i = j = 0; i < q.count; i++)
        if ((filter > 0 && q.elements[i] > 1) ||
            (filter < 0 && q.elements[i] < 0))
          q.elements[j++] = q.elements[i];
      queue_truncate(&q, j);
    }
    return q;
  }
SWIGINTERN Queue Solver_all_decisions(Solver *self,int filter){
    int i, j, cnt;
    Queue q;
    queue_init(&q);
    solver_get_decisionqueue(self, &q);
    if (filter) {
      for (i = j = 0; i < q.count; i++)
        if ((filter > 0 && q.elements[i] > 1) ||
            (filter < 0 && q.elements[i] < 0))
          q.elements[j++] = q.elements[i];
      queue_truncate(&q, j);
    }
    cnt = q.count;
    for (i = 0; i < cnt; i++) {
      Id ruleid, p = q.elements[i];
      int reason;
      if (p == 0 || p == 1)
        continue;       /* ignore system solvable */
      reason = solver_describe_decision(self, p > 0 ? p : -p, &ruleid);
      queue_push(&q, p);
      queue_push2(&q, reason, ruleid);
    }
    queue_deleten(&q, 0, cnt);
    return q;
  }
SWIGINTERN Decision *Solver_get_decision(Solver *self,XSolvable *s){
    Id info;
    int lvl = solver_get_decisionlevel(self, s->id);
    Id p = lvl > 0 ? s->id : -s->id;
    int reason = solver_describe_decision(self, p, &info);
    return new_Decision(self, p, reason, info);
  }
SWIGINTERN Queue Solver_get_learnt(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self, s->id, SOLVER_DECISIONLIST_SOLVABLE, &q);
    return q;
  }
SWIGINTERN Queue Solver_get_decisionlist(Solver *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self, s->id, SOLVER_DECISIONLIST_SOLVABLE, &q);
    return q;
  }
SWIGINTERN Queue Solver_get_recommended(Solver *self,bool noselected){
    Queue q;
    queue_init(&q);
    solver_get_recommendations(self, &q, NULL, noselected);
    return q;
  }
SWIGINTERN Queue Solver_get_suggested(Solver *self,bool noselected){
    Queue q;
    queue_init(&q);
    solver_get_recommendations(self, NULL, &q, noselected);
    return q;
  }
SWIGINTERN Chksum *new_Chksum__SWIG_0(Id type){
    return solv_chksum_create(type);
  }
SWIGINTERN Chksum *new_Chksum__SWIG_1(Id type,char const *hex){
    unsigned char buf[64];
    int l = solv_chksum_len(type);
    if (!l)
      return 0;
    if (solv_hex2bin(&hex, buf, sizeof(buf)) != l || hex[0])
      return 0;
    return solv_chksum_create_from_bin(type, buf);
  }

SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}

SWIGINTERN Chksum *Chksum_from_bin(Id type,unsigned char const *str,size_t len){
    return len == solv_chksum_len(type) ? solv_chksum_create_from_bin(type, str) : 0;
  }
SWIGINTERN void delete_Chksum(Chksum *self){
    solv_chksum_free(self, 0);
  }

  SWIGINTERN Id Chksum_type_get(Chksum *chk) {
    return solv_chksum_get_type(chk);
  }
  
SWIGINTERN void Chksum_add(Chksum *self,unsigned char const *str,size_t len){
    solv_chksum_add(self, str, (int)len);
  }
SWIGINTERN void Chksum_add_fp(Chksum *self,FILE *fp){
    char buf[4096];
    int l;
    while ((l = fread(buf, 1, sizeof(buf), fp)) > 0)
      solv_chksum_add(self, buf, l);
    rewind(fp);         /* convenience */
  }
SWIGINTERN void Chksum_add_fd(Chksum *self,int fd){
    char buf[4096];
    int l;
    while ((l = read(fd, buf, sizeof(buf))) > 0)
      solv_chksum_add(self, buf, l);
    lseek(fd, 0, 0);    /* convenience */
  }
SWIGINTERN void Chksum_add_stat(Chksum *self,char const *filename){
    struct stat stb;
    if (stat(filename, &stb))
      memset(&stb, 0, sizeof(stb));
    solv_chksum_add(self, &stb.st_dev, sizeof(stb.st_dev));
    solv_chksum_add(self, &stb.st_ino, sizeof(stb.st_ino));
    solv_chksum_add(self, &stb.st_size, sizeof(stb.st_size));
    solv_chksum_add(self, &stb.st_mtime, sizeof(stb.st_mtime));
  }
SWIGINTERN void Chksum_add_fstat(Chksum *self,int fd){
    struct stat stb;
    if (fstat(fd, &stb))
      memset(&stb, 0, sizeof(stb));
    solv_chksum_add(self, &stb.st_dev, sizeof(stb.st_dev));
    solv_chksum_add(self, &stb.st_ino, sizeof(stb.st_ino));
    solv_chksum_add(self, &stb.st_size, sizeof(stb.st_size));
    solv_chksum_add(self, &stb.st_mtime, sizeof(stb.st_mtime));
  }
SWIGINTERN BinaryBlob Chksum_raw(Chksum *self){
    BinaryBlob bl;
    int l;
    const unsigned char *b;
    b = solv_chksum_get(self, &l);
    bl.data = b;
    bl.len = l;
    return bl;
  }
SWIGINTERN char *Chksum_hex(Chksum *self){
    int l;
    const unsigned char *b;
    char *ret;

    b = solv_chksum_get(self, &l);
    ret = solv_malloc(2 * l + 1);
    solv_bin2hex(b, l, ret);
    return ret;
  }
SWIGINTERN char const *Chksum_typestr(Chksum *self){
    return solv_chksum_type2str(solv_chksum_get_type(self));
  }
SWIGINTERN bool Chksum___eq__(Chksum *self,Chksum *chk){
    return solv_chksum_cmp(self, chk);
  }
SWIGINTERN bool Chksum___ne__(Chksum *self,Chksum *chk){
    return !solv_chksum_cmp(self, chk);
  }
SWIGINTERN char const *Chksum___str__(Chksum *self){
    const char *str;
    const char *h = 0;
    if (solv_chksum_isfinished(self))
      h = Chksum_hex(self);
    str = solv_dupjoin(solv_chksum_type2str(solv_chksum_get_type(self)), ":", h ? h : "unfinished");
    solv_free((void *)h);
    return str;
  }
SWIGINTERN char const *Chksum___repr__(Chksum *self){
    const char *h = Chksum___str__(self);
    const char *str = solv_dupjoin("<Chksum ", h, ">");
    solv_free((void *)h);
    return str;
  }
SWIGINTERN Solvsig *new_Solvsig(FILE *fp){
    return solvsig_create(fp);
  }
SWIGINTERN void delete_Solvsig(Solvsig *self){
    solvsig_free(self);
  }
SWIGINTERN Chksum *Solvsig_Chksum(Solvsig *self){
    return self->htype ? (Chksum *)solv_chksum_create(self->htype) : 0;
  }
SWIGINTERN XSolvable *Solvsig_verify(Solvsig *self,Repo *repo,Chksum *chksum){
    Id p = solvsig_verify(self, repo, chksum);
    return new_XSolvable(repo->pool, p);
  }
SWIGINTERN void delete_SolvFp(SolvFp *self){
    if (self->fp)
      fclose(self->fp);
    free(self);
  }
SWIGINTERN int SolvFp_fileno(SolvFp *self){
    return self->fp ? fileno(self->fp) : -1;
  }
SWIGINTERN int SolvFp_dup(SolvFp *self){
    return self->fp ? dup(fileno(self->fp)) : -1;
  }
SWIGINTERN bool SolvFp_write(SolvFp *self,unsigned char const *str,size_t len){
    return fwrite(str, len, 1, self->fp) == 1;
  }
SWIGINTERN bool SolvFp_flush(SolvFp *self){
    if (!self->fp)
      return 1;
    return fflush(self->fp) == 0;
  }
SWIGINTERN bool SolvFp_close(SolvFp *self){
    bool ret;
    if (!self->fp)
      return 1;
    ret = fclose(self->fp) == 0;
    self->fp = 0;
    return ret;
  }
SWIGINTERN void SolvFp_cloexec(SolvFp *self,bool state){
    if (!self->fp || fileno(self->fp) == -1)
      return;
    solv_setcloexec(fileno(self->fp), state);
  }

  SWIGINTERN SolvFp *solvfp_xfopen_fd(const char *fn, int fd, const char *mode) {
    SolvFp *sfp;
    FILE *fp;
    fd = dup(fd);
    if (fd == -1)
      return 0;
    solv_setcloexec(fd, 1);
    fp = solv_xfopen_fd(fn, fd, mode);
    if (!fp) {
      close(fd);
      return 0;
    }
    sfp = solv_calloc(1, sizeof(SolvFp));
    sfp->fp = fp;
    return sfp;
  }
  SWIGINTERN SolvFp *solvfp_xfopen(const char *fn, const char *mode) {
    SolvFp *sfp;
    FILE *fp;
    fp = solv_xfopen(fn, mode);
    if (!fp)
      return 0;
    if (fileno(fp) != -1)
      solv_setcloexec(fileno(fp), 1);
    sfp = solv_calloc(1, sizeof(SolvFp));
    sfp->fp = fp;
    return sfp;
  }

SWIGINTERN Problem *new_Problem(Solver *solv,Id id){
    Problem *p;
    p = solv_calloc(1, sizeof(*p));
    p->solv = solv;
    p->id = id;
    return p;
  }
SWIGINTERN XRule *Problem_findproblemrule(Problem *self){
    Id r = solver_findproblemrule(self->solv, self->id);
    return new_XRule(self->solv, r);
  }
SWIGINTERN Queue Problem_findallproblemrules(Problem *self,int unfiltered){
    Solver *solv = self->solv;
    Id probr;
    int i, j;
    Queue q;
    queue_init(&q);
    solver_findallproblemrules(solv, self->id, &q);
    if (!unfiltered)
      {
        for (i = j = 0; i < q.count; i++)
          {
            SolverRuleinfo rclass;
            probr = q.elements[i];
            rclass = solver_ruleclass(solv, probr);
            if (rclass == SOLVER_RULE_UPDATE || rclass == SOLVER_RULE_JOB)
              continue;
            q.elements[j++] = probr;
          }
        if (j)
          queue_truncate(&q, j);
      }
    return q;
  }
SWIGINTERN int Problem_solution_count(Problem *self){
    return solver_solution_count(self->solv, self->id);
  }
SWIGINTERN Queue Problem_solutions(Problem *self){
    Queue q;
    int i, cnt;
    queue_init(&q);
    cnt = solver_solution_count(self->solv, self->id);
    for (i = 1; i <= cnt; i++)
      queue_push(&q, i);
    return q;
  }
SWIGINTERN Queue Problem_get_learnt(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_learnt(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM, &q);
    return q;
  }
SWIGINTERN Queue Problem_get_decisionlist(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM | SOLVER_DECISIONLIST_SORTED, &q);
    return q;
  }
SWIGINTERN Queue Problem_get_decisionsetlist(Problem *self){
    Queue q;
    queue_init(&q);
    solver_get_decisionlist(self->solv, self->id, SOLVER_DECISIONLIST_PROBLEM | SOLVER_DECISIONLIST_SORTED | SOLVER_DECISIONLIST_WITHINFO | SOLVER_DECISIONLIST_MERGEDINFO, &q);
    prepare_decisionset_queue(self->solv, &q);
    return q;
  }
SWIGINTERN char const *Problem___str__(Problem *self){
    return solver_problem2str(self->solv, self->id);
  }
SWIGINTERN Solution *new_Solution(Problem *p,Id id){
    Solution *s;
    s = solv_calloc(1, sizeof(*s));
    s->solv = p->solv;
    s->problemid = p->id;
    s->id = id;
    return s;
  }
SWIGINTERN int Solution_element_count(Solution *self){
    return solver_solutionelement_count(self->solv, self->problemid, self->id);
  }
SWIGINTERN Queue Solution_elements(Solution *self,bool expandreplaces){
    Queue q;
    queue_init(&q);
    solver_all_solutionelements(self->solv, self->problemid, self->id, expandreplaces, &q);
    return q;
  }
SWIGINTERN Solutionelement *new_Solutionelement(Solver *solv,Id problemid,Id solutionid,Id type,Id p,Id rp){
    Solutionelement *e;
    e = solv_calloc(1, sizeof(*e));
    e->solv = solv;
    e->problemid = problemid;
    e->solutionid = solutionid;
    e->type = type;
    e->p = p;
    e->rp = rp;
    return e;
  }
SWIGINTERN char const *Solutionelement_str(Solutionelement *self){
    return solver_solutionelementtype2str(self->solv, self->type, self->p, self->rp);
  }
SWIGINTERN char const *Solutionelement___str__(Solutionelement *self){
    return solver_solutionelementtype2str(self->solv, self->type, self->p, self->rp);
  }
SWIGINTERN Queue Solutionelement_replaceelements(Solutionelement *self){
    Queue q;
    int illegal;

    queue_init(&q);
    if (self->type != SOLVER_SOLUTION_REPLACE || self->p <= 0 || self->rp <= 0)
      illegal = 0;
    else
      illegal = policy_is_illegal(self->solv, self->solv->pool->solvables + self->p, self->solv->pool->solvables + self->rp, 0);
    if ((illegal & POLICY_ILLEGAL_DOWNGRADE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_DOWNGRADE);
    if ((illegal & POLICY_ILLEGAL_ARCHCHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_ARCHCHANGE);
    if ((illegal & POLICY_ILLEGAL_VENDORCHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_VENDORCHANGE);
    if ((illegal & POLICY_ILLEGAL_NAMECHANGE) != 0)
      queue_push(&q, SOLVER_SOLUTION_REPLACE_NAMECHANGE);
    if (!q.count)
      queue_push(&q, self->type);
    return q;
  }
SWIGINTERN int Solutionelement_illegalreplace(Solutionelement *self){
    if (self->type != SOLVER_SOLUTION_REPLACE || self->p <= 0 || self->rp <= 0)
      return 0;
    return policy_is_illegal(self->solv, self->solv->pool->solvables + self->p, self->solv->pool->solvables + self->rp, 0);
  }

    SWIGINTERN XSolvable *Solutionelement_solvable_get(Solutionelement *e) {
      return new_XSolvable(e->solv->pool, e->p);
    }
    SWIGINTERN XSolvable *Solutionelement_replacement_get(Solutionelement *e) {
      return new_XSolvable(e->solv->pool, e->rp);
    }
    SWIGINTERN int Solutionelement_jobidx_get(Solutionelement *e) {
      if (e->type != SOLVER_SOLUTION_JOB && e->type != SOLVER_SOLUTION_POOLJOB)
        return -1;
      return (e->p - 1) / 2;
    }
  
SWIGINTERN Job *Solutionelement_Job(Solutionelement *self){
    Id extraflags = solver_solutionelement_extrajobflags(self->solv, self->problemid, self->solutionid);
    if (self->type == SOLVER_SOLUTION_JOB || self->type == SOLVER_SOLUTION_POOLJOB)
      return new_Job(self->solv->pool, SOLVER_NOOP, 0);
    if (self->type == SOLVER_SOLUTION_INFARCH || self->type == SOLVER_SOLUTION_DISTUPGRADE || self->type == SOLVER_SOLUTION_BEST)
      return new_Job(self->solv->pool, SOLVER_INSTALL|SOLVER_SOLVABLE|SOLVER_NOTBYUSER|extraflags, self->p);
    if (self->type == SOLVER_SOLUTION_REPLACE || self->type == SOLVER_SOLUTION_REPLACE_DOWNGRADE || self->type == SOLVER_SOLUTION_REPLACE_ARCHCHANGE || self->type == SOLVER_SOLUTION_REPLACE_VENDORCHANGE || self->type == SOLVER_SOLUTION_REPLACE_NAMECHANGE)
      return new_Job(self->solv->pool, SOLVER_INSTALL|SOLVER_SOLVABLE|SOLVER_NOTBYUSER|extraflags, self->rp);
    if (self->type == SOLVER_SOLUTION_ERASE)
      return new_Job(self->solv->pool, SOLVER_ERASE|SOLVER_SOLVABLE|extraflags, self->p);
    return 0;
  }
SWIGINTERN void delete_Alternative(Alternative *self){
    queue_free(&self->choices);
    solv_free(self);
  }

    SWIGINTERN XSolvable *Alternative_chosen_get(Alternative *a) {
      return new_XSolvable(a->solv->pool, a->chosen_id);
    }
    SWIGINTERN XRule *Alternative_rule_get(Alternative *a) {
      return new_XRule(a->solv, a->rid);
    }
    SWIGINTERN XSolvable *Alternative_depsolvable_get(Alternative *a) {
      return new_XSolvable(a->solv->pool, a->from_id);
    }
    SWIGINTERN Dep *Alternative_dep_get(Alternative *a) {
      return new_Dep(a->solv->pool, a->dep_id);
    }
  
SWIGINTERN Queue Alternative_choices_raw(Alternative *self){
    Queue q;
    queue_init_clone(&q, &self->choices);
    return q;
  }
SWIGINTERN Queue Alternative_choices(Alternative *self){
    int i;
    Queue q;
    queue_init_clone(&q, &self->choices);
    for (i = 0; i < q.count; i++)
      if (q.elements[i] < 0)
        q.elements[i] = -q.elements[i];
    return q;
  }
SWIGINTERN char const *Alternative___str__(Alternative *self){
    return solver_alternative2str(self->solv, self->type, self->type == SOLVER_ALTERNATIVE_TYPE_RULE ? self->rid : self->dep_id, self->from_id);
  }
SWIGINTERN void delete_Transaction(Transaction *self){
    transaction_free(self);
  }
SWIGINTERN bool Transaction_isempty(Transaction *self){
    return self->steps.count == 0;
  }
SWIGINTERN XSolvable *Transaction_othersolvable(Transaction *self,XSolvable *s){
    Id op = transaction_obs_pkg(self, s->id);
    return new_XSolvable(self->pool, op);
  }
SWIGINTERN Queue Transaction_allothersolvables(Transaction *self,XSolvable *s){
    Queue q;
    queue_init(&q);
    transaction_all_obs_pkgs(self, s->id, &q);
    return q;
  }
SWIGINTERN Queue Transaction_classify(Transaction *self,int mode){
    Queue q;
    queue_init(&q);
    transaction_classify(self, mode, &q);
    return q;
  }
SWIGINTERN Queue Transaction_newpackages(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    queue_truncate(&q, cut);
    return q;
  }
SWIGINTERN Queue Transaction_keptpackages(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    if (cut)
      queue_deleten(&q, 0, cut);
    return q;
  }
SWIGINTERN Queue Transaction_newsolvables(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    queue_truncate(&q, cut);
    return q;
  }
SWIGINTERN Queue Transaction_keptsolvables(Transaction *self){
    Queue q;
    int cut;
    queue_init(&q);
    cut = transaction_installedresult(self, &q);
    if (cut)
      queue_deleten(&q, 0, cut);
    return q;
  }
SWIGINTERN Queue Transaction_steps(Transaction *self){
    Queue q;
    queue_init_clone(&q, &self->steps);
    return q;
  }
SWIGINTERN int Transaction_steptype(Transaction *self,XSolvable *s,int mode){
    return transaction_type(self, s->id, mode);
  }
SWIGINTERN long long Transaction_calc_installsizechange(Transaction *self){
    return transaction_calc_installsizechange(self);
  }

#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyInt_FromLong((long)(value));
}
#endif

SWIGINTERN void Transaction_order(Transaction *self,int flags){
    transaction_order(self, flags);
  }
SWIGINTERN TransactionClass *new_TransactionClass(Transaction *trans,int mode,Id type,int count,Id fromid,Id toid){
    TransactionClass *cl = solv_calloc(1, sizeof(*cl));
    cl->transaction = trans;
    cl->mode = mode;
    cl->type = type;
    cl->count = count;
    cl->fromid = fromid;
    cl->toid = toid;
    return cl;
  }
SWIGINTERN Queue TransactionClass_solvables(TransactionClass *self){
    Queue q;
    queue_init(&q);
    transaction_classify_pkgs(self->transaction, self->mode, self->type, self->fromid, self->toid, &q);
    return q;
  }

    SWIGINTERN const char *TransactionClass_fromstr_get(TransactionClass *cl) {
      return pool_id2str(cl->transaction->pool, cl->fromid);
    }
    SWIGINTERN const char *TransactionClass_tostr_get(TransactionClass *cl) {
      return pool_id2str(cl->transaction->pool, cl->toid);
    }
  
#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_Selection_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_pool_get" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_flags_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_flags_get" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = (int)(int) ((arg1)->flags);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Selection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Selection" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Selection *)new_Selection(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Selection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Selection" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  delete_Selection(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_isempty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_isempty" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = (bool)Selection_isempty(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_clone(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_clone", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_clone" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_clone" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)Selection_clone(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_filter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  Selection *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_filter", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_filter" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_filter" "', argument " "2"" of type '" "Selection *""'"); 
  }
  arg2 = (Selection *)(argp2);
  Selection_filter(arg1,arg2);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  Selection *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_add", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_add" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_add" "', argument " "2"" of type '" "Selection *""'"); 
  }
  arg2 = (Selection *)(argp2);
  Selection_add(arg1,arg2);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_add_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_add_raw", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_add_raw" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_add_raw" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_add_raw" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  Selection_add_raw(arg1,arg2,arg3);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_subtract(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  Selection *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_subtract", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_subtract" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_subtract" "', argument " "2"" of type '" "Selection *""'"); 
  }
  arg2 = (Selection *)(argp2);
  Selection_subtract(arg1,arg2);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_select(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_select", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_select" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_select" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_select" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  Selection_select(arg1,(char const *)arg2,arg3);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_matchdeps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_matchdeps", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchdeps" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_matchdeps" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchdeps" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchdeps" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchdeps" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  Selection_matchdeps(arg1,(char const *)arg2,arg3,arg4,arg5);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_matchdepid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  DepId arg2 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_matchdepid", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchdepid" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_matchdepid" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchdepid" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchdepid" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchdepid" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  Selection_matchdepid(arg1,arg2,arg3,arg4,arg5);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_matchsolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_matchsolvable", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_matchsolvable" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Selection_matchsolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Selection_matchsolvable" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Selection_matchsolvable" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Selection_matchsolvable" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  Selection_matchsolvable(arg1,arg2,arg3,arg4,arg5);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_jobs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Selection_jobs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_jobs" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Selection_jobs" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = Selection_jobs(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 2;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 2)
    {
      Id id = *idp;
#define result resultx
      Job * result = new_Job(arg1->pool, id, idp[1]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection_solvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection_solvables" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = Selection_solvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection___str__" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = (char *)Selection___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Selection___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Selection *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Selection, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Selection___repr__" "', argument " "1"" of type '" "Selection *""'"); 
  }
  arg1 = (Selection *)(argp1);
  result = (char *)Selection___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Selection_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Selection, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Selection_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Dep_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_pool_get" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_id_get" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Dep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Dep", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Dep" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Dep" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Dep *)new_Dep(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep_Rel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  int arg2 ;
  DepId arg3 ;
  bool arg4 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dep_Rel", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Rel" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Rel" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Dep_Rel" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Dep_Rel" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = (bool)(val4);
  }
  result = (Dep *)Dep_Rel(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep_Selection_name(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dep_Selection_name", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Selection_name" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Selection_name" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)Dep_Selection_name(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep_Selection_provides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dep_Selection_provides", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_Selection_provides" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dep_Selection_provides" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)Dep_Selection_provides(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep_str" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (char *)Dep_str(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  Dep *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dep___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___eq__" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Dep___eq__" "', argument " "2"" of type '" "Dep *""'"); 
  }
  arg2 = (Dep *)(argp2);
  result = (bool)Dep___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  Dep *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dep___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___ne__" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Dep___ne__" "', argument " "2"" of type '" "Dep *""'"); 
  }
  arg2 = (Dep *)(argp2);
  result = (bool)Dep___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep___hash__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___hash__" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (int)Dep___hash__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___str__" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (char *)Dep___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dep___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dep___repr__" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  result = (char *)Dep___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Dep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dep *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dep, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Dep" "', argument " "1"" of type '" "Dep *""'"); 
  }
  arg1 = (Dep *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Dep_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Dep, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Dep_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Repo *arg2 = 0 ;
  Id arg3 ;
  Id arg4 ;
  char *arg5 = 0 ;
  int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Dataiterator", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Dataiterator" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Dataiterator" "', argument " "2"" of type '" "Repo *""'"); 
  }
  arg2 = (Repo *)(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Dataiterator" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Dataiterator" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_Dataiterator" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = (char *)(buf5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Dataiterator" "', argument " "6"" of type '" "int""'");
  } 
  arg6 = (int)(val6);
  result = (Dataiterator *)new_Dataiterator(arg1,arg2,arg3,arg4,(char const *)arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_NEW |  0 );
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return resultobj;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dataiterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dataiterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Dataiterator" "', argument " "1"" of type '" "Dataiterator *""'"); 
  }
  arg1 = (Dataiterator *)(argp1);
  delete_Dataiterator(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dataiterator___next__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dataiterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Datamatch *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator___next__" "', argument " "1"" of type '" "Dataiterator *""'"); 
  }
  arg1 = (Dataiterator *)(argp1);
  {
    result = (Datamatch *)Dataiterator___next__(arg1);
    if (!result) {
      PyErr_SetString(PyExc_StopIteration,"no more matches");
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datamatch, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dataiterator_prepend_keyname(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dataiterator *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Dataiterator_prepend_keyname", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator_prepend_keyname" "', argument " "1"" of type '" "Dataiterator *""'"); 
  }
  arg1 = (Dataiterator *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Dataiterator_prepend_keyname" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  Dataiterator_prepend_keyname(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Dataiterator_skip_solvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Dataiterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Dataiterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Dataiterator_skip_solvable" "', argument " "1"" of type '" "Dataiterator *""'"); 
  }
  arg1 = (Dataiterator *)(argp1);
  Dataiterator_skip_solvable(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Dataiterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Dataiterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Dataiterator_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_XSolvable_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_pool_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_id_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_XSolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_XSolvable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XSolvable" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XSolvable" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XSolvable *)new_XSolvable(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_str" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_str(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_str" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (char *)XSolvable_lookup_str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_id", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_id" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Id)XSolvable_lookup_id(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  unsigned long long arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned long long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned long long result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_num", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_num" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_num" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_num" "', argument " "3"" of type '" "unsigned long long""'");
    } 
    arg3 = (unsigned long long)(val3);
  }
  result = (unsigned long long)XSolvable_lookup_num(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_void(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_void", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_void" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_void" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (bool)XSolvable_lookup_void(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_checksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_checksum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_checksum" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_checksum" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Chksum *)XSolvable_lookup_checksum(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_idarray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  Id arg3 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_idarray", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_idarray" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_idarray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_idarray" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
  }
  result = XSolvable_lookup_idarray(arg1,arg2,arg3);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_deparray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  Id arg3 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_lookup_deparray", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_deparray" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_lookup_deparray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_lookup_deparray" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
  }
  result = XSolvable_lookup_deparray(arg1,arg2,arg3);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Dep * result = new_Dep(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_location(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  arg2 = &temp2;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_location" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_lookup_location(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags), 0);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_lookup_sourcepkg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_lookup_sourcepkg" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_lookup_sourcepkg(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_Dataiterator", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_Dataiterator" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_Dataiterator" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XSolvable_Dataiterator" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_Dataiterator" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (Dataiterator *)XSolvable_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_installable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_installable" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (bool)XSolvable_installable(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_isinstalled(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_isinstalled" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (bool)XSolvable_isinstalled(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_name_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_name_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_name_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_name_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  XSolvable_name_set(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_name_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_name_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_nameid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_nameid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_nameid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_nameid_set" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  XSolvable_nameid_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_nameid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_nameid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Id)XSolvable_nameid_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_evr_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_evr_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evr_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_evr_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  XSolvable_evr_set(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_evr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evr_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_evr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_evrid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_evrid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_evrid_set" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  XSolvable_evrid_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_evrid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Id)XSolvable_evrid_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_arch_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_arch_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_arch_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_arch_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  XSolvable_arch_set(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_arch_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_arch_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_arch_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_archid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_archid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_archid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_archid_set" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  XSolvable_archid_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_archid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_archid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Id)XSolvable_archid_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_vendor_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_vendor_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendor_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_vendor_set" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  XSolvable_vendor_set(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_vendor_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendor_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable_vendor_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_vendorid_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_vendorid_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendorid_set" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_vendorid_set" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  XSolvable_vendorid_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_vendorid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_vendorid_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Id)XSolvable_vendorid_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_repo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_repo_get" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (Repo *)XSolvable_repo_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_provides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  Id arg3 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_provides", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_provides" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_provides" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_provides" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
  }
  XSolvable_add_provides(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_obsoletes(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_obsoletes", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_obsoletes" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_obsoletes" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_obsoletes(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_conflicts(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_conflicts", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_conflicts" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_conflicts" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_conflicts(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_requires(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  Id arg3 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_requires", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_requires" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_requires" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_requires" "', argument " "3"" of type '" "Id""'");
    } 
    arg3 = (Id)(val3);
  }
  XSolvable_add_requires(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_recommends(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_recommends", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_recommends" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_recommends" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_recommends(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_suggests(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_suggests", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_suggests" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_suggests" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_suggests(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_supplements(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_supplements", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_supplements" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_supplements" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_supplements(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_enhances(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_enhances", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_enhances" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_enhances" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  XSolvable_add_enhances(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_unset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_unset", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_unset" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_unset" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  XSolvable_unset(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_add_deparray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  DepId arg3 ;
  Id arg4 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_add_deparray", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_add_deparray" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_add_deparray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_add_deparray" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_add_deparray" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
  }
  XSolvable_add_deparray(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_Selection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_Selection", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_Selection" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_Selection" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)XSolvable_Selection(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_identical(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_identical", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_identical" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_identical" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (bool)XSolvable_identical(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_evrcmp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_evrcmp", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_evrcmp" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable_evrcmp" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (int)XSolvable_evrcmp(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable_matchesdep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  Id arg2 ;
  DepId arg3 ;
  Id arg4 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable_matchesdep", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable_matchesdep" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XSolvable_matchesdep" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XSolvable_matchesdep" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XSolvable_matchesdep" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
  }
  result = (bool)XSolvable_matchesdep(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___eq__" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable___eq__" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (bool)XSolvable___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XSolvable___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___ne__" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XSolvable___ne__" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (bool)XSolvable___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable___hash__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___hash__" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (int)XSolvable___hash__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___str__" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XSolvable___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XSolvable___repr__" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  result = (char *)XSolvable___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_XSolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XSolvable *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XSolvable, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XSolvable" "', argument " "1"" of type '" "XSolvable *""'"); 
  }
  arg1 = (XSolvable *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *XSolvable_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_XSolvable, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *XSolvable_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Ruleinfo_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_solv_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_type_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (int)(int) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_dep_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_dep_id_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (Id)(Id) ((arg1)->dep_id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Ruleinfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  Id arg4 ;
  Id arg5 ;
  Id arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  Ruleinfo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Ruleinfo", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Ruleinfo" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Ruleinfo" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Ruleinfo" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Ruleinfo" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Ruleinfo" "', argument " "5"" of type '" "Id""'");
  } 
  arg5 = (Id)(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Ruleinfo" "', argument " "6"" of type '" "Id""'");
  } 
  arg6 = (Id)(val6);
  result = (Ruleinfo *)new_Ruleinfo(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_solvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_solvable_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (XSolvable *)Ruleinfo_solvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_othersolvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_othersolvable_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (XSolvable *)Ruleinfo_othersolvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_dep_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_dep_get" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (Dep *)Ruleinfo_dep_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo_problemstr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo_problemstr" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (char *)Ruleinfo_problemstr(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Ruleinfo___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Ruleinfo___str__" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  result = (char *)Ruleinfo___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Ruleinfo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Ruleinfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Ruleinfo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Ruleinfo" "', argument " "1"" of type '" "Ruleinfo *""'"); 
  }
  arg1 = (Ruleinfo *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Ruleinfo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Ruleinfo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Ruleinfo_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_XRule_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_solv_get" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_id_get" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_XRule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XRule *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_XRule", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XRule" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XRule" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XRule *)new_XRule(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_type_get" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (int)XRule_type_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Ruleinfo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_info" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (Ruleinfo *)XRule_info(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_allinfos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_allinfos" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = XRule_allinfos(arg1);
  {
    int i;
    int cnt = (&result)->count / 4;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 4)
    {
      Id id = *idp;
#define result resultx
      Ruleinfo * result = new_Ruleinfo(arg1->solv, arg1->id, id, idp[1], idp[2], idp[3]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_get_learnt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_learnt" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = XRule_get_learnt(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XRule * result = new_XRule(arg1->solv, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_get_decisionlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_decisionlist" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = XRule_get_decisionlist(arg1);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule_get_decisionsetlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule_get_decisionsetlist" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = XRule_get_decisionsetlist(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Decisionset * result = decisionset_fromids(arg1->solv, idp + id, idp[1] - id + 1);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  XRule *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRule___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___eq__" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRule___eq__" "', argument " "2"" of type '" "XRule *""'"); 
  }
  arg2 = (XRule *)(argp2);
  result = (bool)XRule___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  XRule *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRule___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___ne__" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRule___ne__" "', argument " "2"" of type '" "XRule *""'"); 
  }
  arg2 = (XRule *)(argp2);
  result = (bool)XRule___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule___hash__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___hash__" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (int)XRule___hash__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRule___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRule___repr__" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  result = (char *)XRule___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_XRule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRule *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRule, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XRule" "', argument " "1"" of type '" "XRule *""'"); 
  }
  arg1 = (XRule *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *XRule_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_XRule, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *XRule_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_XRepodata_repo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_repo_get" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  result = (Repo *)(Repo *) ((arg1)->repo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_id_get" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_XRepodata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XRepodata *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_XRepodata", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XRepodata" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_XRepodata" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XRepodata *)new_XRepodata(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_new_handle(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_new_handle" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  result = (Id)XRepodata_new_handle(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  DepId arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  DepId val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_id", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_id" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_id" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsValDepId(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_set_id" "', argument " "4"" of type '" "DepId""'");
  } 
  arg4 = (DepId)(val4);
  XRepodata_set_id(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  unsigned long long arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned long long val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_num", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_num" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_num" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_num" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_set_num" "', argument " "4"" of type '" "unsigned long long""'");
  } 
  arg4 = (unsigned long long)(val4);
  XRepodata_set_num(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_str", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_str" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_str" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_str" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  XRepodata_set_str(arg1,arg2,arg3,(char const *)arg4);
  resultobj = SWIG_Py_Void();
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_void(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_void", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_void" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_void" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_void" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  XRepodata_set_void(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_poolstr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_poolstr", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_poolstr" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_poolstr" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_poolstr" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_poolstr" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  XRepodata_set_poolstr(arg1,arg2,arg3,(char const *)arg4);
  resultobj = SWIG_Py_Void();
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_add_idarray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  DepId arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  DepId val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_add_idarray", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_idarray" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_idarray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_idarray" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsValDepId(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_idarray" "', argument " "4"" of type '" "DepId""'");
  } 
  arg4 = (DepId)(val4);
  XRepodata_add_idarray(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_add_flexarray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  Id arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_add_flexarray", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_flexarray" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_flexarray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_flexarray" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_flexarray" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  XRepodata_add_flexarray(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_checksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  Chksum *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_checksum", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_checksum" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_checksum" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_checksum" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  res4 = SWIG_ConvertPtr(swig_obj[3], &argp4,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_checksum" "', argument " "4"" of type '" "Chksum *""'"); 
  }
  arg4 = (Chksum *)(argp4);
  XRepodata_set_checksum(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_sourcepkg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_sourcepkg", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_sourcepkg" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_sourcepkg" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XRepodata_set_sourcepkg" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  XRepodata_set_sourcepkg(arg1,arg2,(char const *)arg3);
  resultobj = SWIG_Py_Void();
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_set_location(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  unsigned int arg3 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_set_location", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_set_location" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_set_location" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_set_location" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = (unsigned int)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XRepodata_set_location" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  XRepodata_set_location(arg1,arg2,arg3,(char const *)arg4);
  resultobj = SWIG_Py_Void();
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_unset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_unset", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_unset" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_unset" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_unset" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  XRepodata_unset(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_str", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_str" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_str" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (char *)XRepodata_lookup_str(arg1,arg2,arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_id", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_id" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_id" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Id)XRepodata_lookup_id(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  unsigned long long arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned long long val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned long long result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_num", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_num" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_num" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_num" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_lookup_num" "', argument " "4"" of type '" "unsigned long long""'");
    } 
    arg4 = (unsigned long long)(val4);
  }
  result = (unsigned long long)XRepodata_lookup_num(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_void(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_void", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_void" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_void" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_void" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (bool)XRepodata_lookup_void(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_idarray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_idarray", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_idarray" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_idarray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_idarray" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = XRepodata_lookup_idarray(arg1,arg2,arg3);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_lookup_checksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_lookup_checksum", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_lookup_checksum" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_lookup_checksum" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_lookup_checksum" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Chksum *)XRepodata_lookup_checksum(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_internalize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_internalize" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  XRepodata_internalize(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_create_stubs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_create_stubs" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  XRepodata_create_stubs(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_write(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_write", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_write" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_write" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  result = (bool)XRepodata_write(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_str2dir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  char *arg2 = 0 ;
  bool arg3 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_str2dir", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_str2dir" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata_str2dir" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_str2dir" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = (bool)(val3);
  }
  result = (Id)XRepodata_str2dir(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_dir2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_dir2str", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_dir2str" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_dir2str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XRepodata_dir2str" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  result = (char *)XRepodata_dir2str(arg1,arg2,(char const *)arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_add_dirstr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  Id arg4 ;
  char *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  PyObject *swig_obj[5] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_add_dirstr", 5, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_dirstr" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_dirstr" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_dirstr" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XRepodata_add_dirstr" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  res5 = SWIG_AsCharPtrAndSize(swig_obj[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "XRepodata_add_dirstr" "', argument " "5"" of type '" "char const *""'");
  }
  arg5 = (char *)(buf5);
  XRepodata_add_dirstr(arg1,arg2,arg3,arg4,(char const *)arg5);
  resultobj = SWIG_Py_Void();
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return resultobj;
fail:
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_add_solv(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata_add_solv", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_add_solv" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XRepodata_add_solv" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "XRepodata_add_solv" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)XRepodata_add_solv(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata_extend_to_repo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata_extend_to_repo" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  XRepodata_extend_to_repo(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  XRepodata *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___eq__" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata___eq__" "', argument " "2"" of type '" "XRepodata *""'"); 
  }
  arg2 = (XRepodata *)(argp2);
  result = (bool)XRepodata___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  XRepodata *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "XRepodata___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___ne__" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XRepodata___ne__" "', argument " "2"" of type '" "XRepodata *""'"); 
  }
  arg2 = (XRepodata *)(argp2);
  result = (bool)XRepodata___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata___hash__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___hash__" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  result = (int)XRepodata___hash__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_XRepodata___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XRepodata___repr__" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  result = (char *)XRepodata___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_XRepodata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  XRepodata *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_XRepodata, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XRepodata" "', argument " "1"" of type '" "XRepodata *""'"); 
  }
  arg1 = (XRepodata *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *XRepodata_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_XRepodata, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *XRepodata_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Pool_solvable_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Pool_solvable_iterator" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_solvable_iterator *)new_Pool_solvable_iterator(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvable_iterator___next__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___next__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
  }
  arg1 = (Pool_solvable_iterator *)(argp1);
  {
    result = (XSolvable *)Pool_solvable_iterator___next__(arg1);
    if (!result) {
      PyErr_SetString(PyExc_StopIteration,"no more matches");
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvable_iterator___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_solvable_iterator *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_solvable_iterator___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___getitem__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
  }
  arg1 = (Pool_solvable_iterator *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_solvable_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XSolvable *)Pool_solvable_iterator___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvable_iterator___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvable_iterator___len__" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
  }
  arg1 = (Pool_solvable_iterator *)(argp1);
  result = (int)Pool_solvable_iterator___len__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Pool_solvable_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_solvable_iterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool_solvable_iterator" "', argument " "1"" of type '" "Pool_solvable_iterator *""'"); 
  }
  arg1 = (Pool_solvable_iterator *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Pool_solvable_iterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Pool_solvable_iterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Pool_solvable_iterator_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Pool_repo_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_repo_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Pool_repo_iterator" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_repo_iterator *)new_Pool_repo_iterator(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_repo_iterator___next__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_repo_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___next__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
  }
  arg1 = (Pool_repo_iterator *)(argp1);
  {
    result = (Repo *)Pool_repo_iterator___next__(arg1);
    if (!result) {
      PyErr_SetString(PyExc_StopIteration,"no more matches");
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_repo_iterator___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_repo_iterator *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_repo_iterator___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___getitem__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
  }
  arg1 = (Pool_repo_iterator *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_repo_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Repo *)Pool_repo_iterator___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_repo_iterator___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_repo_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_repo_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repo_iterator___len__" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
  }
  arg1 = (Pool_repo_iterator *)(argp1);
  result = (int)Pool_repo_iterator___len__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Pool_repo_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool_repo_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool_repo_iterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool_repo_iterator" "', argument " "1"" of type '" "Pool_repo_iterator *""'"); 
  }
  arg1 = (Pool_repo_iterator *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Pool_repo_iterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Pool_repo_iterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Pool_repo_iterator_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Repo_solvable_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Repo_solvable_iterator" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Repo_solvable_iterator *)new_Repo_solvable_iterator(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_solvable_iterator___next__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___next__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
  }
  arg1 = (Repo_solvable_iterator *)(argp1);
  {
    result = (XSolvable *)Repo_solvable_iterator___next__(arg1);
    if (!result) {
      PyErr_SetString(PyExc_StopIteration,"no more matches");
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_solvable_iterator___getitem__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo_solvable_iterator *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_solvable_iterator___getitem__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___getitem__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
  }
  arg1 = (Repo_solvable_iterator *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_solvable_iterator___getitem__" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XSolvable *)Repo_solvable_iterator___getitem__(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_solvable_iterator___len__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo_solvable_iterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvable_iterator___len__" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
  }
  arg1 = (Repo_solvable_iterator *)(argp1);
  result = (int)Repo_solvable_iterator___len__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Repo_solvable_iterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo_solvable_iterator *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo_solvable_iterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Repo_solvable_iterator" "', argument " "1"" of type '" "Repo_solvable_iterator *""'"); 
  }
  arg1 = (Repo_solvable_iterator *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Repo_solvable_iterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Repo_solvable_iterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Repo_solvable_iterator_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Datamatch_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_pool_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_repo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_repo_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Repo *)(Repo *) ((arg1)->repo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_solvid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_solvid_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Id)(Id) ((arg1)->solvid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Datamatch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Datamatch" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  delete_Datamatch(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_solvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_solvable_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (XSolvable *)Datamatch_solvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_key_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_key_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Id)Datamatch_key_id_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_key_idstr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_key_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (char *)Datamatch_key_idstr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_type_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_type_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Id)Datamatch_type_id_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_type_idstr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_type_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (char *)Datamatch_type_idstr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_id_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Id)Datamatch_id_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_idstr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_idstr_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (char *)Datamatch_idstr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_str_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_str_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (char *)Datamatch_str_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_dep_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_dep_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Dep *)Datamatch_dep_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_binary_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  BinaryBlob result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_binary_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = Datamatch_binary_get(arg1);
  resultobj = (&result)->data ? Py_BuildValue("y#", (&result)->data, (Py_ssize_t)(&result)->len) : SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_num_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_num_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (unsigned long long)Datamatch_num_get(arg1);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_num2_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_num2_get" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (unsigned int)Datamatch_num2_get(arg1);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_pos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Datapos *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_pos" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Datapos *)Datamatch_pos(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch_parentpos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Datapos *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch_parentpos" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (Datapos *)Datamatch_parentpos(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datamatch___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datamatch *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datamatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datamatch___str__" "', argument " "1"" of type '" "Datamatch *""'"); 
  }
  arg1 = (Datamatch *)(argp1);
  result = (char *)Datamatch___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Datamatch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Datamatch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_Datapos_repo_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_repo_get" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  result = (Repo *)(Repo *) ((arg1)->repo);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_id", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_id" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Id)Datapos_lookup_id(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_str" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (char *)Datapos_lookup_str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  unsigned long long arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  unsigned long long val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  unsigned long long result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_num", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_num" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_num" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Datapos_lookup_num" "', argument " "3"" of type '" "unsigned long long""'");
    } 
    arg3 = (unsigned long long)(val3);
  }
  result = (unsigned long long)Datapos_lookup_num(arg1,arg2,arg3);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_void(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_void", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_void" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_void" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (bool)Datapos_lookup_void(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_checksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_checksum", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_checksum" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_checksum" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Chksum *)Datapos_lookup_checksum(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_deltaseq(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_deltaseq" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  result = (char *)Datapos_lookup_deltaseq(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_deltalocation(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  unsigned int *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int temp2 ;
  int res2 = SWIG_TMPOBJ ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  arg2 = &temp2;
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_deltalocation" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  result = (char *)Datapos_lookup_deltalocation(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  if (SWIG_IsTmpObj(res2)) {
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_From_unsigned_SS_int((*arg2)), 0);
  } else {
    int new_flags = SWIG_IsNewObj(res2) ? (SWIG_POINTER_OWN |  0 ) :  0 ;
    resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(arg2), SWIGTYPE_p_unsigned_int, new_flags), 0);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_lookup_idarray(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_lookup_idarray", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_lookup_idarray" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_lookup_idarray" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = Datapos_lookup_idarray(arg1,arg2);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Datapos_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Datapos_Dataiterator", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Datapos_Dataiterator" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Datapos_Dataiterator" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Datapos_Dataiterator" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Datapos_Dataiterator" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (Dataiterator *)Datapos_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Datapos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Datapos *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Datapos, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Datapos" "', argument " "1"" of type '" "Datapos *""'"); 
  }
  arg1 = (Datapos *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Datapos_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Datapos, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_Job_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_pool_get" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_how_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Job_how_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_how_set" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Job_how_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->how = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_how_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_how_get" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (int) ((arg1)->how);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_what_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Job_what_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_what_set" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Job_what_set" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (arg1) (arg1)->what = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_what_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_what_get" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (Id) ((arg1)->what);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Job(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Job *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Job", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Job" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Job" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Job" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Job *)new_Job(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_solvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_solvables" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = Job_solvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job_isemptyupdate(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job_isemptyupdate" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (bool)Job_isemptyupdate(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  Job *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Job___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___eq__" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Job___eq__" "', argument " "2"" of type '" "Job *""'"); 
  }
  arg2 = (Job *)(argp2);
  result = (bool)Job___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  Job *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Job___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___ne__" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Job___ne__" "', argument " "2"" of type '" "Job *""'"); 
  }
  arg2 = (Job *)(argp2);
  result = (bool)Job___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___str__" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (char *)Job___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Job___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Job___repr__" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  result = (char *)Job___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Job(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Job *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Job, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Job" "', argument " "1"" of type '" "Job *""'"); 
  }
  arg1 = (Job *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Job_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Job, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Job_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Pool(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Pool", 0, 0, 0)) SWIG_fail;
  result = (Pool *)new_Pool();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_setdisttype(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_setdisttype", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setdisttype" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_setdisttype" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)Pool_setdisttype(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_debuglevel(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_debuglevel", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_debuglevel" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_debuglevel" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  Pool_set_debuglevel(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_flag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_flag", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_flag" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_flag" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_set_flag" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (int)Pool_set_flag(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_get_flag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_get_flag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_flag" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_get_flag" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)Pool_get_flag(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_rootdir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_rootdir", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_rootdir" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_set_rootdir" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  Pool_set_rootdir(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_get_rootdir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_get_rootdir", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_rootdir" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_get_rootdir" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (char *)Pool_get_rootdir(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_clr_loadcallback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_clr_loadcallback" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  Pool_clr_loadcallback(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_loadcallback(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  PyObject *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_loadcallback", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_loadcallback" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  arg2 = swig_obj[1];
  Pool_set_loadcallback(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Pool(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Pool" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  delete_Pool(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_free(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  disown_helper result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_free" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_free(arg1);
  {
    SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    resultobj = SWIG_From_int((int)(0));
    
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_disown(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  disown_helper result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_disown" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_disown(arg1);
  {
    SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, SWIG_POINTER_DISOWN |  0 );
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    resultobj = SWIG_From_int((int)(0));
    
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_appdata_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  AppObjectPtr arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_appdata_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_set" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    if (swig_obj[1])
    Py_INCREF(swig_obj[1]);
    arg2 = swig_obj[1];
  }
  Pool_appdata_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_appdata_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  AppObjectPtr result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_get" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (AppObjectPtr)Pool_appdata_get(arg1);
  {
    resultobj = result ? result : Py_None;
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_appdata_disown(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_appdata_disown" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  Pool_appdata_disown(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_str2id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  bool arg3 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_str2id", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_str2id" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_str2id" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_str2id" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = (bool)(val3);
  }
  result = (Id)Pool_str2id(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Dep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  bool arg3 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_Dep", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dep" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_Dep" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_bool(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Dep" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = (bool)(val3);
  }
  result = (Dep *)Pool_Dep(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_parserpmrichdep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_parserpmrichdep", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_parserpmrichdep" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_parserpmrichdep" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (Dep *)Pool_parserpmrichdep(arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_id2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_id2str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (char *)Pool_id2str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_dep2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_dep2str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_dep2str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_dep2str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (char *)Pool_dep2str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_rel2id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  int arg4 ;
  bool arg5 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_rel2id", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_rel2id" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_rel2id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_rel2id" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_rel2id" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_bool(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_rel2id" "', argument " "5"" of type '" "bool""'");
    } 
    arg5 = (bool)(val5);
  }
  result = (Id)Pool_rel2id(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_id2langid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  bool arg4 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_id2langid", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2langid" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2langid" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_id2langid" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_id2langid" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = (bool)(val4);
  }
  result = (Id)Pool_id2langid(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_int((int)(result));
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_setarch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_setarch", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setarch" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  if (swig_obj[1]) {
    res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_setarch" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
  }
  Pool_setarch(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_add_repo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_add_repo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_add_repo" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_add_repo" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (Repo *)Pool_add_repo(arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_lookup_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_lookup_str", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_str" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (char *)Pool_lookup_str(arg1,arg2,arg3);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_lookup_id(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Id result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_lookup_id", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_id" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_id" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_id" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Id)Pool_lookup_id(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_lookup_num(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  unsigned long long arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  unsigned long long val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  unsigned long long result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_lookup_num", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_num" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_num" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_num" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_unsigned_SS_long_SS_long(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_lookup_num" "', argument " "4"" of type '" "unsigned long long""'");
    } 
    arg4 = (unsigned long long)(val4);
  }
  result = (unsigned long long)Pool_lookup_num(arg1,arg2,arg3,arg4);
  resultobj = SWIG_From_unsigned_SS_long_SS_long((unsigned long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_lookup_void(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_lookup_void", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_void" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_void" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_void" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (bool)Pool_lookup_void(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_lookup_checksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_lookup_checksum", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_lookup_checksum" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_lookup_checksum" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_lookup_checksum" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Chksum *)Pool_lookup_checksum(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_Dataiterator", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dataiterator" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Dataiterator" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_Dataiterator" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_Dataiterator" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (Dataiterator *)Pool_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Dataiterator_solvid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  char *arg4 = 0 ;
  int arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_Dataiterator_solvid", 3, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Dataiterator_solvid" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Dataiterator_solvid" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Dataiterator_solvid" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  if (swig_obj[3]) {
    res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Pool_Dataiterator_solvid" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
  }
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_Dataiterator_solvid" "', argument " "5"" of type '" "int""'");
    } 
    arg5 = (int)(val5);
  }
  result = (Dataiterator *)Pool_Dataiterator_solvid(arg1,arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvid2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_solvid2str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvid2str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_solvid2str" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (char *)Pool_solvid2str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvidset2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_solvidset2str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvidset2str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      int v;
      int e = SWIG_AsVal_int(o, &v);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only " "integers");
      queue_push(&arg2, v);
    }
  }
  
  result = (char *)Pool_solvidset2str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvableset2str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_solvableset2str", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvableset2str" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      XSolvable * obj;
      int e = SWIG_ConvertPtr(o, (void **)&obj, SWIGTYPE_p_XSolvable, 0 | 0);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""XSolvable *");
      queue_push(&arg2, obj->id);
    }
  }
  
  result = (char *)Pool_solvableset2str(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_addfileprovides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_addfileprovides" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  Pool_addfileprovides(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_addfileprovides_queue(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_addfileprovides_queue" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_addfileprovides_queue(arg1);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_createwhatprovides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_createwhatprovides" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  Pool_createwhatprovides(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_id2solvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_id2solvable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2solvable" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2solvable" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (XSolvable *)Pool_id2solvable(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvables_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvables_get" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_solvable_iterator *)Pool_solvables_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_solvables_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_solvables_iter" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_solvable_iterator *)Pool_solvables_iter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_solvable_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_id2repo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_id2repo", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_id2repo" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_id2repo" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Repo *)Pool_id2repo(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_repos_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_repo_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repos_get" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_repo_iterator *)Pool_repos_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_repos_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool_repo_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_repos_iter" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Pool_repo_iterator *)Pool_repos_iter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool_repo_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_installed_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Repo *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_installed_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_installed_set" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Repo, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_installed_set" "', argument " "2"" of type '" "Repo *""'"); 
  }
  arg2 = (Repo *)(argp2);
  Pool_installed_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_installed_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_installed_get" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Repo *)Pool_installed_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_errstr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_errstr_get" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (char *)Pool_errstr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_matchprovidingids(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_matchprovidingids", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchprovidingids" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchprovidingids" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchprovidingids" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = Pool_matchprovidingids(arg1,(char const *)arg2,arg3);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Job(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 ;
  Id arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Job *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_Job", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Job" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Job" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_Job" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  result = (Job *)Pool_Job(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_whatprovides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_whatprovides", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatprovides" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatprovides" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  result = Pool_whatprovides(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_best_solvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Queue result;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_best_solvables", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_best_solvables" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      XSolvable * obj;
      int e = SWIG_ConvertPtr(o, (void **)&obj, SWIGTYPE_p_XSolvable, 0 | 0);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""XSolvable *");
      queue_push(&arg2, obj->id);
    }
  }
  
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_best_solvables" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = Pool_best_solvables(arg1,arg2,arg3);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_towhatprovides(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  Id result;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_towhatprovides", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_towhatprovides" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      int v;
      int e = SWIG_AsVal_int(o, &v);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only " "integers");
      queue_push(&arg2, v);
    }
  }
  
  result = (Id)Pool_towhatprovides(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_namespaceproviders(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  DepId arg2 ;
  DepId arg3 ;
  bool arg4 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_namespaceproviders", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_namespaceproviders" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_set_namespaceproviders" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_set_namespaceproviders" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_bool(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_set_namespaceproviders" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = (bool)(val4);
  }
  Pool_set_namespaceproviders(arg1,arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_flush_namespaceproviders(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  DepId arg2 ;
  DepId arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_flush_namespaceproviders", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_flush_namespaceproviders" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_flush_namespaceproviders" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_flush_namespaceproviders" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  Pool_flush_namespaceproviders(arg1,arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_whatcontainsdep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  DepId arg3 ;
  Id arg4 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_whatcontainsdep", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatcontainsdep" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatcontainsdep" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_whatcontainsdep" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatcontainsdep" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
  }
  result = Pool_whatcontainsdep(arg1,arg2,arg3,arg4);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_whatmatchesdep(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  DepId arg3 ;
  Id arg4 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  DepId val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_whatmatchesdep", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatmatchesdep" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatmatchesdep" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsValDepId(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_whatmatchesdep" "', argument " "3"" of type '" "DepId""'");
  } 
  arg3 = (DepId)(val3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatmatchesdep" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
  }
  result = Pool_whatmatchesdep(arg1,arg2,arg3,arg4);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_whatmatchessolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Id arg2 ;
  XSolvable *arg3 = 0 ;
  Id arg4 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_whatmatchessolvable", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_whatmatchessolvable" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_whatmatchessolvable" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Pool_whatmatchessolvable" "', argument " "3"" of type '" "XSolvable *""'"); 
  }
  arg3 = (XSolvable *)(argp3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_whatmatchessolvable" "', argument " "4"" of type '" "Id""'");
    } 
    arg4 = (Id)(val4);
  }
  result = Pool_whatmatchessolvable(arg1,arg2,arg3,arg4);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_isknownarch(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  DepId arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_isknownarch", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_isknownarch" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_isknownarch" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  result = (bool)Pool_isknownarch(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Solver(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Solver" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Solver *)Pool_Solver(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Selection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Selection" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = (Selection *)Pool_Selection(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_Selection_all(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_Selection_all", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_Selection_all" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_Selection_all" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)Pool_Selection_all(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_select(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_select", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_select" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_select" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_select" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (Selection *)Pool_select(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_matchdeps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_matchdeps", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchdeps" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchdeps" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchdeps" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchdeps" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchdeps" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  result = (Selection *)Pool_matchdeps(arg1,(char const *)arg2,arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_matchdepid(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  DepId arg2 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  DepId val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_matchdepid", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchdepid" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  ecode2 = SWIG_AsValDepId(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Pool_matchdepid" "', argument " "2"" of type '" "DepId""'");
  } 
  arg2 = (DepId)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchdepid" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchdepid" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchdepid" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  result = (Selection *)Pool_matchdepid(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_matchsolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  int arg3 ;
  Id arg4 ;
  Id arg5 = -1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_matchsolvable", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_matchsolvable" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Pool_matchsolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Pool_matchsolvable" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Pool_matchsolvable" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Pool_matchsolvable" "', argument " "5"" of type '" "Id""'");
    } 
    arg5 = (Id)(val5);
  }
  result = (Selection *)Pool_matchsolvable(arg1,arg2,arg3,arg4,arg5);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_get_considered_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_considered_list" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_get_considered_list(arg1);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_get_disabled_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_get_disabled_list" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_get_disabled_list(arg1);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_considered_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_considered_list", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_considered_list" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      int v;
      int e = SWIG_AsVal_int(o, &v);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only " "integers");
      queue_push(&arg2, v);
    }
  }
  
  Pool_set_considered_list(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_set_disabled_list(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_set_disabled_list", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_set_disabled_list" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      int v;
      int e = SWIG_AsVal_int(o, &v);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only " "integers");
      queue_push(&arg2, v);
    }
  }
  
  Pool_set_disabled_list(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_setpooljobs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Pool_setpooljobs", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_setpooljobs" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      Job * obj;
      int e = SWIG_ConvertPtr(o, (void **)&obj, SWIGTYPE_p_Job, 0 | 0);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""Job *");
      queue_push2(&arg2, obj->how, obj->what);
    }
  }
  
  Pool_setpooljobs(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Pool_getpooljobs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Pool *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Pool, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Pool_getpooljobs" "', argument " "1"" of type '" "Pool *""'"); 
  }
  arg1 = (Pool *)(argp1);
  result = Pool_getpooljobs(arg1);
  {
    int i;
    int cnt = (&result)->count / 2;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 2)
    {
      Id id = *idp;
#define result resultx
      Job * result = new_Job(arg1, id, idp[1]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Pool_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Pool, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Pool_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Repo_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_pool_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_name_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_name_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (char *)(char *) ((arg1)->name);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_priority_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_priority_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_priority_set" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_priority_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->priority = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_priority_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_priority_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (int) ((arg1)->priority);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_subpriority_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_subpriority_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_subpriority_set" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_subpriority_set" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (arg1) (arg1)->subpriority = arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_subpriority_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_subpriority_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (int) ((arg1)->subpriority);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_nsolvables_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_nsolvables_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (int)(int) ((arg1)->nsolvables);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_free(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_free", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_free" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_free" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  Repo_free(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_empty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_empty", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_empty" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_empty" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  Repo_empty(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_isempty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_isempty" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (bool)Repo_isempty(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_appdata_set(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  AppObjectPtr arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_appdata_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_appdata_set" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  {
    if (swig_obj[1])
    Py_INCREF(swig_obj[1]);
    arg2 = swig_obj[1];
  }
  Repo_appdata_set(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_appdata_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  AppObjectPtr result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_appdata_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (AppObjectPtr)Repo_appdata_get(arg1);
  {
    resultobj = result ? result : Py_None;
    Py_INCREF(resultobj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_solv__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solv" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_solv" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_solv" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_solv__SWIG_0(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_solv__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  bool result;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solv" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_solv" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_solv" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_solv__SWIG_1(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_solv(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "Repo_add_solv", 0, 3, argv))) SWIG_fail;
  --argc;
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Repo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsValSolvFpPtr(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_Repo_add_solv__SWIG_1(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Repo_add_solv__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_Repo, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_Repo_add_solv__SWIG_0(self, argc, argv);
        }
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_Repo_add_solv__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'Repo_add_solv'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Repo::add_solv(char const *,int)\n"
    "    Repo::add_solv(FILE *,int)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Repo_add_solvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_solvable" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (XSolvable *)Repo_add_solvable(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_rpmdb(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_rpmdb", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (bool)Repo_add_rpmdb(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_rpmdb_reffp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_rpmdb_reffp", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb_reffp" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb_reffp" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_rpmdb_reffp" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_rpmdb_reffp(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_rpm(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_rpm", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpm" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_rpm" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_rpm" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (XSolvable *)Repo_add_rpm(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_rpmdb_pubkeys(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_rpmdb_pubkeys", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmdb_pubkeys" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmdb_pubkeys" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (bool)Repo_add_rpmdb_pubkeys(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_pubkey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_pubkey", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_pubkey" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_pubkey" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_pubkey" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (XSolvable *)Repo_add_pubkey(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_keyring(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_keyring", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_keyring" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_keyring" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_keyring" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_keyring(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_keydir(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_keydir", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_keydir" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_keydir" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_add_keydir" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_add_keydir" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (bool)Repo_add_keydir(arg1,(char const *)arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_rpmmd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_rpmmd", 3, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_rpmmd" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_rpmmd" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_add_rpmmd" "', argument " "3"" of type '" "char const *""'");
  }
  arg3 = (char *)(buf3);
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_add_rpmmd" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (bool)Repo_add_rpmmd(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_repomdxml(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_repomdxml", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_repomdxml" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_repomdxml" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_repomdxml" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_repomdxml(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_updateinfoxml(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_updateinfoxml", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_updateinfoxml" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_updateinfoxml" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_updateinfoxml" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_updateinfoxml(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_deltainfoxml(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_deltainfoxml", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_deltainfoxml" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_deltainfoxml" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_deltainfoxml" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_deltainfoxml(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_debdb(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_debdb", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_debdb" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_debdb" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (bool)Repo_add_debdb(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_debpackages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_debpackages", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_debpackages" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_debpackages" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_debpackages" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_debpackages(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_deb(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_deb", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_deb" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_deb" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_deb" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (XSolvable *)Repo_add_deb(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_susetags(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  Id arg3 ;
  char *arg4 = 0 ;
  int arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  PyObject *swig_obj[5] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_susetags", 4, 5, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_susetags" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_susetags" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_susetags" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  res4 = SWIG_AsCharPtrAndSize(swig_obj[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Repo_add_susetags" "', argument " "4"" of type '" "char const *""'");
  }
  arg4 = (char *)(buf4);
  if (swig_obj[4]) {
    ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Repo_add_susetags" "', argument " "5"" of type '" "int""'");
    } 
    arg5 = (int)(val5);
  }
  result = (bool)Repo_add_susetags(arg1,arg2,arg3,(char const *)arg4,arg5);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return resultobj;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_content(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_content", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_content" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_content" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_content" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_content(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_products(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_products", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_products" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_products" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_products" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_products(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_mdk(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_mdk", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_mdk" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_mdk" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_mdk" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_mdk(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_mdk_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_mdk_info", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_mdk_info" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_mdk_info" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_mdk_info" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_mdk_info(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_arch_repo(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_arch_repo", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_repo" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_arch_repo" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_repo" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_arch_repo(arg1,arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_arch_local(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_arch_local", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_local" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_arch_local" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_local" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (bool)Repo_add_arch_local(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_arch_pkg(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  int arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_arch_pkg", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_arch_pkg" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_add_arch_pkg" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  if (swig_obj[2]) {
    ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Repo_add_arch_pkg" "', argument " "3"" of type '" "int""'");
    } 
    arg3 = (int)(val3);
  }
  result = (XSolvable *)Repo_add_arch_pkg(arg1,(char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_internalize(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_internalize" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  Repo_internalize(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_write(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_write", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_write" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_write" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  result = (bool)Repo_write(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_write_first_repodata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_write_first_repodata", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_write_first_repodata" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_write_first_repodata" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  result = (bool)Repo_write_first_repodata(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_Dataiterator(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_Dataiterator", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Dataiterator" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Dataiterator" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_Dataiterator" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_Dataiterator" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (Dataiterator *)Repo_Dataiterator(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_Dataiterator_meta(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Id arg2 ;
  char *arg3 = 0 ;
  int arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Dataiterator *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_Dataiterator_meta", 2, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Dataiterator_meta" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Dataiterator_meta" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Repo_Dataiterator_meta" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  if (swig_obj[3]) {
    ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Repo_Dataiterator_meta" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
  }
  result = (Dataiterator *)Repo_Dataiterator_meta(arg1,arg2,(char const *)arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dataiterator, SWIG_POINTER_OWN |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_id_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Id)Repo_id_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_solvables_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvables_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Repo_solvable_iterator *)Repo_solvables_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_meta_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Datapos *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_meta_get" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Datapos *)Repo_meta_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Datapos, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_solvables_iter(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Repo_solvable_iterator *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_solvables_iter" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (Repo_solvable_iterator *)Repo_solvables_iter(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo_solvable_iterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_add_repodata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  XRepodata *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_add_repodata", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_add_repodata" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_add_repodata" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (XRepodata *)Repo_add_repodata(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_create_stubs(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_create_stubs" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  Repo_create_stubs(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_iscontiguous(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_iscontiguous" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (bool)Repo_iscontiguous(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_first_repodata(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XRepodata *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_first_repodata" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (XRepodata *)Repo_first_repodata(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRepodata, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_Selection(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Selection *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_Selection", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_Selection" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Repo_Selection" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = (Selection *)Repo_Selection(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Selection, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_find_pubkey(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_find_pubkey", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_find_pubkey" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_find_pubkey" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (XSolvable *)Repo_find_pubkey(arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_createshadow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  Repo *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_createshadow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_createshadow" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo_createshadow" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (Repo *)Repo_createshadow(arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Repo, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo_moveshadow(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo_moveshadow", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo_moveshadow" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      int v;
      int e = SWIG_AsVal_int(o, &v);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only " "integers");
      queue_push(&arg2, v);
    }
  }
  
  Repo_moveshadow(arg1,arg2);
  resultobj = SWIG_Py_Void();
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Repo *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___eq__" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo___eq__" "', argument " "2"" of type '" "Repo *""'"); 
  }
  arg2 = (Repo *)(argp2);
  result = (bool)Repo___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  Repo *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Repo___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___ne__" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Repo___ne__" "', argument " "2"" of type '" "Repo *""'"); 
  }
  arg2 = (Repo *)(argp2);
  result = (bool)Repo___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo___hash__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___hash__" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (int)Repo___hash__(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___str__" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (char *)Repo___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Repo___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Repo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Repo___repr__" "', argument " "1"" of type '" "Repo *""'"); 
  }
  arg1 = (Repo *)(argp1);
  result = (char *)Repo___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Repo_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Repo, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_Decision_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_solv_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_p_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_p_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (Id)(Id) ((arg1)->p);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_reason_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_reason_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (int)(int) ((arg1)->reason);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_infoid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_infoid_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (Id)(Id) ((arg1)->infoid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Decision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  int arg3 ;
  Id arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject *swig_obj[4] ;
  Decision *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Decision", 4, 4, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Decision" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Decision" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Decision" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Decision" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  result = (Decision *)new_Decision(arg1,arg2,arg3,arg4);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_rule_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XRule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_rule_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (XRule *)Decision_rule_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_solvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_solvable_get" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (XSolvable *)Decision_solvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Ruleinfo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_info" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (Ruleinfo *)Decision_info(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_allinfos(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_allinfos" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = Decision_allinfos(arg1);
  {
    int i;
    int cnt = (&result)->count / 4;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 4)
    {
      Id id = *idp;
#define result resultx
      Ruleinfo * result = new_Ruleinfo(arg1->solv, arg1->infoid, id, idp[1], idp[2], idp[3]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision_reasonstr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Decision_reasonstr", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision_reasonstr" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Decision_reasonstr" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  result = (char *)Decision_reasonstr(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decision___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decision___str__" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  result = (char *)Decision___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Decision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decision *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decision, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Decision" "', argument " "1"" of type '" "Decision *""'"); 
  }
  arg1 = (Decision *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Decision_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Decision, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Decision_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Decisionset_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_solv_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_p_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_p_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Id)(Id) ((arg1)->p);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_reason_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_reason_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (int)(int) ((arg1)->reason);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_infoid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_infoid_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Id)(Id) ((arg1)->infoid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_bits_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_bits_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (int)(int) ((arg1)->bits);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_type_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (int)(int) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_dep_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_dep_id_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Id)(Id) ((arg1)->dep_id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Decisionset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Decisionset *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Decisionset" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  result = (Decisionset *)new_Decisionset(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Decisionset(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Decisionset" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  delete_Decisionset(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_info(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Ruleinfo *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_info" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Ruleinfo *)Decisionset_info(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Ruleinfo, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_dep_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_dep_get" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (Dep *)Decisionset_dep_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_solvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_solvables" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = Decisionset_solvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->solv->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_decisions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_decisions" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = Decisionset_decisions(arg1);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset_reasonstr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  char *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Decisionset_reasonstr", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset_reasonstr" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Decisionset_reasonstr" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  result = (char *)Decisionset_reasonstr(arg1,arg2);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Decisionset___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Decisionset *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Decisionset, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Decisionset___str__" "', argument " "1"" of type '" "Decisionset *""'"); 
  }
  arg1 = (Decisionset *)(argp1);
  result = (char *)Decisionset___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Decisionset_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Decisionset, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Decisionset_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Solver_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_pool_get" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Solver(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solver" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  delete_Solver(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_set_flag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  int arg2 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_set_flag", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_set_flag" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_set_flag" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Solver_set_flag" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (int)Solver_set_flag(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_flag(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_flag", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_flag" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_flag" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  result = (int)Solver_get_flag(arg1,arg2);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_solve(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Queue arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  {
    queue_init(&arg2);
  }
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_solve", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_solve" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  {
    int i, size;
    if (!PyList_Check(swig_obj[1]))
    SWIG_exception_fail(SWIG_TypeError, "argument 2 is not a list");
    size = PyList_Size(swig_obj[1]);
    for (i = 0; i < size; i++) {
      PyObject *o = PyList_GetItem(swig_obj[1],i);
      Job * obj;
      int e = SWIG_ConvertPtr(o, (void **)&obj, SWIGTYPE_p_Job, 0 | 0);
      if (!SWIG_IsOK(e))
      SWIG_exception_fail(SWIG_ArgError(e), "list in argument 2 must contain only ""Job *");
      queue_push2(&arg2, obj->how, obj->what);
    }
  }
  
  result = Solver_solve(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Problem * result = new_Problem(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Problem, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  {
    queue_free(&arg2);
  }
  return resultobj;
fail:
  {
    queue_free(&arg2);
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_transaction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Transaction *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_transaction" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  result = (Transaction *)Solver_transaction(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Transaction, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_describe_decision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  XRule **arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  XRule *temp3 ;
  PyObject *swig_obj[2] ;
  int result;
  
  arg3 = &temp3;
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_describe_decision", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_describe_decision" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_describe_decision" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (int)Solver_describe_decision(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  resultobj = SWIG_Python_AppendOutput(resultobj, SWIG_NewPointerObj((void*)(*arg3), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 ), 0);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_describe_weakdep_decision_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_describe_weakdep_decision_raw", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_describe_weakdep_decision_raw" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_describe_weakdep_decision_raw" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = Solver_describe_weakdep_decision_raw(arg1,arg2);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_alternatives_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_alternatives_count" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  result = (int)Solver_alternatives_count(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_alternative(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Alternative *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_alternative", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_alternative" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_alternative" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Alternative *)Solver_get_alternative(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Alternative, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_alternatives(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_alternatives" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  result = Solver_alternatives(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Alternative * result = Solver_get_alternative(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Alternative, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_write_testcase(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_write_testcase", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_write_testcase" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_write_testcase" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (bool)Solver_write_testcase(arg1,(char const *)arg2);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_raw_decisions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_raw_decisions", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_raw_decisions" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_raw_decisions" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = Solver_raw_decisions(arg1,arg2);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_all_decisions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_all_decisions", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_all_decisions" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_all_decisions" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = Solver_all_decisions(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Decision * result = new_Decision(arg1, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_decision(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  Decision *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_decision", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_decision" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_decision" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (Decision *)Solver_get_decision(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_learnt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_learnt", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_learnt" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_learnt" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = Solver_get_learnt(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XRule * result = new_XRule(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_decisionlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_decisionlist", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_decisionlist" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solver_get_decisionlist" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = Solver_get_decisionlist(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Decision * result = new_Decision(arg1, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_recommended(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_recommended", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_recommended" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_recommended" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  result = Solver_get_recommended(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solver_get_suggested(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solver_get_suggested", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solver_get_suggested" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solver_get_suggested" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  result = Solver_get_suggested(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Solver_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Solver, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Chksum__SWIG_0(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Id arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  Chksum *result = 0 ;
  
  (void)self;
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Chksum" "', argument " "1"" of type '" "Id""'");
  } 
  arg1 = (Id)(val1);
  result = (Chksum *)new_Chksum__SWIG_0(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Chksum__SWIG_1(PyObject *self, Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  Id arg1 ;
  char *arg2 = 0 ;
  int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  Chksum *result = 0 ;
  
  (void)self;
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Chksum" "', argument " "1"" of type '" "Id""'");
  } 
  arg1 = (Id)(val1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Chksum" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  result = (Chksum *)new_Chksum__SWIG_1(arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_NEW |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Chksum(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_Chksum", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 1) {
    int _v = 0;
    {
      int res = SWIG_AsVal_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_Chksum__SWIG_0(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v = 0;
    {
      int res = SWIG_AsVal_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_Chksum__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_Chksum'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    Chksum::Chksum(Id)\n"
    "    Chksum::Chksum(Id,char const *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_Chksum_from_bin(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Id arg1 ;
  unsigned char *arg2 = 0 ;
  size_t arg3 ;
  int val1 ;
  int ecode1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  size_t size2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_from_bin", 2, 2, swig_obj)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Chksum_from_bin" "', argument " "1"" of type '" "Id""'");
  } 
  arg1 = (Id)(val1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, &size2, &alloc2);
  if (buf2 && size2)
  size2--;
  
  if (!SWIG_IsOK(res2)) {
    const void *pybuf = 0;
    Py_ssize_t pysize = 0;
#if PY_VERSION_HEX >= 0x03000000
    res2 = PyBytes_AsStringAndSize(swig_obj[1], (char **)&pybuf, &pysize);
#else
    res2 = PyObject_AsReadBuffer(swig_obj[1], &pybuf, &pysize);
#endif
    if (res2 < 0) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_from_bin" "', argument " "2"" of type '" "BinaryBlob""'");
    } else {
      buf2 = (void *)pybuf;
      size2 = pysize;
    }
    
    
    
  }
  arg2 = (unsigned char *)buf2;
  arg3 = size2;
  result = (Chksum *)Chksum_from_bin(arg1,(unsigned char const *)arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Chksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Chksum" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  delete_Chksum(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_type_get" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = (Id)Chksum_type_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_add(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  size_t size2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_add", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, &size2, &alloc2);
  if (buf2 && size2)
  size2--;
  
  if (!SWIG_IsOK(res2)) {
    const void *pybuf = 0;
    Py_ssize_t pysize = 0;
#if PY_VERSION_HEX >= 0x03000000
    res2 = PyBytes_AsStringAndSize(swig_obj[1], (char **)&pybuf, &pysize);
#else
    res2 = PyObject_AsReadBuffer(swig_obj[1], &pybuf, &pysize);
#endif
    if (res2 < 0) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_add" "', argument " "2"" of type '" "BinaryBlob""'");
    } else {
      buf2 = (void *)pybuf;
      size2 = pysize;
    }
    
    
    
  }
  arg2 = (unsigned char *)buf2;
  arg3 = size2;
  Chksum_add(arg1,(unsigned char const *)arg2,arg3);
  {
    resultobj = swig_obj[0];
    Py_INCREF(resultobj);
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_add_fp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  FILE *val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_add_fp", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fp" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  ecode2 = SWIG_AsValSolvFpPtr(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fp" "', argument " "2"" of type '" "FILE *""'");
  } 
  arg2 = (FILE *)(val2);
  Chksum_add_fp(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_add_fd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_add_fd", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fd" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fd" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  Chksum_add_fd(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_add_stat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_add_stat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_stat" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum_add_stat" "', argument " "2"" of type '" "char const *""'");
  }
  arg2 = (char *)(buf2);
  Chksum_add_stat(arg1,(char const *)arg2);
  resultobj = SWIG_Py_Void();
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_add_fstat(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum_add_fstat", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_add_fstat" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Chksum_add_fstat" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  Chksum_add_fstat(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  BinaryBlob result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_raw" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = Chksum_raw(arg1);
  resultobj = (&result)->data ? Py_BuildValue("y#", (&result)->data, (Py_ssize_t)(&result)->len) : SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_hex(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_hex" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = (char *)Chksum_hex(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum_typestr(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum_typestr" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = (char *)Chksum_typestr(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum___eq__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  Chksum *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum___eq__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___eq__" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum___eq__" "', argument " "2"" of type '" "Chksum *""'"); 
  }
  arg2 = (Chksum *)(argp2);
  result = (bool)Chksum___eq__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum___ne__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  Chksum *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Chksum___ne__", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___ne__" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Chksum___ne__" "', argument " "2"" of type '" "Chksum *""'"); 
  }
  arg2 = (Chksum *)(argp2);
  result = (bool)Chksum___ne__(arg1,arg2);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___str__" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = (char *)Chksum___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Chksum___repr__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Chksum *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Chksum___repr__" "', argument " "1"" of type '" "Chksum *""'"); 
  }
  arg1 = (Chksum *)(argp1);
  result = (char *)Chksum___repr__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  free((char*)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Chksum_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Chksum, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Chksum_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Solvsig_htype_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_htype_get" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  result = (Id)(Id) ((arg1)->htype);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solvsig_created_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_created_get" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  result = (unsigned int)(unsigned int) ((arg1)->created);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solvsig_expires_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_expires_get" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  result = (unsigned int)(unsigned int) ((arg1)->expires);
  resultobj = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solvsig_keyid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_keyid_get" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  result = (char *)(char *) ((arg1)->keyid);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Solvsig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  FILE *arg1 = 0 ;
  FILE *val1 ;
  int ecode1 = 0 ;
  PyObject *swig_obj[1] ;
  Solvsig *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  ecode1 = SWIG_AsValSolvFpPtr(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Solvsig" "', argument " "1"" of type '" "FILE *""'");
  } 
  arg1 = (FILE *)(val1);
  result = (Solvsig *)new_Solvsig(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solvsig, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Solvsig(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solvsig" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  delete_Solvsig(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solvsig_Chksum(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Chksum *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_Chksum" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  result = (Chksum *)Solvsig_Chksum(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Chksum, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solvsig_verify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solvsig *arg1 = 0 ;
  Repo *arg2 = 0 ;
  Chksum *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject *swig_obj[3] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solvsig_verify", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solvsig, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solvsig_verify" "', argument " "1"" of type '" "Solvsig *""'"); 
  }
  arg1 = (Solvsig *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_Repo, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Solvsig_verify" "', argument " "2"" of type '" "Repo *""'"); 
  }
  arg2 = (Repo *)(argp2);
  res3 = SWIG_ConvertPtr(swig_obj[2], &argp3,SWIGTYPE_p_Chksum, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Solvsig_verify" "', argument " "3"" of type '" "Chksum *""'"); 
  }
  arg3 = (Chksum *)(argp3);
  result = (XSolvable *)Solvsig_verify(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Solvsig_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Solvsig, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Solvsig_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_SolvFp(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SolvFp" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  delete_SolvFp(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_fileno(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_fileno" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  result = (int)SolvFp_fileno(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_dup(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_dup" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  result = (int)SolvFp_dup(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_write(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  unsigned char *arg2 = 0 ;
  size_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  size_t size2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  bool result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SolvFp_write", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_write" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, &size2, &alloc2);
  if (buf2 && size2)
  size2--;
  
  if (!SWIG_IsOK(res2)) {
    const void *pybuf = 0;
    Py_ssize_t pysize = 0;
#if PY_VERSION_HEX >= 0x03000000
    res2 = PyBytes_AsStringAndSize(swig_obj[1], (char **)&pybuf, &pysize);
#else
    res2 = PyObject_AsReadBuffer(swig_obj[1], &pybuf, &pysize);
#endif
    if (res2 < 0) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SolvFp_write" "', argument " "2"" of type '" "BinaryBlob""'");
    } else {
      buf2 = (void *)pybuf;
      size2 = pysize;
    }
    
    
    
  }
  arg2 = (unsigned char *)buf2;
  arg3 = size2;
  result = (bool)SolvFp_write(arg1,(unsigned char const *)arg2,arg3);
  resultobj = SWIG_From_bool((bool)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_flush(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_flush" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  result = (bool)SolvFp_flush(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_close(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_close" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  result = (bool)SolvFp_close(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SolvFp_cloexec(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  SolvFp *arg1 = 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "SolvFp_cloexec", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_SolvFp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SolvFp_cloexec" "', argument " "1"" of type '" "SolvFp *""'"); 
  }
  arg1 = (SolvFp *)(argp1);
  ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SolvFp_cloexec" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = (bool)(val2);
  SolvFp_cloexec(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *SolvFp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_SolvFp, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_xfopen(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = 0 ;
  char *arg2 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  PyObject *swig_obj[2] ;
  SolvFp *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "xfopen", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "xfopen" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  if (swig_obj[1]) {
    res2 = SWIG_AsCharPtrAndSize(swig_obj[1], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "xfopen" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
  }
  result = (SolvFp *)solvfp_xfopen((char const *)arg1,(char const *)arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_SolvFp, SWIG_POINTER_OWN |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return NULL;
}


SWIGINTERN PyObject *_wrap_xfopen_fd(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  char *arg1 = 0 ;
  int arg2 ;
  char *arg3 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  PyObject *swig_obj[3] ;
  SolvFp *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "xfopen_fd", 2, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_AsCharPtrAndSize(swig_obj[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "xfopen_fd" "', argument " "1"" of type '" "char const *""'");
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "xfopen_fd" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  if (swig_obj[2]) {
    res3 = SWIG_AsCharPtrAndSize(swig_obj[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "xfopen_fd" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
  }
  result = (SolvFp *)solvfp_xfopen_fd((char const *)arg1,arg2,(char const *)arg3);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_SolvFp, SWIG_POINTER_OWN |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return resultobj;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solv_get" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_id_get" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Problem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Problem *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Problem", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Problem" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Problem" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Problem *)new_Problem(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Problem, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_findproblemrule(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XRule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_findproblemrule" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = (XRule *)Problem_findproblemrule(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_findallproblemrules(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Problem_findallproblemrules", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_findallproblemrules" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Problem_findallproblemrules" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = Problem_findallproblemrules(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XRule * result = new_XRule(arg1->solv, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_solution_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solution_count" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = (int)Problem_solution_count(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_solutions(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_solutions" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = Problem_solutions(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Solution * result = new_Solution(arg1, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solution, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_get_learnt(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_learnt" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = Problem_get_learnt(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XRule * result = new_XRule(arg1->solv, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_get_decisionlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_decisionlist" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = Problem_get_decisionlist(arg1);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Decision * result = new_Decision(arg1->solv, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decision, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem_get_decisionsetlist(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem_get_decisionsetlist" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = Problem_get_decisionsetlist(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Decisionset * result = decisionset_fromids(arg1->solv, idp + id, idp[1] - id + 1);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Decisionset, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Problem___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Problem___str__" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  result = (char *)Problem___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Problem(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Problem" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Problem_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Problem, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Problem_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Solution_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_solv_get" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solution_problemid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_problemid_get" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  result = (Id)(Id) ((arg1)->problemid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solution_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_id_get" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  result = (Id)(Id) ((arg1)->id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Solution(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Problem *arg1 = 0 ;
  Id arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Solution *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Solution", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Problem, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Solution" "', argument " "1"" of type '" "Problem *""'"); 
  }
  arg1 = (Problem *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Solution" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  result = (Solution *)new_Solution(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solution, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solution_element_count(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_element_count" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  result = (int)Solution_element_count(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solution_elements(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  bool arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Solution_elements", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solution_elements" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_bool(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Solution_elements" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = (bool)(val2);
  }
  result = Solution_elements(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 3;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 3)
    {
      Id id = *idp;
#define result resultx
      Solutionelement * result = new_Solutionelement(arg1->solv, arg1->problemid, arg1->id, id, idp[1], idp[2]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Solution(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solution *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solution, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solution" "', argument " "1"" of type '" "Solution *""'"); 
  }
  arg1 = (Solution *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Solution_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Solution, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Solution_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Solutionelement_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solv_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_problemid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_problemid_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (Id)(Id) ((arg1)->problemid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_solutionid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solutionid_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (Id)(Id) ((arg1)->solutionid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_type_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (Id)(Id) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Solutionelement(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solver *arg1 = 0 ;
  Id arg2 ;
  Id arg3 ;
  Id arg4 ;
  Id arg5 ;
  Id arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  Solutionelement *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_Solutionelement", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solver, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Solutionelement" "', argument " "1"" of type '" "Solver *""'"); 
  }
  arg1 = (Solver *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Solutionelement" "', argument " "2"" of type '" "Id""'");
  } 
  arg2 = (Id)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Solutionelement" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Solutionelement" "', argument " "4"" of type '" "Id""'");
  } 
  arg4 = (Id)(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Solutionelement" "', argument " "5"" of type '" "Id""'");
  } 
  arg5 = (Id)(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_Solutionelement" "', argument " "6"" of type '" "Id""'");
  } 
  arg6 = (Id)(val6);
  result = (Solutionelement *)new_Solutionelement(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_str(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_str" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (char *)Solutionelement_str(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement___str__" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (char *)Solutionelement___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_replaceelements(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_replaceelements" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = Solutionelement_replaceelements(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      Solutionelement * result = new_Solutionelement(arg1->solv, arg1->problemid, arg1->solutionid, id, arg1->p, arg1->rp);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solutionelement, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_illegalreplace(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_illegalreplace" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (int)Solutionelement_illegalreplace(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_solvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_solvable_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (XSolvable *)Solutionelement_solvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_replacement_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_replacement_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (XSolvable *)Solutionelement_replacement_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_jobidx_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_jobidx_get" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (int)Solutionelement_jobidx_get(arg1);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Solutionelement_Job(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Job *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Solutionelement_Job" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  result = (Job *)Solutionelement_Job(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Job, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Solutionelement(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Solutionelement *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Solutionelement, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Solutionelement" "', argument " "1"" of type '" "Solutionelement *""'"); 
  }
  arg1 = (Solutionelement *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Solutionelement_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Solutionelement, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Solutionelement_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_Alternative_solv_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Solver *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_solv_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (Solver *)(Solver *) ((arg1)->solv);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Solver, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_type_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (Id)(Id) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_dep_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_dep_id_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (Id)(Id) ((arg1)->dep_id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_chosen_id_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_chosen_id_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (Id)(Id) ((arg1)->chosen_id);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_level_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_level_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (int)(int) ((arg1)->level);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Alternative(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Alternative" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  delete_Alternative(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_chosen_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_chosen_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (XSolvable *)Alternative_chosen_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_rule_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XRule *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_rule_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (XRule *)Alternative_rule_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XRule, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_depsolvable_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_depsolvable_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (XSolvable *)Alternative_depsolvable_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_dep_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Dep *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_dep_get" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (Dep *)Alternative_dep_get(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Dep, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_choices_raw(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_choices_raw" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = Alternative_choices_raw(arg1);
  {
    int i;
    PyObject *o = PyList_New((&result)->count);
    for (i = 0; i < (&result)->count; i++)
    PyList_SetItem(o, i, SWIG_From_int((&result)->elements[i]));
    queue_free(&result);
    resultobj = o;
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative_choices(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative_choices" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = Alternative_choices(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->solv->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, 0 |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Alternative___str__(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Alternative *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Alternative, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Alternative___str__" "', argument " "1"" of type '" "Alternative *""'"); 
  }
  arg1 = (Alternative *)(argp1);
  result = (char *)Alternative___str__(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Alternative_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Alternative, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_Transaction_pool_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Pool *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_pool_get" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = (Pool *)(Pool *) ((arg1)->pool);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Pool, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Transaction(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Transaction" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  delete_Transaction(arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_isempty(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_isempty" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = (bool)Transaction_isempty(arg1);
  resultobj = SWIG_From_bool((bool)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_othersolvable(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  XSolvable *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Transaction_othersolvable", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_othersolvable" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_othersolvable" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = (XSolvable *)Transaction_othersolvable(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_allothersolvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Transaction_allothersolvables", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_allothersolvables" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_allothersolvables" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  result = Transaction_allothersolvables(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_classify(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  Queue result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Transaction_classify", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_classify" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transaction_classify" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  result = Transaction_classify(arg1,arg2);
  {
    int i;
    int cnt = (&result)->count / 4;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 4)
    {
      Id id = *idp;
#define result resultx
      TransactionClass * result = new_TransactionClass(arg1, arg2, id, idp[1], idp[2], idp[3]);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TransactionClass, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_newpackages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_newpackages" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = Transaction_newpackages(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_keptpackages(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_keptpackages" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = Transaction_keptpackages(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_newsolvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_newsolvables" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = Transaction_newsolvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_keptsolvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_keptsolvables" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = Transaction_keptsolvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_steps(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_steps" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = Transaction_steps(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_steptype(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  XSolvable *arg2 = 0 ;
  int arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject *swig_obj[3] ;
  int result;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Transaction_steptype", 3, 3, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_steptype" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_XSolvable, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transaction_steptype" "', argument " "2"" of type '" "XSolvable *""'"); 
  }
  arg2 = (XSolvable *)(argp2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Transaction_steptype" "', argument " "3"" of type '" "int""'");
  } 
  arg3 = (int)(val3);
  result = (int)Transaction_steptype(arg1,arg2,arg3);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_calc_installsizechange(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  long long result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_calc_installsizechange" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  result = (long long)Transaction_calc_installsizechange(arg1);
  resultobj = SWIG_From_long_SS_long((long long)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Transaction_order(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject *swig_obj[2] ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "Transaction_order", 1, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transaction_order" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  if (swig_obj[1]) {
    ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transaction_order" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
  }
  Transaction_order(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Transaction_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_Transaction, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_TransactionClass_transaction_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Transaction *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_transaction_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (Transaction *)(Transaction *) ((arg1)->transaction);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_Transaction, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_type_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_type_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (Id)(Id) ((arg1)->type);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_fromid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_fromid_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (Id)(Id) ((arg1)->fromid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_toid_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Id result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_toid_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (Id)(Id) ((arg1)->toid);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_count_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_count_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (int)(int) ((arg1)->count);
  resultobj = SWIG_From_int((int)(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_TransactionClass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  Transaction *arg1 = 0 ;
  int arg2 ;
  Id arg3 ;
  int arg4 ;
  Id arg5 ;
  Id arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int val6 ;
  int ecode6 = 0 ;
  PyObject *swig_obj[6] ;
  TransactionClass *result = 0 ;
  
  (void)self;
  if (!SWIG_Python_UnpackTuple(args, "new_TransactionClass", 6, 6, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_Transaction, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TransactionClass" "', argument " "1"" of type '" "Transaction *""'"); 
  }
  arg1 = (Transaction *)(argp1);
  ecode2 = SWIG_AsVal_int(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_TransactionClass" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = (int)(val2);
  ecode3 = SWIG_AsVal_int(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_TransactionClass" "', argument " "3"" of type '" "Id""'");
  } 
  arg3 = (Id)(val3);
  ecode4 = SWIG_AsVal_int(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_TransactionClass" "', argument " "4"" of type '" "int""'");
  } 
  arg4 = (int)(val4);
  ecode5 = SWIG_AsVal_int(swig_obj[4], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_TransactionClass" "', argument " "5"" of type '" "Id""'");
  } 
  arg5 = (Id)(val5);
  ecode6 = SWIG_AsVal_int(swig_obj[5], &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_TransactionClass" "', argument " "6"" of type '" "Id""'");
  } 
  arg6 = (Id)(val6);
  result = (TransactionClass *)new_TransactionClass(arg1,arg2,arg3,arg4,arg5,arg6);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_TransactionClass, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_solvables(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  Queue result;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_solvables" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = TransactionClass_solvables(arg1);
  {
    int i;
    int cnt = (&result)->count / 1;
    Id *idp = (&result)->elements;
    PyObject *o = PyList_New(cnt);
    for (i = 0; i < cnt; i++, idp += 1)
    {
      Id id = *idp;
#define result resultx
      XSolvable * result = new_XSolvable(arg1->transaction->pool, id);
      resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_XSolvable, SWIG_POINTER_OWN |  0 );
      PyList_SetItem(o, i, resultobj);
#undef result
    }
    queue_free(&result);
    resultobj = o;
  }
  
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_fromstr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_fromstr_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (char *)TransactionClass_fromstr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TransactionClass_tostr_get(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  char *result = 0 ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TransactionClass_tostr_get" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  result = (char *)TransactionClass_tostr_get(arg1);
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TransactionClass(PyObject *self, PyObject *args) {
  PyObject *resultobj = 0;
  TransactionClass *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  (void)self;
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_TransactionClass, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TransactionClass" "', argument " "1"" of type '" "TransactionClass *""'"); 
  }
  arg1 = (TransactionClass *)(argp1);
  free((char *) arg1);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TransactionClass_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj = NULL;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_TransactionClass, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *TransactionClass_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

static PyMethodDef SwigMethods[] = {
	 { "Selection_pool_get", _wrap_Selection_pool_get, METH_O, NULL},
	 { "Selection_flags_get", _wrap_Selection_flags_get, METH_O, NULL},
	 { "new_Selection", _wrap_new_Selection, METH_O, NULL},
	 { "delete_Selection", _wrap_delete_Selection, METH_O, NULL},
	 { "Selection_isempty", _wrap_Selection_isempty, METH_O, NULL},
	 { "Selection_clone", _wrap_Selection_clone, METH_VARARGS, NULL},
	 { "Selection_filter", _wrap_Selection_filter, METH_VARARGS, NULL},
	 { "Selection_add", _wrap_Selection_add, METH_VARARGS, NULL},
	 { "Selection_add_raw", _wrap_Selection_add_raw, METH_VARARGS, NULL},
	 { "Selection_subtract", _wrap_Selection_subtract, METH_VARARGS, NULL},
	 { "Selection_select", _wrap_Selection_select, METH_VARARGS, NULL},
	 { "Selection_matchdeps", _wrap_Selection_matchdeps, METH_VARARGS, NULL},
	 { "Selection_matchdepid", _wrap_Selection_matchdepid, METH_VARARGS, NULL},
	 { "Selection_matchsolvable", _wrap_Selection_matchsolvable, METH_VARARGS, NULL},
	 { "Selection_jobs", _wrap_Selection_jobs, METH_VARARGS, NULL},
	 { "Selection_solvables", _wrap_Selection_solvables, METH_O, NULL},
	 { "Selection___str__", _wrap_Selection___str__, METH_O, NULL},
	 { "Selection___repr__", _wrap_Selection___repr__, METH_O, NULL},
	 { "Selection_swigregister", Selection_swigregister, METH_O, NULL},
	 { "Selection_swiginit", Selection_swiginit, METH_VARARGS, NULL},
	 { "Dep_pool_get", _wrap_Dep_pool_get, METH_O, NULL},
	 { "Dep_id_get", _wrap_Dep_id_get, METH_O, NULL},
	 { "new_Dep", _wrap_new_Dep, METH_VARARGS, NULL},
	 { "Dep_Rel", _wrap_Dep_Rel, METH_VARARGS, NULL},
	 { "Dep_Selection_name", _wrap_Dep_Selection_name, METH_VARARGS, NULL},
	 { "Dep_Selection_provides", _wrap_Dep_Selection_provides, METH_VARARGS, NULL},
	 { "Dep_str", _wrap_Dep_str, METH_O, NULL},
	 { "Dep___eq__", _wrap_Dep___eq__, METH_VARARGS, NULL},
	 { "Dep___ne__", _wrap_Dep___ne__, METH_VARARGS, NULL},
	 { "Dep___hash__", _wrap_Dep___hash__, METH_O, NULL},
	 { "Dep___str__", _wrap_Dep___str__, METH_O, NULL},
	 { "Dep___repr__", _wrap_Dep___repr__, METH_O, NULL},
	 { "delete_Dep", _wrap_delete_Dep, METH_O, NULL},
	 { "Dep_swigregister", Dep_swigregister, METH_O, NULL},
	 { "Dep_swiginit", Dep_swiginit, METH_VARARGS, NULL},
	 { "new_Dataiterator", _wrap_new_Dataiterator, METH_VARARGS, NULL},
	 { "delete_Dataiterator", _wrap_delete_Dataiterator, METH_O, NULL},
	 { "Dataiterator___next__", _wrap_Dataiterator___next__, METH_O, NULL},
	 { "Dataiterator_prepend_keyname", _wrap_Dataiterator_prepend_keyname, METH_VARARGS, NULL},
	 { "Dataiterator_skip_solvable", _wrap_Dataiterator_skip_solvable, METH_O, NULL},
	 { "Dataiterator_swigregister", Dataiterator_swigregister, METH_O, NULL},
	 { "Dataiterator_swiginit", Dataiterator_swiginit, METH_VARARGS, NULL},
	 { "XSolvable_pool_get", _wrap_XSolvable_pool_get, METH_O, NULL},
	 { "XSolvable_id_get", _wrap_XSolvable_id_get, METH_O, NULL},
	 { "new_XSolvable", _wrap_new_XSolvable, METH_VARARGS, NULL},
	 { "XSolvable_str", _wrap_XSolvable_str, METH_O, NULL},
	 { "XSolvable_lookup_str", _wrap_XSolvable_lookup_str, METH_VARARGS, NULL},
	 { "XSolvable_lookup_id", _wrap_XSolvable_lookup_id, METH_VARARGS, NULL},
	 { "XSolvable_lookup_num", _wrap_XSolvable_lookup_num, METH_VARARGS, NULL},
	 { "XSolvable_lookup_void", _wrap_XSolvable_lookup_void, METH_VARARGS, NULL},
	 { "XSolvable_lookup_checksum", _wrap_XSolvable_lookup_checksum, METH_VARARGS, NULL},
	 { "XSolvable_lookup_idarray", _wrap_XSolvable_lookup_idarray, METH_VARARGS, NULL},
	 { "XSolvable_lookup_deparray", _wrap_XSolvable_lookup_deparray, METH_VARARGS, NULL},
	 { "XSolvable_lookup_location", _wrap_XSolvable_lookup_location, METH_O, NULL},
	 { "XSolvable_lookup_sourcepkg", _wrap_XSolvable_lookup_sourcepkg, METH_O, NULL},
	 { "XSolvable_Dataiterator", _wrap_XSolvable_Dataiterator, METH_VARARGS, NULL},
	 { "XSolvable_installable", _wrap_XSolvable_installable, METH_O, NULL},
	 { "XSolvable_isinstalled", _wrap_XSolvable_isinstalled, METH_O, NULL},
	 { "XSolvable_name_set", _wrap_XSolvable_name_set, METH_VARARGS, NULL},
	 { "XSolvable_name_get", _wrap_XSolvable_name_get, METH_O, NULL},
	 { "XSolvable_nameid_set", _wrap_XSolvable_nameid_set, METH_VARARGS, NULL},
	 { "XSolvable_nameid_get", _wrap_XSolvable_nameid_get, METH_O, NULL},
	 { "XSolvable_evr_set", _wrap_XSolvable_evr_set, METH_VARARGS, NULL},
	 { "XSolvable_evr_get", _wrap_XSolvable_evr_get, METH_O, NULL},
	 { "XSolvable_evrid_set", _wrap_XSolvable_evrid_set, METH_VARARGS, NULL},
	 { "XSolvable_evrid_get", _wrap_XSolvable_evrid_get, METH_O, NULL},
	 { "XSolvable_arch_set", _wrap_XSolvable_arch_set, METH_VARARGS, NULL},
	 { "XSolvable_arch_get", _wrap_XSolvable_arch_get, METH_O, NULL},
	 { "XSolvable_archid_set", _wrap_XSolvable_archid_set, METH_VARARGS, NULL},
	 { "XSolvable_archid_get", _wrap_XSolvable_archid_get, METH_O, NULL},
	 { "XSolvable_vendor_set", _wrap_XSolvable_vendor_set, METH_VARARGS, NULL},
	 { "XSolvable_vendor_get", _wrap_XSolvable_vendor_get, METH_O, NULL},
	 { "XSolvable_vendorid_set", _wrap_XSolvable_vendorid_set, METH_VARARGS, NULL},
	 { "XSolvable_vendorid_get", _wrap_XSolvable_vendorid_get, METH_O, NULL},
	 { "XSolvable_repo_get", _wrap_XSolvable_repo_get, METH_O, NULL},
	 { "XSolvable_add_provides", _wrap_XSolvable_add_provides, METH_VARARGS, NULL},
	 { "XSolvable_add_obsoletes", _wrap_XSolvable_add_obsoletes, METH_VARARGS, NULL},
	 { "XSolvable_add_conflicts", _wrap_XSolvable_add_conflicts, METH_VARARGS, NULL},
	 { "XSolvable_add_requires", _wrap_XSolvable_add_requires, METH_VARARGS, NULL},
	 { "XSolvable_add_recommends", _wrap_XSolvable_add_recommends, METH_VARARGS, NULL},
	 { "XSolvable_add_suggests", _wrap_XSolvable_add_suggests, METH_VARARGS, NULL},
	 { "XSolvable_add_supplements", _wrap_XSolvable_add_supplements, METH_VARARGS, NULL},
	 { "XSolvable_add_enhances", _wrap_XSolvable_add_enhances, METH_VARARGS, NULL},
	 { "XSolvable_unset", _wrap_XSolvable_unset, METH_VARARGS, NULL},
	 { "XSolvable_add_deparray", _wrap_XSolvable_add_deparray, METH_VARARGS, NULL},
	 { "XSolvable_Selection", _wrap_XSolvable_Selection, METH_VARARGS, NULL},
	 { "XSolvable_identical", _wrap_XSolvable_identical, METH_VARARGS, NULL},
	 { "XSolvable_evrcmp", _wrap_XSolvable_evrcmp, METH_VARARGS, NULL},
	 { "XSolvable_matchesdep", _wrap_XSolvable_matchesdep, METH_VARARGS, NULL},
	 { "XSolvable___eq__", _wrap_XSolvable___eq__, METH_VARARGS, NULL},
	 { "XSolvable___ne__", _wrap_XSolvable___ne__, METH_VARARGS, NULL},
	 { "XSolvable___hash__", _wrap_XSolvable___hash__, METH_O, NULL},
	 { "XSolvable___str__", _wrap_XSolvable___str__, METH_O, NULL},
	 { "XSolvable___repr__", _wrap_XSolvable___repr__, METH_O, NULL},
	 { "delete_XSolvable", _wrap_delete_XSolvable, METH_O, NULL},
	 { "XSolvable_swigregister", XSolvable_swigregister, METH_O, NULL},
	 { "XSolvable_swiginit", XSolvable_swiginit, METH_VARARGS, NULL},
	 { "Ruleinfo_solv_get", _wrap_Ruleinfo_solv_get, METH_O, NULL},
	 { "Ruleinfo_type_get", _wrap_Ruleinfo_type_get, METH_O, NULL},
	 { "Ruleinfo_dep_id_get", _wrap_Ruleinfo_dep_id_get, METH_O, NULL},
	 { "new_Ruleinfo", _wrap_new_Ruleinfo, METH_VARARGS, NULL},
	 { "Ruleinfo_solvable_get", _wrap_Ruleinfo_solvable_get, METH_O, NULL},
	 { "Ruleinfo_othersolvable_get", _wrap_Ruleinfo_othersolvable_get, METH_O, NULL},
	 { "Ruleinfo_dep_get", _wrap_Ruleinfo_dep_get, METH_O, NULL},
	 { "Ruleinfo_problemstr", _wrap_Ruleinfo_problemstr, METH_O, NULL},
	 { "Ruleinfo___str__", _wrap_Ruleinfo___str__, METH_O, NULL},
	 { "delete_Ruleinfo", _wrap_delete_Ruleinfo, METH_O, NULL},
	 { "Ruleinfo_swigregister", Ruleinfo_swigregister, METH_O, NULL},
	 { "Ruleinfo_swiginit", Ruleinfo_swiginit, METH_VARARGS, NULL},
	 { "XRule_solv_get", _wrap_XRule_solv_get, METH_O, NULL},
	 { "XRule_id_get", _wrap_XRule_id_get, METH_O, NULL},
	 { "new_XRule", _wrap_new_XRule, METH_VARARGS, NULL},
	 { "XRule_type_get", _wrap_XRule_type_get, METH_O, NULL},
	 { "XRule_info", _wrap_XRule_info, METH_O, NULL},
	 { "XRule_allinfos", _wrap_XRule_allinfos, METH_O, NULL},
	 { "XRule_get_learnt", _wrap_XRule_get_learnt, METH_O, NULL},
	 { "XRule_get_decisionlist", _wrap_XRule_get_decisionlist, METH_O, NULL},
	 { "XRule_get_decisionsetlist", _wrap_XRule_get_decisionsetlist, METH_O, NULL},
	 { "XRule___eq__", _wrap_XRule___eq__, METH_VARARGS, NULL},
	 { "XRule___ne__", _wrap_XRule___ne__, METH_VARARGS, NULL},
	 { "XRule___hash__", _wrap_XRule___hash__, METH_O, NULL},
	 { "XRule___repr__", _wrap_XRule___repr__, METH_O, NULL},
	 { "delete_XRule", _wrap_delete_XRule, METH_O, NULL},
	 { "XRule_swigregister", XRule_swigregister, METH_O, NULL},
	 { "XRule_swiginit", XRule_swiginit, METH_VARARGS, NULL},
	 { "XRepodata_repo_get", _wrap_XRepodata_repo_get, METH_O, NULL},
	 { "XRepodata_id_get", _wrap_XRepodata_id_get, METH_O, NULL},
	 { "new_XRepodata", _wrap_new_XRepodata, METH_VARARGS, NULL},
	 { "XRepodata_new_handle", _wrap_XRepodata_new_handle, METH_O, NULL},
	 { "XRepodata_set_id", _wrap_XRepodata_set_id, METH_VARARGS, NULL},
	 { "XRepodata_set_num", _wrap_XRepodata_set_num, METH_VARARGS, NULL},
	 { "XRepodata_set_str", _wrap_XRepodata_set_str, METH_VARARGS, NULL},
	 { "XRepodata_set_void", _wrap_XRepodata_set_void, METH_VARARGS, NULL},
	 { "XRepodata_set_poolstr", _wrap_XRepodata_set_poolstr, METH_VARARGS, NULL},
	 { "XRepodata_add_idarray", _wrap_XRepodata_add_idarray, METH_VARARGS, NULL},
	 { "XRepodata_add_flexarray", _wrap_XRepodata_add_flexarray, METH_VARARGS, NULL},
	 { "XRepodata_set_checksum", _wrap_XRepodata_set_checksum, METH_VARARGS, NULL},
	 { "XRepodata_set_sourcepkg", _wrap_XRepodata_set_sourcepkg, METH_VARARGS, NULL},
	 { "XRepodata_set_location", _wrap_XRepodata_set_location, METH_VARARGS, NULL},
	 { "XRepodata_unset", _wrap_XRepodata_unset, METH_VARARGS, NULL},
	 { "XRepodata_lookup_str", _wrap_XRepodata_lookup_str, METH_VARARGS, NULL},
	 { "XRepodata_lookup_id", _wrap_XRepodata_lookup_id, METH_VARARGS, NULL},
	 { "XRepodata_lookup_num", _wrap_XRepodata_lookup_num, METH_VARARGS, NULL},
	 { "XRepodata_lookup_void", _wrap_XRepodata_lookup_void, METH_VARARGS, NULL},
	 { "XRepodata_lookup_idarray", _wrap_XRepodata_lookup_idarray, METH_VARARGS, NULL},
	 { "XRepodata_lookup_checksum", _wrap_XRepodata_lookup_checksum, METH_VARARGS, NULL},
	 { "XRepodata_internalize", _wrap_XRepodata_internalize, METH_O, NULL},
	 { "XRepodata_create_stubs", _wrap_XRepodata_create_stubs, METH_O, NULL},
	 { "XRepodata_write", _wrap_XRepodata_write, METH_VARARGS, NULL},
	 { "XRepodata_str2dir", _wrap_XRepodata_str2dir, METH_VARARGS, NULL},
	 { "XRepodata_dir2str", _wrap_XRepodata_dir2str, METH_VARARGS, NULL},
	 { "XRepodata_add_dirstr", _wrap_XRepodata_add_dirstr, METH_VARARGS, NULL},
	 { "XRepodata_add_solv", _wrap_XRepodata_add_solv, METH_VARARGS, NULL},
	 { "XRepodata_extend_to_repo", _wrap_XRepodata_extend_to_repo, METH_O, NULL},
	 { "XRepodata___eq__", _wrap_XRepodata___eq__, METH_VARARGS, NULL},
	 { "XRepodata___ne__", _wrap_XRepodata___ne__, METH_VARARGS, NULL},
	 { "XRepodata___hash__", _wrap_XRepodata___hash__, METH_O, NULL},
	 { "XRepodata___repr__", _wrap_XRepodata___repr__, METH_O, NULL},
	 { "delete_XRepodata", _wrap_delete_XRepodata, METH_O, NULL},
	 { "XRepodata_swigregister", XRepodata_swigregister, METH_O, NULL},
	 { "XRepodata_swiginit", XRepodata_swiginit, METH_VARARGS, NULL},
	 { "new_Pool_solvable_iterator", _wrap_new_Pool_solvable_iterator, METH_O, NULL},
	 { "Pool_solvable_iterator___next__", _wrap_Pool_solvable_iterator___next__, METH_O, NULL},
	 { "Pool_solvable_iterator___getitem__", _wrap_Pool_solvable_iterator___getitem__, METH_VARARGS, NULL},
	 { "Pool_solvable_iterator___len__", _wrap_Pool_solvable_iterator___len__, METH_O, NULL},
	 { "delete_Pool_solvable_iterator", _wrap_delete_Pool_solvable_iterator, METH_O, NULL},
	 { "Pool_solvable_iterator_swigregister", Pool_solvable_iterator_swigregister, METH_O, NULL},
	 { "Pool_solvable_iterator_swiginit", Pool_solvable_iterator_swiginit, METH_VARARGS, NULL},
	 { "new_Pool_repo_iterator", _wrap_new_Pool_repo_iterator, METH_O, NULL},
	 { "Pool_repo_iterator___next__", _wrap_Pool_repo_iterator___next__, METH_O, NULL},
	 { "Pool_repo_iterator___getitem__", _wrap_Pool_repo_iterator___getitem__, METH_VARARGS, NULL},
	 { "Pool_repo_iterator___len__", _wrap_Pool_repo_iterator___len__, METH_O, NULL},
	 { "delete_Pool_repo_iterator", _wrap_delete_Pool_repo_iterator, METH_O, NULL},
	 { "Pool_repo_iterator_swigregister", Pool_repo_iterator_swigregister, METH_O, NULL},
	 { "Pool_repo_iterator_swiginit", Pool_repo_iterator_swiginit, METH_VARARGS, NULL},
	 { "new_Repo_solvable_iterator", _wrap_new_Repo_solvable_iterator, METH_O, NULL},
	 { "Repo_solvable_iterator___next__", _wrap_Repo_solvable_iterator___next__, METH_O, NULL},
	 { "Repo_solvable_iterator___getitem__", _wrap_Repo_solvable_iterator___getitem__, METH_VARARGS, NULL},
	 { "Repo_solvable_iterator___len__", _wrap_Repo_solvable_iterator___len__, METH_O, NULL},
	 { "delete_Repo_solvable_iterator", _wrap_delete_Repo_solvable_iterator, METH_O, NULL},
	 { "Repo_solvable_iterator_swigregister", Repo_solvable_iterator_swigregister, METH_O, NULL},
	 { "Repo_solvable_iterator_swiginit", Repo_solvable_iterator_swiginit, METH_VARARGS, NULL},
	 { "Datamatch_pool_get", _wrap_Datamatch_pool_get, METH_O, NULL},
	 { "Datamatch_repo_get", _wrap_Datamatch_repo_get, METH_O, NULL},
	 { "Datamatch_solvid_get", _wrap_Datamatch_solvid_get, METH_O, NULL},
	 { "delete_Datamatch", _wrap_delete_Datamatch, METH_O, NULL},
	 { "Datamatch_solvable_get", _wrap_Datamatch_solvable_get, METH_O, NULL},
	 { "Datamatch_key_id_get", _wrap_Datamatch_key_id_get, METH_O, NULL},
	 { "Datamatch_key_idstr_get", _wrap_Datamatch_key_idstr_get, METH_O, NULL},
	 { "Datamatch_type_id_get", _wrap_Datamatch_type_id_get, METH_O, NULL},
	 { "Datamatch_type_idstr_get", _wrap_Datamatch_type_idstr_get, METH_O, NULL},
	 { "Datamatch_id_get", _wrap_Datamatch_id_get, METH_O, NULL},
	 { "Datamatch_idstr_get", _wrap_Datamatch_idstr_get, METH_O, NULL},
	 { "Datamatch_str_get", _wrap_Datamatch_str_get, METH_O, NULL},
	 { "Datamatch_dep_get", _wrap_Datamatch_dep_get, METH_O, NULL},
	 { "Datamatch_binary_get", _wrap_Datamatch_binary_get, METH_O, NULL},
	 { "Datamatch_num_get", _wrap_Datamatch_num_get, METH_O, NULL},
	 { "Datamatch_num2_get", _wrap_Datamatch_num2_get, METH_O, NULL},
	 { "Datamatch_pos", _wrap_Datamatch_pos, METH_O, NULL},
	 { "Datamatch_parentpos", _wrap_Datamatch_parentpos, METH_O, NULL},
	 { "Datamatch___str__", _wrap_Datamatch___str__, METH_O, NULL},
	 { "Datamatch_swigregister", Datamatch_swigregister, METH_O, NULL},
	 { "Datapos_repo_get", _wrap_Datapos_repo_get, METH_O, NULL},
	 { "Datapos_lookup_id", _wrap_Datapos_lookup_id, METH_VARARGS, NULL},
	 { "Datapos_lookup_str", _wrap_Datapos_lookup_str, METH_VARARGS, NULL},
	 { "Datapos_lookup_num", _wrap_Datapos_lookup_num, METH_VARARGS, NULL},
	 { "Datapos_lookup_void", _wrap_Datapos_lookup_void, METH_VARARGS, NULL},
	 { "Datapos_lookup_checksum", _wrap_Datapos_lookup_checksum, METH_VARARGS, NULL},
	 { "Datapos_lookup_deltaseq", _wrap_Datapos_lookup_deltaseq, METH_O, NULL},
	 { "Datapos_lookup_deltalocation", _wrap_Datapos_lookup_deltalocation, METH_O, NULL},
	 { "Datapos_lookup_idarray", _wrap_Datapos_lookup_idarray, METH_VARARGS, NULL},
	 { "Datapos_Dataiterator", _wrap_Datapos_Dataiterator, METH_VARARGS, NULL},
	 { "delete_Datapos", _wrap_delete_Datapos, METH_O, NULL},
	 { "Datapos_swigregister", Datapos_swigregister, METH_O, NULL},
	 { "Job_pool_get", _wrap_Job_pool_get, METH_O, NULL},
	 { "Job_how_set", _wrap_Job_how_set, METH_VARARGS, NULL},
	 { "Job_how_get", _wrap_Job_how_get, METH_O, NULL},
	 { "Job_what_set", _wrap_Job_what_set, METH_VARARGS, NULL},
	 { "Job_what_get", _wrap_Job_what_get, METH_O, NULL},
	 { "new_Job", _wrap_new_Job, METH_VARARGS, NULL},
	 { "Job_solvables", _wrap_Job_solvables, METH_O, NULL},
	 { "Job_isemptyupdate", _wrap_Job_isemptyupdate, METH_O, NULL},
	 { "Job___eq__", _wrap_Job___eq__, METH_VARARGS, NULL},
	 { "Job___ne__", _wrap_Job___ne__, METH_VARARGS, NULL},
	 { "Job___str__", _wrap_Job___str__, METH_O, NULL},
	 { "Job___repr__", _wrap_Job___repr__, METH_O, NULL},
	 { "delete_Job", _wrap_delete_Job, METH_O, NULL},
	 { "Job_swigregister", Job_swigregister, METH_O, NULL},
	 { "Job_swiginit", Job_swiginit, METH_VARARGS, NULL},
	 { "new_Pool", _wrap_new_Pool, METH_NOARGS, NULL},
	 { "Pool_setdisttype", _wrap_Pool_setdisttype, METH_VARARGS, NULL},
	 { "Pool_set_debuglevel", _wrap_Pool_set_debuglevel, METH_VARARGS, NULL},
	 { "Pool_set_flag", _wrap_Pool_set_flag, METH_VARARGS, NULL},
	 { "Pool_get_flag", _wrap_Pool_get_flag, METH_VARARGS, NULL},
	 { "Pool_set_rootdir", _wrap_Pool_set_rootdir, METH_VARARGS, NULL},
	 { "Pool_get_rootdir", _wrap_Pool_get_rootdir, METH_VARARGS, NULL},
	 { "Pool_clr_loadcallback", _wrap_Pool_clr_loadcallback, METH_O, NULL},
	 { "Pool_set_loadcallback", _wrap_Pool_set_loadcallback, METH_VARARGS, NULL},
	 { "delete_Pool", _wrap_delete_Pool, METH_O, NULL},
	 { "Pool_free", _wrap_Pool_free, METH_O, NULL},
	 { "Pool_disown", _wrap_Pool_disown, METH_O, NULL},
	 { "Pool_appdata_set", _wrap_Pool_appdata_set, METH_VARARGS, NULL},
	 { "Pool_appdata_get", _wrap_Pool_appdata_get, METH_O, NULL},
	 { "Pool_appdata_disown", _wrap_Pool_appdata_disown, METH_O, NULL},
	 { "Pool_str2id", _wrap_Pool_str2id, METH_VARARGS, NULL},
	 { "Pool_Dep", _wrap_Pool_Dep, METH_VARARGS, NULL},
	 { "Pool_parserpmrichdep", _wrap_Pool_parserpmrichdep, METH_VARARGS, NULL},
	 { "Pool_id2str", _wrap_Pool_id2str, METH_VARARGS, NULL},
	 { "Pool_dep2str", _wrap_Pool_dep2str, METH_VARARGS, NULL},
	 { "Pool_rel2id", _wrap_Pool_rel2id, METH_VARARGS, NULL},
	 { "Pool_id2langid", _wrap_Pool_id2langid, METH_VARARGS, NULL},
	 { "Pool_setarch", _wrap_Pool_setarch, METH_VARARGS, NULL},
	 { "Pool_add_repo", _wrap_Pool_add_repo, METH_VARARGS, NULL},
	 { "Pool_lookup_str", _wrap_Pool_lookup_str, METH_VARARGS, NULL},
	 { "Pool_lookup_id", _wrap_Pool_lookup_id, METH_VARARGS, NULL},
	 { "Pool_lookup_num", _wrap_Pool_lookup_num, METH_VARARGS, NULL},
	 { "Pool_lookup_void", _wrap_Pool_lookup_void, METH_VARARGS, NULL},
	 { "Pool_lookup_checksum", _wrap_Pool_lookup_checksum, METH_VARARGS, NULL},
	 { "Pool_Dataiterator", _wrap_Pool_Dataiterator, METH_VARARGS, NULL},
	 { "Pool_Dataiterator_solvid", _wrap_Pool_Dataiterator_solvid, METH_VARARGS, NULL},
	 { "Pool_solvid2str", _wrap_Pool_solvid2str, METH_VARARGS, NULL},
	 { "Pool_solvidset2str", _wrap_Pool_solvidset2str, METH_VARARGS, NULL},
	 { "Pool_solvableset2str", _wrap_Pool_solvableset2str, METH_VARARGS, NULL},
	 { "Pool_addfileprovides", _wrap_Pool_addfileprovides, METH_O, NULL},
	 { "Pool_addfileprovides_queue", _wrap_Pool_addfileprovides_queue, METH_O, NULL},
	 { "Pool_createwhatprovides", _wrap_Pool_createwhatprovides, METH_O, NULL},
	 { "Pool_id2solvable", _wrap_Pool_id2solvable, METH_VARARGS, NULL},
	 { "Pool_solvables_get", _wrap_Pool_solvables_get, METH_O, NULL},
	 { "Pool_solvables_iter", _wrap_Pool_solvables_iter, METH_O, NULL},
	 { "Pool_id2repo", _wrap_Pool_id2repo, METH_VARARGS, NULL},
	 { "Pool_repos_get", _wrap_Pool_repos_get, METH_O, NULL},
	 { "Pool_repos_iter", _wrap_Pool_repos_iter, METH_O, NULL},
	 { "Pool_installed_set", _wrap_Pool_installed_set, METH_VARARGS, NULL},
	 { "Pool_installed_get", _wrap_Pool_installed_get, METH_O, NULL},
	 { "Pool_errstr_get", _wrap_Pool_errstr_get, METH_O, NULL},
	 { "Pool_matchprovidingids", _wrap_Pool_matchprovidingids, METH_VARARGS, NULL},
	 { "Pool_Job", _wrap_Pool_Job, METH_VARARGS, NULL},
	 { "Pool_whatprovides", _wrap_Pool_whatprovides, METH_VARARGS, NULL},
	 { "Pool_best_solvables", _wrap_Pool_best_solvables, METH_VARARGS, NULL},
	 { "Pool_towhatprovides", _wrap_Pool_towhatprovides, METH_VARARGS, NULL},
	 { "Pool_set_namespaceproviders", _wrap_Pool_set_namespaceproviders, METH_VARARGS, NULL},
	 { "Pool_flush_namespaceproviders", _wrap_Pool_flush_namespaceproviders, METH_VARARGS, NULL},
	 { "Pool_whatcontainsdep", _wrap_Pool_whatcontainsdep, METH_VARARGS, NULL},
	 { "Pool_whatmatchesdep", _wrap_Pool_whatmatchesdep, METH_VARARGS, NULL},
	 { "Pool_whatmatchessolvable", _wrap_Pool_whatmatchessolvable, METH_VARARGS, NULL},
	 { "Pool_isknownarch", _wrap_Pool_isknownarch, METH_VARARGS, NULL},
	 { "Pool_Solver", _wrap_Pool_Solver, METH_O, NULL},
	 { "Pool_Selection", _wrap_Pool_Selection, METH_O, NULL},
	 { "Pool_Selection_all", _wrap_Pool_Selection_all, METH_VARARGS, NULL},
	 { "Pool_select", _wrap_Pool_select, METH_VARARGS, NULL},
	 { "Pool_matchdeps", _wrap_Pool_matchdeps, METH_VARARGS, NULL},
	 { "Pool_matchdepid", _wrap_Pool_matchdepid, METH_VARARGS, NULL},
	 { "Pool_matchsolvable", _wrap_Pool_matchsolvable, METH_VARARGS, NULL},
	 { "Pool_get_considered_list", _wrap_Pool_get_considered_list, METH_O, NULL},
	 { "Pool_get_disabled_list", _wrap_Pool_get_disabled_list, METH_O, NULL},
	 { "Pool_set_considered_list", _wrap_Pool_set_considered_list, METH_VARARGS, NULL},
	 { "Pool_set_disabled_list", _wrap_Pool_set_disabled_list, METH_VARARGS, NULL},
	 { "Pool_setpooljobs", _wrap_Pool_setpooljobs, METH_VARARGS, NULL},
	 { "Pool_getpooljobs", _wrap_Pool_getpooljobs, METH_O, NULL},
	 { "Pool_swigregister", Pool_swigregister, METH_O, NULL},
	 { "Pool_swiginit", Pool_swiginit, METH_VARARGS, NULL},
	 { "Repo_pool_get", _wrap_Repo_pool_get, METH_O, NULL},
	 { "Repo_name_get", _wrap_Repo_name_get, METH_O, NULL},
	 { "Repo_priority_set", _wrap_Repo_priority_set, METH_VARARGS, NULL},
	 { "Repo_priority_get", _wrap_Repo_priority_get, METH_O, NULL},
	 { "Repo_subpriority_set", _wrap_Repo_subpriority_set, METH_VARARGS, NULL},
	 { "Repo_subpriority_get", _wrap_Repo_subpriority_get, METH_O, NULL},
	 { "Repo_nsolvables_get", _wrap_Repo_nsolvables_get, METH_O, NULL},
	 { "Repo_free", _wrap_Repo_free, METH_VARARGS, NULL},
	 { "Repo_empty", _wrap_Repo_empty, METH_VARARGS, NULL},
	 { "Repo_isempty", _wrap_Repo_isempty, METH_O, NULL},
	 { "Repo_appdata_set", _wrap_Repo_appdata_set, METH_VARARGS, NULL},
	 { "Repo_appdata_get", _wrap_Repo_appdata_get, METH_O, NULL},
	 { "Repo_add_solv", _wrap_Repo_add_solv, METH_VARARGS, NULL},
	 { "Repo_add_solvable", _wrap_Repo_add_solvable, METH_O, NULL},
	 { "Repo_add_rpmdb", _wrap_Repo_add_rpmdb, METH_VARARGS, NULL},
	 { "Repo_add_rpmdb_reffp", _wrap_Repo_add_rpmdb_reffp, METH_VARARGS, NULL},
	 { "Repo_add_rpm", _wrap_Repo_add_rpm, METH_VARARGS, NULL},
	 { "Repo_add_rpmdb_pubkeys", _wrap_Repo_add_rpmdb_pubkeys, METH_VARARGS, NULL},
	 { "Repo_add_pubkey", _wrap_Repo_add_pubkey, METH_VARARGS, NULL},
	 { "Repo_add_keyring", _wrap_Repo_add_keyring, METH_VARARGS, NULL},
	 { "Repo_add_keydir", _wrap_Repo_add_keydir, METH_VARARGS, NULL},
	 { "Repo_add_rpmmd", _wrap_Repo_add_rpmmd, METH_VARARGS, NULL},
	 { "Repo_add_repomdxml", _wrap_Repo_add_repomdxml, METH_VARARGS, NULL},
	 { "Repo_add_updateinfoxml", _wrap_Repo_add_updateinfoxml, METH_VARARGS, NULL},
	 { "Repo_add_deltainfoxml", _wrap_Repo_add_deltainfoxml, METH_VARARGS, NULL},
	 { "Repo_add_debdb", _wrap_Repo_add_debdb, METH_VARARGS, NULL},
	 { "Repo_add_debpackages", _wrap_Repo_add_debpackages, METH_VARARGS, NULL},
	 { "Repo_add_deb", _wrap_Repo_add_deb, METH_VARARGS, NULL},
	 { "Repo_add_susetags", _wrap_Repo_add_susetags, METH_VARARGS, NULL},
	 { "Repo_add_content", _wrap_Repo_add_content, METH_VARARGS, NULL},
	 { "Repo_add_products", _wrap_Repo_add_products, METH_VARARGS, NULL},
	 { "Repo_add_mdk", _wrap_Repo_add_mdk, METH_VARARGS, NULL},
	 { "Repo_add_mdk_info", _wrap_Repo_add_mdk_info, METH_VARARGS, NULL},
	 { "Repo_add_arch_repo", _wrap_Repo_add_arch_repo, METH_VARARGS, NULL},
	 { "Repo_add_arch_local", _wrap_Repo_add_arch_local, METH_VARARGS, NULL},
	 { "Repo_add_arch_pkg", _wrap_Repo_add_arch_pkg, METH_VARARGS, NULL},
	 { "Repo_internalize", _wrap_Repo_internalize, METH_O, NULL},
	 { "Repo_write", _wrap_Repo_write, METH_VARARGS, NULL},
	 { "Repo_write_first_repodata", _wrap_Repo_write_first_repodata, METH_VARARGS, NULL},
	 { "Repo_Dataiterator", _wrap_Repo_Dataiterator, METH_VARARGS, NULL},
	 { "Repo_Dataiterator_meta", _wrap_Repo_Dataiterator_meta, METH_VARARGS, NULL},
	 { "Repo_id_get", _wrap_Repo_id_get, METH_O, NULL},
	 { "Repo_solvables_get", _wrap_Repo_solvables_get, METH_O, NULL},
	 { "Repo_meta_get", _wrap_Repo_meta_get, METH_O, NULL},
	 { "Repo_solvables_iter", _wrap_Repo_solvables_iter, METH_O, NULL},
	 { "Repo_add_repodata", _wrap_Repo_add_repodata, METH_VARARGS, NULL},
	 { "Repo_create_stubs", _wrap_Repo_create_stubs, METH_O, NULL},
	 { "Repo_iscontiguous", _wrap_Repo_iscontiguous, METH_O, NULL},
	 { "Repo_first_repodata", _wrap_Repo_first_repodata, METH_O, NULL},
	 { "Repo_Selection", _wrap_Repo_Selection, METH_VARARGS, NULL},
	 { "Repo_find_pubkey", _wrap_Repo_find_pubkey, METH_VARARGS, NULL},
	 { "Repo_createshadow", _wrap_Repo_createshadow, METH_VARARGS, NULL},
	 { "Repo_moveshadow", _wrap_Repo_moveshadow, METH_VARARGS, NULL},
	 { "Repo___eq__", _wrap_Repo___eq__, METH_VARARGS, NULL},
	 { "Repo___ne__", _wrap_Repo___ne__, METH_VARARGS, NULL},
	 { "Repo___hash__", _wrap_Repo___hash__, METH_O, NULL},
	 { "Repo___str__", _wrap_Repo___str__, METH_O, NULL},
	 { "Repo___repr__", _wrap_Repo___repr__, METH_O, NULL},
	 { "Repo_swigregister", Repo_swigregister, METH_O, NULL},
	 { "Decision_solv_get", _wrap_Decision_solv_get, METH_O, NULL},
	 { "Decision_p_get", _wrap_Decision_p_get, METH_O, NULL},
	 { "Decision_reason_get", _wrap_Decision_reason_get, METH_O, NULL},
	 { "Decision_infoid_get", _wrap_Decision_infoid_get, METH_O, NULL},
	 { "new_Decision", _wrap_new_Decision, METH_VARARGS, NULL},
	 { "Decision_rule_get", _wrap_Decision_rule_get, METH_O, NULL},
	 { "Decision_solvable_get", _wrap_Decision_solvable_get, METH_O, NULL},
	 { "Decision_info", _wrap_Decision_info, METH_O, NULL},
	 { "Decision_allinfos", _wrap_Decision_allinfos, METH_O, NULL},
	 { "Decision_reasonstr", _wrap_Decision_reasonstr, METH_VARARGS, NULL},
	 { "Decision___str__", _wrap_Decision___str__, METH_O, NULL},
	 { "delete_Decision", _wrap_delete_Decision, METH_O, NULL},
	 { "Decision_swigregister", Decision_swigregister, METH_O, NULL},
	 { "Decision_swiginit", Decision_swiginit, METH_VARARGS, NULL},
	 { "Decisionset_solv_get", _wrap_Decisionset_solv_get, METH_O, NULL},
	 { "Decisionset_p_get", _wrap_Decisionset_p_get, METH_O, NULL},
	 { "Decisionset_reason_get", _wrap_Decisionset_reason_get, METH_O, NULL},
	 { "Decisionset_infoid_get", _wrap_Decisionset_infoid_get, METH_O, NULL},
	 { "Decisionset_bits_get", _wrap_Decisionset_bits_get, METH_O, NULL},
	 { "Decisionset_type_get", _wrap_Decisionset_type_get, METH_O, NULL},
	 { "Decisionset_dep_id_get", _wrap_Decisionset_dep_id_get, METH_O, NULL},
	 { "new_Decisionset", _wrap_new_Decisionset, METH_O, NULL},
	 { "delete_Decisionset", _wrap_delete_Decisionset, METH_O, NULL},
	 { "Decisionset_info", _wrap_Decisionset_info, METH_O, NULL},
	 { "Decisionset_dep_get", _wrap_Decisionset_dep_get, METH_O, NULL},
	 { "Decisionset_solvables", _wrap_Decisionset_solvables, METH_O, NULL},
	 { "Decisionset_decisions", _wrap_Decisionset_decisions, METH_O, NULL},
	 { "Decisionset_reasonstr", _wrap_Decisionset_reasonstr, METH_VARARGS, NULL},
	 { "Decisionset___str__", _wrap_Decisionset___str__, METH_O, NULL},
	 { "Decisionset_swigregister", Decisionset_swigregister, METH_O, NULL},
	 { "Decisionset_swiginit", Decisionset_swiginit, METH_VARARGS, NULL},
	 { "Solver_pool_get", _wrap_Solver_pool_get, METH_O, NULL},
	 { "delete_Solver", _wrap_delete_Solver, METH_O, NULL},
	 { "Solver_set_flag", _wrap_Solver_set_flag, METH_VARARGS, NULL},
	 { "Solver_get_flag", _wrap_Solver_get_flag, METH_VARARGS, NULL},
	 { "Solver_solve", _wrap_Solver_solve, METH_VARARGS, NULL},
	 { "Solver_transaction", _wrap_Solver_transaction, METH_O, NULL},
	 { "Solver_describe_decision", _wrap_Solver_describe_decision, METH_VARARGS, NULL},
	 { "Solver_describe_weakdep_decision_raw", _wrap_Solver_describe_weakdep_decision_raw, METH_VARARGS, NULL},
	 { "Solver_alternatives_count", _wrap_Solver_alternatives_count, METH_O, NULL},
	 { "Solver_get_alternative", _wrap_Solver_get_alternative, METH_VARARGS, NULL},
	 { "Solver_alternatives", _wrap_Solver_alternatives, METH_O, NULL},
	 { "Solver_write_testcase", _wrap_Solver_write_testcase, METH_VARARGS, NULL},
	 { "Solver_raw_decisions", _wrap_Solver_raw_decisions, METH_VARARGS, NULL},
	 { "Solver_all_decisions", _wrap_Solver_all_decisions, METH_VARARGS, NULL},
	 { "Solver_get_decision", _wrap_Solver_get_decision, METH_VARARGS, NULL},
	 { "Solver_get_learnt", _wrap_Solver_get_learnt, METH_VARARGS, NULL},
	 { "Solver_get_decisionlist", _wrap_Solver_get_decisionlist, METH_VARARGS, NULL},
	 { "Solver_get_recommended", _wrap_Solver_get_recommended, METH_VARARGS, NULL},
	 { "Solver_get_suggested", _wrap_Solver_get_suggested, METH_VARARGS, NULL},
	 { "Solver_swigregister", Solver_swigregister, METH_O, NULL},
	 { "new_Chksum", _wrap_new_Chksum, METH_VARARGS, NULL},
	 { "Chksum_from_bin", _wrap_Chksum_from_bin, METH_VARARGS, NULL},
	 { "delete_Chksum", _wrap_delete_Chksum, METH_O, NULL},
	 { "Chksum_type_get", _wrap_Chksum_type_get, METH_O, NULL},
	 { "Chksum_add", _wrap_Chksum_add, METH_VARARGS, NULL},
	 { "Chksum_add_fp", _wrap_Chksum_add_fp, METH_VARARGS, NULL},
	 { "Chksum_add_fd", _wrap_Chksum_add_fd, METH_VARARGS, NULL},
	 { "Chksum_add_stat", _wrap_Chksum_add_stat, METH_VARARGS, NULL},
	 { "Chksum_add_fstat", _wrap_Chksum_add_fstat, METH_VARARGS, NULL},
	 { "Chksum_raw", _wrap_Chksum_raw, METH_O, NULL},
	 { "Chksum_hex", _wrap_Chksum_hex, METH_O, NULL},
	 { "Chksum_typestr", _wrap_Chksum_typestr, METH_O, NULL},
	 { "Chksum___eq__", _wrap_Chksum___eq__, METH_VARARGS, NULL},
	 { "Chksum___ne__", _wrap_Chksum___ne__, METH_VARARGS, NULL},
	 { "Chksum___str__", _wrap_Chksum___str__, METH_O, NULL},
	 { "Chksum___repr__", _wrap_Chksum___repr__, METH_O, NULL},
	 { "Chksum_swigregister", Chksum_swigregister, METH_O, NULL},
	 { "Chksum_swiginit", Chksum_swiginit, METH_VARARGS, NULL},
	 { "Solvsig_htype_get", _wrap_Solvsig_htype_get, METH_O, NULL},
	 { "Solvsig_created_get", _wrap_Solvsig_created_get, METH_O, NULL},
	 { "Solvsig_expires_get", _wrap_Solvsig_expires_get, METH_O, NULL},
	 { "Solvsig_keyid_get", _wrap_Solvsig_keyid_get, METH_O, NULL},
	 { "new_Solvsig", _wrap_new_Solvsig, METH_O, NULL},
	 { "delete_Solvsig", _wrap_delete_Solvsig, METH_O, NULL},
	 { "Solvsig_Chksum", _wrap_Solvsig_Chksum, METH_O, NULL},
	 { "Solvsig_verify", _wrap_Solvsig_verify, METH_VARARGS, NULL},
	 { "Solvsig_swigregister", Solvsig_swigregister, METH_O, NULL},
	 { "Solvsig_swiginit", Solvsig_swiginit, METH_VARARGS, NULL},
	 { "delete_SolvFp", _wrap_delete_SolvFp, METH_O, NULL},
	 { "SolvFp_fileno", _wrap_SolvFp_fileno, METH_O, NULL},
	 { "SolvFp_dup", _wrap_SolvFp_dup, METH_O, NULL},
	 { "SolvFp_write", _wrap_SolvFp_write, METH_VARARGS, NULL},
	 { "SolvFp_flush", _wrap_SolvFp_flush, METH_O, NULL},
	 { "SolvFp_close", _wrap_SolvFp_close, METH_O, NULL},
	 { "SolvFp_cloexec", _wrap_SolvFp_cloexec, METH_VARARGS, NULL},
	 { "SolvFp_swigregister", SolvFp_swigregister, METH_O, NULL},
	 { "xfopen", _wrap_xfopen, METH_VARARGS, NULL},
	 { "xfopen_fd", _wrap_xfopen_fd, METH_VARARGS, NULL},
	 { "Problem_solv_get", _wrap_Problem_solv_get, METH_O, NULL},
	 { "Problem_id_get", _wrap_Problem_id_get, METH_O, NULL},
	 { "new_Problem", _wrap_new_Problem, METH_VARARGS, NULL},
	 { "Problem_findproblemrule", _wrap_Problem_findproblemrule, METH_O, NULL},
	 { "Problem_findallproblemrules", _wrap_Problem_findallproblemrules, METH_VARARGS, NULL},
	 { "Problem_solution_count", _wrap_Problem_solution_count, METH_O, NULL},
	 { "Problem_solutions", _wrap_Problem_solutions, METH_O, NULL},
	 { "Problem_get_learnt", _wrap_Problem_get_learnt, METH_O, NULL},
	 { "Problem_get_decisionlist", _wrap_Problem_get_decisionlist, METH_O, NULL},
	 { "Problem_get_decisionsetlist", _wrap_Problem_get_decisionsetlist, METH_O, NULL},
	 { "Problem___str__", _wrap_Problem___str__, METH_O, NULL},
	 { "delete_Problem", _wrap_delete_Problem, METH_O, NULL},
	 { "Problem_swigregister", Problem_swigregister, METH_O, NULL},
	 { "Problem_swiginit", Problem_swiginit, METH_VARARGS, NULL},
	 { "Solution_solv_get", _wrap_Solution_solv_get, METH_O, NULL},
	 { "Solution_problemid_get", _wrap_Solution_problemid_get, METH_O, NULL},
	 { "Solution_id_get", _wrap_Solution_id_get, METH_O, NULL},
	 { "new_Solution", _wrap_new_Solution, METH_VARARGS, NULL},
	 { "Solution_element_count", _wrap_Solution_element_count, METH_O, NULL},
	 { "Solution_elements", _wrap_Solution_elements, METH_VARARGS, NULL},
	 { "delete_Solution", _wrap_delete_Solution, METH_O, NULL},
	 { "Solution_swigregister", Solution_swigregister, METH_O, NULL},
	 { "Solution_swiginit", Solution_swiginit, METH_VARARGS, NULL},
	 { "Solutionelement_solv_get", _wrap_Solutionelement_solv_get, METH_O, NULL},
	 { "Solutionelement_problemid_get", _wrap_Solutionelement_problemid_get, METH_O, NULL},
	 { "Solutionelement_solutionid_get", _wrap_Solutionelement_solutionid_get, METH_O, NULL},
	 { "Solutionelement_type_get", _wrap_Solutionelement_type_get, METH_O, NULL},
	 { "new_Solutionelement", _wrap_new_Solutionelement, METH_VARARGS, NULL},
	 { "Solutionelement_str", _wrap_Solutionelement_str, METH_O, NULL},
	 { "Solutionelement___str__", _wrap_Solutionelement___str__, METH_O, NULL},
	 { "Solutionelement_replaceelements", _wrap_Solutionelement_replaceelements, METH_O, NULL},
	 { "Solutionelement_illegalreplace", _wrap_Solutionelement_illegalreplace, METH_O, NULL},
	 { "Solutionelement_solvable_get", _wrap_Solutionelement_solvable_get, METH_O, NULL},
	 { "Solutionelement_replacement_get", _wrap_Solutionelement_replacement_get, METH_O, NULL},
	 { "Solutionelement_jobidx_get", _wrap_Solutionelement_jobidx_get, METH_O, NULL},
	 { "Solutionelement_Job", _wrap_Solutionelement_Job, METH_O, NULL},
	 { "delete_Solutionelement", _wrap_delete_Solutionelement, METH_O, NULL},
	 { "Solutionelement_swigregister", Solutionelement_swigregister, METH_O, NULL},
	 { "Solutionelement_swiginit", Solutionelement_swiginit, METH_VARARGS, NULL},
	 { "Alternative_solv_get", _wrap_Alternative_solv_get, METH_O, NULL},
	 { "Alternative_type_get", _wrap_Alternative_type_get, METH_O, NULL},
	 { "Alternative_dep_id_get", _wrap_Alternative_dep_id_get, METH_O, NULL},
	 { "Alternative_chosen_id_get", _wrap_Alternative_chosen_id_get, METH_O, NULL},
	 { "Alternative_level_get", _wrap_Alternative_level_get, METH_O, NULL},
	 { "delete_Alternative", _wrap_delete_Alternative, METH_O, NULL},
	 { "Alternative_chosen_get", _wrap_Alternative_chosen_get, METH_O, NULL},
	 { "Alternative_rule_get", _wrap_Alternative_rule_get, METH_O, NULL},
	 { "Alternative_depsolvable_get", _wrap_Alternative_depsolvable_get, METH_O, NULL},
	 { "Alternative_dep_get", _wrap_Alternative_dep_get, METH_O, NULL},
	 { "Alternative_choices_raw", _wrap_Alternative_choices_raw, METH_O, NULL},
	 { "Alternative_choices", _wrap_Alternative_choices, METH_O, NULL},
	 { "Alternative___str__", _wrap_Alternative___str__, METH_O, NULL},
	 { "Alternative_swigregister", Alternative_swigregister, METH_O, NULL},
	 { "Transaction_pool_get", _wrap_Transaction_pool_get, METH_O, NULL},
	 { "delete_Transaction", _wrap_delete_Transaction, METH_O, NULL},
	 { "Transaction_isempty", _wrap_Transaction_isempty, METH_O, NULL},
	 { "Transaction_othersolvable", _wrap_Transaction_othersolvable, METH_VARARGS, NULL},
	 { "Transaction_allothersolvables", _wrap_Transaction_allothersolvables, METH_VARARGS, NULL},
	 { "Transaction_classify", _wrap_Transaction_classify, METH_VARARGS, NULL},
	 { "Transaction_newpackages", _wrap_Transaction_newpackages, METH_O, NULL},
	 { "Transaction_keptpackages", _wrap_Transaction_keptpackages, METH_O, NULL},
	 { "Transaction_newsolvables", _wrap_Transaction_newsolvables, METH_O, NULL},
	 { "Transaction_keptsolvables", _wrap_Transaction_keptsolvables, METH_O, NULL},
	 { "Transaction_steps", _wrap_Transaction_steps, METH_O, NULL},
	 { "Transaction_steptype", _wrap_Transaction_steptype, METH_VARARGS, NULL},
	 { "Transaction_calc_installsizechange", _wrap_Transaction_calc_installsizechange, METH_O, NULL},
	 { "Transaction_order", _wrap_Transaction_order, METH_VARARGS, NULL},
	 { "Transaction_swigregister", Transaction_swigregister, METH_O, NULL},
	 { "TransactionClass_transaction_get", _wrap_TransactionClass_transaction_get, METH_O, NULL},
	 { "TransactionClass_type_get", _wrap_TransactionClass_type_get, METH_O, NULL},
	 { "TransactionClass_fromid_get", _wrap_TransactionClass_fromid_get, METH_O, NULL},
	 { "TransactionClass_toid_get", _wrap_TransactionClass_toid_get, METH_O, NULL},
	 { "TransactionClass_count_get", _wrap_TransactionClass_count_get, METH_O, NULL},
	 { "new_TransactionClass", _wrap_new_TransactionClass, METH_VARARGS, NULL},
	 { "TransactionClass_solvables", _wrap_TransactionClass_solvables, METH_O, NULL},
	 { "TransactionClass_fromstr_get", _wrap_TransactionClass_fromstr_get, METH_O, NULL},
	 { "TransactionClass_tostr_get", _wrap_TransactionClass_tostr_get, METH_O, NULL},
	 { "delete_TransactionClass", _wrap_delete_TransactionClass, METH_O, NULL},
	 { "TransactionClass_swigregister", TransactionClass_swigregister, METH_O, NULL},
	 { "TransactionClass_swiginit", TransactionClass_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_Alternative = {"_p_Alternative", "Alternative *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Chksum = {"_p_Chksum", "Chksum *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Dataiterator = {"_p_Dataiterator", "Dataiterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Datamatch = {"_p_Datamatch", "Datamatch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Datapos = {"_p_Datapos", "Datapos *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Decision = {"_p_Decision", "Decision *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Decisionset = {"_p_Decisionset", "Decisionset *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Dep = {"_p_Dep", "Dep *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_FILE = {"_p_FILE", "FILE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Job = {"_p_Job", "Job *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Pool = {"_p_Pool", "Pool *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Pool_repo_iterator = {"_p_Pool_repo_iterator", "Pool_repo_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Pool_solvable_iterator = {"_p_Pool_solvable_iterator", "Pool_solvable_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Problem = {"_p_Problem", "Problem *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Queue = {"_p_Queue", "Queue *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Repo = {"_p_Repo", "Repo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Repo_solvable_iterator = {"_p_Repo_solvable_iterator", "Repo_solvable_iterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Ruleinfo = {"_p_Ruleinfo", "Ruleinfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Selection = {"_p_Selection", "Selection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Solution = {"_p_Solution", "Solution *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Solutionelement = {"_p_Solutionelement", "Solutionelement *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_SolvFp = {"_p_SolvFp", "SolvFp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Solver = {"_p_Solver", "Solver *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Solvsig = {"_p_Solvsig", "Solvsig *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Transaction = {"_p_Transaction", "Transaction *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_TransactionClass = {"_p_TransactionClass", "TransactionClass *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_XRepodata = {"_p_XRepodata", "XRepodata *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_XRule = {"_p_XRule", "XRule *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_XSolvable = {"_p_XSolvable", "XSolvable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_int = {"_p_int", "Id *|int *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_XRule = {"_p_p_XRule", "XRule **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_Alternative,
  &_swigt__p_Chksum,
  &_swigt__p_Dataiterator,
  &_swigt__p_Datamatch,
  &_swigt__p_Datapos,
  &_swigt__p_Decision,
  &_swigt__p_Decisionset,
  &_swigt__p_Dep,
  &_swigt__p_FILE,
  &_swigt__p_Job,
  &_swigt__p_Pool,
  &_swigt__p_Pool_repo_iterator,
  &_swigt__p_Pool_solvable_iterator,
  &_swigt__p_Problem,
  &_swigt__p_Queue,
  &_swigt__p_Repo,
  &_swigt__p_Repo_solvable_iterator,
  &_swigt__p_Ruleinfo,
  &_swigt__p_Selection,
  &_swigt__p_Solution,
  &_swigt__p_Solutionelement,
  &_swigt__p_SolvFp,
  &_swigt__p_Solver,
  &_swigt__p_Solvsig,
  &_swigt__p_Transaction,
  &_swigt__p_TransactionClass,
  &_swigt__p_XRepodata,
  &_swigt__p_XRule,
  &_swigt__p_XSolvable,
  &_swigt__p_char,
  &_swigt__p_int,
  &_swigt__p_p_XRule,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
};

static swig_cast_info _swigc__p_Alternative[] = {  {&_swigt__p_Alternative, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Chksum[] = {  {&_swigt__p_Chksum, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Dataiterator[] = {  {&_swigt__p_Dataiterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Datamatch[] = {  {&_swigt__p_Datamatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Datapos[] = {  {&_swigt__p_Datapos, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Decision[] = {  {&_swigt__p_Decision, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Decisionset[] = {  {&_swigt__p_Decisionset, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Dep[] = {  {&_swigt__p_Dep, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_FILE[] = {  {&_swigt__p_FILE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Job[] = {  {&_swigt__p_Job, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool[] = {  {&_swigt__p_Pool, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool_repo_iterator[] = {  {&_swigt__p_Pool_repo_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Pool_solvable_iterator[] = {  {&_swigt__p_Pool_solvable_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Problem[] = {  {&_swigt__p_Problem, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Queue[] = {  {&_swigt__p_Queue, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Repo[] = {  {&_swigt__p_Repo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Repo_solvable_iterator[] = {  {&_swigt__p_Repo_solvable_iterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Ruleinfo[] = {  {&_swigt__p_Ruleinfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Selection[] = {  {&_swigt__p_Selection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solution[] = {  {&_swigt__p_Solution, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solutionelement[] = {  {&_swigt__p_Solutionelement, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_SolvFp[] = {  {&_swigt__p_SolvFp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solver[] = {  {&_swigt__p_Solver, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Solvsig[] = {  {&_swigt__p_Solvsig, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Transaction[] = {  {&_swigt__p_Transaction, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_TransactionClass[] = {  {&_swigt__p_TransactionClass, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XRepodata[] = {  {&_swigt__p_XRepodata, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XRule[] = {  {&_swigt__p_XRule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_XSolvable[] = {  {&_swigt__p_XSolvable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_int[] = {  {&_swigt__p_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_XRule[] = {  {&_swigt__p_p_XRule, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_Alternative,
  _swigc__p_Chksum,
  _swigc__p_Dataiterator,
  _swigc__p_Datamatch,
  _swigc__p_Datapos,
  _swigc__p_Decision,
  _swigc__p_Decisionset,
  _swigc__p_Dep,
  _swigc__p_FILE,
  _swigc__p_Job,
  _swigc__p_Pool,
  _swigc__p_Pool_repo_iterator,
  _swigc__p_Pool_solvable_iterator,
  _swigc__p_Problem,
  _swigc__p_Queue,
  _swigc__p_Repo,
  _swigc__p_Repo_solvable_iterator,
  _swigc__p_Ruleinfo,
  _swigc__p_Selection,
  _swigc__p_Solution,
  _swigc__p_Solutionelement,
  _swigc__p_SolvFp,
  _swigc__p_Solver,
  _swigc__p_Solvsig,
  _swigc__p_Transaction,
  _swigc__p_TransactionClass,
  _swigc__p_XRepodata,
  _swigc__p_XRule,
  _swigc__p_XSolvable,
  _swigc__p_char,
  _swigc__p_int,
  _swigc__p_p_XRule,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        SWIG_Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------
   * Patch %callback methods' docstrings to hold the callback ptrs
   * -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods, const swig_const_info *const_table, swig_type_info **types, swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        const swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif




/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

SWIGINTERN int SWIG_mod_exec(PyObject *module);

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
#if PY_VERSION_HEX >= 0x03000000
  static PyModuleDef_Slot SwigSlots[] = {
    {
      Py_mod_exec, (void *)SWIG_mod_exec 
    },
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
    {
      Py_mod_gil, Py_MOD_GIL_NOT_USED 
    },
#endif
#endif
    {
      0, NULL 
    }
  };
  
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    0,
    SwigMethods,
    SwigSlots,
    NULL,
    NULL,
    NULL
  };
  
  return PyModuleDef_Init(&SWIG_module);
#else
  PyObject *m = Py_InitModule(SWIG_name, SwigMethods);
  if (m && SWIG_mod_exec(m) != 0) {
    Py_DECREF(m);
  }
#endif
}

SWIGINTERN int SWIG_mod_exec(PyObject *m) {
  PyObject *d, *md, *globals;
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType_Type();
  assert(metatype);
  
  SwigPyStaticVar_Type();
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_runtime_data_module();
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_Type();
  SwigVarLink_Type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_Type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#ifdef SWIGPYTHON_NOGIL
#ifdef Py_GIL_DISABLED
  PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
#endif
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
    return -1;
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_Type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  SWIG_Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  SWIG_Python_SetConstant(d, "ID_NULL",SWIG_From_int((int)(ID_NULL)));
  SWIG_Python_SetConstant(d, "ID_EMPTY",SWIG_From_int((int)(ID_EMPTY)));
  SWIG_Python_SetConstant(d, "SOLVABLE_NAME",SWIG_From_int((int)(SOLVABLE_NAME)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ARCH",SWIG_From_int((int)(SOLVABLE_ARCH)));
  SWIG_Python_SetConstant(d, "SOLVABLE_EVR",SWIG_From_int((int)(SOLVABLE_EVR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_VENDOR",SWIG_From_int((int)(SOLVABLE_VENDOR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PROVIDES",SWIG_From_int((int)(SOLVABLE_PROVIDES)));
  SWIG_Python_SetConstant(d, "SOLVABLE_OBSOLETES",SWIG_From_int((int)(SOLVABLE_OBSOLETES)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CONFLICTS",SWIG_From_int((int)(SOLVABLE_CONFLICTS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_REQUIRES",SWIG_From_int((int)(SOLVABLE_REQUIRES)));
  SWIG_Python_SetConstant(d, "SOLVABLE_RECOMMENDS",SWIG_From_int((int)(SOLVABLE_RECOMMENDS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SUGGESTS",SWIG_From_int((int)(SOLVABLE_SUGGESTS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SUPPLEMENTS",SWIG_From_int((int)(SOLVABLE_SUPPLEMENTS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ENHANCES",SWIG_From_int((int)(SOLVABLE_ENHANCES)));
  SWIG_Python_SetConstant(d, "RPM_RPMDBID",SWIG_From_int((int)(RPM_RPMDBID)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PREREQMARKER",SWIG_From_int((int)(SOLVABLE_PREREQMARKER)));
  SWIG_Python_SetConstant(d, "SOLVABLE_FILEMARKER",SWIG_From_int((int)(SOLVABLE_FILEMARKER)));
  SWIG_Python_SetConstant(d, "NAMESPACE_INSTALLED",SWIG_From_int((int)(NAMESPACE_INSTALLED)));
  SWIG_Python_SetConstant(d, "NAMESPACE_MODALIAS",SWIG_From_int((int)(NAMESPACE_MODALIAS)));
  SWIG_Python_SetConstant(d, "NAMESPACE_SPLITPROVIDES",SWIG_From_int((int)(NAMESPACE_SPLITPROVIDES)));
  SWIG_Python_SetConstant(d, "NAMESPACE_LANGUAGE",SWIG_From_int((int)(NAMESPACE_LANGUAGE)));
  SWIG_Python_SetConstant(d, "NAMESPACE_FILESYSTEM",SWIG_From_int((int)(NAMESPACE_FILESYSTEM)));
  SWIG_Python_SetConstant(d, "NAMESPACE_OTHERPROVIDERS",SWIG_From_int((int)(NAMESPACE_OTHERPROVIDERS)));
  SWIG_Python_SetConstant(d, "SYSTEM_SYSTEM",SWIG_From_int((int)(SYSTEM_SYSTEM)));
  SWIG_Python_SetConstant(d, "ARCH_SRC",SWIG_From_int((int)(ARCH_SRC)));
  SWIG_Python_SetConstant(d, "ARCH_NOSRC",SWIG_From_int((int)(ARCH_NOSRC)));
  SWIG_Python_SetConstant(d, "ARCH_NOARCH",SWIG_From_int((int)(ARCH_NOARCH)));
  SWIG_Python_SetConstant(d, "ARCH_ALL",SWIG_From_int((int)(ARCH_ALL)));
  SWIG_Python_SetConstant(d, "ARCH_ANY",SWIG_From_int((int)(ARCH_ANY)));
  SWIG_Python_SetConstant(d, "REPOSITORY_SOLVABLES",SWIG_From_int((int)(REPOSITORY_SOLVABLES)));
  SWIG_Python_SetConstant(d, "REPOSITORY_DELTAINFO",SWIG_From_int((int)(REPOSITORY_DELTAINFO)));
  SWIG_Python_SetConstant(d, "REPOSITORY_EXTERNAL",SWIG_From_int((int)(REPOSITORY_EXTERNAL)));
  SWIG_Python_SetConstant(d, "REPOSITORY_KEYS",SWIG_From_int((int)(REPOSITORY_KEYS)));
  SWIG_Python_SetConstant(d, "REPOSITORY_LOCATION",SWIG_From_int((int)(REPOSITORY_LOCATION)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_VOID",SWIG_From_int((int)(REPOKEY_TYPE_VOID)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_CONSTANT",SWIG_From_int((int)(REPOKEY_TYPE_CONSTANT)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_CONSTANTID",SWIG_From_int((int)(REPOKEY_TYPE_CONSTANTID)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_ID",SWIG_From_int((int)(REPOKEY_TYPE_ID)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_NUM",SWIG_From_int((int)(REPOKEY_TYPE_NUM)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_DIR",SWIG_From_int((int)(REPOKEY_TYPE_DIR)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_STR",SWIG_From_int((int)(REPOKEY_TYPE_STR)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_BINARY",SWIG_From_int((int)(REPOKEY_TYPE_BINARY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_IDARRAY",SWIG_From_int((int)(REPOKEY_TYPE_IDARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_REL_IDARRAY",SWIG_From_int((int)(REPOKEY_TYPE_REL_IDARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_DIRSTRARRAY",SWIG_From_int((int)(REPOKEY_TYPE_DIRSTRARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_DIRNUMNUMARRAY",SWIG_From_int((int)(REPOKEY_TYPE_DIRNUMNUMARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_MD5",SWIG_From_int((int)(REPOKEY_TYPE_MD5)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_SHA1",SWIG_From_int((int)(REPOKEY_TYPE_SHA1)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_SHA224",SWIG_From_int((int)(REPOKEY_TYPE_SHA224)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_SHA256",SWIG_From_int((int)(REPOKEY_TYPE_SHA256)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_SHA384",SWIG_From_int((int)(REPOKEY_TYPE_SHA384)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_SHA512",SWIG_From_int((int)(REPOKEY_TYPE_SHA512)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_FIXARRAY",SWIG_From_int((int)(REPOKEY_TYPE_FIXARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_FLEXARRAY",SWIG_From_int((int)(REPOKEY_TYPE_FLEXARRAY)));
  SWIG_Python_SetConstant(d, "REPOKEY_TYPE_DELETED",SWIG_From_int((int)(REPOKEY_TYPE_DELETED)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SUMMARY",SWIG_From_int((int)(SOLVABLE_SUMMARY)));
  SWIG_Python_SetConstant(d, "SOLVABLE_DESCRIPTION",SWIG_From_int((int)(SOLVABLE_DESCRIPTION)));
  SWIG_Python_SetConstant(d, "SOLVABLE_DISTRIBUTION",SWIG_From_int((int)(SOLVABLE_DISTRIBUTION)));
  SWIG_Python_SetConstant(d, "SOLVABLE_AUTHORS",SWIG_From_int((int)(SOLVABLE_AUTHORS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PACKAGER",SWIG_From_int((int)(SOLVABLE_PACKAGER)));
  SWIG_Python_SetConstant(d, "SOLVABLE_GROUP",SWIG_From_int((int)(SOLVABLE_GROUP)));
  SWIG_Python_SetConstant(d, "SOLVABLE_URL",SWIG_From_int((int)(SOLVABLE_URL)));
  SWIG_Python_SetConstant(d, "SOLVABLE_KEYWORDS",SWIG_From_int((int)(SOLVABLE_KEYWORDS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_LICENSE",SWIG_From_int((int)(SOLVABLE_LICENSE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_BUILDTIME",SWIG_From_int((int)(SOLVABLE_BUILDTIME)));
  SWIG_Python_SetConstant(d, "SOLVABLE_BUILDHOST",SWIG_From_int((int)(SOLVABLE_BUILDHOST)));
  SWIG_Python_SetConstant(d, "SOLVABLE_EULA",SWIG_From_int((int)(SOLVABLE_EULA)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CPEID",SWIG_From_int((int)(SOLVABLE_CPEID)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MESSAGEINS",SWIG_From_int((int)(SOLVABLE_MESSAGEINS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MESSAGEDEL",SWIG_From_int((int)(SOLVABLE_MESSAGEDEL)));
  SWIG_Python_SetConstant(d, "SOLVABLE_INSTALLSIZE",SWIG_From_int((int)(SOLVABLE_INSTALLSIZE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_DISKUSAGE",SWIG_From_int((int)(SOLVABLE_DISKUSAGE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_FILELIST",SWIG_From_int((int)(SOLVABLE_FILELIST)));
  SWIG_Python_SetConstant(d, "SOLVABLE_INSTALLTIME",SWIG_From_int((int)(SOLVABLE_INSTALLTIME)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MEDIADIR",SWIG_From_int((int)(SOLVABLE_MEDIADIR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MEDIAFILE",SWIG_From_int((int)(SOLVABLE_MEDIAFILE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MEDIANR",SWIG_From_int((int)(SOLVABLE_MEDIANR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MEDIABASE",SWIG_From_int((int)(SOLVABLE_MEDIABASE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_DOWNLOADSIZE",SWIG_From_int((int)(SOLVABLE_DOWNLOADSIZE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SOURCEARCH",SWIG_From_int((int)(SOLVABLE_SOURCEARCH)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SOURCENAME",SWIG_From_int((int)(SOLVABLE_SOURCENAME)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SOURCEEVR",SWIG_From_int((int)(SOLVABLE_SOURCEEVR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ISVISIBLE",SWIG_From_int((int)(SOLVABLE_ISVISIBLE)));
  SWIG_Python_SetConstant(d, "SOLVABLE_TRIGGERS",SWIG_From_int((int)(SOLVABLE_TRIGGERS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CHECKSUM",SWIG_From_int((int)(SOLVABLE_CHECKSUM)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PKGID",SWIG_From_int((int)(SOLVABLE_PKGID)));
  SWIG_Python_SetConstant(d, "SOLVABLE_HDRID",SWIG_From_int((int)(SOLVABLE_HDRID)));
  SWIG_Python_SetConstant(d, "SOLVABLE_LEADSIGID",SWIG_From_int((int)(SOLVABLE_LEADSIGID)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PATCHCATEGORY",SWIG_From_int((int)(SOLVABLE_PATCHCATEGORY)));
  SWIG_Python_SetConstant(d, "SOLVABLE_HEADEREND",SWIG_From_int((int)(SOLVABLE_HEADEREND)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CHANGELOG",SWIG_From_int((int)(SOLVABLE_CHANGELOG)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CHANGELOG_AUTHOR",SWIG_From_int((int)(SOLVABLE_CHANGELOG_AUTHOR)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CHANGELOG_TIME",SWIG_From_int((int)(SOLVABLE_CHANGELOG_TIME)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CHANGELOG_TEXT",SWIG_From_int((int)(SOLVABLE_CHANGELOG_TEXT)));
  SWIG_Python_SetConstant(d, "SOLVABLE_INSTALLSTATUS",SWIG_From_int((int)(SOLVABLE_INSTALLSTATUS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_PREREQ_IGNOREINST",SWIG_From_int((int)(SOLVABLE_PREREQ_IGNOREINST)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CATEGORY",SWIG_From_int((int)(SOLVABLE_CATEGORY)));
  SWIG_Python_SetConstant(d, "SOLVABLE_INCLUDES",SWIG_From_int((int)(SOLVABLE_INCLUDES)));
  SWIG_Python_SetConstant(d, "SOLVABLE_EXTENDS",SWIG_From_int((int)(SOLVABLE_EXTENDS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ICON",SWIG_From_int((int)(SOLVABLE_ICON)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ORDER",SWIG_From_int((int)(SOLVABLE_ORDER)));
  SWIG_Python_SetConstant(d, "UPDATE_REBOOT",SWIG_From_int((int)(UPDATE_REBOOT)));
  SWIG_Python_SetConstant(d, "UPDATE_RESTART",SWIG_From_int((int)(UPDATE_RESTART)));
  SWIG_Python_SetConstant(d, "UPDATE_RELOGIN",SWIG_From_int((int)(UPDATE_RELOGIN)));
  SWIG_Python_SetConstant(d, "UPDATE_MESSAGE",SWIG_From_int((int)(UPDATE_MESSAGE)));
  SWIG_Python_SetConstant(d, "UPDATE_SEVERITY",SWIG_From_int((int)(UPDATE_SEVERITY)));
  SWIG_Python_SetConstant(d, "UPDATE_RIGHTS",SWIG_From_int((int)(UPDATE_RIGHTS)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION",SWIG_From_int((int)(UPDATE_COLLECTION)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION_NAME",SWIG_From_int((int)(UPDATE_COLLECTION_NAME)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION_EVR",SWIG_From_int((int)(UPDATE_COLLECTION_EVR)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION_ARCH",SWIG_From_int((int)(UPDATE_COLLECTION_ARCH)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION_FILENAME",SWIG_From_int((int)(UPDATE_COLLECTION_FILENAME)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTION_FLAGS",SWIG_From_int((int)(UPDATE_COLLECTION_FLAGS)));
  SWIG_Python_SetConstant(d, "UPDATE_REFERENCE",SWIG_From_int((int)(UPDATE_REFERENCE)));
  SWIG_Python_SetConstant(d, "UPDATE_REFERENCE_TYPE",SWIG_From_int((int)(UPDATE_REFERENCE_TYPE)));
  SWIG_Python_SetConstant(d, "UPDATE_REFERENCE_HREF",SWIG_From_int((int)(UPDATE_REFERENCE_HREF)));
  SWIG_Python_SetConstant(d, "UPDATE_REFERENCE_ID",SWIG_From_int((int)(UPDATE_REFERENCE_ID)));
  SWIG_Python_SetConstant(d, "UPDATE_REFERENCE_TITLE",SWIG_From_int((int)(UPDATE_REFERENCE_TITLE)));
  SWIG_Python_SetConstant(d, "PRODUCT_REFERENCEFILE",SWIG_From_int((int)(PRODUCT_REFERENCEFILE)));
  SWIG_Python_SetConstant(d, "PRODUCT_SHORTLABEL",SWIG_From_int((int)(PRODUCT_SHORTLABEL)));
  SWIG_Python_SetConstant(d, "PRODUCT_DISTPRODUCT",SWIG_From_int((int)(PRODUCT_DISTPRODUCT)));
  SWIG_Python_SetConstant(d, "PRODUCT_DISTVERSION",SWIG_From_int((int)(PRODUCT_DISTVERSION)));
  SWIG_Python_SetConstant(d, "PRODUCT_TYPE",SWIG_From_int((int)(PRODUCT_TYPE)));
  SWIG_Python_SetConstant(d, "PRODUCT_URL",SWIG_From_int((int)(PRODUCT_URL)));
  SWIG_Python_SetConstant(d, "PRODUCT_URL_TYPE",SWIG_From_int((int)(PRODUCT_URL_TYPE)));
  SWIG_Python_SetConstant(d, "PRODUCT_FLAGS",SWIG_From_int((int)(PRODUCT_FLAGS)));
  SWIG_Python_SetConstant(d, "PRODUCT_PRODUCTLINE",SWIG_From_int((int)(PRODUCT_PRODUCTLINE)));
  SWIG_Python_SetConstant(d, "PRODUCT_REGISTER_TARGET",SWIG_From_int((int)(PRODUCT_REGISTER_TARGET)));
  SWIG_Python_SetConstant(d, "PRODUCT_REGISTER_FLAVOR",SWIG_From_int((int)(PRODUCT_REGISTER_FLAVOR)));
  SWIG_Python_SetConstant(d, "PRODUCT_REGISTER_RELEASE",SWIG_From_int((int)(PRODUCT_REGISTER_RELEASE)));
  SWIG_Python_SetConstant(d, "PRODUCT_UPDATES_REPOID",SWIG_From_int((int)(PRODUCT_UPDATES_REPOID)));
  SWIG_Python_SetConstant(d, "PRODUCT_UPDATES",SWIG_From_int((int)(PRODUCT_UPDATES)));
  SWIG_Python_SetConstant(d, "PRODUCT_ENDOFLIFE",SWIG_From_int((int)(PRODUCT_ENDOFLIFE)));
  SWIG_Python_SetConstant(d, "SUSETAGS_DATADIR",SWIG_From_int((int)(SUSETAGS_DATADIR)));
  SWIG_Python_SetConstant(d, "SUSETAGS_DESCRDIR",SWIG_From_int((int)(SUSETAGS_DESCRDIR)));
  SWIG_Python_SetConstant(d, "SUSETAGS_DEFAULTVENDOR",SWIG_From_int((int)(SUSETAGS_DEFAULTVENDOR)));
  SWIG_Python_SetConstant(d, "SUSETAGS_FILE",SWIG_From_int((int)(SUSETAGS_FILE)));
  SWIG_Python_SetConstant(d, "SUSETAGS_FILE_NAME",SWIG_From_int((int)(SUSETAGS_FILE_NAME)));
  SWIG_Python_SetConstant(d, "SUSETAGS_FILE_TYPE",SWIG_From_int((int)(SUSETAGS_FILE_TYPE)));
  SWIG_Python_SetConstant(d, "SUSETAGS_FILE_CHECKSUM",SWIG_From_int((int)(SUSETAGS_FILE_CHECKSUM)));
  SWIG_Python_SetConstant(d, "SUSETAGS_SHARE_NAME",SWIG_From_int((int)(SUSETAGS_SHARE_NAME)));
  SWIG_Python_SetConstant(d, "SUSETAGS_SHARE_EVR",SWIG_From_int((int)(SUSETAGS_SHARE_EVR)));
  SWIG_Python_SetConstant(d, "SUSETAGS_SHARE_ARCH",SWIG_From_int((int)(SUSETAGS_SHARE_ARCH)));
  SWIG_Python_SetConstant(d, "REPOSITORY_ADDEDFILEPROVIDES",SWIG_From_int((int)(REPOSITORY_ADDEDFILEPROVIDES)));
  SWIG_Python_SetConstant(d, "REPOSITORY_RPMDBCOOKIE",SWIG_From_int((int)(REPOSITORY_RPMDBCOOKIE)));
  SWIG_Python_SetConstant(d, "REPOSITORY_FILTEREDFILELIST",SWIG_From_int((int)(REPOSITORY_FILTEREDFILELIST)));
  SWIG_Python_SetConstant(d, "REPOSITORY_TIMESTAMP",SWIG_From_int((int)(REPOSITORY_TIMESTAMP)));
  SWIG_Python_SetConstant(d, "REPOSITORY_EXPIRE",SWIG_From_int((int)(REPOSITORY_EXPIRE)));
  SWIG_Python_SetConstant(d, "REPOSITORY_UPDATES",SWIG_From_int((int)(REPOSITORY_UPDATES)));
  SWIG_Python_SetConstant(d, "REPOSITORY_DISTROS",SWIG_From_int((int)(REPOSITORY_DISTROS)));
  SWIG_Python_SetConstant(d, "REPOSITORY_PRODUCT_LABEL",SWIG_From_int((int)(REPOSITORY_PRODUCT_LABEL)));
  SWIG_Python_SetConstant(d, "REPOSITORY_PRODUCT_CPEID",SWIG_From_int((int)(REPOSITORY_PRODUCT_CPEID)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOID",SWIG_From_int((int)(REPOSITORY_REPOID)));
  SWIG_Python_SetConstant(d, "REPOSITORY_KEYWORDS",SWIG_From_int((int)(REPOSITORY_KEYWORDS)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REVISION",SWIG_From_int((int)(REPOSITORY_REVISION)));
  SWIG_Python_SetConstant(d, "REPOSITORY_TOOLVERSION",SWIG_From_int((int)(REPOSITORY_TOOLVERSION)));
  SWIG_Python_SetConstant(d, "DELTA_PACKAGE_NAME",SWIG_From_int((int)(DELTA_PACKAGE_NAME)));
  SWIG_Python_SetConstant(d, "DELTA_PACKAGE_EVR",SWIG_From_int((int)(DELTA_PACKAGE_EVR)));
  SWIG_Python_SetConstant(d, "DELTA_PACKAGE_ARCH",SWIG_From_int((int)(DELTA_PACKAGE_ARCH)));
  SWIG_Python_SetConstant(d, "DELTA_LOCATION_DIR",SWIG_From_int((int)(DELTA_LOCATION_DIR)));
  SWIG_Python_SetConstant(d, "DELTA_LOCATION_NAME",SWIG_From_int((int)(DELTA_LOCATION_NAME)));
  SWIG_Python_SetConstant(d, "DELTA_LOCATION_EVR",SWIG_From_int((int)(DELTA_LOCATION_EVR)));
  SWIG_Python_SetConstant(d, "DELTA_LOCATION_SUFFIX",SWIG_From_int((int)(DELTA_LOCATION_SUFFIX)));
  SWIG_Python_SetConstant(d, "DELTA_DOWNLOADSIZE",SWIG_From_int((int)(DELTA_DOWNLOADSIZE)));
  SWIG_Python_SetConstant(d, "DELTA_CHECKSUM",SWIG_From_int((int)(DELTA_CHECKSUM)));
  SWIG_Python_SetConstant(d, "DELTA_BASE_EVR",SWIG_From_int((int)(DELTA_BASE_EVR)));
  SWIG_Python_SetConstant(d, "DELTA_SEQ_NAME",SWIG_From_int((int)(DELTA_SEQ_NAME)));
  SWIG_Python_SetConstant(d, "DELTA_SEQ_EVR",SWIG_From_int((int)(DELTA_SEQ_EVR)));
  SWIG_Python_SetConstant(d, "DELTA_SEQ_NUM",SWIG_From_int((int)(DELTA_SEQ_NUM)));
  SWIG_Python_SetConstant(d, "DELTA_LOCATION_BASE",SWIG_From_int((int)(DELTA_LOCATION_BASE)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD",SWIG_From_int((int)(REPOSITORY_REPOMD)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_TYPE",SWIG_From_int((int)(REPOSITORY_REPOMD_TYPE)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_LOCATION",SWIG_From_int((int)(REPOSITORY_REPOMD_LOCATION)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_TIMESTAMP",SWIG_From_int((int)(REPOSITORY_REPOMD_TIMESTAMP)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_CHECKSUM",SWIG_From_int((int)(REPOSITORY_REPOMD_CHECKSUM)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_OPENCHECKSUM",SWIG_From_int((int)(REPOSITORY_REPOMD_OPENCHECKSUM)));
  SWIG_Python_SetConstant(d, "REPOSITORY_REPOMD_SIZE",SWIG_From_int((int)(REPOSITORY_REPOMD_SIZE)));
  SWIG_Python_SetConstant(d, "PUBKEY_KEYID",SWIG_From_int((int)(PUBKEY_KEYID)));
  SWIG_Python_SetConstant(d, "PUBKEY_FINGERPRINT",SWIG_From_int((int)(PUBKEY_FINGERPRINT)));
  SWIG_Python_SetConstant(d, "PUBKEY_EXPIRES",SWIG_From_int((int)(PUBKEY_EXPIRES)));
  SWIG_Python_SetConstant(d, "PUBKEY_SIGNATURES",SWIG_From_int((int)(PUBKEY_SIGNATURES)));
  SWIG_Python_SetConstant(d, "PUBKEY_DATA",SWIG_From_int((int)(PUBKEY_DATA)));
  SWIG_Python_SetConstant(d, "PUBKEY_SUBKEYOF",SWIG_From_int((int)(PUBKEY_SUBKEYOF)));
  SWIG_Python_SetConstant(d, "SIGNATURE_ISSUER",SWIG_From_int((int)(SIGNATURE_ISSUER)));
  SWIG_Python_SetConstant(d, "SIGNATURE_TIME",SWIG_From_int((int)(SIGNATURE_TIME)));
  SWIG_Python_SetConstant(d, "SIGNATURE_EXPIRES",SWIG_From_int((int)(SIGNATURE_EXPIRES)));
  SWIG_Python_SetConstant(d, "SIGNATURE_DATA",SWIG_From_int((int)(SIGNATURE_DATA)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE",SWIG_From_int((int)(UPDATE_MODULE)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE_NAME",SWIG_From_int((int)(UPDATE_MODULE_NAME)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE_STREAM",SWIG_From_int((int)(UPDATE_MODULE_STREAM)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE_VERSION",SWIG_From_int((int)(UPDATE_MODULE_VERSION)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE_CONTEXT",SWIG_From_int((int)(UPDATE_MODULE_CONTEXT)));
  SWIG_Python_SetConstant(d, "UPDATE_MODULE_ARCH",SWIG_From_int((int)(UPDATE_MODULE_ARCH)));
  SWIG_Python_SetConstant(d, "SOLVABLE_BUILDVERSION",SWIG_From_int((int)(SOLVABLE_BUILDVERSION)));
  SWIG_Python_SetConstant(d, "SOLVABLE_BUILDFLAVOR",SWIG_From_int((int)(SOLVABLE_BUILDFLAVOR)));
  SWIG_Python_SetConstant(d, "UPDATE_STATUS",SWIG_From_int((int)(UPDATE_STATUS)));
  SWIG_Python_SetConstant(d, "LIBSOLV_SELF_DESTRUCT_PKG",SWIG_From_int((int)(LIBSOLV_SELF_DESTRUCT_PKG)));
  SWIG_Python_SetConstant(d, "SOLVABLE_CONSTRAINS",SWIG_From_int((int)(SOLVABLE_CONSTRAINS)));
  SWIG_Python_SetConstant(d, "SOLVABLE_TRACK_FEATURES",SWIG_From_int((int)(SOLVABLE_TRACK_FEATURES)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ISDEFAULT",SWIG_From_int((int)(SOLVABLE_ISDEFAULT)));
  SWIG_Python_SetConstant(d, "SOLVABLE_LANGONLY",SWIG_From_int((int)(SOLVABLE_LANGONLY)));
  SWIG_Python_SetConstant(d, "UPDATE_COLLECTIONLIST",SWIG_From_int((int)(UPDATE_COLLECTIONLIST)));
  SWIG_Python_SetConstant(d, "SOLVABLE_MULTIARCH",SWIG_From_int((int)(SOLVABLE_MULTIARCH)));
  SWIG_Python_SetConstant(d, "SOLVABLE_SIGNATUREDATA",SWIG_From_int((int)(SOLVABLE_SIGNATUREDATA)));
  SWIG_Python_SetConstant(d, "SOLVABLE_ORDERWITHREQUIRES",SWIG_From_int((int)(SOLVABLE_ORDERWITHREQUIRES)));
  SWIG_Python_SetConstant(d, "ID_NUM_INTERNAL",SWIG_From_int((int)(ID_NUM_INTERNAL)));
  SWIG_Python_SetConstant(d, "SOLVID_META",SWIG_From_int((int)(SOLVID_META)));
  SWIG_Python_SetConstant(d, "SOLVID_POS",SWIG_From_int((int)(SOLVID_POS)));
  SWIG_Python_SetConstant(d, "REL_EQ",SWIG_From_int((int)(REL_EQ)));
  SWIG_Python_SetConstant(d, "REL_GT",SWIG_From_int((int)(REL_GT)));
  SWIG_Python_SetConstant(d, "REL_LT",SWIG_From_int((int)(REL_LT)));
  SWIG_Python_SetConstant(d, "REL_AND",SWIG_From_int((int)(REL_AND)));
  SWIG_Python_SetConstant(d, "REL_OR",SWIG_From_int((int)(REL_OR)));
  SWIG_Python_SetConstant(d, "REL_WITH",SWIG_From_int((int)(REL_WITH)));
  SWIG_Python_SetConstant(d, "REL_NAMESPACE",SWIG_From_int((int)(REL_NAMESPACE)));
  SWIG_Python_SetConstant(d, "REL_ARCH",SWIG_From_int((int)(REL_ARCH)));
  SWIG_Python_SetConstant(d, "REL_FILECONFLICT",SWIG_From_int((int)(REL_FILECONFLICT)));
  SWIG_Python_SetConstant(d, "REL_COND",SWIG_From_int((int)(REL_COND)));
  SWIG_Python_SetConstant(d, "REL_COMPAT",SWIG_From_int((int)(REL_COMPAT)));
  SWIG_Python_SetConstant(d, "REL_KIND",SWIG_From_int((int)(REL_KIND)));
  SWIG_Python_SetConstant(d, "REL_MULTIARCH",SWIG_From_int((int)(REL_MULTIARCH)));
  SWIG_Python_SetConstant(d, "REL_ELSE",SWIG_From_int((int)(REL_ELSE)));
  SWIG_Python_SetConstant(d, "REL_ERROR",SWIG_From_int((int)(REL_ERROR)));
  SWIG_Python_SetConstant(d, "REL_WITHOUT",SWIG_From_int((int)(REL_WITHOUT)));
  SWIG_Python_SetConstant(d, "REL_UNLESS",SWIG_From_int((int)(REL_UNLESS)));
  SWIG_Python_SetConstant(d, "REL_CONDA",SWIG_From_int((int)(REL_CONDA)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_NAME",SWIG_From_int((int)(SELECTION_NAME)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_PROVIDES",SWIG_From_int((int)(SELECTION_PROVIDES)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_FILELIST",SWIG_From_int((int)(SELECTION_FILELIST)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_CANON",SWIG_From_int((int)(SELECTION_CANON)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_DOTARCH",SWIG_From_int((int)(SELECTION_DOTARCH)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_REL",SWIG_From_int((int)(SELECTION_REL)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_INSTALLED_ONLY",SWIG_From_int((int)(SELECTION_INSTALLED_ONLY)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_GLOB",SWIG_From_int((int)(SELECTION_GLOB)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_FLAT",SWIG_From_int((int)(SELECTION_FLAT)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_NOCASE",SWIG_From_int((int)(SELECTION_NOCASE)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_SKIP_KIND",SWIG_From_int((int)(SELECTION_SKIP_KIND)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_MATCH_DEPSTR",SWIG_From_int((int)(SELECTION_MATCH_DEPSTR)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_SOURCE_ONLY",SWIG_From_int((int)(SELECTION_SOURCE_ONLY)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_WITH_SOURCE",SWIG_From_int((int)(SELECTION_WITH_SOURCE)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_WITH_DISABLED",SWIG_From_int((int)(SELECTION_WITH_DISABLED)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_WITH_BADARCH",SWIG_From_int((int)(SELECTION_WITH_BADARCH)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_WITH_ALL",SWIG_From_int((int)(SELECTION_WITH_ALL)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_ADD",SWIG_From_int((int)(SELECTION_ADD)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_SUBTRACT",SWIG_From_int((int)(SELECTION_SUBTRACT)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_FILTER",SWIG_From_int((int)(SELECTION_FILTER)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_FILTER_KEEP_IFEMPTY",SWIG_From_int((int)(SELECTION_FILTER_KEEP_IFEMPTY)));
  SWIG_Python_SetConstant(d, "Selection_SELECTION_FILTER_SWAPPED",SWIG_From_int((int)(SELECTION_FILTER_SWAPPED)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_STRING",SWIG_From_int((int)(SEARCH_STRING)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_STRINGSTART",SWIG_From_int((int)(SEARCH_STRINGSTART)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_STRINGEND",SWIG_From_int((int)(SEARCH_STRINGEND)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_SUBSTRING",SWIG_From_int((int)(SEARCH_SUBSTRING)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_GLOB",SWIG_From_int((int)(SEARCH_GLOB)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_REGEX",SWIG_From_int((int)(SEARCH_REGEX)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_NOCASE",SWIG_From_int((int)(SEARCH_NOCASE)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_FILES",SWIG_From_int((int)(SEARCH_FILES)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_COMPLETE_FILELIST",SWIG_From_int((int)(SEARCH_COMPLETE_FILELIST)));
  SWIG_Python_SetConstant(d, "Dataiterator_SEARCH_CHECKSUMS",SWIG_From_int((int)(SEARCH_CHECKSUMS)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE",SWIG_From_int((int)(SOLVER_SOLVABLE)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE_NAME",SWIG_From_int((int)(SOLVER_SOLVABLE_NAME)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE_PROVIDES",SWIG_From_int((int)(SOLVER_SOLVABLE_PROVIDES)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE_ONE_OF",SWIG_From_int((int)(SOLVER_SOLVABLE_ONE_OF)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE_REPO",SWIG_From_int((int)(SOLVER_SOLVABLE_REPO)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SOLVABLE_ALL",SWIG_From_int((int)(SOLVER_SOLVABLE_ALL)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SELECTMASK",SWIG_From_int((int)(SOLVER_SELECTMASK)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_NOOP",SWIG_From_int((int)(SOLVER_NOOP)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_INSTALL",SWIG_From_int((int)(SOLVER_INSTALL)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_ERASE",SWIG_From_int((int)(SOLVER_ERASE)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_UPDATE",SWIG_From_int((int)(SOLVER_UPDATE)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_WEAKENDEPS",SWIG_From_int((int)(SOLVER_WEAKENDEPS)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_MULTIVERSION",SWIG_From_int((int)(SOLVER_MULTIVERSION)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_LOCK",SWIG_From_int((int)(SOLVER_LOCK)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_DISTUPGRADE",SWIG_From_int((int)(SOLVER_DISTUPGRADE)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_VERIFY",SWIG_From_int((int)(SOLVER_VERIFY)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_DROP_ORPHANED",SWIG_From_int((int)(SOLVER_DROP_ORPHANED)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_USERINSTALLED",SWIG_From_int((int)(SOLVER_USERINSTALLED)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_ALLOWUNINSTALL",SWIG_From_int((int)(SOLVER_ALLOWUNINSTALL)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_FAVOR",SWIG_From_int((int)(SOLVER_FAVOR)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_DISFAVOR",SWIG_From_int((int)(SOLVER_DISFAVOR)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_EXCLUDEFROMWEAK",SWIG_From_int((int)(SOLVER_EXCLUDEFROMWEAK)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_JOBMASK",SWIG_From_int((int)(SOLVER_JOBMASK)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_WEAK",SWIG_From_int((int)(SOLVER_WEAK)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_ESSENTIAL",SWIG_From_int((int)(SOLVER_ESSENTIAL)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_CLEANDEPS",SWIG_From_int((int)(SOLVER_CLEANDEPS)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_FORCEBEST",SWIG_From_int((int)(SOLVER_FORCEBEST)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_TARGETED",SWIG_From_int((int)(SOLVER_TARGETED)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_NOTBYUSER",SWIG_From_int((int)(SOLVER_NOTBYUSER)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETEV",SWIG_From_int((int)(SOLVER_SETEV)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETEVR",SWIG_From_int((int)(SOLVER_SETEVR)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETARCH",SWIG_From_int((int)(SOLVER_SETARCH)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETVENDOR",SWIG_From_int((int)(SOLVER_SETVENDOR)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETREPO",SWIG_From_int((int)(SOLVER_SETREPO)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETNAME",SWIG_From_int((int)(SOLVER_SETNAME)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_NOAUTOSET",SWIG_From_int((int)(SOLVER_NOAUTOSET)));
  SWIG_Python_SetConstant(d, "Job_SOLVER_SETMASK",SWIG_From_int((int)(SOLVER_SETMASK)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_PROMOTEEPOCH",SWIG_From_int((int)(POOL_FLAG_PROMOTEEPOCH)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_FORBIDSELFCONFLICTS",SWIG_From_int((int)(POOL_FLAG_FORBIDSELFCONFLICTS)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_OBSOLETEUSESPROVIDES",SWIG_From_int((int)(POOL_FLAG_OBSOLETEUSESPROVIDES)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_IMPLICITOBSOLETEUSESPROVIDES",SWIG_From_int((int)(POOL_FLAG_IMPLICITOBSOLETEUSESPROVIDES)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_OBSOLETEUSESCOLORS",SWIG_From_int((int)(POOL_FLAG_OBSOLETEUSESCOLORS)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_IMPLICITOBSOLETEUSESCOLORS",SWIG_From_int((int)(POOL_FLAG_IMPLICITOBSOLETEUSESCOLORS)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_NOINSTALLEDOBSOLETES",SWIG_From_int((int)(POOL_FLAG_NOINSTALLEDOBSOLETES)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_HAVEDISTEPOCH",SWIG_From_int((int)(POOL_FLAG_HAVEDISTEPOCH)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_NOOBSOLETESMULTIVERSION",SWIG_From_int((int)(POOL_FLAG_NOOBSOLETESMULTIVERSION)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_ADDFILEPROVIDESFILTERED",SWIG_From_int((int)(POOL_FLAG_ADDFILEPROVIDESFILTERED)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_NOWHATPROVIDESAUX",SWIG_From_int((int)(POOL_FLAG_NOWHATPROVIDESAUX)));
  SWIG_Python_SetConstant(d, "Pool_POOL_FLAG_WHATPROVIDESWITHDISABLED",SWIG_From_int((int)(POOL_FLAG_WHATPROVIDESWITHDISABLED)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_RPM",SWIG_From_int((int)(DISTTYPE_RPM)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_DEB",SWIG_From_int((int)(DISTTYPE_DEB)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_ARCH",SWIG_From_int((int)(DISTTYPE_ARCH)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_HAIKU",SWIG_From_int((int)(DISTTYPE_HAIKU)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_CONDA",SWIG_From_int((int)(DISTTYPE_CONDA)));
  SWIG_Python_SetConstant(d, "Pool_DISTTYPE_APK",SWIG_From_int((int)(DISTTYPE_APK)));
  SWIG_Python_SetConstant(d, "Repo_REPO_REUSE_REPODATA",SWIG_From_int((int)(REPO_REUSE_REPODATA)));
  SWIG_Python_SetConstant(d, "Repo_REPO_NO_INTERNALIZE",SWIG_From_int((int)(REPO_NO_INTERNALIZE)));
  SWIG_Python_SetConstant(d, "Repo_REPO_LOCALPOOL",SWIG_From_int((int)(REPO_LOCALPOOL)));
  SWIG_Python_SetConstant(d, "Repo_REPO_USE_LOADING",SWIG_From_int((int)(REPO_USE_LOADING)));
  SWIG_Python_SetConstant(d, "Repo_REPO_EXTEND_SOLVABLES",SWIG_From_int((int)(REPO_EXTEND_SOLVABLES)));
  SWIG_Python_SetConstant(d, "Repo_REPO_USE_ROOTDIR",SWIG_From_int((int)(REPO_USE_ROOTDIR)));
  SWIG_Python_SetConstant(d, "Repo_REPO_NO_LOCATION",SWIG_From_int((int)(REPO_NO_LOCATION)));
  SWIG_Python_SetConstant(d, "Repo_SOLV_ADD_NO_STUBS",SWIG_From_int((int)(SOLV_ADD_NO_STUBS)));
  SWIG_Python_SetConstant(d, "Repo_SUSETAGS_RECORD_SHARES",SWIG_From_int((int)(SUSETAGS_RECORD_SHARES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_UNKNOWN",SWIG_From_int((int)(SOLVER_RULE_UNKNOWN)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG",SWIG_From_int((int)(SOLVER_RULE_PKG)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_NOT_INSTALLABLE",SWIG_From_int((int)(SOLVER_RULE_PKG_NOT_INSTALLABLE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_NOTHING_PROVIDES_DEP",SWIG_From_int((int)(SOLVER_RULE_PKG_NOTHING_PROVIDES_DEP)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_REQUIRES",SWIG_From_int((int)(SOLVER_RULE_PKG_REQUIRES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_SELF_CONFLICT",SWIG_From_int((int)(SOLVER_RULE_PKG_SELF_CONFLICT)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_CONFLICTS",SWIG_From_int((int)(SOLVER_RULE_PKG_CONFLICTS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_SAME_NAME",SWIG_From_int((int)(SOLVER_RULE_PKG_SAME_NAME)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_OBSOLETES",SWIG_From_int((int)(SOLVER_RULE_PKG_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_IMPLICIT_OBSOLETES",SWIG_From_int((int)(SOLVER_RULE_PKG_IMPLICIT_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_INSTALLED_OBSOLETES",SWIG_From_int((int)(SOLVER_RULE_PKG_INSTALLED_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_RECOMMENDS",SWIG_From_int((int)(SOLVER_RULE_PKG_RECOMMENDS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_CONSTRAINS",SWIG_From_int((int)(SOLVER_RULE_PKG_CONSTRAINS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_PKG_SUPPLEMENTS",SWIG_From_int((int)(SOLVER_RULE_PKG_SUPPLEMENTS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_UPDATE",SWIG_From_int((int)(SOLVER_RULE_UPDATE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_FEATURE",SWIG_From_int((int)(SOLVER_RULE_FEATURE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_JOB",SWIG_From_int((int)(SOLVER_RULE_JOB)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_JOB_NOTHING_PROVIDES_DEP",SWIG_From_int((int)(SOLVER_RULE_JOB_NOTHING_PROVIDES_DEP)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_JOB_PROVIDED_BY_SYSTEM",SWIG_From_int((int)(SOLVER_RULE_JOB_PROVIDED_BY_SYSTEM)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_JOB_UNKNOWN_PACKAGE",SWIG_From_int((int)(SOLVER_RULE_JOB_UNKNOWN_PACKAGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_JOB_UNSUPPORTED",SWIG_From_int((int)(SOLVER_RULE_JOB_UNSUPPORTED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_DISTUPGRADE",SWIG_From_int((int)(SOLVER_RULE_DISTUPGRADE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_INFARCH",SWIG_From_int((int)(SOLVER_RULE_INFARCH)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_CHOICE",SWIG_From_int((int)(SOLVER_RULE_CHOICE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_LEARNT",SWIG_From_int((int)(SOLVER_RULE_LEARNT)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_BEST",SWIG_From_int((int)(SOLVER_RULE_BEST)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_YUMOBS",SWIG_From_int((int)(SOLVER_RULE_YUMOBS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_RECOMMENDS",SWIG_From_int((int)(SOLVER_RULE_RECOMMENDS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_BLACK",SWIG_From_int((int)(SOLVER_RULE_BLACK)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_STRICT_REPO_PRIORITY",SWIG_From_int((int)(SOLVER_RULE_STRICT_REPO_PRIORITY)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_JOB",SWIG_From_int((int)(SOLVER_SOLUTION_JOB)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_POOLJOB",SWIG_From_int((int)(SOLVER_SOLUTION_POOLJOB)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_INFARCH",SWIG_From_int((int)(SOLVER_SOLUTION_INFARCH)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_DISTUPGRADE",SWIG_From_int((int)(SOLVER_SOLUTION_DISTUPGRADE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_BEST",SWIG_From_int((int)(SOLVER_SOLUTION_BEST)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_ERASE",SWIG_From_int((int)(SOLVER_SOLUTION_ERASE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_REPLACE",SWIG_From_int((int)(SOLVER_SOLUTION_REPLACE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_REPLACE_DOWNGRADE",SWIG_From_int((int)(SOLVER_SOLUTION_REPLACE_DOWNGRADE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_REPLACE_ARCHCHANGE",SWIG_From_int((int)(SOLVER_SOLUTION_REPLACE_ARCHCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_REPLACE_VENDORCHANGE",SWIG_From_int((int)(SOLVER_SOLUTION_REPLACE_VENDORCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_SOLUTION_REPLACE_NAMECHANGE",SWIG_From_int((int)(SOLVER_SOLUTION_REPLACE_NAMECHANGE)));
  SWIG_Python_SetConstant(d, "Solver_POLICY_ILLEGAL_DOWNGRADE",SWIG_From_int((int)(POLICY_ILLEGAL_DOWNGRADE)));
  SWIG_Python_SetConstant(d, "Solver_POLICY_ILLEGAL_ARCHCHANGE",SWIG_From_int((int)(POLICY_ILLEGAL_ARCHCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_POLICY_ILLEGAL_VENDORCHANGE",SWIG_From_int((int)(POLICY_ILLEGAL_VENDORCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_POLICY_ILLEGAL_NAMECHANGE",SWIG_From_int((int)(POLICY_ILLEGAL_NAMECHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ALLOW_DOWNGRADE",SWIG_From_int((int)(SOLVER_FLAG_ALLOW_DOWNGRADE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ALLOW_ARCHCHANGE",SWIG_From_int((int)(SOLVER_FLAG_ALLOW_ARCHCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ALLOW_VENDORCHANGE",SWIG_From_int((int)(SOLVER_FLAG_ALLOW_VENDORCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ALLOW_NAMECHANGE",SWIG_From_int((int)(SOLVER_FLAG_ALLOW_NAMECHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ALLOW_UNINSTALL",SWIG_From_int((int)(SOLVER_FLAG_ALLOW_UNINSTALL)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_NO_UPDATEPROVIDE",SWIG_From_int((int)(SOLVER_FLAG_NO_UPDATEPROVIDE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_SPLITPROVIDES",SWIG_From_int((int)(SOLVER_FLAG_SPLITPROVIDES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_IGNORE_RECOMMENDED",SWIG_From_int((int)(SOLVER_FLAG_IGNORE_RECOMMENDED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ADD_ALREADY_RECOMMENDED",SWIG_From_int((int)(SOLVER_FLAG_ADD_ALREADY_RECOMMENDED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_NO_INFARCHCHECK",SWIG_From_int((int)(SOLVER_FLAG_NO_INFARCHCHECK)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_BEST_OBEY_POLICY",SWIG_From_int((int)(SOLVER_FLAG_BEST_OBEY_POLICY)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_NO_AUTOTARGET",SWIG_From_int((int)(SOLVER_FLAG_NO_AUTOTARGET)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_DUP_ALLOW_DOWNGRADE",SWIG_From_int((int)(SOLVER_FLAG_DUP_ALLOW_DOWNGRADE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_DUP_ALLOW_ARCHCHANGE",SWIG_From_int((int)(SOLVER_FLAG_DUP_ALLOW_ARCHCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_DUP_ALLOW_VENDORCHANGE",SWIG_From_int((int)(SOLVER_FLAG_DUP_ALLOW_VENDORCHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_DUP_ALLOW_NAMECHANGE",SWIG_From_int((int)(SOLVER_FLAG_DUP_ALLOW_NAMECHANGE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_KEEP_ORPHANS",SWIG_From_int((int)(SOLVER_FLAG_KEEP_ORPHANS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_BREAK_ORPHANS",SWIG_From_int((int)(SOLVER_FLAG_BREAK_ORPHANS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_FOCUS_INSTALLED",SWIG_From_int((int)(SOLVER_FLAG_FOCUS_INSTALLED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_YUM_OBSOLETES",SWIG_From_int((int)(SOLVER_FLAG_YUM_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_NEED_UPDATEPROVIDE",SWIG_From_int((int)(SOLVER_FLAG_NEED_UPDATEPROVIDE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_FOCUS_BEST",SWIG_From_int((int)(SOLVER_FLAG_FOCUS_BEST)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_STRONG_RECOMMENDS",SWIG_From_int((int)(SOLVER_FLAG_STRONG_RECOMMENDS)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_INSTALL_ALSO_UPDATES",SWIG_From_int((int)(SOLVER_FLAG_INSTALL_ALSO_UPDATES)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_ONLY_NAMESPACE_RECOMMENDED",SWIG_From_int((int)(SOLVER_FLAG_ONLY_NAMESPACE_RECOMMENDED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_STRICT_REPO_PRIORITY",SWIG_From_int((int)(SOLVER_FLAG_STRICT_REPO_PRIORITY)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_FLAG_FOCUS_NEW",SWIG_From_int((int)(SOLVER_FLAG_FOCUS_NEW)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_UNRELATED",SWIG_From_int((int)(SOLVER_REASON_UNRELATED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_UNIT_RULE",SWIG_From_int((int)(SOLVER_REASON_UNIT_RULE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_KEEP_INSTALLED",SWIG_From_int((int)(SOLVER_REASON_KEEP_INSTALLED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_RESOLVE_JOB",SWIG_From_int((int)(SOLVER_REASON_RESOLVE_JOB)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_UPDATE_INSTALLED",SWIG_From_int((int)(SOLVER_REASON_UPDATE_INSTALLED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_CLEANDEPS_ERASE",SWIG_From_int((int)(SOLVER_REASON_CLEANDEPS_ERASE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_RESOLVE",SWIG_From_int((int)(SOLVER_REASON_RESOLVE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_WEAKDEP",SWIG_From_int((int)(SOLVER_REASON_WEAKDEP)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_RESOLVE_ORPHAN",SWIG_From_int((int)(SOLVER_REASON_RESOLVE_ORPHAN)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_RECOMMENDED",SWIG_From_int((int)(SOLVER_REASON_RECOMMENDED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_SUPPLEMENTED",SWIG_From_int((int)(SOLVER_REASON_SUPPLEMENTED)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_UNSOLVABLE",SWIG_From_int((int)(SOLVER_REASON_UNSOLVABLE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_REASON_PREMISE",SWIG_From_int((int)(SOLVER_REASON_PREMISE)));
  SWIG_Python_SetConstant(d, "Solver_SOLVER_RULE_RPM",SWIG_From_int((int)(SOLVER_RULE_RPM)));
  SWIG_Python_SetConstant(d, "Alternative_SOLVER_ALTERNATIVE_TYPE_RULE",SWIG_From_int((int)(SOLVER_ALTERNATIVE_TYPE_RULE)));
  SWIG_Python_SetConstant(d, "Alternative_SOLVER_ALTERNATIVE_TYPE_RECOMMENDS",SWIG_From_int((int)(SOLVER_ALTERNATIVE_TYPE_RECOMMENDS)));
  SWIG_Python_SetConstant(d, "Alternative_SOLVER_ALTERNATIVE_TYPE_SUGGESTS",SWIG_From_int((int)(SOLVER_ALTERNATIVE_TYPE_SUGGESTS)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_IGNORE",SWIG_From_int((int)(SOLVER_TRANSACTION_IGNORE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_ERASE",SWIG_From_int((int)(SOLVER_TRANSACTION_ERASE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_REINSTALLED",SWIG_From_int((int)(SOLVER_TRANSACTION_REINSTALLED)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_DOWNGRADED",SWIG_From_int((int)(SOLVER_TRANSACTION_DOWNGRADED)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_CHANGED",SWIG_From_int((int)(SOLVER_TRANSACTION_CHANGED)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_UPGRADED",SWIG_From_int((int)(SOLVER_TRANSACTION_UPGRADED)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_OBSOLETED",SWIG_From_int((int)(SOLVER_TRANSACTION_OBSOLETED)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_INSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_INSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_REINSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_REINSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_DOWNGRADE",SWIG_From_int((int)(SOLVER_TRANSACTION_DOWNGRADE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_CHANGE",SWIG_From_int((int)(SOLVER_TRANSACTION_CHANGE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_UPGRADE",SWIG_From_int((int)(SOLVER_TRANSACTION_UPGRADE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_OBSOLETES",SWIG_From_int((int)(SOLVER_TRANSACTION_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_MULTIINSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_MULTIINSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_MULTIREINSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_MULTIREINSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_MAXTYPE",SWIG_From_int((int)(SOLVER_TRANSACTION_MAXTYPE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_SHOW_ACTIVE",SWIG_From_int((int)(SOLVER_TRANSACTION_SHOW_ACTIVE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_SHOW_ALL",SWIG_From_int((int)(SOLVER_TRANSACTION_SHOW_ALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_SHOW_OBSOLETES",SWIG_From_int((int)(SOLVER_TRANSACTION_SHOW_OBSOLETES)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_SHOW_MULTIINSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_SHOW_MULTIINSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_CHANGE_IS_REINSTALL",SWIG_From_int((int)(SOLVER_TRANSACTION_CHANGE_IS_REINSTALL)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_OBSOLETE_IS_UPGRADE",SWIG_From_int((int)(SOLVER_TRANSACTION_OBSOLETE_IS_UPGRADE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_MERGE_VENDORCHANGES",SWIG_From_int((int)(SOLVER_TRANSACTION_MERGE_VENDORCHANGES)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_MERGE_ARCHCHANGES",SWIG_From_int((int)(SOLVER_TRANSACTION_MERGE_ARCHCHANGES)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_RPM_ONLY",SWIG_From_int((int)(SOLVER_TRANSACTION_RPM_ONLY)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_ARCHCHANGE",SWIG_From_int((int)(SOLVER_TRANSACTION_ARCHCHANGE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_VENDORCHANGE",SWIG_From_int((int)(SOLVER_TRANSACTION_VENDORCHANGE)));
  SWIG_Python_SetConstant(d, "Transaction_SOLVER_TRANSACTION_KEEP_ORDERDATA",SWIG_From_int((int)(SOLVER_TRANSACTION_KEEP_ORDERDATA)));
  return 0;
}

